@@ -78,6 +78,7 @@ canonical = System.fetch_env!("CANONICAL")
       DynamicSupervisor,
       GenServer,
       Node,
+      PartitionSupervisor,
       Process,
       Registry,
       Supervisor,
@@ -96,6 +97,7 @@ canonical = System.fetch_env!("CANONICAL")
     ],
     "Code & Macros": [
       Code,
+      Code.Fragment,
       Kernel.ParallelCompiler,
       Macro,
       Macro.Env
