@@ -57,7 +57,7 @@ defmodule Calendar do
   representing the microseconds to external format. If the precision is 0,
   it means microseconds must be skipped.
   """
-  @type microsecond :: {non_neg_integer, non_neg_integer}
+  @type microsecond :: {value :: non_neg_integer, precision :: non_neg_integer}
 
   @typedoc "A calendar implementation"
   @type calendar :: module
@@ -189,7 +189,7 @@ defmodule Calendar do
   @doc """
   Calculates the year and era from the given `year`.
   """
-  @callback year_of_era(year) :: {year, era}
+  @callback year_of_era(year, month, day) :: {year, era}
 
   @doc """
   Calculates the day and era from the given `year`, `month`, and `day`.
@@ -365,7 +365,7 @@ defmodule Calendar do
   """
   @doc since: "1.8.0"
   @spec put_time_zone_database(time_zone_database()) :: :ok
-  def put_time_zone_database(database) do
+  def put_time_zone_database(database) when is_atom(database) do
     Application.put_env(:elixir, :time_zone_database, database)
   end
 
@@ -375,7 +375,7 @@ defmodule Calendar do
   @doc since: "1.8.0"
   @spec get_time_zone_database() :: time_zone_database()
   def get_time_zone_database() do
-    Application.get_env(:elixir, :time_zone_database, Calendar.UTCOnlyTimeZoneDatabase)
+    Application.fetch_env!(:elixir, :time_zone_database)
   end
 
   @doc """
@@ -544,8 +544,8 @@ defmodule Calendar do
     parse_modifiers(rest, width, "", parser_data)
   end
 
-  defp parse_modifiers("0" <> rest, width, nil, parser_data) do
-    parse_modifiers(rest, width, ?0, parser_data)
+  defp parse_modifiers("0" <> rest, nil, nil, parser_data) do
+    parse_modifiers(rest, nil, ?0, parser_data)
   end
 
   defp parse_modifiers("_" <> rest, width, nil, parser_data) do
