@@ -2,15 +2,24 @@ defmodule DateTime do
   @moduledoc """
   A datetime implementation with a time zone.
 
-  This datetime can be seen as an ephemeral snapshot
-  of a datetime at a given time zone. For such purposes,
-  it also includes both UTC and Standard offsets, as
-  well as the zone abbreviation field used exclusively
-  for formatting purposes.
+  This datetime can be seen as a snapshot of a date and time
+  at a given time zone. For such purposes, it also includes both
+  UTC and Standard offsets, as well as the zone abbreviation
+  field used exclusively for formatting purposes. Note future
+  datetimes are not necessarily guaranteed to exist, as time
+  zones may change any time in the future due to geopolitical
+  reasons. See the "Datetimes as snapshots" section for more
+  information.
 
   Remember, comparisons in Elixir using `==/2`, `>/2`, `</2` and friends
   are structural and based on the DateTime struct fields. For proper
-  comparison between datetimes, use the `compare/2` function.
+  comparison between datetimes, use the `compare/2` function. The
+  existence of the `compare/2` function in this module also allows
+  using `Enum.min/2` and `Enum.max/2` functions to get the minimum and
+  maximum datetime of an `Enum`. For example:
+
+      iex> Enum.min([~U[2022-01-12 00:01:00.00Z], ~U[2021-01-12 00:01:00.00Z]], DateTime)
+      ~U[2021-01-12 00:01:00.00Z]
 
   Developers should avoid creating the `DateTime` struct directly
   and instead rely on the functions provided by this module as
@@ -25,22 +34,74 @@ defmodule DateTime do
   datetimes and returns `{:error, :utc_only_time_zone_database}`
   for any other time zone.
 
-  Other time zone databases can also be configured. For example,
-  two of the available options are:
+  Other time zone databases can also be configured. Here are some
+  available options and libraries:
 
-    * [`tz`](https://hexdocs.pm/tz/)
-    * [`tzdata`](https://hexdocs.pm/tzdata/)
+    * [`tz`](https://github.com/mathieuprog/tz)
+    * [`tzdata`](https://github.com/lau/tzdata)
+    * [`zoneinfo`](https://github.com/smartrent/zoneinfo) -
+      recommended for embedded devices
 
   To use them, first make sure it is added as a dependency in `mix.exs`.
   It can then be configured either via configuration:
 
-      config :elixir, :time_zone_database, Tzdata.TimeZoneDatabase
+      config :elixir, :time_zone_database, Tz.TimeZoneDatabase
 
   or by calling `Calendar.put_time_zone_database/1`:
 
-      Calendar.put_time_zone_database(Tzdata.TimeZoneDatabase)
+      Calendar.put_time_zone_database(Tz.TimeZoneDatabase)
 
   See the proper names in the library installation instructions.
+
+  ## Datetimes as snapshots
+
+  In the first section, we described datetimes as a "snapshot of
+  a date and time at a given time zone". To understand precisely
+  what we mean, let's see an example.
+
+  Imagine someone in Poland wants to schedule a meeting with someone
+  in Brazil in the next year. The meeting will happen at 2:30 AM
+  in the Polish time zone. At what time will the meeting happen in
+  Brazil?
+
+  You can consult the time zone database today, one year before,
+  using the API in this module and it will give you an answer that
+  is valid right now. However, this answer may not be valid in the
+  future. Why? Because both Brazil and Poland may change their timezone
+  rules, ultimately affecting the result. For example, a country may
+  choose to enter or abandon "Daylight Saving Time", which is a
+  process where we adjust the clock one hour forward or one hour
+  back once per year. Whenener the rules change, the exact instant
+  that 2:30 AM in Polish time will be in Brazil may change.
+
+  In other words, whenever working with future DateTimes, there is
+  no guarantee the results you get will always be correct, until
+  the event actually happens. Therefore, when you ask for a future
+  time, the answers you get are a snapshot that reflects the current
+  state of the time zone rules. For datetimes in the past, this is
+  not a problem, because time zone rules do not change for past
+  events.
+
+  To make matters worse, it may be that the 2:30 AM in Polish time
+  does not actually even exist or it is ambiguous. If a certain
+  time zone observes "Daylight Saving Time", they will move their
+  clock forward once a year. When this happens, there is a whole
+  hour that does not exist. Then, when they move the clock back,
+  there is a certain hour that will happen twice. So if you want
+  to schedule a meeting when this shift back happens, you would
+  need to explicitly say which of the 2:30 AM you precisely mean.
+  Applications that are date and time sensitive, need to take
+  these scenarios into account and correctly communicate them to
+  users.
+
+  The good news is: Elixir contains all of the building blocks
+  necessary to tackle those problems. The default timezone database
+  used by Elixir, `Calendar.UTCOnlyTimeZoneDatabase`, only works
+  with UTC, which does not observe those issues. Once you bring
+  a proper time zone database, the functions in this module will
+  query the database and return the relevant information. For
+  example, look at how `DateTime.new/4` returns different results
+  based on the scenarios described in this section.
   """
 
   @enforce_keys [:year, :month, :day, :hour, :minute, :second] ++
@@ -82,6 +143,9 @@ defmodule DateTime do
   @doc """
   Returns the current datetime in UTC.
 
+  If you want the current time in Unix seconds,
+  use `System.os_time/1` instead.
+
   ## Examples
 
       iex> datetime = DateTime.utc_now()
@@ -450,7 +514,7 @@ defmodule DateTime do
         # we get the last microsecond just before.
         before_naive =
           first_period_until_wall
-          |> Map.put(:microsecond, {999_999, 6})
+          |> Map.replace!(:microsecond, {999_999, 6})
           |> NaiveDateTime.add(-1)
 
         after_naive = second_period_from_wall
@@ -753,10 +817,6 @@ defmodule DateTime do
   It will return the integer with the given unit,
   according to `System.convert_time_unit/3`.
 
-  If you want to get the current time in Unix seconds,
-  do not do `DateTime.utc_now() |> DateTime.to_unix()`.
-  Simply call `System.os_time(:second)` instead.
-
   ## Examples
 
       iex> 1_464_096_368 |> DateTime.from_unix!() |> DateTime.to_unix()
@@ -801,6 +861,8 @@ defmodule DateTime do
 
   """
   @spec to_naive(Calendar.datetime()) :: NaiveDateTime.t()
+  def to_naive(datetime)
+
   def to_naive(%{
         calendar: calendar,
         year: year,
@@ -840,6 +902,8 @@ defmodule DateTime do
 
   """
   @spec to_date(Calendar.datetime()) :: Date.t()
+  def to_date(datetime)
+
   def to_date(%{
         year: year,
         month: month,
@@ -870,6 +934,8 @@ defmodule DateTime do
 
   """
   @spec to_time(Calendar.datetime()) :: Time.t()
+  def to_time(datetime)
+
   def to_time(%{
         year: _,
         month: _,
@@ -1053,6 +1119,10 @@ defmodule DateTime do
       iex> datetime
       ~U[-2015-01-23 21:20:07.123Z]
 
+      iex> {:ok, datetime, 9000} = DateTime.from_iso8601("20150123T235007.123+0230", :basic)
+      iex> datetime
+      ~U[2015-01-23 21:20:07.123Z]
+
       iex> DateTime.from_iso8601("2015-01-23P23:50:07")
       {:error, :invalid_format}
       iex> DateTime.from_iso8601("2015-01-23T23:50:07")
@@ -1066,11 +1136,37 @@ defmodule DateTime do
 
   """
   @doc since: "1.4.0"
-  @spec from_iso8601(String.t(), Calendar.calendar()) ::
+  @spec from_iso8601(String.t(), Calendar.calendar(), :extended | :basic) ::
           {:ok, t, Calendar.utc_offset()} | {:error, atom}
-  def from_iso8601(string, calendar \\ Calendar.ISO) do
+
+  def from_iso8601(string, format_or_calendar \\ Calendar.ISO)
+
+  def from_iso8601(string, format) when format in [:basic, :extended] do
+    from_iso8601(string, Calendar.ISO, format)
+  end
+
+  def from_iso8601(string, calendar) when is_atom(calendar) do
+    from_iso8601(string, calendar, :extended)
+  end
+
+  @doc """
+  Converts to ISO8601 specifying both a calendar and a mode.
+
+  See `from_iso8601/2` for more information.
+
+  ## Examples
+
+      iex> {:ok, datetime, 9000} = DateTime.from_iso8601("2015-01-23T23:50:07,123+02:30", Calendar.ISO, :extended)
+      iex> datetime
+      ~U[2015-01-23 21:20:07.123Z]
+
+      iex> {:ok, datetime, 9000} = DateTime.from_iso8601("20150123T235007.123+0230", Calendar.ISO, :basic)
+      iex> datetime
+      ~U[2015-01-23 21:20:07.123Z]
+  """
+  def from_iso8601(string, calendar, format) do
     with {:ok, {year, month, day, hour, minute, second, microsecond}, offset} <-
-           Calendar.ISO.parse_utc_datetime(string) do
+           Calendar.ISO.parse_utc_datetime(string, format) do
       datetime = %DateTime{
         year: year,
         month: month,
@@ -1291,12 +1387,11 @@ defmodule DateTime do
   @doc """
   Subtracts `datetime2` from `datetime1`.
 
-  The answer can be returned in any `unit` available from `t:System.time_unit/0`.
+  The answer can be returned in any `:day`, `:hour`, `:minute`, or any `unit`
+  available from `t:System.time_unit/0`. The unit is measured according to
+  `Calendar.ISO` and defaults to `:second`.
 
-  Leap seconds are not taken into account.
-
-  This function returns the difference in seconds where seconds are measured
-  according to `Calendar.ISO`.
+  Fractional results are not supported and are truncated.
 
   ## Examples
 
@@ -1310,14 +1405,36 @@ defmodule DateTime do
       18000
       iex> DateTime.diff(dt2, dt1)
       -18000
+      iex> DateTime.diff(dt1, dt2, :hour)
+      5
+      iex> DateTime.diff(dt2, dt1, :hour)
+      -5
 
   """
   @doc since: "1.5.0"
-  @spec diff(Calendar.datetime(), Calendar.datetime(), System.time_unit()) :: integer()
+  @spec diff(
+          Calendar.datetime(),
+          Calendar.datetime(),
+          :day | :hour | :minute | System.time_unit()
+        ) :: integer()
+  def diff(datetime1, datetime2, unit \\ :second)
+
+  def diff(datetime1, datetime2, :day) do
+    diff(datetime1, datetime2, :second) |> div(86400)
+  end
+
+  def diff(datetime1, datetime2, :hour) do
+    diff(datetime1, datetime2, :second) |> div(3600)
+  end
+
+  def diff(datetime1, datetime2, :minute) do
+    diff(datetime1, datetime2, :second) |> div(60)
+  end
+
   def diff(
         %{utc_offset: utc_offset1, std_offset: std_offset1} = datetime1,
         %{utc_offset: utc_offset2, std_offset: std_offset2} = datetime2,
-        unit \\ :second
+        unit
       ) do
     naive_diff =
       (datetime1 |> to_iso_days() |> Calendar.ISO.iso_days_to_unit(unit)) -
@@ -1330,15 +1447,30 @@ defmodule DateTime do
   @doc """
   Adds a specified amount of time to a `DateTime`.
 
-  Accepts an `amount_to_add` in any `unit` available from `t:System.time_unit/0`.
-  Negative values will move backwards in time.
+  Accepts an `amount_to_add` in any `unit`. `unit` can be `:day`,
+  `:hour`, `:minute`, `:second` or any subsecond precision from
+  `t:System.time_unit/0`. It defaults to `:second`. Negative values
+  will move backwards in time.
 
-  Takes changes such as summer time/DST into account. This means that adding time
-  can cause the wall time to "go backwards" during "fall back" during autumn.
-  Adding just a few seconds to a datetime just before "spring forward" can cause wall
-  time to increase by more than an hour.
+  This function always consider the unit to be computed according
+  to the `Calendar.ISO`.
 
-  Fractional second precision stays the same in a similar way to `NaiveDateTime.add/2`.
+  This function uses relies on a contiguous representation of time,
+  ignoring the wall time and timezone changes. For example, if you add
+  one day when there are summer time/daylight saving time changes,
+  it will also change the time forward or backward by one hour,
+  so the ellapsed time is precisely 24 hours. Similarly, adding just
+  a few seconds to a datetime just before "spring forward" can cause
+  wall time to increase by more than an hour.
+
+  While this means this function is precise in terms of ellapsed time,
+  its result may be misleading in certain use cases. For example, if a
+  user requests a meeting to happen every day at 15:00 and you use this
+  function to compute all future meetings by adding day after day, this
+  function may change the meeting time to 14:00 or 16:00 if there are
+  changes to the current timezone. Computing of recurring datetimes is
+  not currently supported in Elixir's standard library but it is available
+  by third-party libraries.
 
   ### Examples
 
@@ -1349,15 +1481,26 @@ defmodule DateTime do
       iex> DateTime.add(~U[2018-11-15 10:00:00Z], 3600, :second)
       ~U[2018-11-15 11:00:00Z]
 
-  When adding 3 seconds just before "spring forward" we go from 1:59:59 to 3:00:02
+  When adding 3 seconds just before "spring forward" we go from 1:59:59 to 3:00:02:
 
       iex> dt = DateTime.from_naive!(~N[2019-03-31 01:59:59.123], "Europe/Copenhagen", FakeTimeZoneDatabase)
       iex> dt |> DateTime.add(3, :second, FakeTimeZoneDatabase)
       #DateTime<2019-03-31 03:00:02.123+02:00 CEST Europe/Copenhagen>
 
+  When adding 1 day during "spring forward", the hour also changes:
+
+      iex> dt = DateTime.from_naive!(~N[2019-03-31 01:00:00], "Europe/Copenhagen", FakeTimeZoneDatabase)
+      iex> dt |> DateTime.add(1, :day, FakeTimeZoneDatabase)
+      #DateTime<2019-04-01 02:00:00+02:00 CEST Europe/Copenhagen>
+
   """
   @doc since: "1.8.0"
-  @spec add(Calendar.datetime(), integer, System.time_unit(), Calendar.time_zone_database()) ::
+  @spec add(
+          Calendar.datetime(),
+          integer,
+          :day | :hour | :minute | System.time_unit(),
+          Calendar.time_zone_database()
+        ) ::
           t()
   def add(
         datetime,
@@ -1365,7 +1508,20 @@ defmodule DateTime do
         unit \\ :second,
         time_zone_database \\ Calendar.get_time_zone_database()
       )
-      when is_integer(amount_to_add) do
+
+  def add(datetime, amount_to_add, :day, time_zone_database) when is_integer(amount_to_add) do
+    add(datetime, amount_to_add * 86400, :second, time_zone_database)
+  end
+
+  def add(datetime, amount_to_add, :hour, time_zone_database) when is_integer(amount_to_add) do
+    add(datetime, amount_to_add * 3600, :second, time_zone_database)
+  end
+
+  def add(datetime, amount_to_add, :minute, time_zone_database) when is_integer(amount_to_add) do
+    add(datetime, amount_to_add * 60, :second, time_zone_database)
+  end
+
+  def add(datetime, amount_to_add, unit, time_zone_database) when is_integer(amount_to_add) do
     %{
       utc_offset: utc_offset,
       std_offset: std_offset,
