@@ -22,7 +22,7 @@ defmodule Code.Formatter do
   @no_newline_binary_operators [:\\, :in]
 
   # Left associative operators that start on the next line in case of breaks (always pipes)
-  @pipeline_operators [:|>, :~>>, :<<~, :~>, :<~, :<~>, :<|>]
+  @pipeline_operators [:|>, :~>>, :<<~, :~>, :<~, :<~>, :"<|>"]
 
   # Right associative operators that start on the next line in case of breaks
   @right_new_line_before_binary_operators [:|, :when]
@@ -43,8 +43,8 @@ defmodule Code.Formatter do
     :<<~,
     :~>>,
     :<~>,
-    :<|>,
-    :^^^,
+    :"<|>",
+    :"^^^",
     :+++,
     :---,
     :in,
@@ -141,172 +141,113 @@ defmodule Code.Formatter do
 
   @do_end_keywords [:rescue, :catch, :else, :after]
 
+  @bitstring_modifiers [
+    :integer,
+    :float,
+    :bits,
+    :bitstring,
+    :binary,
+    :bytes,
+    :utf8,
+    :utf16,
+    :utf32,
+    :signed,
+    :unsigned,
+    :little,
+    :big,
+    :native,
+    :_
+  ]
+
   @doc """
-  Checks if two strings are equivalent.
+  Converts the quoted expression into an algebra document.
   """
-  def equivalent(string1, string2) when is_binary(string1) and is_binary(string2) do
-    quoted1 = :elixir.string_to_quoted!(to_charlist(string1), 1, 1, "nofile", [])
-    quoted2 = :elixir.string_to_quoted!(to_charlist(string2), 1, 1, "nofile", [])
+  def to_algebra(quoted, opts \\ []) do
+    comments = Keyword.get(opts, :comments, [])
 
-    case not_equivalent(quoted1, quoted2) do
-      {left, right} -> {:error, left, right}
-      nil -> :ok
-    end
-  end
+    state =
+      comments
+      |> Enum.map(&format_comment/1)
+      |> gather_comments()
+      |> state(opts)
 
-  defp not_equivalent({:__block__, _, [left]}, right) do
-    not_equivalent(left, right)
-  end
-
-  defp not_equivalent(left, {:__block__, _, [right]}) do
-    not_equivalent(left, right)
-  end
-
-  defp not_equivalent({:__block__, _, []}, nil) do
-    nil
-  end
-
-  defp not_equivalent(nil, {:__block__, _, []}) do
-    nil
-  end
-
-  defp not_equivalent([left | lefties], [right | righties]) do
-    not_equivalent(left, right) || not_equivalent(lefties, righties)
-  end
-
-  defp not_equivalent({left_name, _, left_args}, {right_name, _, right_args}) do
-    not_equivalent(left_name, right_name) || not_equivalent(left_args, right_args)
-  end
-
-  defp not_equivalent({left1, left2}, {right1, right2}) do
-    not_equivalent(left1, right1) || not_equivalent(left2, right2)
-  end
-
-  defp not_equivalent(side, side) do
-    nil
-  end
-
-  defp not_equivalent(left, right) do
-    {left, right}
+    {doc, _} = block_to_algebra(quoted, @min_line, @max_line, state)
+    doc
   end
 
   @doc """
-  Converts `string` to an algebra document.
-
-  Returns `{:ok, doc}` or `{:error, parser_error}`.
-
-  See `Code.format_string!/2` for the list of options.
+  Lists all default locals without parens.
   """
-  def to_algebra(string, opts \\ []) when is_binary(string) and is_list(opts) do
-    file = Keyword.get(opts, :file, "nofile")
-    line = Keyword.get(opts, :line, 1)
-    charlist = String.to_charlist(string)
-
-    Process.put(:code_formatter_comments, [])
-
-    tokenizer_options = [
-      unescape: false,
-      preserve_comments: &preserve_comments/5,
-      warn_on_unnecessary_quotes: false
-    ]
-
-    parser_options = [
-      literal_encoder: &{:ok, {:__block__, &2, [&1]}},
-      token_metadata: true
-    ]
-
-    with {:ok, tokens} <- :elixir.string_to_tokens(charlist, line, 1, file, tokenizer_options),
-         {:ok, forms} <- :elixir.tokens_to_quoted(tokens, file, parser_options) do
-      state =
-        Process.get(:code_formatter_comments)
-        |> Enum.reverse()
-        |> gather_comments()
-        |> state(opts)
-
-      {doc, _} = block_to_algebra(forms, @min_line, @max_line, state)
-      {:ok, doc}
-    end
-  after
-    Process.delete(:code_formatter_comments)
+  def locals_without_parens do
+    @locals_without_parens
   end
 
   @doc """
-  Converts `string` to an algebra document.
-
-  Raises if the `string` cannot be parsed.
-
-  See `Code.format_string!/2` for the list of options.
+  Checks if a function is a local without parens.
   """
-  def to_algebra!(string, opts \\ []) do
-    case to_algebra(string, opts) do
-      {:ok, doc} ->
-        doc
-
-      {:error, {location, error, token}} ->
-        :elixir_errors.parse_error(location, Keyword.get(opts, :file, "nofile"), error, token)
-    end
+  def local_without_parens?(fun, arity, locals_without_parens) do
+    arity > 0 and
+      Enum.any?(locals_without_parens, fn {key, val} ->
+        key == fun and (val == :* or val == arity)
+      end)
   end
 
   defp state(comments, opts) do
     force_do_end_blocks = Keyword.get(opts, :force_do_end_blocks, false)
+    locals_without_parens = Keyword.get(opts, :locals_without_parens, [])
+    file = Keyword.get(opts, :file, nil)
+    sigils = Keyword.get(opts, :sigils, [])
+    normalize_bitstring_modifiers = Keyword.get(opts, :normalize_bitstring_modifiers, true)
 
-    locals_without_parens =
-      Keyword.get(opts, :locals_without_parens, []) ++ @locals_without_parens
+    sigils =
+      Map.new(sigils, fn {key, value} ->
+        with true <- is_atom(key) and is_function(value, 2),
+             [char] <- Atom.to_charlist(key),
+             true <- char in ?A..?Z do
+          {char, value}
+        else
+          _ ->
+            raise ArgumentError,
+                  ":sigils must be a keyword list with a single uppercased letter as key and an " <>
+                    "anonymous function expecting two arguments as value, got: #{inspect(sigils)}"
+        end
+      end)
 
     %{
       force_do_end_blocks: force_do_end_blocks,
-      locals_without_parens: locals_without_parens,
+      locals_without_parens: locals_without_parens ++ locals_without_parens(),
       operand_nesting: 2,
       skip_eol: false,
-      comments: comments
+      comments: comments,
+      sigils: sigils,
+      file: file,
+      normalize_bitstring_modifiers: normalize_bitstring_modifiers
     }
   end
 
-  # Code comment handling
-
-  defp preserve_comments(line, _column, tokens, comment, rest) do
-    comments = Process.get(:code_formatter_comments)
-    comment = {line, {previous_eol(tokens), next_eol(rest, 0)}, format_comment(comment, [])}
-    Process.put(:code_formatter_comments, [comment | comments])
+  defp format_comment(%{text: text} = comment) do
+    %{comment | text: format_comment_text(text)}
   end
 
-  defp next_eol('\s' ++ rest, count), do: next_eol(rest, count)
-  defp next_eol('\t' ++ rest, count), do: next_eol(rest, count)
-  defp next_eol('\n' ++ rest, count), do: next_eol(rest, count + 1)
-  defp next_eol('\r\n' ++ rest, count), do: next_eol(rest, count + 1)
-  defp next_eol(_, count), do: count
-
-  defp previous_eol([{token, {_, _, count}} | _])
-       when token in [:eol, :",", :";"] and count > 0 do
-    count
-  end
-
-  defp previous_eol([]), do: 1
-  defp previous_eol(_), do: nil
-
-  defp format_comment('##' ++ rest, acc), do: format_comment([?# | rest], [?# | acc])
-
-  defp format_comment('#!', acc), do: reverse_to_string(acc, '#!')
-  defp format_comment('#! ' ++ _ = rest, acc), do: reverse_to_string(acc, rest)
-  defp format_comment('#!' ++ rest, acc), do: reverse_to_string(acc, [?#, ?!, ?\s, rest])
-
-  defp format_comment('#', acc), do: reverse_to_string(acc, '#')
-  defp format_comment('# ' ++ _ = rest, acc), do: reverse_to_string(acc, rest)
-  defp format_comment('#' ++ rest, acc), do: reverse_to_string(acc, [?#, ?\s, rest])
-
-  defp reverse_to_string(acc, prefix) do
-    acc |> Enum.reverse(prefix) |> List.to_string()
-  end
+  defp format_comment_text("#"), do: "#"
+  defp format_comment_text("#!" <> rest), do: "#!" <> rest
+  defp format_comment_text("##" <> rest), do: "#" <> format_comment_text("#" <> rest)
+  defp format_comment_text("# " <> rest), do: "# " <> rest
+  defp format_comment_text("#" <> rest), do: "# " <> rest
 
   # If there is a no new line before, we can't gather all followup comments.
-  defp gather_comments([{line, {nil, next_eol}, doc} | comments]) do
-    comment = {line, {@newlines, next_eol}, doc}
+  defp gather_comments([%{previous_eol_count: 0} = comment | comments]) do
+    comment = %{comment | previous_eol_count: @newlines}
     [comment | gather_comments(comments)]
   end
 
-  defp gather_comments([{line, {previous_eol, next_eol}, doc} | comments]) do
-    {next_eol, comments, doc} = gather_followup_comments(line + 1, next_eol, comments, doc)
-    comment = {line, {previous_eol, next_eol}, doc}
+  defp gather_comments([comment | comments]) do
+    %{line: line, next_eol_count: next_eol_count, text: doc} = comment
+
+    {next_eol_count, comments, doc} =
+      gather_followup_comments(line + 1, next_eol_count, comments, doc)
+
+    comment = %{comment | next_eol_count: next_eol_count, text: doc}
     [comment | gather_comments(comments)]
   end
 
@@ -314,13 +255,14 @@ defmodule Code.Formatter do
     []
   end
 
-  defp gather_followup_comments(line, _, [{line, {previous_eol, next_eol}, text} | comments], doc)
-       when previous_eol != nil do
-    gather_followup_comments(line + 1, next_eol, comments, line(doc, text))
+  defp gather_followup_comments(line, _, [%{line: line} = comment | comments], doc)
+       when comment.previous_eol_count != 0 do
+    %{next_eol_count: next_eol_count, text: text} = comment
+    gather_followup_comments(line + 1, next_eol_count, comments, line(doc, text))
   end
 
-  defp gather_followup_comments(_line, next_eol, comments, doc) do
-    {next_eol, comments, doc}
+  defp gather_followup_comments(_line, next_eol_count, comments, doc) do
+    {next_eol_count, comments, doc}
   end
 
   # Special AST nodes from compiler feedback
@@ -350,7 +292,7 @@ defmodule Code.Formatter do
         {doc, state} =
           entries
           |> prepend_heredoc_line()
-          |> interpolation_to_algebra(:heredoc, state, @double_heredoc, @double_heredoc)
+          |> interpolation_to_algebra(~s["""], state, @double_heredoc, @double_heredoc)
 
         {force_unfit(doc), state}
 
@@ -372,7 +314,7 @@ defmodule Code.Formatter do
         {doc, state} =
           entries
           |> prepend_heredoc_line()
-          |> list_interpolation_to_algebra(:heredoc, state, @single_heredoc, @single_heredoc)
+          |> list_interpolation_to_algebra(~s['''], state, @single_heredoc, @single_heredoc)
 
         {force_unfit(doc), state}
 
@@ -435,7 +377,7 @@ defmodule Code.Formatter do
   defp quoted_to_algebra({:__block__, meta, [list]}, _context, state) when is_list(list) do
     case meta[:delimiter] do
       ~s['''] ->
-        string = list |> List.to_string() |> escape_heredoc()
+        string = list |> List.to_string() |> escape_heredoc(~s['''])
         {@single_heredoc |> concat(string) |> concat(@single_heredoc) |> force_unfit(), state}
 
       ~s['] ->
@@ -449,7 +391,7 @@ defmodule Code.Formatter do
 
   defp quoted_to_algebra({:__block__, meta, [string]}, _context, state) when is_binary(string) do
     if meta[:delimiter] == ~s["""] do
-      string = escape_heredoc(string)
+      string = escape_heredoc(string, ~s["""])
       {@double_heredoc |> concat(string) |> concat(@double_heredoc) |> force_unfit(), state}
     else
       string = escape_string(string, @double_quote)
@@ -457,8 +399,8 @@ defmodule Code.Formatter do
     end
   end
 
-  defp quoted_to_algebra({:__block__, _, [atom]}, _context, state) when is_atom(atom) do
-    {atom_to_algebra(atom), state}
+  defp quoted_to_algebra({:__block__, meta, [atom]}, _context, state) when is_atom(atom) do
+    {atom_to_algebra(atom, meta), state}
   end
 
   defp quoted_to_algebra({:__block__, meta, [integer]}, _context, state)
@@ -523,6 +465,15 @@ defmodule Code.Formatter do
     binary_op_to_algebra(:in, "not in", meta, left, right, context, state)
   end
 
+  # ..
+  defp quoted_to_algebra({:.., _meta, []}, context, state) do
+    if context in [:no_parens_arg, :no_parens_one_arg] do
+      {"(..)", state}
+    else
+      {"..", state}
+    end
+  end
+
   # 1..2//3
   defp quoted_to_algebra({:"..//", meta, [left, middle, right]}, context, state) do
     quoted_to_algebra({:"//", meta, [{:.., meta, [left, middle]}, right]}, context, state)
@@ -567,12 +518,10 @@ defmodule Code.Formatter do
 
             {:__block__, _, [atom]} when is_atom(atom) ->
               key =
-                case Code.Identifier.classify(atom) do
-                  type when type in [:callable_local, :callable_operator, :not_callable] ->
-                    IO.iodata_to_binary([Atom.to_string(atom), ?:])
-
-                  _ ->
-                    IO.iodata_to_binary([?", Atom.to_string(atom), ?", ?:])
+                if Macro.classify_atom(atom) in [:identifier, :unquoted] do
+                  IO.iodata_to_binary([Atom.to_string(atom), ?:])
+                else
+                  IO.iodata_to_binary([?", Atom.to_string(atom), ?", ?:])
                 end
 
               {string(key), state}
@@ -598,6 +547,11 @@ defmodule Code.Formatter do
     {doc, state}
   end
 
+  # #PID's and #Ref's may appear on regular AST
+  defp quoted_to_algebra(unknown, _context, state) do
+    {inspect(unknown), state}
+  end
+
   ## Blocks
 
   defp block_to_algebra([{:->, _, _} | _] = type_fun, min_line, max_line, state) do
@@ -940,7 +894,7 @@ defmodule Code.Formatter do
   # @foo(bar)
   defp module_attribute_to_algebra(meta, {name, call_meta, [_] = args} = expr, context, state)
        when is_atom(name) and name not in [:__block__, :__aliases__] do
-    if Code.Identifier.classify(name) == :callable_local do
+    if Macro.classify_atom(name) == :identifier do
       {{call_doc, state}, wrap_in_parens?} =
         call_args_to_algebra(args, call_meta, context, :skip_unless_many_args, false, state)
 
@@ -985,7 +939,7 @@ defmodule Code.Formatter do
        )
        when is_atom(fun) and is_integer(arity) do
     {target_doc, state} = remote_target_to_algebra(target, state)
-    fun = Code.Identifier.inspect_as_function(fun)
+    fun = Macro.inspect_atom(:remote_call, fun)
     {target_doc |> nest(1) |> concat(string(".#{fun}/#{arity}")), state}
   end
 
@@ -1030,7 +984,7 @@ defmodule Code.Formatter do
   defp remote_to_algebra({{:., _, [target, fun]}, meta, args}, context, state)
        when is_atom(fun) do
     {target_doc, state} = remote_target_to_algebra(target, state)
-    fun = Code.Identifier.inspect_as_function(fun)
+    fun = Macro.inspect_atom(:remote_call, fun)
     remote_doc = target_doc |> concat(".") |> concat(string(fun))
 
     if args == [] and not remote_target_is_a_module?(target) and not meta?(meta, :closing) do
@@ -1080,9 +1034,14 @@ defmodule Code.Formatter do
   defp local_to_algebra(fun, meta, args, context, state) when is_atom(fun) do
     skip_parens =
       cond do
-        meta?(meta, :closing) -> :skip_if_only_do_end
-        local_without_parens?(fun, args, state) -> :skip_unless_many_args
-        true -> :skip_if_do_end
+        meta?(meta, :closing) ->
+          :skip_if_only_do_end
+
+        local_without_parens?(fun, length(args), state.locals_without_parens) ->
+          :skip_unless_many_args
+
+        true ->
+          :skip_if_do_end
       end
 
     {{call_doc, state}, wrap_in_parens?} =
@@ -1261,15 +1220,6 @@ defmodule Code.Formatter do
     end
   end
 
-  defp local_without_parens?(fun, args, %{locals_without_parens: locals_without_parens}) do
-    length = length(args)
-
-    length > 0 and
-      Enum.any?(locals_without_parens, fn {key, val} ->
-        key == fun and (val == :* or val == length)
-      end)
-  end
-
   defp no_generators?(args) do
     not Enum.any?(args, &match?({:<-, _, [_, _]}, &1))
   end
@@ -1347,7 +1297,7 @@ defmodule Code.Formatter do
 
   defp list_interpolation_to_algebra([entry | entries], escape, state, acc, last) do
     {{:., _, [Kernel, :to_string]}, _meta, [quoted]} = entry
-    {doc, state} = interpolation_to_string(quoted, state)
+    {doc, state} = interpolation_to_algebra(quoted, state)
     list_interpolation_to_algebra(entries, escape, state, concat(acc, doc), last)
   end
 
@@ -1363,7 +1313,7 @@ defmodule Code.Formatter do
 
   defp interpolation_to_algebra([entry | entries], escape, state, acc, last) do
     {:"::", _, [{{:., _, [Kernel, :to_string]}, _meta, [quoted]}, {:binary, _, _}]} = entry
-    {doc, state} = interpolation_to_string(quoted, state)
+    {doc, state} = interpolation_to_algebra(quoted, state)
     interpolation_to_algebra(entries, escape, state, concat(acc, doc), last)
   end
 
@@ -1371,21 +1321,9 @@ defmodule Code.Formatter do
     {concat(acc, last), state}
   end
 
-  defp interpolation_to_string(quoted, %{skip_eol: skip_eol} = state) do
+  defp interpolation_to_algebra(quoted, %{skip_eol: skip_eol} = state) do
     {doc, state} = block_to_algebra(quoted, @max_line, @min_line, %{state | skip_eol: true})
-    doc = interpolation_to_string(surround("\#{", doc, "}"))
-    {doc, %{state | skip_eol: skip_eol}}
-  end
-
-  defp interpolation_to_string(doc) do
-    [head | tail] =
-      doc
-      |> format_to_string()
-      |> String.split("\n")
-
-    Enum.reduce(tail, string(head), fn line, acc ->
-      concat([acc, line(), string(line)])
-    end)
+    {no_limit(surround("\#{", doc, "}")), %{state | skip_eol: skip_eol}}
   end
 
   ## Sigils
@@ -1396,13 +1334,37 @@ defmodule Code.Formatter do
          opening_delimiter when not is_nil(opening_delimiter) <- meta[:delimiter] do
       doc = <<?~, name, opening_delimiter::binary>>
 
+      entries =
+        case state.sigils do
+          %{^name => callback} ->
+            metadata = [
+              file: state.file,
+              line: meta[:line],
+              sigil: List.to_atom([name]),
+              modifiers: modifiers,
+              opening_delimiter: opening_delimiter
+            ]
+
+            case callback.(hd(entries), metadata) do
+              iodata when is_binary(iodata) or is_list(iodata) ->
+                [IO.iodata_to_binary(iodata)]
+
+              other ->
+                raise ArgumentError,
+                      "expected sigil callback to return iodata, got: #{inspect(other)}"
+            end
+
+          %{} ->
+            entries
+        end
+
       if opening_delimiter in [@double_heredoc, @single_heredoc] do
         closing_delimiter = concat(opening_delimiter, List.to_string(modifiers))
 
         {doc, state} =
           entries
           |> prepend_heredoc_line()
-          |> interpolation_to_algebra(:heredoc, state, doc, closing_delimiter)
+          |> interpolation_to_algebra(opening_delimiter, state, doc, closing_delimiter)
 
         {force_unfit(doc), state}
       else
@@ -1470,14 +1432,30 @@ defmodule Code.Formatter do
 
   defp bitstring_spec_to_algebra({op, _, [left, right]}, state) when op in [:-, :*] do
     {left, state} = bitstring_spec_to_algebra(left, state)
-    {right, state} = quoted_to_algebra_with_parens_if_operator(right, :parens_arg, state)
+    {right, state} = bitstring_spec_element_to_algebra(right, state)
     {concat(concat(left, Atom.to_string(op)), right), state}
   end
 
   defp bitstring_spec_to_algebra(spec, state) do
-    quoted_to_algebra_with_parens_if_operator(spec, :parens_arg, state)
+    bitstring_spec_element_to_algebra(spec, state)
   end
 
+  defp bitstring_spec_element_to_algebra(
+         {atom, meta, empty_args},
+         state = %{normalize_bitstring_modifiers: true}
+       )
+       when is_atom(atom) and empty_args in [nil, []] do
+    empty_args = bitstring_spec_normalize_empty_args(atom)
+    quoted_to_algebra_with_parens_if_operator({atom, meta, empty_args}, :parens_arg, state)
+  end
+
+  defp bitstring_spec_element_to_algebra(spec_element, state) do
+    quoted_to_algebra_with_parens_if_operator(spec_element, :parens_arg, state)
+  end
+
+  defp bitstring_spec_normalize_empty_args(atom) when atom in @bitstring_modifiers, do: nil
+  defp bitstring_spec_normalize_empty_args(_atom), do: []
+
   defp bitstring_wrap_parens(doc, i, last) when i == 0 or i == last do
     string = format_to_string(doc)
 
@@ -1545,25 +1523,36 @@ defmodule Code.Formatter do
     end
   end
 
-  defp atom_to_algebra(atom) when atom in [nil, true, false] do
+  defp atom_to_algebra(atom, _) when atom in [nil, true, false] do
     Atom.to_string(atom)
   end
 
   # TODO: Remove this clause in v1.16 when we no longer quote operator :..//
-  defp atom_to_algebra(:"..//") do
+  defp atom_to_algebra(:"..//", _) do
     string(":\"..//\"")
   end
 
-  defp atom_to_algebra(atom) do
+  defp atom_to_algebra(:\\, meta) do
+    # Since we parse strings without unescaping, the atoms
+    # :\\ and :"\\" have the same representation, so we need
+    # to check the delimiter and handle them accordingly.
+    string =
+      case Keyword.get(meta, :delimiter) do
+        "\"" -> ":\"\\\\\""
+        _ -> ":\\\\"
+      end
+
+    string(string)
+  end
+
+  defp atom_to_algebra(atom, _) do
     string = Atom.to_string(atom)
 
     iodata =
-      case Code.Identifier.classify(atom) do
-        type when type in [:callable_local, :callable_operator, :not_callable] ->
-          [?:, string]
-
-        _ ->
-          [?:, ?", String.replace(string, "\"", "\\\""), ?"]
+      if Macro.classify_atom(atom) in [:unquoted, :identifier] do
+        [?:, string]
+      else
+        [?:, ?", String.replace(string, "\"", "\\\""), ?"]
       end
 
     iodata |> IO.iodata_to_binary() |> string()
@@ -1611,11 +1600,13 @@ defmodule Code.Formatter do
     end
   end
 
-  defp escape_heredoc(string) do
+  defp escape_heredoc(string, escape) do
+    string = String.replace(string, escape, "\\" <> escape)
     heredoc_to_algebra(["" | String.split(string, "\n")])
   end
 
-  defp escape_string(string, :heredoc) do
+  defp escape_string(string, <<_, _, _>> = escape) do
+    string = String.replace(string, escape, "\\" <> escape)
     heredoc_to_algebra(String.split(string, "\n"))
   end
 
@@ -1708,12 +1699,15 @@ defmodule Code.Formatter do
        ) do
     min_line = line(meta)
     {body_doc, state} = block_to_algebra(body, min_line, max_line, state)
+    break_or_line = clause_break_or_line(clauses, state)
 
     doc =
       "fn ->"
-      |> glue(body_doc)
+      |> concat(break_or_line)
+      |> concat(body_doc)
       |> nest(2)
-      |> glue("end")
+      |> concat(break_or_line)
+      |> concat("end")
       |> maybe_force_clauses(clauses, state)
       |> group()
 
@@ -1742,12 +1736,16 @@ defmodule Code.Formatter do
       |> nest(:cursor)
       |> group()
 
+    break_or_line = clause_break_or_line(clauses, state)
+
     doc =
       "fn "
       |> concat(head)
-      |> glue(body_doc)
+      |> concat(break_or_line)
+      |> concat(body_doc)
       |> nest(2)
-      |> glue("end")
+      |> concat(break_or_line)
+      |> concat("end")
       |> maybe_force_clauses(clauses, state)
       |> group()
 
@@ -1789,13 +1787,14 @@ defmodule Code.Formatter do
     min_line = line(meta)
     {args_doc, state} = clause_args_to_algebra(args, min_line, state)
     {body_doc, state} = block_to_algebra(body, min_line, max_line, state)
+    break_or_line = clause_break_or_line(clauses, state)
 
     doc =
       args_doc
       |> ungroup_if_group()
       |> concat(" ->")
       |> group()
-      |> concat(break() |> concat(body_doc) |> nest(2))
+      |> concat(break_or_line |> concat(body_doc) |> nest(2))
       |> wrap_in_parens()
       |> maybe_force_clauses(clauses, state)
       |> group()
@@ -1816,12 +1815,21 @@ defmodule Code.Formatter do
 
   ## Clauses
 
+  defp multi_line_clauses?(clauses, state) do
+    Enum.any?(clauses, fn {:->, meta, [_, block]} ->
+      eol?(meta, state) or multi_line_block?(block)
+    end)
+  end
+
+  defp multi_line_block?({:__block__, _, [_, _ | _]}), do: true
+  defp multi_line_block?(_), do: false
+
+  defp clause_break_or_line(clauses, state) do
+    if multi_line_clauses?(clauses, state), do: line(), else: break()
+  end
+
   defp maybe_force_clauses(doc, clauses, state) do
-    if Enum.any?(clauses, fn {:->, meta, _} -> eol?(meta, state) end) do
-      force_unfit(doc)
-    else
-      doc
-    end
+    if multi_line_clauses?(clauses, state), do: force_unfit(doc), else: doc
   end
 
   defp clauses_to_algebra([{:->, _, _} | _] = clauses, min_line, max_line, state) do
@@ -1920,7 +1928,7 @@ defmodule Code.Formatter do
   defp quoted_to_algebra_with_comments(args, acc, min_line, max_line, state, fun) do
     {pre_comments, state} =
       get_and_update_in(state.comments, fn comments ->
-        Enum.split_while(comments, fn {line, _, _} -> line <= min_line end)
+        Enum.split_while(comments, fn %{line: line} -> line <= min_line end)
       end)
 
     {docs, comments?, state} =
@@ -1936,23 +1944,29 @@ defmodule Code.Formatter do
   end
 
   defp each_quoted_to_algebra_with_comments([arg | args], acc, max_line, state, comments?, fun) do
-    {doc_start, doc_end} = traverse_line(arg, {@max_line, @min_line})
+    case traverse_line(arg, {@max_line, @min_line}) do
+      {@max_line, @min_line} ->
+        {doc_triplet, state} = fun.(arg, args, state)
+        acc = [doc_triplet | acc]
+        each_quoted_to_algebra_with_comments(args, acc, max_line, state, comments?, fun)
 
-    {acc, comments, comments?} =
-      extract_comments_before(doc_start, acc, state.comments, comments?)
+      {doc_start, doc_end} ->
+        {acc, comments, comments?} =
+          extract_comments_before(doc_start, acc, state.comments, comments?)
 
-    {doc_triplet, state} = fun.(arg, args, %{state | comments: comments})
+        {doc_triplet, state} = fun.(arg, args, %{state | comments: comments})
 
-    {acc, comments, comments?} =
-      extract_comments_trailing(doc_start, doc_end, acc, state.comments, comments?)
+        {acc, comments, comments?} =
+          extract_comments_trailing(doc_start, doc_end, acc, state.comments, comments?)
 
-    acc = [adjust_trailing_newlines(doc_triplet, doc_end, comments) | acc]
-    state = %{state | comments: comments}
-    each_quoted_to_algebra_with_comments(args, acc, max_line, state, comments?, fun)
+        acc = [adjust_trailing_newlines(doc_triplet, doc_end, comments) | acc]
+        state = %{state | comments: comments}
+        each_quoted_to_algebra_with_comments(args, acc, max_line, state, comments?, fun)
+    end
   end
 
-  defp extract_comments_before(max, acc, [{line, _, _} = comment | rest], _) when line < max do
-    {_, {previous, next}, doc} = comment
+  defp extract_comments_before(max, acc, [%{line: line} = comment | rest], _) when line < max do
+    %{previous_eol_count: previous, next_eol_count: next, text: doc} = comment
     acc = [{doc, @empty, next} | add_previous_to_acc(acc, previous)]
     extract_comments_before(max, acc, rest, true)
   end
@@ -1967,7 +1981,7 @@ defmodule Code.Formatter do
   defp add_previous_to_acc(acc, _previous),
     do: acc
 
-  defp extract_comments_trailing(min, max, acc, [{line, _, doc_comment} | rest], _)
+  defp extract_comments_trailing(min, max, acc, [%{line: line, text: doc_comment} | rest], _)
        when line >= min and line <= max do
     acc = [{doc_comment, @empty, 1} | acc]
     extract_comments_trailing(min, max, acc, rest, true)
@@ -1979,7 +1993,7 @@ defmodule Code.Formatter do
 
   # If the document is immediately followed by comment which is followed by newlines,
   # its newlines wouldn't have considered the comment, so we need to adjust it.
-  defp adjust_trailing_newlines({doc, next_line, newlines}, doc_end, [{line, _, _} | _])
+  defp adjust_trailing_newlines({doc, next_line, newlines}, doc_end, [%{line: line} | _])
        when newlines > 1 and line == doc_end + 1 do
     {doc, next_line, 1}
   end
@@ -2097,7 +2111,7 @@ defmodule Code.Formatter do
 
   defp module_attribute_read?({:@, _, [{var, _, var_context}]})
        when is_atom(var) and is_atom(var_context) do
-    Code.Identifier.classify(var) == :callable_local
+    Macro.classify_atom(var) == :identifier
   end
 
   defp module_attribute_read?(_), do: false
@@ -2186,7 +2200,7 @@ defmodule Code.Formatter do
       (
         min_line = line(meta)
         max_line = closing_line(meta)
-        Enum.any?(comments, fn {line, _, _} -> line > min_line and line < max_line end)
+        Enum.any?(comments, fn %{line: line} -> line > min_line and line < max_line end)
       )
   end
 
@@ -2211,8 +2225,8 @@ defmodule Code.Formatter do
         {{_, arg_meta, _}, _} = hd(arg)
         first_line = line(arg_meta)
 
-        case Enum.drop_while(comments, fn {line, _, _} -> line <= block_line end) do
-          [{line, _, _} | _] when line <= first_line ->
+        case Enum.drop_while(comments, fn %{line: line} -> line <= block_line end) do
+          [%{line: line} | _] when line <= first_line ->
             {false, block}
 
           _ ->
@@ -2230,7 +2244,7 @@ defmodule Code.Formatter do
   end
 
   defp force_args?(args) do
-    match?([_, _ | _], args) and force_args?(args, MapSet.new())
+    match?([_ | _], args) and force_args?(args, %{})
   end
 
   defp force_args?([[arg | _] | args], lines) do
@@ -2240,18 +2254,19 @@ defmodule Code.Formatter do
   defp force_args?([arg | args], lines) do
     line =
       case arg do
-        {{_, meta, _}, _} -> line(meta)
-        {_, meta, _} -> line(meta)
+        {{_, meta, _}, _} -> meta[:line]
+        {_, meta, _} -> meta[:line]
       end
 
-    if MapSet.member?(lines, line) do
-      false
-    else
-      force_args?(args, MapSet.put(lines, line))
+    cond do
+      # Line may be missing from non-formatter AST
+      is_nil(line) -> force_args?(args, lines)
+      Map.has_key?(lines, line) -> false
+      true -> force_args?(args, Map.put(lines, line, true))
     end
   end
 
-  defp force_args?([], _lines), do: true
+  defp force_args?([], lines), do: map_size(lines) >= 2
 
   defp force_keyword(doc, arg) do
     if force_args?(arg), do: force_unfit(doc), else: doc
