@@ -15,7 +15,7 @@ defmodule Float do
   ## Known issues
 
   There are some very well known problems with floating-point numbers
-  and arithmetics due to the fact most decimal fractions cannot be
+  and arithmetic due to the fact most decimal fractions cannot be
   represented by a floating-point binary and most operations are not exact,
   but operate on approximations. Those issues are not specific
   to Elixir, they are a property of floating point representation itself.
@@ -46,6 +46,71 @@ defmodule Float do
   @precision_range 0..15
   @type precision_range :: 0..15
 
+  @min_finite then(<<0xFFEFFFFFFFFFFFFF::64>>, fn <<num::float>> -> num end)
+  @max_finite then(<<0x7FEFFFFFFFFFFFFF::64>>, fn <<num::float>> -> num end)
+
+  @doc """
+  Returns the maximum finite value for a float.
+
+  ## Examples
+
+      iex> Float.max_finite()
+      1.7976931348623157e308
+
+  """
+  def max_finite, do: @max_finite
+
+  @doc """
+  Returns the minimum finite value for a float.
+
+  ## Examples
+
+      iex> Float.min_finite()
+      -1.7976931348623157e308
+
+  """
+  def min_finite, do: @min_finite
+
+  @doc """
+  Computes `base` raised to power of `exponent`.
+
+  `base` must be a float and `exponent` can be any number.
+  However, if a negative base and a fractional exponent
+  are given, it raises `ArithmeticError`.
+
+  It always returns a float. See `Integer.pow/2` for
+  exponentiation that returns integers.
+
+  ## Examples
+
+      iex> Float.pow(2.0, 0)
+      1.0
+      iex> Float.pow(2.0, 1)
+      2.0
+      iex> Float.pow(2.0, 10)
+      1024.0
+      iex> Float.pow(2.0, -1)
+      0.5
+      iex> Float.pow(2.0, -3)
+      0.125
+
+      iex> Float.pow(3.0, 1.5)
+      5.196152422706632
+
+      iex> Float.pow(-2.0, 3)
+      -8.0
+      iex> Float.pow(-2.0, 4)
+      16.0
+
+      iex> Float.pow(-1.0, 0.5)
+      ** (ArithmeticError) bad argument in arithmetic expression
+
+  """
+  @doc since: "1.12.0"
+  @spec pow(float, number) :: float
+  def pow(base, exponent) when is_float(base) and is_number(exponent),
+    do: :math.pow(base, exponent)
+
   @doc """
   Parses a binary into a float.
 
@@ -54,7 +119,8 @@ defmodule Float do
   returned.
 
   If the size of float exceeds the maximum size of `1.7976931348623157e+308`,
-  the `ArgumentError` exception is raised.
+  `:error` is returned even though the textual representation itself might be
+  well formed.
 
   If you want to convert a string-formatted float directly to a float,
   `String.to_float/1` can be used instead.
@@ -70,6 +136,8 @@ defmodule Float do
 
       iex> Float.parse("pi")
       :error
+      iex> Float.parse("1.7976931348623159e+308")
+      :error
 
   """
   @spec parse(binary) :: {float, binary} | :error
@@ -107,7 +175,18 @@ defmodule Float do
        when exp_marker in 'eE' and sign in '-+' and digit in ?0..?9,
        do: parse_unsigned(rest, true, true, <<add_dot(acc, dot?)::binary, ?e, sign, digit>>)
 
-  defp parse_unsigned(rest, dot?, _e?, acc),
+  # When floats are expressed in scientific notation, :erlang.binary_to_float/1 can raise an
+  # ArgumentError if the e exponent is too big. For example, "1.0e400". Because of this, we
+  # rescue the ArgumentError here and return an error.
+  defp parse_unsigned(rest, dot?, true = _e?, acc) do
+    :erlang.binary_to_float(add_dot(acc, dot?))
+  rescue
+    ArgumentError -> :error
+  else
+    float -> {float, rest}
+  end
+
+  defp parse_unsigned(rest, dot?, false = _e?, acc),
     do: {:erlang.binary_to_float(add_dot(acc, dot?)), rest}
 
   defp add_dot(acc, true), do: acc
@@ -268,11 +347,11 @@ defmodule Float do
     raise ArgumentError, invalid_precision_message(precision)
   end
 
-  defp round(0.0, _precision, _rounding), do: 0.0
+  defp round(0.0 = num, _precision, _rounding), do: num
 
   defp round(float, precision, rounding) do
     <<sign::1, exp::11, significant::52-bitstring>> = <<float::float>>
-    {num, count, _} = decompose(significant, 1)
+    {num, count} = decompose(significant, 1)
     count = count - exp + 1023
 
     cond do
@@ -324,6 +403,22 @@ defmodule Float do
     end
   end
 
+  defp decompose(significant, initial) do
+    decompose(significant, 1, 0, initial)
+  end
+
+  defp decompose(<<1::1, bits::bitstring>>, count, last_count, acc) do
+    decompose(bits, count + 1, count, (acc <<< (count - last_count)) + 1)
+  end
+
+  defp decompose(<<0::1, bits::bitstring>>, count, last_count, acc) do
+    decompose(bits, count + 1, last_count, acc)
+  end
+
+  defp decompose(<<>>, _count, last_count, acc) do
+    {acc, last_count}
+  end
+
   defp scale_up(num, boundary, exp) when num >= boundary, do: {num, exp}
   defp scale_up(num, boundary, exp), do: scale_up(num <<< 1, boundary, exp - 1)
 
@@ -403,63 +498,70 @@ defmodule Float do
   def ratio(0.0), do: {0, 1}
 
   def ratio(float) when is_float(float) do
-    case <<float::float>> do
-      <<sign::1, 0::11, significant::52-bitstring>> ->
-        {num, _, den} = decompose(significant, 0)
-        {sign(sign, num), shift_left(den, 1022)}
+    <<sign::1, exp::11, mantissa::52>> = <<float::float>>
 
-      <<sign::1, exp::11, significant::52-bitstring>> ->
-        {num, _, den} = decompose(significant, 1)
-        num = sign(sign, num)
+    {num, den_exp} =
+      if exp != 0 do
+        # Floats are expressed like this:
+        # (2**52 + mantissa) * 2**(-52 + exp - 1023)
+        #
+        # We compute the root factors of the mantissa so we have this:
+        # (2**52 + mantissa * 2**count) * 2**(-52 + exp - 1023)
+        {mantissa, count} = root_factors(mantissa, 0)
 
-        case exp - 1023 do
-          exp when exp > 0 ->
-            {den, exp} = shift_right(den, exp)
-            {shift_left(num, exp), den}
-
-          exp when exp < 0 ->
-            {num, shift_left(den, -exp)}
-
-          0 ->
-            {num, den}
+        # Now we can move the count around so we have this:
+        # (2**(52-count) + mantissa) * 2**(count + -52 + exp - 1023)
+        if mantissa == 0 do
+          {1, exp - 1023}
+        else
+          num = (1 <<< (52 - count)) + mantissa
+          den_exp = count - 52 + exp - 1023
+          {num, den_exp}
         end
+      else
+        # Subnormals are expressed like this:
+        # (mantissa) * 2**(-52 + 1 - 1023)
+        #
+        # So we compute it to this:
+        # (mantissa * 2**(count)) * 2**(-52 + 1 - 1023)
+        #
+        # Which becomes:
+        # mantissa * 2**(count-1074)
+        root_factors(mantissa, -1074)
+      end
+
+    if den_exp > 0 do
+      {sign(sign, num <<< den_exp), 1}
+    else
+      {sign(sign, num), 1 <<< -den_exp}
     end
   end
 
-  defp decompose(significant, initial) do
-    decompose(significant, 1, 0, 2, 1, initial)
-  end
+  defp root_factors(mantissa, count) when mantissa != 0 and (mantissa &&& 1) == 0,
+    do: root_factors(mantissa >>> 1, count + 1)
 
-  defp decompose(<<1::1, bits::bitstring>>, count, last_count, power, _last_power, acc) do
-    decompose(bits, count + 1, count, power <<< 1, power, shift_left(acc, count - last_count) + 1)
-  end
+  defp root_factors(mantissa, count),
+    do: {mantissa, count}
 
-  defp decompose(<<0::1, bits::bitstring>>, count, last_count, power, last_power, acc) do
-    decompose(bits, count + 1, last_count, power <<< 1, last_power, acc)
-  end
-
-  defp decompose(<<>>, _count, last_count, _power, last_power, acc) do
-    {acc, last_count, last_power}
-  end
-
-  @compile {:inline, sign: 2, shift_left: 2}
+  @compile {:inline, sign: 2}
   defp sign(0, num), do: num
   defp sign(1, num), do: -num
 
-  defp shift_left(num, times), do: num <<< times
-
-  defp shift_right(num, 0), do: {num, 0}
-  defp shift_right(1, times), do: {1, times}
-  defp shift_right(num, times), do: shift_right(num >>> 1, times - 1)
-
   @doc """
-  Returns a charlist which corresponds to the text representation
+  Returns a charlist which corresponds to the shortest text representation
   of the given float.
 
-  It uses the shortest representation according to algorithm described
-  in "Printing Floating-Point Numbers Quickly and Accurately" in
-  Proceedings of the SIGPLAN '96 Conference on Programming Language
-  Design and Implementation.
+  The underlying algorithm changes depending on the Erlang/OTP version:
+
+    * For OTP >= 24, it uses the algorithm presented in "Ryū: fast
+      float-to-string conversion" in Proceedings of the SIGPLAN '2018
+      Conference on Programming Language Design and Implementation.
+
+    * For OTP < 24, it uses the algorithm presented in "Printing Floating-Point
+      Numbers Quickly and Accurately" in Proceedings of the SIGPLAN '1996
+      Conference on Programming Language Design and Implementation.
+
+  For a configurable representation, use `:erlang.float_to_list/2`.
 
   ## Examples
 
@@ -473,13 +575,20 @@ defmodule Float do
   end
 
   @doc """
-  Returns a binary which corresponds to the text representation
+  Returns a binary which corresponds to the shortest text representation
   of the given float.
 
-  It uses the shortest representation according to algorithm described
-  in "Printing Floating-Point Numbers Quickly and Accurately" in
-  Proceedings of the SIGPLAN '96 Conference on Programming Language
-  Design and Implementation.
+  The underlying algorithm changes depending on the Erlang/OTP version:
+
+    * For OTP >= 24, it uses the algorithm presented in "Ryū: fast
+      float-to-string conversion" in Proceedings of the SIGPLAN '2018
+      Conference on Programming Language Design and Implementation.
+
+    * For OTP < 24, it uses the algorithm presented in "Printing Floating-Point
+      Numbers Quickly and Accurately" in Proceedings of the SIGPLAN '1996
+      Conference on Programming Language Design and Implementation.
+
+  For a configurable representation, use `:erlang.float_to_binary/2`.
 
   ## Examples
 
