@@ -3,6 +3,12 @@ defmodule Kernel.ParallelCompiler do
   A module responsible for compiling and requiring files in parallel.
   """
 
+  @typedoc "The line. 0 indicates no line."
+  @type line() :: non_neg_integer()
+  @type location() :: line() | {pos_integer(), column :: non_neg_integer}
+  @type warning() :: {file :: Path.t(), location(), message :: String.t()}
+  @type error() :: {file :: Path.t(), location(), message :: String.t()}
+
   @doc """
   Starts a task for parallel compilation.
 
@@ -15,24 +21,29 @@ defmodule Kernel.ParallelCompiler do
   always awaited on by calling `Task.await/1`
   """
   @doc since: "1.6.0"
-  def async(fun) when is_function(fun) do
-    if parent = :erlang.get(:elixir_compiler_pid) do
-      file = :erlang.get(:elixir_compiler_file)
-      dest = :erlang.get(:elixir_compiler_dest)
-      {:error_handler, error_handler} = :erlang.process_info(self(), :error_handler)
+  def async(fun) when is_function(fun, 0) do
+    case :erlang.get(:elixir_compiler_info) do
+      {compiler, _} ->
+        file = :erlang.get(:elixir_compiler_file)
+        dest = :erlang.get(:elixir_compiler_dest)
 
-      Task.async(fn ->
-        send(parent, {:async, self()})
-        :erlang.put(:elixir_compiler_pid, parent)
-        :erlang.put(:elixir_compiler_file, file)
-        dest != :undefined and :erlang.put(:elixir_compiler_dest, dest)
-        :erlang.process_flag(:error_handler, error_handler)
-        fun.()
-      end)
-    else
-      raise ArgumentError,
-            "cannot spawn parallel compiler task because " <>
-              "the current file is not being compiled/required"
+        {:error_handler, error_handler} = :erlang.process_info(self(), :error_handler)
+        {_parent, checker} = Module.ParallelChecker.get()
+
+        Task.async(fn ->
+          send(compiler, {:async, self()})
+          Module.ParallelChecker.put(compiler, checker)
+          :erlang.put(:elixir_compiler_info, {compiler, self()})
+          :erlang.put(:elixir_compiler_file, file)
+          dest != :undefined and :erlang.put(:elixir_compiler_dest, dest)
+          :erlang.process_flag(:error_handler, error_handler)
+          fun.()
+        end)
+
+      :undefined ->
+        raise ArgumentError,
+              "cannot spawn parallel compiler task because " <>
+                "the current file is not being compiled/required"
     end
   end
 
@@ -63,12 +74,16 @@ defmodule Kernel.ParallelCompiler do
 
     * `:each_cycle` - after the given files are compiled, invokes this function
       that should return the following values:
-      * `{:compile, modules}` - to continue compilation with a list of further modules to compile
-      * `{:runtime, modules}` - to stop compilation and verify the list of modules because
-        dependent modules have changed
+      * `{:compile, modules, warnings}` - to continue compilation with a list of
+        further modules to compile
+      * `{:runtime, modules, warnings}` - to stop compilation and verify the list
+        of modules because dependent modules have changed
 
-    * `:long_compilation_threshold` - the timeout (in seconds) after the
-      `:each_long_compilation` callback is invoked; defaults to `15`
+    * `:long_compilation_threshold` - the timeout (in seconds) to check for modules
+      taking too long to compile. For each file that exceeds the threshold, the
+      `:each_long_compilation` callback is invoked. From Elixir v1.11, only the time
+      spent compiling the actual module is taken into account by the threshold, the
+      time spent waiting is not considered. Defaults to `10` seconds.
 
     * `:profile` - if set to `:time` measure the compilation time of each compilation cycle
        and group pass checker
@@ -82,11 +97,19 @@ defmodule Kernel.ParallelCompiler do
 
   """
   @doc since: "1.6.0"
+  @spec compile([Path.t()], keyword()) :: {:ok, [atom], [warning]} | {:error, [error], [warning]}
   def compile(files, options \\ []) when is_list(options) do
     spawn_workers(files, :compile, options)
   end
 
+  @doc """
+  Compiles the given files and writes resulting BEAM files into path.
+
+  See `compile/2` for more information.
+  """
   @doc since: "1.6.0"
+  @spec compile_to_path([Path.t()], Path.t(), keyword()) ::
+          {:ok, [atom], [warning]} | {:error, [error], [warning]}
   def compile_to_path(files, path, options \\ []) when is_binary(path) and is_list(options) do
     spawn_workers(files, {:compile, path}, options)
   end
@@ -112,10 +135,21 @@ defmodule Kernel.ParallelCompiler do
 
   """
   @doc since: "1.6.0"
+  @spec require([Path.t()], keyword()) ::
+          {:ok, [atom], [warning]} | {:error, [error], [warning]}
   def require(files, options \\ []) when is_list(options) do
     spawn_workers(files, :require, options)
   end
 
+  @doc """
+  Prints a warning returned by the compiler.
+  """
+  @doc since: "1.13.0"
+  @spec print_warning(warning) :: :ok
+  def print_warning({file, location, warning}) do
+    :elixir_errors.print_warning(location, file, warning)
+  end
+
   @doc false
   @deprecated "Use Kernel.ParallelCompiler.compile/2 instead"
   def files(files, options \\ []) when is_list(options) do
@@ -136,47 +170,62 @@ defmodule Kernel.ParallelCompiler do
 
   defp spawn_workers(files, output, options) do
     {:module, _} = :code.ensure_loaded(Kernel.ErrorHandler)
-    compiler_pid = self()
-    :elixir_code_server.cast({:reset_warnings, compiler_pid})
     schedulers = max(:erlang.system_info(:schedulers_online), 2)
-    beam_timestamp = Keyword.get(options, :beam_timestamp)
+    {:ok, checker} = Module.ParallelChecker.start_link(schedulers)
 
-    outcome =
-      spawn_workers(files, 0, [], [], %{}, [], %{
-        dest: Keyword.get(options, :dest),
-        each_cycle: Keyword.get(options, :each_cycle, fn -> {:runtime, []} end),
-        each_file: Keyword.get(options, :each_file, fn _, _ -> :ok end) |> each_file(),
-        each_long_compilation: Keyword.get(options, :each_long_compilation, fn _file -> :ok end),
-        each_module: Keyword.get(options, :each_module, fn _file, _module, _binary -> :ok end),
-        long_compilation_threshold: Keyword.get(options, :long_compilation_threshold, 15),
-        profile: Keyword.get(options, :profile),
-        cycle_start: System.monotonic_time(),
-        module_counter: 0,
-        output: output,
-        schedulers: schedulers
-      })
-
-    # In case --warning-as-errors is enabled and there was a warning,
-    # compilation status will be set to error.
-    compilation_status = :elixir_code_server.call({:compilation_status, compiler_pid})
-
-    case {outcome, compilation_status} do
-      {{:ok, _, warnings}, :error} ->
+    try do
+      outcome = spawn_workers(schedulers, checker, files, output, options)
+      {outcome, Code.get_compiler_option(:warnings_as_errors)}
+    else
+      {{:ok, _, [_ | _] = warnings}, true} ->
         message = "Compilation failed due to warnings while using the --warnings-as-errors option"
         IO.puts(:stderr, message)
         {:error, warnings, []}
 
-      {{:error, errors, warnings}, :error} ->
-        {:error, errors ++ warnings, []}
-
       {{:ok, outcome, warnings}, _} ->
+        beam_timestamp = Keyword.get(options, :beam_timestamp)
         {:ok, write_module_binaries(outcome, output, beam_timestamp), warnings}
 
+      {{:error, errors, warnings}, true} ->
+        {:error, errors ++ warnings, []}
+
       {{:error, errors, warnings}, _} ->
         {:error, errors, warnings}
+    after
+      Module.ParallelChecker.stop(checker)
     end
   end
 
+  defp spawn_workers(schedulers, checker, files, output, options) do
+    threshold = Keyword.get(options, :long_compilation_threshold, 10) * 1000
+    timer_ref = Process.send_after(self(), :threshold_check, threshold)
+
+    {outcome, state} =
+      spawn_workers(files, 0, [], [], %{}, [], %{
+        dest: Keyword.get(options, :dest),
+        each_cycle: Keyword.get(options, :each_cycle, fn -> {:runtime, [], []} end),
+        each_file: Keyword.get(options, :each_file, fn _, _ -> :ok end) |> each_file(),
+        each_long_compilation: Keyword.get(options, :each_long_compilation, fn _file -> :ok end),
+        each_module: Keyword.get(options, :each_module, fn _file, _module, _binary -> :ok end),
+        profile: profile_init(Keyword.get(options, :profile)),
+        output: output,
+        timer_ref: timer_ref,
+        long_compilation_threshold: threshold,
+        schedulers: schedulers,
+        checker: checker
+      })
+
+    Process.cancel_timer(state.timer_ref)
+
+    receive do
+      :threshold_check -> :ok
+    after
+      0 -> :ok
+    end
+
+    outcome
+  end
+
   defp each_file(fun) when is_function(fun, 1), do: fn file, _ -> fun.(file) end
   defp each_file(fun) when is_function(fun, 2), do: fun
 
@@ -192,7 +241,7 @@ defmodule Kernel.ParallelCompiler do
 
   defp write_module_binaries(result, {:compile, path}, timestamp) do
     Enum.flat_map(result, fn
-      {{:module, module}, {binary, _map}} ->
+      {{:module, module}, binary} ->
         full_path = Path.join(path, Atom.to_string(module) <> ".beam")
         File.write!(full_path, binary)
         if timestamp, do: File.touch!(full_path, timestamp)
@@ -212,39 +261,31 @@ defmodule Kernel.ParallelCompiler do
   defp verify_modules(result, warnings, dependent_modules, state) do
     checker_warnings = maybe_check_modules(result, dependent_modules, state)
     warnings = Enum.reverse(warnings, checker_warnings)
-    {:ok, result, warnings}
+    {{:ok, result, warnings}, state}
   end
 
   defp maybe_check_modules(result, runtime_modules, state) do
-    %{schedulers: schedulers, profile: profile} = state
+    %{profile: profile, checker: checker} = state
 
-    if :elixir_config.get(:bootstrap) do
-      []
-    else
-      compiled_modules = checker_compiled_modules(result)
-      runtime_modules = checker_runtime_modules(runtime_modules)
+    compiled_modules =
+      for {{:module, module}, _} <- result,
+          do: module
 
-      profile_checker(profile, compiled_modules, runtime_modules, fn ->
-        Module.ParallelChecker.verify(compiled_modules, runtime_modules, schedulers)
-      end)
-    end
+    runtime_modules =
+      for module <- runtime_modules,
+          path = :code.which(module),
+          is_list(path) and path != [],
+          do: {module, path}
+
+    profile_checker(profile, compiled_modules, runtime_modules, fn ->
+      Module.ParallelChecker.verify(checker, runtime_modules)
+    end)
   end
 
-  defp checker_compiled_modules(result) do
-    for {{:module, _module}, {binary, module_map}} <- result do
-      {module_map, binary}
-    end
-  end
+  defp profile_init(:time), do: {:time, System.monotonic_time(), 0}
+  defp profile_init(nil), do: :none
 
-  defp checker_runtime_modules(modules) do
-    for module <- modules,
-        path = :code.which(module),
-        is_list(path) and path != [] do
-      {module, File.read!(path)}
-    end
-  end
-
-  defp profile_checker(_profile = :time, compiled_modules, runtime_modules, fun) do
+  defp profile_checker({:time, _, _}, compiled_modules, runtime_modules, fun) do
     {time, result} = :timer.tc(fun)
     time = div(time, 1000)
     num_modules = length(compiled_modules) + length(runtime_modules)
@@ -252,7 +293,7 @@ defmodule Kernel.ParallelCompiler do
     result
   end
 
-  defp profile_checker(_profile = nil, _compiled_modules, _runtime_modules, fun) do
+  defp profile_checker(:none, _compiled_modules, _runtime_modules, fun) do
     fun.()
   end
 
@@ -274,29 +315,30 @@ defmodule Kernel.ParallelCompiler do
 
   # Release waiting processes
   defp spawn_workers([{ref, found} | t], spawned, waiting, files, result, warnings, state) do
-    waiting =
+    {files, waiting} =
       case List.keytake(waiting, ref, 2) do
-        {{_kind, pid, ^ref, _on, _defining, _deadlock}, waiting} ->
+        {{_kind, pid, ^ref, file_pid, _on, _defining, _deadlock}, waiting} ->
           send(pid, {ref, found})
-          waiting
+          {update_timing(files, file_pid, :waiting), waiting}
 
         nil ->
           # In case the waiting process died (for example, it was an async process),
           # it will no longer be on the list. So we need to take it into account here.
-          waiting
+          {files, waiting}
       end
 
     spawn_workers(t, spawned, waiting, files, result, warnings, state)
   end
 
   defp spawn_workers([file | queue], spawned, waiting, files, result, warnings, state) do
-    %{output: output, long_compilation_threshold: threshold, dest: dest} = state
+    %{output: output, dest: dest, checker: checker} = state
     parent = self()
     file = Path.expand(file)
 
     {pid, ref} =
       :erlang.spawn_monitor(fn ->
-        :erlang.put(:elixir_compiler_pid, parent)
+        Module.ParallelChecker.put(parent, checker)
+        :erlang.put(:elixir_compiler_info, {parent, self()})
         :erlang.put(:elixir_compiler_file, file)
 
         try do
@@ -313,24 +355,34 @@ defmodule Kernel.ParallelCompiler do
         exit(:shutdown)
       end)
 
-    timer_ref = Process.send_after(self(), {:timed_out, pid}, threshold * 1000)
-    files = [{pid, ref, file, timer_ref} | files]
+    file_data = %{
+      pid: pid,
+      ref: ref,
+      file: file,
+      timestamp: System.monotonic_time(),
+      compiling: 0,
+      waiting: 0,
+      warned: false
+    }
+
+    files = [file_data | files]
     spawn_workers(queue, spawned + 1, waiting, files, result, warnings, state)
   end
 
   # No more queue, nothing waiting, this cycle is done
   defp spawn_workers([], 0, [], [], result, warnings, state) do
+    cycle_return = each_cycle_return(state.each_cycle.())
     state = cycle_timing(result, state)
 
-    case each_cycle_return(state.each_cycle.()) do
-      {:runtime, dependent_modules} ->
-        verify_modules(result, warnings, dependent_modules, state)
+    case cycle_return do
+      {:runtime, dependent_modules, extra_warnings} ->
+        verify_modules(result, extra_warnings ++ warnings, dependent_modules, state)
 
-      {:compile, []} ->
-        verify_modules(result, warnings, [], state)
+      {:compile, [], extra_warnings} ->
+        verify_modules(result, extra_warnings ++ warnings, [], state)
 
-      {:compile, more} ->
-        spawn_workers(more, 0, [], [], result, warnings, state)
+      {:compile, more, extra_warnings} ->
+        spawn_workers(more, 0, [], [], result, extra_warnings ++ warnings, state)
     end
   end
 
@@ -340,8 +392,8 @@ defmodule Kernel.ParallelCompiler do
   defp spawn_workers(
          [],
          1,
-         [{_, pid, ref, _, _, _}] = waiting,
-         [{pid, _, _, _}] = files,
+         [{_, pid, ref, _, _, _, _}] = waiting,
+         [%{pid: pid}] = files,
          result,
          warnings,
          state
@@ -355,24 +407,31 @@ defmodule Kernel.ParallelCompiler do
     # There is potentially a deadlock. We will release modules with
     # the following order:
     #
-    #   1. Code.ensure_compiled/1 checks (deadlock = soft)
-    #   2. Struct checks (deadlock = hard)
-    #   3. Modules without a known definition
-    #   4. Code invocation (deadlock = raise)
+    #   1. Code.ensure_compiled/1 checks without a known definition (deadlock = soft)
+    #   2. Code.ensure_compiled/1 checks with a known definition (deadlock = soft)
+    #   3. Struct/import/require/ensure_compiled! checks without a known definition (deadlock = hard)
+    #   4. Modules without a known definition
+    #   5. Code invocation (deadlock = raise)
     #
-    # In theory there is no difference between hard and raise, the
+    # The reason for step 3 and 4 is to not treat typos as deadlocks and
+    # help developers handle those sooner. However, this can have false
+    # positives in case multiple modules are defined in the same file
+    # and the module we are waiting for is defined later on.
+    #
+    # Finally, note there is no difference between hard and raise, the
     # difference is where the raise is happening, inside the compiler
     # or in the caller.
-    cond do
-      deadlocked = deadlocked(waiting, :soft) || deadlocked(waiting, :hard) ->
-        spawn_workers(deadlocked, spawned, waiting, files, result, warnings, state)
 
-      without_definition = without_definition(waiting, files) ->
-        spawn_workers(without_definition, spawned, waiting, files, result, warnings, state)
+    deadlocked =
+      deadlocked(waiting, :soft, false) ||
+        deadlocked(waiting, :soft, true) || deadlocked(waiting, :hard, false) ||
+        without_definition(waiting, files)
 
-      true ->
-        errors = handle_deadlock(waiting, files)
-        {:error, errors, warnings}
+    if deadlocked do
+      spawn_workers(deadlocked, spawned, waiting, files, result, warnings, state)
+    else
+      errors = handle_deadlock(waiting, files)
+      {{:error, errors, warnings}, state}
     end
   end
 
@@ -398,8 +457,11 @@ defmodule Kernel.ParallelCompiler do
     end
   end
 
-  defp cycle_timing(result, %{profile: :time} = state) do
-    %{cycle_start: cycle_start, module_counter: module_counter} = state
+  defp cycle_timing(_result, %{profile: :none} = state) do
+    state
+  end
+
+  defp cycle_timing(result, %{profile: {:time, cycle_start, module_counter}} = state) do
     num_modules = count_modules(result)
     diff_modules = num_modules - module_counter
     now = System.monotonic_time()
@@ -410,39 +472,52 @@ defmodule Kernel.ParallelCompiler do
       "[profile] Finished compilation cycle of #{diff_modules} modules in #{time}ms"
     )
 
-    %{state | cycle_start: now, module_counter: num_modules}
-  end
-
-  defp cycle_timing(_result, %{profile: nil} = state) do
-    state
+    %{state | profile: {:time, now, num_modules}}
   end
 
   defp count_modules(result) do
     Enum.count(result, &match?({{:module, _}, _}, &1))
   end
 
-  # TODO: Deprecate on v1.14
-  defp each_cycle_return(modules) when is_list(modules), do: {:compile, modules}
-  defp each_cycle_return(other), do: other
+  defp each_cycle_return({kind, modules, warnings}), do: {kind, modules, warnings}
+
+  defp each_cycle_return(other) do
+    IO.warn(
+      "the :each_cycle callback must return a tuple of format {:compile | :runtime, modules, warnings}"
+    )
+
+    case other do
+      {kind, modules} -> {kind, modules, []}
+      modules when is_list(modules) -> {:compile, modules, []}
+    end
+  end
 
   # The goal of this function is to find leaves in the dependency graph,
   # i.e. to find code that depends on code that we know is not being defined.
-  # Note that not all files have been compile yet, so they may not be in waiting.
+  # Note that not all files have been compiled yet, so they may not be in waiting.
   defp without_definition(waiting, files) do
-    nillify_empty(
-      for {pid, _, _, _} <- files,
-          {_, ^pid, ref, on, _, _} <- List.wrap(List.keyfind(waiting, pid, 1)),
-          not Enum.any?(waiting, fn {_, _, _, _, defining, _} -> on in defining end),
+    nilify_empty(
+      for %{pid: pid} <- files,
+          {_, _, ref, ^pid, on, _, _} <- waiting,
+          not defining?(on, waiting),
           do: {ref, :not_found}
     )
   end
 
-  defp deadlocked(waiting, type) do
-    nillify_empty(for {_, _, ref, _, _, ^type} <- waiting, do: {ref, :deadlock})
+  defp deadlocked(waiting, type, defining?) do
+    nilify_empty(
+      for {_, _, ref, _, on, _, ^type} <- waiting,
+          defining?(on, waiting) == defining?,
+          do: {ref, :deadlock}
+    )
   end
 
-  defp nillify_empty([]), do: nil
-  defp nillify_empty([_ | _] = list), do: list
+  defp defining?(on, waiting) do
+    Enum.any?(waiting, fn {_, _, _, _, _, defining, _} -> on in defining end)
+  end
+
+  defp nilify_empty([]), do: nil
+  defp nilify_empty([_ | _] = list), do: list
 
   # Wait for messages from child processes
   defp wait_for_messages(queue, spawned, waiting, files, result, warnings, state) do
@@ -455,66 +530,73 @@ defmodule Kernel.ParallelCompiler do
 
       {:available, kind, module} ->
         available =
-          for {^kind, _, ref, ^module, _defining, _deadlock} <- waiting,
+          for {^kind, _, ref, _, ^module, _defining, _deadlock} <- waiting,
               do: {ref, :found}
 
         result = Map.put(result, {kind, module}, true)
         spawn_workers(available ++ queue, spawned, waiting, files, result, warnings, state)
 
-      {:module_available, child, ref, file, module, binary, module_map} ->
+      {:module_available, child, ref, file, module, binary} ->
         state.each_module.(file, module, binary)
 
         # Release the module loader which is waiting for an ack
         send(child, {ref, :ack})
 
         available =
-          for {:module, _, ref, ^module, _defining, _deadlock} <- waiting,
+          for {:module, _, ref, _, ^module, _defining, _deadlock} <- waiting,
               do: {ref, :found}
 
-        cancel_waiting_timer(files, child)
-        result = Map.put(result, {:module, module}, {binary, module_map})
+        result = Map.put(result, {:module, module}, binary)
         spawn_workers(available ++ queue, spawned, waiting, files, result, warnings, state)
 
       # If we are simply requiring files, we do not add to waiting.
-      {:waiting, _kind, child, ref, _on, _defining, _deadlock} when output == :require ->
+      {:waiting, _kind, child, ref, _file_pid, _on, _defining, _deadlock} when output == :require ->
         send(child, {ref, :not_found})
         spawn_workers(queue, spawned, waiting, files, result, warnings, state)
 
-      {:waiting, kind, child, ref, on, defining, deadlock?} ->
+      {:waiting, kind, child_pid, ref, file_pid, on, defining, deadlock?} ->
         # If we already got what we were waiting for, do not put it on waiting.
-        # Alternatively, we're waiting on ourselves,
-        # send :found so that we can crash with a better error.
-        waiting =
+        # If we're waiting on ourselves, send :found so that we can crash with
+        # a better error.
+        {files, waiting} =
           if Map.has_key?(result, {kind, on}) or on in defining do
-            send(child, {ref, :found})
-            waiting
+            send(child_pid, {ref, :found})
+            {files, waiting}
           else
-            [{kind, child, ref, on, defining, deadlock?} | waiting]
+            files = update_timing(files, file_pid, :compiling)
+            {files, [{kind, child_pid, ref, file_pid, on, defining, deadlock?} | waiting]}
           end
 
         spawn_workers(queue, spawned, waiting, files, result, warnings, state)
 
-      {:timed_out, child} ->
-        case List.keyfind(files, child, 0) do
-          {^child, _, file, _} -> state.each_long_compilation.(file)
-          _ -> :ok
-        end
+      :threshold_check ->
+        files =
+          for data <- files do
+            if data.warned or List.keymember?(waiting, data.pid, 1) do
+              data
+            else
+              data = update_timing(data, :compiling)
+              data = maybe_warn_long_compilation(data, state)
+              data
+            end
+          end
 
+        timer_ref = Process.send_after(self(), :threshold_check, state.long_compilation_threshold)
+        state = %{state | timer_ref: timer_ref}
         spawn_workers(queue, spawned, waiting, files, result, warnings, state)
 
-      {:warning, file, line, message} ->
+      {:warning, file, location, message} ->
         file = file && Path.absname(file)
         message = :unicode.characters_to_binary(message)
-        warning = {file, line, message}
+        warning = {file, location, message}
         wait_for_messages(queue, spawned, waiting, files, result, [warning | warnings], state)
 
       {:file_ok, child_pid, ref, file, lexical} ->
         state.each_file.(file, lexical)
         send(child_pid, ref)
-        cancel_waiting_timer(files, child_pid)
 
         discard_down(child_pid)
-        new_files = List.keydelete(files, child_pid, 0)
+        new_files = discard_and_maybe_log_file(files, child_pid, state)
 
         # Sometimes we may have spurious entries in the waiting list
         # because someone invoked try/rescue UndefinedFunctionError
@@ -522,28 +604,84 @@ defmodule Kernel.ParallelCompiler do
         spawn_workers(queue, spawned - 1, new_waiting, new_files, result, warnings, state)
 
       {:file_cancel, child_pid} ->
-        cancel_waiting_timer(files, child_pid)
         discard_down(child_pid)
-        new_files = List.keydelete(files, child_pid, 0)
+        new_files = Enum.reject(files, &(&1.pid == child_pid))
         spawn_workers(queue, spawned - 1, waiting, new_files, result, warnings, state)
 
       {:file_error, child_pid, file, {kind, reason, stack}} ->
         print_error(file, kind, reason, stack)
-        cancel_waiting_timer(files, child_pid)
         discard_down(child_pid)
-        files |> List.keydelete(child_pid, 0) |> terminate()
-        {:error, [to_error(file, kind, reason, stack)], warnings}
+        files |> Enum.reject(&(&1.pid == child_pid)) |> terminate()
+        {{:error, [to_error(file, kind, reason, stack)], warnings}, state}
 
       {:DOWN, ref, :process, pid, reason} ->
         waiting = List.keydelete(waiting, pid, 1)
 
         case handle_down(files, ref, reason) do
           :ok -> wait_for_messages(queue, spawned - 1, waiting, files, result, warnings, state)
-          {:error, errors} -> {:error, errors, warnings}
+          {:error, errors} -> {{:error, errors, warnings}, state}
         end
     end
   end
 
+  defp update_timing(files, pid, key) do
+    Enum.map(files, fn data ->
+      if data.pid == pid do
+        time = System.monotonic_time()
+        %{data | key => data[key] + time - data.timestamp, timestamp: time}
+      else
+        data
+      end
+    end)
+  end
+
+  defp update_timing(data, key) do
+    time = System.monotonic_time()
+    %{data | key => data[key] + time - data.timestamp, timestamp: time}
+  end
+
+  defp maybe_warn_long_compilation(data, state) do
+    compiling = System.convert_time_unit(data.compiling, :native, :millisecond)
+
+    if not data.warned and compiling >= state.long_compilation_threshold do
+      state.each_long_compilation.(data.file)
+      %{data | warned: true}
+    else
+      data
+    end
+  end
+
+  defp discard_and_maybe_log_file(files, pid, state) do
+    Enum.reject(files, fn data ->
+      if data.pid == pid do
+        data = update_timing(data, :compiling)
+        data = maybe_warn_long_compilation(data, state)
+
+        if state.profile != :none do
+          compiling = to_padded_ms(data.compiling)
+          waiting = to_padded_ms(data.waiting)
+          relative = Path.relative_to_cwd(data.file)
+
+          IO.puts(
+            :stderr,
+            "[profile] #{compiling}ms compiling + #{waiting}ms waiting for #{relative}"
+          )
+        end
+
+        true
+      else
+        false
+      end
+    end)
+  end
+
+  defp to_padded_ms(time) do
+    time
+    |> System.convert_time_unit(:native, :millisecond)
+    |> Integer.to_string()
+    |> String.pad_leading(6, " ")
+  end
+
   defp discard_down(pid) do
     receive do
       {:DOWN, _, :process, ^pid, _} -> :ok
@@ -555,28 +693,24 @@ defmodule Kernel.ParallelCompiler do
   end
 
   defp handle_down(files, ref, reason) do
-    case List.keyfind(files, ref, 1) do
-      {child_pid, ^ref, file, _timer_ref} ->
+    case Enum.find(files, &(&1.ref == ref)) do
+      %{pid: pid, file: file} ->
         print_error(file, :exit, reason, [])
-
-        files
-        |> List.keydelete(child_pid, 0)
-        |> terminate()
-
+        files |> Enum.reject(&(&1.pid == pid)) |> terminate()
         {:error, [to_error(file, :exit, reason, [])]}
 
-      _ ->
+      nil ->
         :ok
     end
   end
 
   defp handle_deadlock(waiting, files) do
     deadlock =
-      for {pid, _, file, _} <- files do
+      for %{pid: pid, file: file} <- files do
         {:current_stacktrace, stacktrace} = Process.info(pid, :current_stacktrace)
         Process.exit(pid, :kill)
 
-        {kind, ^pid, _, on, _, _} = List.keyfind(waiting, pid, 1)
+        {kind, ^pid, _, _, on, _, _} = List.keyfind(waiting, pid, 1)
         description = "deadlocked waiting on #{kind} #{inspect(on)}"
         error = CompileError.exception(description: description, file: nil, line: nil)
         print_error(file, :error, error, stacktrace)
@@ -607,8 +741,8 @@ defmodule Kernel.ParallelCompiler do
   end
 
   defp terminate(files) do
-    for {pid, _, _, _} <- files, do: Process.exit(pid, :kill)
-    for {pid, _, _, _} <- files, do: discard_down(pid)
+    for %{pid: pid} <- files, do: Process.exit(pid, :kill)
+    for %{pid: pid} <- files, do: discard_down(pid)
     :ok
   end
 
@@ -619,27 +753,16 @@ defmodule Kernel.ParallelCompiler do
     ])
   end
 
-  defp cancel_waiting_timer(files, child_pid) do
-    case List.keyfind(files, child_pid, 0) do
-      {^child_pid, _ref, _file, timer_ref} ->
-        Process.cancel_timer(timer_ref)
-        # Let's flush the message in case it arrived before we canceled the timeout.
-        receive do
-          {:timed_out, ^child_pid} -> :ok
-        after
-          0 -> :ok
-        end
-
-      nil ->
-        :ok
-    end
-  end
-
   defp to_error(file, kind, reason, stack) do
     line = get_line(file, reason, stack)
     file = Path.absname(file)
     message = :unicode.characters_to_binary(Kernel.CLI.format_error(kind, reason, stack))
-    {file, line, message}
+    {file, line || 0, message}
+  end
+
+  defp get_line(_file, %{line: line, column: column}, _stack)
+       when is_integer(line) and line > 0 and is_integer(column) and column >= 0 do
+    {line, column}
   end
 
   defp get_line(_file, %{line: line}, _stack) when is_integer(line) and line > 0 do
@@ -652,6 +775,13 @@ defmodule Kernel.ParallelCompiler do
     end
   end
 
+  defp get_line(file, _reason, [{_, _, _, [file: expanding]}, {_, _, _, info} | _])
+       when expanding in ['expanding macro', 'expanding struct'] do
+    if Keyword.get(info, :file) == to_charlist(Path.relative_to_cwd(file)) do
+      Keyword.get(info, :line)
+    end
+  end
+
   defp get_line(file, _reason, [{_, _, _, info} | _]) do
     if Keyword.get(info, :file) == to_charlist(Path.relative_to_cwd(file)) do
       Keyword.get(info, :line)
