@@ -104,7 +104,7 @@ defmodule Kernel.Typespec do
   @doc """
   Defines a typespec.
 
-  Invoked by `Kernel.@/1` expansion.
+  Invoked by `@/1` expansion.
   """
   def deftypespec(:spec, expr, _line, _file, module, pos) do
     {_set, bag} = :elixir_module.data_tables(module)
@@ -194,14 +194,14 @@ defmodule Kernel.Typespec do
 
   defp get_doc_info(set, attr, line) do
     case :ets.take(set, attr) do
-      [{^attr, {line, doc}, _}] -> {line, doc}
+      [{^attr, {line, doc}, _, _}] -> {line, doc}
       [] -> {line, nil}
     end
   end
 
   defp get_doc_meta(spec_meta, doc_kind, set) do
     case :ets.take(set, {doc_kind, :meta}) do
-      [{{^doc_kind, :meta}, metadata, _}] -> Map.merge(metadata, spec_meta)
+      [{{^doc_kind, :meta}, metadata}] -> Map.merge(metadata, spec_meta)
       [] -> spec_meta
     end
   end
@@ -553,9 +553,7 @@ defmodule Kernel.Typespec do
   end
 
   defp typespec({:%, _, [name, {:%{}, meta, fields}]}, vars, caller, state) do
-    # We cannot set a function name to avoid tracking
-    # as a compile time dependency, because for structs it actually is one.
-    module = Macro.expand(name, caller)
+    module = Macro.expand(name, %{caller | function: {:__info__, 1}})
 
     struct =
       module
@@ -569,7 +567,10 @@ defmodule Kernel.Typespec do
 
     types =
       :lists.map(
-        fn {field, _} -> {field, Keyword.get(fields, field, quote(do: term()))} end,
+        fn
+          {:__exception__ = field, true} -> {field, Keyword.get(fields, field, true)}
+          {field, _} -> {field, Keyword.get(fields, field, quote(do: term()))}
+        end,
         :lists.sort(struct)
       )
 
@@ -577,7 +578,7 @@ defmodule Kernel.Typespec do
       unless Keyword.has_key?(struct, field) do
         compile_error(
           caller,
-          "undefined field #{inspect(field)} on struct #{Macro.to_string(name)}"
+          "undefined field #{inspect(field)} on struct #{inspect(module)}"
         )
       end
     end
@@ -680,10 +681,20 @@ defmodule Kernel.Typespec do
     end
   end
 
-  defp typespec({:"::", meta, [left, right]} = expr, vars, caller, state) do
+  defp typespec({:"::", meta, [left, right]}, vars, caller, state) do
     message =
-      "invalid type annotation. When using the | operator to represent the union of types, " <>
-        "make sure to wrap type annotations in parentheses: #{Macro.to_string(expr)}"
+      "invalid type annotation. The left side of :: must be a variable, got: #{Macro.to_string(left)}"
+
+    message =
+      case left do
+        {:|, _, _} ->
+          message <>
+            ". Note \"left | right :: ann\" is the same as \"(left | right) :: ann\". " <>
+            "To solve this, use parentheses around the union operands: \"left | (right :: ann)\""
+
+        _ ->
+          message
+      end
 
     # TODO: Make this an error on v2.0, and remove the code below and
     # the :undefined_type_error_enabled? key from the state
