@@ -22,16 +22,52 @@ defmodule Kernel.Utils do
   defp destructure_nil(count), do: [nil | destructure_nil(count - 1)]
 
   @doc """
-  Callback for defdelegate.
+  Callback for defdelegate entry point.
   """
-  def defdelegate(fun, opts) when is_list(opts) do
+  def defdelegate_all(funs, opts, env) do
+    to = Keyword.get(opts, :to) || raise ArgumentError, "expected to: to be given as argument"
+    as = Keyword.get(opts, :as)
+
+    if to == env.module and is_nil(as) do
+      raise ArgumentError,
+            "defdelegate function is calling itself, which will lead to an infinite loop. You should either change the value of the :to option or specify the :as option"
+    end
+
+    if is_list(funs) do
+      IO.warn(
+        "passing a list to Kernel.defdelegate/2 is deprecated, please define each delegate separately",
+        Macro.Env.stacktrace(env)
+      )
+    end
+
+    if Keyword.has_key?(opts, :append_first) do
+      IO.warn(
+        "Kernel.defdelegate/2 :append_first option is deprecated",
+        Macro.Env.stacktrace(env)
+      )
+    end
+
+    to
+  end
+
+  @doc """
+  Callback for each function in defdelegate.
+  """
+  def defdelegate_each(fun, opts) when is_list(opts) do
     # TODO: Remove on v2.0
     append_first? = Keyword.get(opts, :append_first, false)
 
     {name, args} =
-      case Macro.decompose_call(fun) do
-        {_, _} = pair -> pair
-        _ -> raise ArgumentError, "invalid syntax in defdelegate #{Macro.to_string(fun)}"
+      case fun do
+        {:when, _, [_left, right]} ->
+          raise ArgumentError,
+                "guards are not allowed in defdelegate/2, got: when #{Macro.to_string(right)}"
+
+        _ ->
+          case Macro.decompose_call(fun) do
+            {_, _} = pair -> pair
+            _ -> raise ArgumentError, "invalid syntax in defdelegate #{Macro.to_string(fun)}"
+          end
       end
 
     as = Keyword.get(opts, :as, name)
@@ -64,7 +100,15 @@ defmodule Kernel.Utils do
   @doc """
   Callback for defstruct.
   """
-  def defstruct(module, fields) do
+  def defstruct(module, fields, bootstrapped?) do
+    {set, bag} = :elixir_module.data_tables(module)
+
+    if :ets.member(set, :__struct__) do
+      raise ArgumentError,
+            "defstruct has already been called for " <>
+              "#{Kernel.inspect(module)}, defstruct can only be called once per module"
+    end
+
     case fields do
       fs when is_list(fs) ->
         :ok
@@ -76,7 +120,7 @@ defmodule Kernel.Utils do
     mapper = fn
       {key, val} when is_atom(key) ->
         try do
-          Macro.escape(val)
+          :elixir_quote.escape(val, false, :none)
         rescue
           e in [ArgumentError] ->
             raise ArgumentError, "invalid value for struct field #{key}, " <> Exception.message(e)
@@ -92,7 +136,13 @@ defmodule Kernel.Utils do
     end
 
     fields = :lists.map(mapper, fields)
-    enforce_keys = List.wrap(Module.get_attribute(module, :enforce_keys))
+
+    enforce_keys =
+      case :ets.take(set, :enforce_keys) do
+        [{_, enforce_keys, _, _}] when is_list(enforce_keys) -> enforce_keys
+        [{_, enforce_key, _, _}] -> [enforce_key]
+        [] -> []
+      end
 
     # TODO: Make it raise on v2.0
     warn_on_duplicate_struct_key(:lists.keysort(1, fields))
@@ -106,9 +156,68 @@ defmodule Kernel.Utils do
     end
 
     :lists.foreach(foreach, enforce_keys)
-
     struct = :maps.put(:__struct__, module, :maps.from_list(fields))
-    {struct, enforce_keys, Module.get_attribute(module, :derive)}
+
+    body =
+      case bootstrapped? do
+        true ->
+          case enforce_keys do
+            [] ->
+              quote do
+                Enum.reduce(kv, @__struct__, fn {key, val}, map ->
+                  %{map | key => val}
+                end)
+              end
+
+            _ ->
+              quote do
+                {map, keys} =
+                  Enum.reduce(kv, {@__struct__, unquote(enforce_keys)}, fn
+                    {key, val}, {map, keys} ->
+                      {%{map | key => val}, List.delete(keys, key)}
+                  end)
+
+                case keys do
+                  [] ->
+                    map
+
+                  _ ->
+                    raise ArgumentError,
+                          "the following keys must also be given when building " <>
+                            "struct #{inspect(__MODULE__)}: #{inspect(keys)}"
+                end
+              end
+          end
+
+        false ->
+          quote do
+            :lists.foldl(
+              fn {key, val}, acc -> %{acc | key => val} end,
+              @__struct__,
+              kv
+            )
+          end
+      end
+
+    case enforce_keys -- :maps.keys(struct) do
+      [] ->
+        # The __struct__ field is used for expansion and for loading remote structs
+        :ets.insert(set, {:__struct__, struct, nil, []})
+
+        # Store all field metadata to go into __info__(:struct)
+        mapper = fn {key, val} ->
+          %{field: key, default: val, required: :lists.member(key, enforce_keys)}
+        end
+
+        :ets.insert(set, {{:elixir, :struct}, :lists.map(mapper, fields)})
+        derive = :lists.map(fn {_, value} -> value end, :ets.take(bag, {:accumulate, :derive}))
+        {struct, :lists.reverse(derive), quote(do: kv), body}
+
+      error_keys ->
+        raise ArgumentError,
+              "@enforce_keys required keys (#{inspect(error_keys)}) that are not defined in defstruct: " <>
+                "#{inspect(fields)}"
+    end
   end
 
   defp warn_on_duplicate_struct_key([]) do
@@ -128,9 +237,9 @@ defmodule Kernel.Utils do
   Announcing callback for defstruct.
   """
   def announce_struct(module) do
-    case :erlang.get(:elixir_compiler_pid) do
+    case :erlang.get(:elixir_compiler_info) do
       :undefined -> :ok
-      pid -> send(pid, {:available, :struct, module})
+      {pid, _} -> send(pid, {:available, :struct, module})
     end
   end
 
@@ -210,7 +319,7 @@ defmodule Kernel.Utils do
   def defguard(args, expr, env) do
     {^args, vars} = extract_refs_from_args(args)
     env = :elixir_env.with_vars(%{env | context: :guard}, vars)
-    {expr, _scope} = :elixir_expand.expand(expr, env)
+    {expr, _, _} = :elixir_expand.expand(expr, :elixir_env.env_to_ex(env), env)
 
     quote do
       case Macro.Env.in_guard?(__CALLER__) do
@@ -258,7 +367,7 @@ defmodule Kernel.Utils do
                   {new_var, acc}
 
                 %{} ->
-                  generated = String.to_atom("arg" <> Integer.to_string(map_size(acc)))
+                  generated = String.to_atom("arg" <> Integer.to_string(map_size(acc) + 1))
                   new_var = Macro.var(generated, Elixir)
                   {new_var, Map.put(acc, pair, {new_var, var})}
               end
