@@ -13,45 +13,43 @@ defmodule Keyword do
 
       [{:exit_on_close, true}, {:active, :once}, {:packet_size, 1024}]
 
-  Elixir provides a special and more concise syntax for keyword lists
-  that looks like this:
+  Elixir provides a special and more concise syntax for keyword lists:
 
       [exit_on_close: true, active: :once, packet_size: 1024]
 
-  The two syntaxes are completely equivalent. Like atoms, keyword
-  lists keys must be composed of Unicode characters such as letters,
-  numbers, underscore, and `@`. If the keyword has a character that
-  does not belong to the category above, such as spaces, you can wrap
-  it in quotes:
+  The two syntaxes return the exact same value.
+
+  A *key* can be any atom, consisting of Unicode letters, numbers,
+  an underscore or the `@` sign. If the *key* should have any other
+  characters, such as spaces, you can wrap it in quotes:
 
       iex> ["exit on close": true]
       ["exit on close": true]
 
-  Wrapping a keyword in quotes does not make it a string. Keyword lists
-  keys are always atoms. If you use quotes around the key when quoting
-  is not necessary, Elixir will warn.
+  Wrapping an atom in quotes does not make it a string. Keyword list
+  *keys* are always atoms. Quotes should only be used when necessary
+  or Elixir will issue a warning.
 
   ## Duplicate keys and ordering
 
-  A keyword may have duplicated keys so it is not strictly a key-value
-  data type. However most of the functions in this module behave exactly
-  as a key-value so they work similarly to the functions you would find
-  in the `Map` module. For example, `Keyword.get/3` will get the first
-  entry matching the given key, regardless if duplicated entries exist.
-  Similarly, `Keyword.put/3` and `Keyword.delete/2` ensure all duplicated
+  A keyword may have duplicate keys so it is not strictly a key-value
+  data type. However most of the functions in this module work on a
+  key-value structure and behave similar to the functions you would
+  find in the `Map` module. For example, `Keyword.get/3` will get the first
+  entry matching the given key, regardless if duplicate entries exist.
+  Similarly, `Keyword.put/3` and `Keyword.delete/2` ensure all duplicate
   entries for a given key are removed when invoked. Note, however, that
   keyword list operations need to traverse the whole list in order to find
   keys, so these operations are slower than their map counterparts.
 
-  A handful of functions exist to handle duplicated keys, for example,
+  A handful of functions exist to handle duplicate keys, for example,
   `get_values/2` returns all values for a given key and `delete_first/2`
-  deletes just one of the existing entries.
+  deletes just the first entry of the existing ones.
 
-  Even though lists preserve the user ordering, the functions in
+  Even though lists preserve the existing order, the functions in
   `Keyword` do not guarantee any ordering. For example, if you invoke
-  `Keyword.put(opts, new_key, new_value)`, there is no guarantee to
-  where `new_key` will be added (to the front, to the end, or
-  anywhere else).
+  `Keyword.put(opts, new_key, new_value)`, there is no guarantee for
+  where `new_key` will be added to (the front, the end or anywhere else).
 
   Given ordering is not guaranteed, it is not recommended to pattern
   match on keyword lists either. For example, a function such as:
@@ -67,7 +65,7 @@ defmodule Keyword do
       my_function([another_key: :bar, some_key: :foo])
 
   Most of the functions in this module work in linear time. This means
-  that, the time it takes to perform an operation grows at the same
+  that the time it takes to perform an operation grows at the same
   rate as the length of the list.
 
   ## Call syntax
@@ -83,8 +81,8 @@ defmodule Keyword do
 
       String.split("1-0", "-", trim: true, parts: 2)
 
-  Since tuples, lists, maps, and others are treated the same as function
-  calls in Elixir syntax, this property is also available to them:
+  Since tuples, lists and maps are treated similarly to function
+  arguments in Elixir syntax, this property is also available to them:
 
       iex> {1, 2, foo: :bar}
       {1, 2, [{:foo, :bar}]}
@@ -94,6 +92,7 @@ defmodule Keyword do
 
       iex> %{1 => 2, foo: :bar}
       %{1 => 2, :foo => :bar}
+
   """
 
   @compile :inline_list_funcs
@@ -105,7 +104,24 @@ defmodule Keyword do
   @type t(value) :: [{key, value}]
 
   @doc """
-  Returns `true` if `term` is a keyword list; otherwise returns `false`.
+  Builds a keyword from the given `keys` and the fixed `value`.
+
+  ## Examples
+
+      iex> Keyword.from_keys([:foo, :bar, :baz], :atom)
+      [foo: :atom, bar: :atom, baz: :atom]
+
+  """
+  @doc since: "1.14.0"
+  @spec from_keys([key], value) :: t(value)
+  def from_keys(keys, value) when is_list(keys) do
+    :lists.map(&{&1, value}, keys)
+  end
+
+  @doc """
+  Returns `true` if `term` is a keyword list, otherwise `false`.
+
+  When `term` is a list it is traversed to the end.
 
   ## Examples
 
@@ -145,7 +161,7 @@ defmodule Keyword do
   @doc """
   Creates a keyword list from an enumerable.
 
-  Duplicated entries are removed, the latest one prevails.
+  Removes duplicate entries and the last one prevails.
   Unlike `Enum.into(enumerable, [])`, `Keyword.new(enumerable)`
   guarantees the keys are unique.
 
@@ -158,7 +174,7 @@ defmodule Keyword do
       [a: 3]
 
   """
-  @spec new(Enum.t()) :: t
+  @spec new(Enumerable.t()) :: t
   def new(pairs) do
     new(pairs, fn pair -> pair end)
   end
@@ -166,7 +182,7 @@ defmodule Keyword do
   @doc """
   Creates a keyword list from an enumerable via the transformation function.
 
-  Duplicated entries are removed, the latest one prevails.
+  Removes duplicate entries and the last one prevails.
   Unlike `Enum.into(enumerable, [], fun)`,
   `Keyword.new(enumerable, fun)` guarantees the keys are unique.
 
@@ -176,7 +192,7 @@ defmodule Keyword do
       [a: :a, b: :b]
 
   """
-  @spec new(Enum.t(), (term -> {key, value})) :: t
+  @spec new(Enumerable.t(), (term -> {key, value})) :: t
   def new(pairs, transform) when is_function(transform, 1) do
     fun = fn el, acc ->
       {k, v} = transform.(el)
@@ -187,12 +203,166 @@ defmodule Keyword do
   end
 
   @doc """
-  Gets the value for a specific `key`.
+  Ensures the given `keyword` has only the keys given in `values`.
 
-  If `key` does not exist, return the default value
-  (`nil` if no default value).
+  The second argument must be a list of atoms, specifying
+  a given key, or tuples specifying a key and a default value.
 
-  If duplicated entries exist, the first one is returned.
+  If the keyword list has only the given keys, it returns
+  `{:ok, keyword}` with default values applied. Otherwise it
+  returns `{:error, invalid_keys}` with invalid keys.
+
+  See also: `validate!/2`.
+
+  ## Examples
+
+      iex> {:ok, result} = Keyword.validate([], [one: 1, two: 2])
+      iex> Enum.sort(result)
+      [one: 1, two: 2]
+
+      iex> {:ok, result} = Keyword.validate([two: 3], [one: 1, two: 2])
+      iex> Enum.sort(result)
+      [one: 1, two: 3]
+
+  If atoms are given, they are supported as keys but do not
+  provide a default value:
+
+      iex> {:ok, result} = Keyword.validate([], [:one, two: 2])
+      iex> Enum.sort(result)
+      [two: 2]
+
+      iex> {:ok, result} = Keyword.validate([one: 1], [:one, two: 2])
+      iex> Enum.sort(result)
+      [one: 1, two: 2]
+
+  Passing unknown keys returns an error:
+
+      iex> Keyword.validate([three: 3, four: 4], [one: 1, two: 2])
+      {:error, [:four, :three]}
+
+  Passing the same key multiple times also errors:
+
+      iex> Keyword.validate([one: 1, two: 2, one: 1], [:one, :two])
+      {:error, [:one]}
+
+  """
+  @doc since: "1.13.0"
+  @spec validate(keyword(), values :: [atom() | {atom(), term()}]) ::
+          {:ok, keyword()} | {:error, [atom]}
+  def validate(keyword, values) when is_list(keyword) and is_list(values) do
+    validate(keyword, values, [], [], [])
+  end
+
+  defp validate([{key, _} = pair | keyword], values1, values2, acc, bad_keys) when is_atom(key) do
+    case find_key!(key, values1, values2) do
+      {values1, values2} ->
+        validate(keyword, values1, values2, [pair | acc], bad_keys)
+
+      :error ->
+        case find_key!(key, values2, values1) do
+          {values1, values2} ->
+            validate(keyword, values1, values2, [pair | acc], bad_keys)
+
+          :error ->
+            validate(keyword, values1, values2, acc, [key | bad_keys])
+        end
+    end
+  end
+
+  defp validate([], values1, values2, acc, []) do
+    {:ok, move_pairs!(values1, move_pairs!(values2, acc))}
+  end
+
+  defp validate([], _values1, _values2, _acc, bad_keys) do
+    {:error, bad_keys}
+  end
+
+  defp validate([pair | _], _values1, _values2, _acc, []) do
+    raise ArgumentError,
+          "expected a keyword list as first argument, got invalid entry: #{inspect(pair)}"
+  end
+
+  defp find_key!(key, [key | rest], acc), do: {rest, acc}
+  defp find_key!(key, [{key, _} | rest], acc), do: {rest, acc}
+  defp find_key!(key, [head | tail], acc), do: find_key!(key, tail, [head | acc])
+  defp find_key!(_key, [], _acc), do: :error
+
+  defp move_pairs!([key | rest], acc) when is_atom(key),
+    do: move_pairs!(rest, acc)
+
+  defp move_pairs!([{key, _} = pair | rest], acc) when is_atom(key),
+    do: move_pairs!(rest, [pair | acc])
+
+  defp move_pairs!([], acc),
+    do: acc
+
+  defp move_pairs!([other | _], _) do
+    raise ArgumentError,
+          "expected the second argument to be a list of atoms or tuples, got: #{inspect(other)}"
+  end
+
+  @doc """
+  Similar to `validate/2` but returns the keyword or raises an error.
+
+  ## Examples
+
+      iex> Keyword.validate!([], [one: 1, two: 2]) |> Enum.sort()
+      [one: 1, two: 2]
+      iex> Keyword.validate!([two: 3], [one: 1, two: 2]) |> Enum.sort()
+      [one: 1, two: 3]
+
+  If atoms are given, they are supported as keys but do not
+  provide a default value:
+
+      iex> Keyword.validate!([], [:one, two: 2]) |> Enum.sort()
+      [two: 2]
+      iex> Keyword.validate!([one: 1], [:one, two: 2]) |> Enum.sort()
+      [one: 1, two: 2]
+
+  Passing unknown keys raises an error:
+
+      iex> Keyword.validate!([three: 3], [one: 1, two: 2])
+      ** (ArgumentError) unknown keys [:three] in [three: 3], the allowed keys are: [:one, :two]
+
+  Passing the same key multiple times also errors:
+
+      iex> Keyword.validate!([one: 1, two: 2, one: 1], [:one, :two])
+      ** (ArgumentError) duplicate keys [:one] in [one: 1, two: 2, one: 1]
+
+  """
+  @doc since: "1.13.0"
+  @spec validate!(keyword(), values :: [atom() | {atom(), term()}]) :: keyword()
+  def validate!(keyword, values) do
+    case validate(keyword, values) do
+      {:ok, kw} ->
+        kw
+
+      {:error, invalid_keys} ->
+        keys =
+          for value <- values,
+              do: if(is_atom(value), do: value, else: elem(value, 0))
+
+        message =
+          case Enum.split_with(invalid_keys, &(&1 in keys)) do
+            {_, [_ | _] = unknown} ->
+              "unknown keys #{inspect(unknown)} in #{inspect(keyword)}, " <>
+                "the allowed keys are: #{inspect(keys)}"
+
+            {[_ | _] = known, _} ->
+              "duplicate keys #{inspect(known)} in #{inspect(keyword)}"
+          end
+
+        raise ArgumentError, message
+    end
+  end
+
+  @doc """
+  Gets the value under the given `key`.
+
+  Returns the default value if `key` does not exist
+  (`nil` if no default value is provided).
+
+  If duplicate entries exist, it returns the first one.
   Use `get_values/2` to retrieve all entries.
 
   ## Examples
@@ -206,7 +376,7 @@ defmodule Keyword do
       iex> Keyword.get([a: 1], :b, 3)
       3
 
-  With duplicated keys:
+  With duplicate keys:
 
       iex> Keyword.get([a: 1, a: 2], :a, 3)
       1
@@ -223,14 +393,14 @@ defmodule Keyword do
   end
 
   @doc """
-  Gets the value for a specific `key`.
+  Gets the value under the given `key`.
 
   If `key` does not exist, lazily evaluates `fun` and returns its result.
 
   This is useful if the default value is very expensive to calculate or
-  generally difficult to setup and teardown again.
+  generally difficult to set up and tear down again.
 
-  If duplicated entries exist, the first one is returned.
+  If duplicate entries exist, it returns the first one.
   Use `get_values/2` to retrieve all entries.
 
   ## Examples
@@ -258,14 +428,14 @@ defmodule Keyword do
   @doc """
   Gets the value from `key` and updates it, all in one pass.
 
-  This `fun` argument receives the value of `key` (or `nil` if `key`
+  The `fun` argument receives the value of `key` (or `nil` if `key`
   is not present) and must return a two-element tuple: the current value
   (the retrieved value, which can be operated on before being returned)
   and the new value to be stored under `key`. The `fun` may also
   return `:pop`, implying the current value shall be removed from the
   keyword list and returned.
 
-  The returned value is a tuple with the current value returned by
+  Returns a tuple that contains the current value returned by
   `fun` and a new keyword list with the updated value under `key`.
 
   ## Examples
@@ -280,6 +450,11 @@ defmodule Keyword do
       ...> end)
       {nil, [b: "new value!", a: 1]}
 
+      iex> Keyword.get_and_update([a: 2], :a, fn number ->
+      ...>   {2 * number, 3 * number}
+      ...> end)
+      {4, [a: 6]}
+
       iex> Keyword.get_and_update([a: 1], :a, fn _ -> :pop end)
       {1, []}
 
@@ -323,15 +498,15 @@ defmodule Keyword do
   end
 
   @doc """
-  Gets the value from `key` and updates it. Raises if there is no `key`.
+  Gets the value under `key` and updates it. Raises if there is no `key`.
 
-  This `fun` argument receives the value of `key` and must return a
+  The `fun` argument receives the value under `key` and must return a
   two-element tuple: the current value (the retrieved value, which can be
   operated on before being returned) and the new value to be stored under
   `key`.
 
-  The returned value is a tuple with the current value returned by `fun` and a new
-  keyword list with the updated value under `key`.
+  Returns a tuple that contains the current value returned by
+  `fun` and a new keyword list with the updated value under `key`.
 
   ## Examples
 
@@ -376,13 +551,13 @@ defmodule Keyword do
   end
 
   defp get_and_update!([], key, _fun, acc) when is_atom(key) do
-    raise(KeyError, key: key, term: acc)
+    raise KeyError, key: key, term: acc
   end
 
   @doc """
   Fetches the value for a specific `key` and returns it in a tuple.
 
-  If the `key` does not exist, returns `:error`.
+  If the `key` does not exist, it returns `:error`.
 
   ## Examples
 
@@ -403,7 +578,7 @@ defmodule Keyword do
   @doc """
   Fetches the value for specific `key`.
 
-  If `key` does not exist, a `KeyError` is raised.
+  If the `key` does not exist, it raises a `KeyError`.
 
   ## Examples
 
@@ -417,12 +592,12 @@ defmodule Keyword do
   def fetch!(keywords, key) when is_list(keywords) and is_atom(key) do
     case :lists.keyfind(key, 1, keywords) do
       {^key, value} -> value
-      false -> raise(KeyError, key: key, term: keywords)
+      false -> raise KeyError, key: key, term: keywords
     end
   end
 
   @doc """
-  Gets all values for a specific `key`.
+  Gets all values under a specific `key`.
 
   ## Examples
 
@@ -446,7 +621,7 @@ defmodule Keyword do
   @doc """
   Returns all keys from the keyword list.
 
-  Duplicated keys appear duplicated in the final list of keys.
+  Keeps duplicate keys in the resulting list of keys.
 
   ## Examples
 
@@ -462,26 +637,24 @@ defmodule Keyword do
   """
   @spec keys(t) :: [key]
   def keys(keywords) when is_list(keywords) do
-    try do
-      :lists.map(
-        fn
-          {key, _} when is_atom(key) -> key
-          element -> throw(element)
-        end,
-        keywords
-      )
-    catch
-      element ->
-        raise ArgumentError,
-              "expected a keyword list, but an entry in the list is not a two-element tuple with an atom as its first element, " <>
-                "got: #{inspect(element)}"
-    end
+    :lists.map(
+      fn
+        {key, _} when is_atom(key) -> key
+        element -> throw(element)
+      end,
+      keywords
+    )
+  catch
+    element ->
+      raise ArgumentError,
+            "expected a keyword list, but an entry in the list is not a two-element tuple " <>
+              "with an atom as its first element, got: #{inspect(element)}"
   end
 
   @doc """
   Returns all values from the keyword list.
 
-  Values from duplicated keys will be kept in the final list of values.
+  Keeps values from duplicate keys in the resulting list of values.
 
   ## Examples
 
@@ -518,11 +691,11 @@ defmodule Keyword do
   end
 
   @doc """
-  Deletes the entries in the keyword list for a specific `key`.
+  Deletes the entries in the keyword list under a specific `key`.
 
-  If the `key` does not exist, returns the keyword list unchanged.
+  If the `key` does not exist, it returns the keyword list unchanged.
   Use `delete_first/2` to delete just the first entry in case of
-  duplicated keys.
+  duplicate keys.
 
   ## Examples
 
@@ -548,9 +721,9 @@ defmodule Keyword do
   defp delete_key([], _key), do: []
 
   @doc """
-  Deletes the first entry in the keyword list for a specific `key`.
+  Deletes the first entry in the keyword list under a specific `key`.
 
-  If the `key` does not exist, returns the keyword list unchanged.
+  If the `key` does not exist, it returns the keyword list unchanged.
 
   ## Examples
 
@@ -581,10 +754,10 @@ defmodule Keyword do
   end
 
   @doc """
-  Puts the given `value` under `key`.
+  Puts the given `value` under the specified `key`.
 
-  If a previous value is already stored, all entries are
-  removed and the value is overridden.
+  If a value under `key` already exists, it overrides the value
+  and removes all duplicate entries.
 
   ## Examples
 
@@ -606,19 +779,19 @@ defmodule Keyword do
   in keyword list unless `key` is already present.
 
   This is useful if the value is very expensive to calculate or
-  generally difficult to setup and teardown again.
+  generally difficult to set up and tear down again.
 
   ## Examples
 
       iex> keyword = [a: 1]
       iex> fun = fn ->
       ...>   # some expensive operation here
-      ...>   3
+      ...>   13
       ...> end
       iex> Keyword.put_new_lazy(keyword, :a, fun)
       [a: 1]
       iex> Keyword.put_new_lazy(keyword, :b, fun)
-      [b: 3, a: 1]
+      [b: 13, a: 1]
 
   """
   @spec put_new_lazy(t, key, (() -> value)) :: t
@@ -631,8 +804,7 @@ defmodule Keyword do
   end
 
   @doc """
-  Puts the given `value` under `key` unless the entry `key`
-  already exists.
+  Puts the given `value` under `key`, unless the entry `key` already exists.
 
   ## Examples
 
@@ -653,8 +825,8 @@ defmodule Keyword do
   @doc """
   Puts a value under `key` only if the `key` already exists in `keywords`.
 
-  In the case a value is stored multiple times in the keyword list,
-  later occurrences are removed.
+  In case a key exists multiple times in the keyword list,
+  it removes later occurrences.
 
   ## Examples
 
@@ -686,7 +858,7 @@ defmodule Keyword do
   @doc """
   Puts a value under `key` only if the `key` already exists in `keywords`.
 
-  If `key` is not present in `keywords`, a `KeyError` exception is raised.
+  If `key` is not present in `keywords`, it raises a `KeyError`.
 
   ## Examples
 
@@ -714,13 +886,50 @@ defmodule Keyword do
   end
 
   defp replace!([], key, _value, original) do
-    raise(KeyError, key: key, term: original)
+    raise KeyError, key: key, term: original
   end
 
+  @doc """
+  Replaces the value under `key` using the given function only if
+  `key` already exists in `keywords`.
+
+  In comparison to `replace/3`, this can be useful when it's expensive to calculate the value.
+
+  If `key` does not exist, the original keyword list is returned unchanged.
+
+  ## Examples
+
+      iex> Keyword.replace_lazy([a: 1, b: 2], :a, fn v -> v * 4 end)
+      [a: 4, b: 2]
+
+      iex> Keyword.replace_lazy([a: 2, b: 2, a: 1], :a, fn v -> v * 4 end)
+      [a: 8, b: 2]
+
+      iex> Keyword.replace_lazy([a: 1, b: 2], :c, fn v -> v * 4 end)
+      [a: 1, b: 2]
+
+  """
+  @doc since: "1.14.0"
+  @spec replace_lazy(t, key, (existing_value :: value -> new_value :: value)) :: t
+  def replace_lazy(keywords, key, fun)
+      when is_list(keywords) and is_atom(key) and is_function(fun, 1) do
+    do_replace_lazy(keywords, key, fun)
+  end
+
+  defp do_replace_lazy([{key, value} | keywords], key, fun) do
+    [{key, fun.(value)} | delete(keywords, key)]
+  end
+
+  defp do_replace_lazy([{_, _} = e | keywords], key, fun) do
+    [e | do_replace_lazy(keywords, key, fun)]
+  end
+
+  defp do_replace_lazy([], _key, _value), do: []
+
   @doc """
   Checks if two keywords are equal.
 
-  Two keywords are considered to be equal if they contain
+  Considers two keywords to be equal if they contain
   the same keys and those keys contain the same values.
 
   ## Examples
@@ -747,10 +956,10 @@ defmodule Keyword do
   @doc """
   Merges two keyword lists into one.
 
-  All keys, including duplicated keys, given in `keywords2` will be added
-  to `keywords1`, overriding any existing one.
+  Adds all keys, including duplicate keys, given in `keywords2`
+  to `keywords1`, overriding any existing ones.
 
-  There are no guarantees about the order of keys in the returned keyword.
+  There are no guarantees about the order of the keys in the returned keyword.
 
   ## Examples
 
@@ -791,13 +1000,13 @@ defmodule Keyword do
   @doc """
   Merges two keyword lists into one.
 
-  All keys, including duplicated keys, given in `keywords2` will be added
-  to `keywords1`. The given function will be invoked to solve conflicts.
+  Adds all keys, including duplicate keys, given in `keywords2`
+  to `keywords1`. Invokes the given function to solve conflicts.
 
-  If `keywords2` has duplicate keys, the given function will be invoked
+  If `keywords2` has duplicate keys, it invokes the given function
   for each matching pair in `keywords1`.
 
-  There are no guarantees about the order of keys in the returned keyword.
+  There are no guarantees about the order of the keys in the returned keyword.
 
   ## Examples
 
@@ -871,12 +1080,11 @@ defmodule Keyword do
   end
 
   @doc """
-  Updates the `key` with the given function.
+  Updates the value under `key` using the given function.
 
-  If the `key` does not exist, raises `KeyError`.
+  Raises `KeyError` if the `key` does not exist.
 
-  If there are duplicated keys, they are all removed and only the first one
-  is updated.
+  Removes all duplicate keys and only updates the first one.
 
   ## Examples
 
@@ -904,18 +1112,16 @@ defmodule Keyword do
   end
 
   defp update!([], key, _fun, original) do
-    raise(KeyError, key: key, term: original)
+    raise KeyError, key: key, term: original
   end
 
   @doc """
-  Updates the `key` in `keywords` with the given function.
+  Updates the value under `key` in `keywords` using the given function.
 
   If the `key` does not exist, it inserts the given `default` value.
+  Does not pass the `default` value through the update function.
 
-  If there are duplicated keys, they are all removed and only the first one
-  is updated.
-
-  The default value will not be passed through the update function.
+  Removes all duplicate keys and only updates the first one.
 
   ## Examples
 
@@ -948,14 +1154,14 @@ defmodule Keyword do
   end
 
   @doc """
-  Takes all entries corresponding to the given keys and extracts them into a
+  Takes all entries corresponding to the given `keys` and extracts them into a
   separate keyword list.
 
   Returns a tuple with the new list and the old list with removed keys.
 
-  Keys for which there are no entries in the keyword list are ignored.
+  Ignores keys for which there are no entries in the keyword list.
 
-  Entries with duplicated keys end up in the same keyword list.
+  Entries with duplicate keys end up in the same keyword list.
 
   ## Examples
 
@@ -980,10 +1186,10 @@ defmodule Keyword do
   end
 
   @doc """
-  Takes all entries corresponding to the given keys and returns them in a new
+  Takes all entries corresponding to the given `keys` and returns them as a new
   keyword list.
 
-  Duplicated keys are preserved in the new keyword list.
+  Preserves duplicate keys in the new keyword list.
 
   ## Examples
 
@@ -995,16 +1201,18 @@ defmodule Keyword do
   """
   @spec take(t, [key]) :: t
   def take(keywords, keys) when is_list(keywords) and is_list(keys) do
-    :lists.filter(fn {k, _} -> k in keys end, keywords)
+    :lists.filter(fn {k, _} -> :lists.member(k, keys) end, keywords)
   end
 
   @doc """
-  Drops the given keys from the keyword list.
+  Drops the given `keys` from the keyword list.
 
-  Duplicated keys are preserved in the new keyword list.
+  Removes duplicate keys from the new keyword list.
 
   ## Examples
 
+      iex> Keyword.drop([a: 1, a: 2], [:a])
+      []
       iex> Keyword.drop([a: 1, b: 2, c: 3], [:b, :d])
       [a: 1, c: 3]
       iex> Keyword.drop([a: 1, b: 2, b: 3, c: 3, a: 5], [:b, :d])
@@ -1013,7 +1221,7 @@ defmodule Keyword do
   """
   @spec drop(t, [key]) :: t
   def drop(keywords, keys) when is_list(keywords) and is_list(keys) do
-    :lists.filter(fn {key, _} -> key not in keys end, keywords)
+    :lists.filter(fn {k, _} -> k not in keys end, keywords)
   end
 
   @doc """
@@ -1021,11 +1229,10 @@ defmodule Keyword do
 
   It returns a tuple where the first element is the first value for `key` and the
   second element is a keyword list with all entries associated with `key` removed.
-  If the `key` is not present in the keyword list, `{default, keyword_list}` is
-  returned.
+  If the `key` is not present in the keyword list, it returns `{default, keyword_list}`.
 
   If you don't want to remove all the entries associated with `key` use `pop_first/3`
-  instead, that function will remove only the first entry.
+  instead, which will remove only the first entry.
 
   ## Examples
 
@@ -1051,7 +1258,7 @@ defmodule Keyword do
   Returns the first value for `key` and removes all associated entries in the keyword list,
   raising if `key` is not present.
 
-  This function behaves like `pop/3`, but raises in cases the `key` is not present in the
+  This function behaves like `pop/3`, but raises in case the `key` is not present in the
   given `keywords`.
 
   ## Examples
@@ -1078,11 +1285,10 @@ defmodule Keyword do
 
   It returns a tuple where the first element is a list of values for `key` and the
   second element is a keyword list with all entries associated with `key` removed.
-  If the `key` is not present in the keyword list, `{[], keyword_list}` is
-  returned.
+  If the `key` is not present in the keyword list, it returns `{[], keyword_list}`.
 
   If you don't want to remove all the entries associated with `key` use `pop_first/3`
-  instead, that function will remove only the first entry.
+  instead, which will remove only the first entry.
 
   ## Examples
 
@@ -1113,10 +1319,9 @@ defmodule Keyword do
   Lazily returns and removes all values associated with `key` in the keyword list.
 
   This is useful if the default value is very expensive to calculate or
-  generally difficult to setup and teardown again.
+  generally difficult to set up and tear down again.
 
-  All duplicated keys are removed. See `pop_first/3` for
-  removing only the first entry.
+  Removes all duplicate keys. See `pop_first/3` for removing only the first entry.
 
   ## Examples
 
@@ -1135,18 +1340,15 @@ defmodule Keyword do
   def pop_lazy(keywords, key, fun)
       when is_list(keywords) and is_atom(key) and is_function(fun, 0) do
     case fetch(keywords, key) do
-      {:ok, value} ->
-        {value, delete(keywords, key)}
-
-      :error ->
-        {fun.(), keywords}
+      {:ok, value} -> {value, delete(keywords, key)}
+      :error -> {fun.(), keywords}
     end
   end
 
   @doc """
   Returns and removes the first value associated with `key` in the keyword list.
 
-  Duplicated keys are not removed.
+  Keeps duplicate keys in the resulting keyword list.
 
   ## Examples
 
@@ -1178,13 +1380,76 @@ defmodule Keyword do
 
   """
   @spec to_list(t) :: t
-  def to_list(keyword) when is_list(keyword) do
-    keyword
+  def to_list(keywords) when is_list(keywords) do
+    keywords
   end
 
   @doc false
   @deprecated "Use Kernel.length/1 instead"
-  def size(keyword) do
-    length(keyword)
+  def size(keywords) do
+    length(keywords)
+  end
+
+  @doc """
+  Returns a keyword list containing only the entries from `keywords`
+  for which the function `fun` returns a truthy value.
+
+  See also `reject/2` which discards all entries where the function
+  returns a truthy value.
+
+  ## Examples
+
+      iex> Keyword.filter([one: 1, two: 2, three: 3], fn {_key, val} -> rem(val, 2) == 1 end)
+      [one: 1, three: 3]
+
+  """
+  @doc since: "1.13.0"
+  @spec filter(t, ({key, value} -> as_boolean(term))) :: t
+  def filter(keywords, fun) when is_list(keywords) and is_function(fun, 1) do
+    do_filter(keywords, fun)
+  end
+
+  defp do_filter([], _fun), do: []
+
+  defp do_filter([{_, _} = entry | entries], fun) do
+    if fun.(entry) do
+      [entry | do_filter(entries, fun)]
+    else
+      do_filter(entries, fun)
+    end
+  end
+
+  @doc """
+  Returns a keyword list excluding the entries from `keywords`
+  for which the function `fun` returns a truthy value.
+
+  See also `filter/2`.
+
+  ## Examples
+
+      iex> Keyword.reject([one: 1, two: 2, three: 3], fn {_key, val} -> rem(val, 2) == 1 end)
+      [two: 2]
+
+  """
+  @doc since: "1.13.0"
+  @spec reject(t, ({key, value} -> as_boolean(term))) :: t
+  def reject(keywords, fun) when is_list(keywords) and is_function(fun, 1) do
+    do_reject(keywords, fun)
+  end
+
+  defp do_reject([], _fun), do: []
+
+  defp do_reject([{_, _} = entry | entries], fun) do
+    if fun.(entry) do
+      do_reject(entries, fun)
+    else
+      [entry | do_reject(entries, fun)]
+    end
+  end
+
+  @doc false
+  @deprecated "Use Keyword.new/2 instead"
+  def map(keywords, fun) when is_list(keywords) do
+    Enum.map(keywords, fn {k, v} -> {k, fun.({k, v})} end)
   end
 end
