@@ -8,7 +8,7 @@ defmodule List do
       [1, "two", 3, :four]
 
   Two lists can be concatenated and subtracted using the
-  `Kernel.++/2` and `Kernel.--/2` operators:
+  `++/2` and `--/2` operators:
 
       iex> [1, 2, 3] ++ [4, 5, 6]
       [1, 2, 3, 4, 5, 6]
@@ -230,7 +230,7 @@ defmodule List do
 
   @doc """
   Folds (reduces) the given list from the left with
-  a function. Requires an accumulator.
+  a function. Requires an accumulator, which can be any value.
 
   ## Examples
 
@@ -240,6 +240,9 @@ defmodule List do
       iex> List.foldl([1, 2, 3, 4], 0, fn x, acc -> x - acc end)
       2
 
+      iex> List.foldl([1, 2, 3], {0, 0}, fn x, {a1, a2} -> {a1 + x, a2 - x} end)
+      {6, -6}
+
   """
   @spec foldl([elem], acc, (elem, acc -> acc)) :: acc when elem: var, acc: var
   def foldl(list, acc, fun) when is_list(list) and is_function(fun) do
@@ -248,13 +251,16 @@ defmodule List do
 
   @doc """
   Folds (reduces) the given list from the right with
-  a function. Requires an accumulator.
+  a function. Requires an accumulator, which can be any value.
 
   ## Examples
 
       iex> List.foldr([1, 2, 3, 4], 0, fn x, acc -> x - acc end)
       -2
 
+      iex> List.foldr([1, 2, 3, 4], %{sum: 0, product: 1}, fn x, %{sum: a1, product: a2} -> %{sum: a1 + x, product: a2 * x} end)
+      %{product: 24, sum: 10}
+
   """
   @spec foldr([elem], acc, (elem, acc -> acc)) :: acc when elem: var, acc: var
   def foldr(list, acc, fun) when is_list(list) and is_function(fun) do
@@ -333,12 +339,52 @@ defmodule List do
       iex> List.keyfind([a: 1, b: 2], :c, 0)
       nil
 
+  This function works for any list of tuples:
+
+      iex> List.keyfind([{22, "SSH"}, {80, "HTTP"}], 22, 0)
+      {22, "SSH"}
+
   """
   @spec keyfind([tuple], any, non_neg_integer, any) :: any
-  def keyfind(list, key, position, default \\ nil) do
+  def keyfind(list, key, position, default \\ nil) when is_integer(position) do
     :lists.keyfind(key, position + 1, list) || default
   end
 
+  @doc """
+  Receives a list of tuples and returns the first tuple
+  where the element at `position` in the tuple matches the
+  given `key`.
+
+  If no matching tuple is found, an error is raised.
+
+  ## Examples
+
+      iex> List.keyfind!([a: 1, b: 2], :a, 0)
+      {:a, 1}
+
+      iex> List.keyfind!([a: 1, b: 2], 2, 1)
+      {:b, 2}
+
+      iex> List.keyfind!([a: 1, b: 2], :c, 0)
+      ** (KeyError) key :c at position 0 not found in: [a: 1, b: 2]
+
+  This function works for any list of tuples:
+
+      iex> List.keyfind!([{22, "SSH"}, {80, "HTTP"}], 22, 0)
+      {22, "SSH"}
+
+  """
+  @doc since: "1.13.0"
+  @spec keyfind!([tuple], any, non_neg_integer) :: any
+  def keyfind!(list, key, position) when is_integer(position) do
+    :lists.keyfind(key, position + 1, list) ||
+      raise KeyError,
+        key: key,
+        term: list,
+        message:
+          "key #{inspect(key)} at position #{inspect(position)} not found in: #{inspect(list)}"
+  end
+
   @doc """
   Receives a list of tuples and returns `true` if there is
   a tuple where the element at `position` in the tuple matches
@@ -355,9 +401,14 @@ defmodule List do
       iex> List.keymember?([a: 1, b: 2], :c, 0)
       false
 
+  This function works for any list of tuples:
+
+      iex> List.keymember?([{22, "SSH"}, {80, "HTTP"}], 22, 0)
+      true
+
   """
   @spec keymember?([tuple], any, non_neg_integer) :: boolean
-  def keymember?(list, key, position) do
+  def keymember?(list, key, position) when is_integer(position) do
     :lists.keymember(key, position + 1, list)
   end
 
@@ -373,15 +424,26 @@ defmodule List do
       iex> List.keyreplace([a: 1, b: 2], :a, 1, {:a, 3})
       [a: 1, b: 2]
 
+  This function works for any list of tuples:
+
+      iex> List.keyreplace([{22, "SSH"}, {80, "HTTP"}], 22, 0, {22, "Secure Shell"})
+      [{22, "Secure Shell"}, {80, "HTTP"}]
+
   """
   @spec keyreplace([tuple], any, non_neg_integer, tuple) :: [tuple]
-  def keyreplace(list, key, position, new_tuple) do
+  def keyreplace(list, key, position, new_tuple) when is_integer(position) do
     :lists.keyreplace(key, position + 1, list, new_tuple)
   end
 
   @doc """
   Receives a list of tuples and sorts the elements
-  at `position` of the tuples. The sort is stable.
+  at `position` of the tuples.
+
+  The sort is stable.
+
+  A `sorter` argument is available since Elixir v1.14.0. Similar to
+  `Enum.sort/2`, the sorter can be an anonymous function, the atoms
+  `:asc` or `:desc`, or module that implements a compare function.
 
   ## Examples
 
@@ -391,12 +453,69 @@ defmodule List do
       iex> List.keysort([a: 5, c: 1, b: 3], 0)
       [a: 5, b: 3, c: 1]
 
+  To sort in descending order:
+
+      iex> List.keysort([a: 5, c: 1, b: 3], 0, :desc)
+      [c: 1, b: 3, a: 5]
+
+  As in `Enum.sort/2`, avoid using the default sorting function to sort
+  structs, as by default it performs structural comparison instead of a
+  semantic one. In such cases, you shall pass a sorting function as third
+  element or any module that implements a `compare/2` function. For example,
+  if you have tuples with user names and their birthday, and you want to
+  sort on their birthday, in both ascending and descending order, you should
+  do:
+
+      iex> users = [
+      ...>   {"Ellis", ~D[1943-05-11]},
+      ...>   {"Lovelace", ~D[1815-12-10]},
+      ...>   {"Turing", ~D[1912-06-23]}
+      ...> ]
+      iex> List.keysort(users, 1, Date)
+      [
+        {"Lovelace", ~D[1815-12-10]},
+        {"Turing", ~D[1912-06-23]},
+        {"Ellis", ~D[1943-05-11]}
+      ]
+      iex> List.keysort(users, 1, {:desc, Date})
+      [
+        {"Ellis", ~D[1943-05-11]},
+        {"Turing", ~D[1912-06-23]},
+        {"Lovelace", ~D[1815-12-10]}
+      ]
+
   """
-  @spec keysort([tuple], non_neg_integer) :: [tuple]
-  def keysort(list, position) do
+  @doc since: "1.14.0"
+  @spec keysort(
+          [tuple],
+          non_neg_integer,
+          (any, any -> boolean) | :asc | :desc | module() | {:asc | :desc, module()}
+        ) :: [tuple]
+  def keysort(list, position, sorter \\ :asc)
+
+  def keysort(list, position, :asc) when is_list(list) and is_integer(position) do
     :lists.keysort(position + 1, list)
   end
 
+  def keysort(list, position, sorter) when is_list(list) and is_integer(position) do
+    :lists.sort(keysort_fun(sorter, position + 1), list)
+  end
+
+  defp keysort_fun(sorter, position) when is_function(sorter, 2),
+    do: &sorter.(:erlang.element(position, &1), :erlang.element(position, &2))
+
+  defp keysort_fun(:desc, position),
+    do: &(:erlang.element(position, &1) >= :erlang.element(position, &2))
+
+  defp keysort_fun(module, position) when is_atom(module),
+    do: &(module.compare(:erlang.element(position, &1), :erlang.element(position, &2)) != :gt)
+
+  defp keysort_fun({:asc, module}, position) when is_atom(module),
+    do: &(module.compare(:erlang.element(position, &1), :erlang.element(position, &2)) != :gt)
+
+  defp keysort_fun({:desc, module}, position) when is_atom(module),
+    do: &(module.compare(:erlang.element(position, &1), :erlang.element(position, &2)) != :lt)
+
   @doc """
   Receives a `list` of tuples and replaces the element
   identified by `key` at `position` with `new_tuple`.
@@ -411,9 +530,14 @@ defmodule List do
       iex> List.keystore([a: 1, b: 2], :c, 0, {:c, 3})
       [a: 1, b: 2, c: 3]
 
+  This function works for any list of tuples:
+
+      iex> List.keystore([{22, "SSH"}], 80, 0, {80, "HTTP"})
+      [{22, "SSH"}, {80, "HTTP"}]
+
   """
   @spec keystore([tuple], any, non_neg_integer, tuple) :: [tuple, ...]
-  def keystore(list, key, position, new_tuple) do
+  def keystore(list, key, position, new_tuple) when is_integer(position) do
     :lists.keystore(key, position + 1, list, new_tuple)
   end
 
@@ -433,9 +557,14 @@ defmodule List do
       iex> List.keydelete([a: 1, b: 2], :c, 0)
       [a: 1, b: 2]
 
+  This function works for any list of tuples:
+
+      iex> List.keydelete([{22, "SSH"}, {80, "HTTP"}], 80, 0)
+      [{22, "SSH"}]
+
   """
   @spec keydelete([tuple], any, non_neg_integer) :: [tuple]
-  def keydelete(list, key, position) do
+  def keydelete(list, key, position) when is_integer(position) do
     :lists.keydelete(key, position + 1, list)
   end
 
@@ -457,9 +586,14 @@ defmodule List do
       iex> List.keytake([a: 1, b: 2], :c, 0)
       nil
 
+  This function works for any list of tuples:
+
+      iex> List.keytake([{22, "SSH"}, {80, "HTTP"}], 80, 0)
+      {{80, "HTTP"}, [{22, "SSH"}]}
+
   """
   @spec keytake([tuple], any, non_neg_integer) :: {tuple, [tuple]} | nil
-  def keytake(list, key, position) do
+  def keytake(list, key, position) when is_integer(position) do
     case :lists.keytake(key, position + 1, list) do
       {:value, element, list} -> {element, list}
       false -> nil
@@ -813,14 +947,22 @@ defmodule List do
   end
 
   @doc """
-  Converts a charlist to an existing atom. Raises an `ArgumentError`
-  if the atom does not exist.
+  Converts a charlist to an existing atom.
 
   Elixir supports conversions from charlists which contains any Unicode
-  code point.
+  code point. Raises an `ArgumentError` if the atom does not exist.
 
   Inlined by the compiler.
 
+  > #### Atoms and modules {: .info}
+  >
+  > Since Elixir is a compiled language, the atoms defined in a module
+  > will only exist after said module is loaded, which typically happens
+  > whenever a function in the module is executed. Therefore, it is
+  > generally recommended to call `List.to_existing_atom/1` only to
+  > convert atoms defined within the module making the function call
+  > to `to_existing_atom/1`.
+
   ## Examples
 
       iex> _ = :my_atom
