@@ -88,12 +88,18 @@ defmodule Map do
       %{1 => :one, 2 => :two, 3 => :three}
 
   Maps also support a specific update syntax to update the value stored under
-  *existing* atom keys:
+  *existing* keys. You can update using the atom keys syntax:
 
       iex> map = %{one: 1, two: 2}
       iex> %{map | one: "one"}
       %{one: "one", two: 2}
 
+  Or any other key:
+
+      iex> other_map = %{"three" => 3, "four" => 4}
+      iex> %{other_map | "three" => "three"}
+      %{"four" => 4, "three" => "three"}
+
   When a key that does not exist in the map is updated a `KeyError` exception will be raised:
 
       %{map | three: 3}
@@ -117,6 +123,27 @@ defmodule Map do
   @type value :: any
   @compile {:inline, fetch: 2, fetch!: 2, get: 2, put: 3, delete: 2, has_key?: 2, replace!: 3}
 
+  # TODO: Remove conditional on Erlang/OTP 24+
+  @compile {:no_warn_undefined, {:maps, :from_keys, 2}}
+  @doc """
+  Builds a map from the given `keys` and the fixed `value`.
+
+  ## Examples
+
+      iex> Map.from_keys([1, 2, 3], :number)
+      %{1 => :number, 2 => :number, 3 => :number}
+
+  """
+  @doc since: "1.14.0"
+  @spec from_keys([key], value) :: map
+  def from_keys(keys, value) do
+    if function_exported?(:maps, :from_keys, 2) do
+      :maps.from_keys(keys, value)
+    else
+      :maps.from_list(:lists.map(&{&1, value}, keys))
+    end
+  end
+
   @doc """
   Returns all keys from `map`.
 
@@ -214,7 +241,24 @@ defmodule Map do
 
   """
   @spec new(Enumerable.t(), (term -> {key, value})) :: map
-  def new(enumerable, transform) when is_function(transform, 1) do
+  def new(enumerable, transform)
+  def new(%_{} = enumerable, transform), do: new_from_enum(enumerable, transform)
+  def new(%{} = map, transform), do: new_from_map(map, transform)
+  def new(enumerable, transform), do: new_from_enum(enumerable, transform)
+
+  defp new_from_map(map, transform) when is_function(transform, 1) do
+    iter = :maps.iterator(map)
+    next = :maps.next(iter)
+    :maps.from_list(do_map(next, transform))
+  end
+
+  defp do_map(:none, _fun), do: []
+
+  defp do_map({key, value, iter}, transform) do
+    [transform.({key, value}) | do_map(:maps.next(iter), transform)]
+  end
+
+  defp new_from_enum(enumerable, transform) when is_function(transform, 1) do
     enumerable
     |> Enum.map(transform)
     |> :maps.from_list()
@@ -318,7 +362,7 @@ defmodule Map do
   def replace(map, key, value) do
     case map do
       %{^key => _value} ->
-        put(map, key, value)
+        %{map | key => value}
 
       %{} ->
         map
@@ -350,6 +394,32 @@ defmodule Map do
     :maps.update(key, value, map)
   end
 
+  @doc """
+  Replaces the value under `key` using the given function only if
+  `key` already exists in `map`.
+
+  In comparison to `replace/3`, this can be useful when it's expensive to calculate the value.
+
+  If `key` does not exist, the original map is returned unchanged.
+
+  ## Examples
+
+      iex> Map.replace_lazy(%{a: 1, b: 2}, :a, fn v -> v * 4 end)
+      %{a: 4, b: 2}
+
+      iex> Map.replace_lazy(%{a: 1, b: 2}, :c, fn v -> v * 4 end)
+      %{a: 1, b: 2}
+
+  """
+  @doc since: "1.14.0"
+  @spec replace_lazy(map, key, (existing_value :: value -> new_value :: value)) :: map
+  def replace_lazy(map, key, fun) when is_map(map) and is_function(fun, 1) do
+    case map do
+      %{^key => val} -> %{map | key => fun.(val)}
+      %{} -> map
+    end
+  end
+
   @doc """
   Evaluates `fun` and puts the result under `key`
   in `map` unless `key` is already present.
@@ -546,7 +616,7 @@ defmodule Map do
   If you have a struct and you would like to merge a set of keys into the
   struct, do not use this function, as it would merge all keys on the right
   side into the struct, even if the key is not part of the struct. Instead,
-  use `Kernel.struct/2`.
+  use `struct/2`.
 
   Inlined by the compiler.
 
@@ -614,7 +684,7 @@ defmodule Map do
   def update(map, key, default, fun) when is_function(fun, 1) do
     case map do
       %{^key => value} ->
-        put(map, key, fun.(value))
+        %{map | key => fun.(value)}
 
       %{} ->
         put(map, key, default)
@@ -810,7 +880,7 @@ defmodule Map do
   @spec update!(map, key, (existing_value :: value -> new_value :: value)) :: map
   def update!(map, key, fun) when is_function(fun, 1) do
     value = fetch!(map, key)
-    put(map, key, fun.(value))
+    %{map | key => fun.(value)}
   end
 
   @doc """
@@ -895,7 +965,7 @@ defmodule Map do
 
     case fun.(value) do
       {get, update} ->
-        {get, put(map, key, update)}
+        {get, %{map | key => update}}
 
       :pop ->
         {value, delete(map, key)}
@@ -971,4 +1041,79 @@ defmodule Map do
   def size(map) do
     map_size(map)
   end
+
+  @doc """
+  Returns a map containing only those pairs from `map`
+  for which `fun` returns a truthy value.
+
+  `fun` receives the key and value of each of the
+  elements in the map as a key-value pair.
+
+  See also `reject/2` which discards all elements where the
+  function returns a truthy value.
+
+  > Note: if you find yourself doing multiple calls to `Map.filter/2`
+  > and `Map.reject/2` in a pipeline, it is likely more efficient
+  > to use `Enum.map/2` and `Enum.filter/2` instead and convert to
+  > a map at the end using `Map.new/1`.
+
+  ## Examples
+
+      iex> Map.filter(%{one: 1, two: 2, three: 3}, fn {_key, val} -> rem(val, 2) == 1 end)
+      %{one: 1, three: 3}
+
+  """
+  @doc since: "1.13.0"
+  @spec filter(map, ({key, value} -> as_boolean(term))) :: map
+  def filter(map, fun) when is_map(map) and is_function(fun, 1) do
+    iter = :maps.iterator(map)
+    next = :maps.next(iter)
+    :maps.from_list(do_filter(next, fun))
+  end
+
+  defp do_filter(:none, _fun), do: []
+
+  defp do_filter({key, value, iter}, fun) do
+    if fun.({key, value}) do
+      [{key, value} | do_filter(:maps.next(iter), fun)]
+    else
+      do_filter(:maps.next(iter), fun)
+    end
+  end
+
+  @doc """
+  Returns map excluding the pairs from `map` for which `fun` returns
+  a truthy value.
+
+  See also `filter/2`.
+
+  ## Examples
+
+      iex> Map.reject(%{one: 1, two: 2, three: 3}, fn {_key, val} -> rem(val, 2) == 1 end)
+      %{two: 2}
+
+  """
+  @doc since: "1.13.0"
+  @spec reject(map, ({key, value} -> as_boolean(term))) :: map
+  def reject(map, fun) when is_map(map) and is_function(fun, 1) do
+    iter = :maps.iterator(map)
+    next = :maps.next(iter)
+    :maps.from_list(do_reject(next, fun))
+  end
+
+  defp do_reject(:none, _fun), do: []
+
+  defp do_reject({key, value, iter}, fun) do
+    if fun.({key, value}) do
+      do_reject(:maps.next(iter), fun)
+    else
+      [{key, value} | do_reject(:maps.next(iter), fun)]
+    end
+  end
+
+  @doc false
+  @deprecated "Use Map.new/2 instead (invoke Map.from_struct/1 before if you have a struct)"
+  def map(map, fun) when is_map(map) do
+    :maps.map(fn k, v -> fun.({k, v}) end, map)
+  end
 end
