@@ -1,44 +1,158 @@
 defmodule Module.ParallelChecker do
   @moduledoc false
 
+  import Kernel, except: [spawn: 3]
+
   @type cache() :: {pid(), :ets.tid()}
   @type warning() :: term()
   @type kind() :: :def | :defmacro
   @type mode() :: :elixir | :erlang
 
   @doc """
-  Receives pairs of module maps and BEAM binaries. In parallel it verifies
-  the modules and adds the ExCk chunk to the binaries. Returns the updated
-  binaries and a list of warnings from the verification.
+  Initializes the parallel checker process.
   """
-  @spec verify([{map(), binary()}], [{module(), binary()}], pos_integer() | nil) :: [warning()]
-  def verify(compiled_modules, runtime_binaries, schedulers \\ nil) do
-    compiled_maps = Enum.map(compiled_modules, fn {map, _binary} -> {map.module, map} end)
+  def start_link(schedulers \\ nil) do
+    :gen_server.start_link(__MODULE__, schedulers, [])
+  end
 
-    case compiled_maps ++ runtime_binaries do
-      [] ->
-        []
+  @doc """
+  Stops the parallel checker process.
+  """
+  def stop(checker) do
+    send(checker, {__MODULE__, :stop})
+    :ok
+  end
 
-      check_modules ->
-        schedulers = schedulers || max(:erlang.system_info(:schedulers_online), 2)
+  @doc """
+  Gets the parallel checker data from pdict.
+  """
+  def get do
+    case :erlang.get(:elixir_checker_info) do
+      {parent, nil} ->
+        {:ok, checker} = start_link()
+        put(parent, checker)
+        {parent, checker}
 
-        {:ok, server} =
-          :gen_server.start_link(__MODULE__, [check_modules, self(), schedulers], [])
-
-        preload_cache(get_ets(server), check_modules)
-        start(server)
-        collect_results(length(check_modules), [])
+      {parent, checker} ->
+        {parent, checker}
     end
   end
 
-  defp collect_results(0, warnings) do
+  @doc """
+  Stores the parallel checker information.
+  """
+  def put(pid, checker) when is_pid(pid) and is_pid(checker) do
+    :erlang.put(:elixir_checker_info, {pid, checker})
+  end
+
+  @doc """
+  Spawns a process that runs the parallel checker.
+  """
+  def spawn({pid, checker}, module, info) do
+    ref = make_ref()
+
+    spawned =
+      spawn(fn ->
+        Process.link(pid)
+        mon_ref = Process.monitor(pid)
+
+        receive do
+          {^ref, :cache, ets} ->
+            module_map =
+              if is_map(info) do
+                info
+              else
+                info |> File.read!() |> fetch_module_map!(module)
+              end
+
+            cache_from_module_map(ets, module_map)
+            send(checker, {ref, :cached})
+
+            receive do
+              {^ref, :check} ->
+                # Set the compiler info so we can collect warnings
+                :erlang.put(:elixir_compiler_info, {pid, self()})
+                warnings = check_module(module_map, {checker, ets})
+                send(pid, {__MODULE__, module, warnings})
+                send(checker, {__MODULE__, :done})
+            end
+
+          {:DOWN, ^mon_ref, _, _, _} ->
+            :ok
+        end
+      end)
+
+    register(checker, spawned, ref)
+    :ok
+  end
+
+  @doc """
+  Verifies the given compilation function
+  by starting a checker if one does not exist.
+  See `verify/3`.
+  """
+  def verify(fun) do
+    case :erlang.get(:elixir_compiler_info) do
+      :undefined ->
+        previous = :erlang.put(:elixir_checker_info, {self(), nil})
+
+        try do
+          result = fun.()
+
+          case :erlang.get(:elixir_checker_info) do
+            {_, nil} -> :ok
+            {_, checker} -> verify(checker, [])
+          end
+
+          result
+        after
+          {_, checker} = :erlang.get(:elixir_checker_info)
+
+          if previous != :undefined do
+            :erlang.put(:elixir_checker_info, previous)
+          else
+            :erlang.erase(:elixir_checker_info)
+          end
+
+          checker && stop(checker)
+        end
+
+      _ ->
+        # If we are during compilation, then they will be
+        # reported to the compiler, which will validate them.
+        fun.()
+    end
+  end
+
+  @doc """
+  Receives pairs of module maps and BEAM binaries. In parallel it verifies
+  the modules and adds the ExCk chunk to the binaries. Returns the updated
+  list of warnings from the verification.
+  """
+  @spec verify(pid(), [{module(), binary()}]) :: [warning()]
+  def verify(checker, runtime_files) do
+    for {module, file} <- runtime_files do
+      spawn({self(), checker}, module, file)
+    end
+
+    modules = :gen_server.call(checker, :start)
+    collect_results(modules, [])
+  end
+
+  defp collect_results([], warnings) do
     warnings
   end
 
-  defp collect_results(count, warnings) do
+  defp collect_results(modules, warnings) do
     receive do
+      {:warning, file, location, message} ->
+        file = file && Path.absname(file)
+        message = :unicode.characters_to_binary(message)
+        warning = {file, location, message}
+        collect_results(modules, [warning | warnings])
+
       {__MODULE__, _module, new_warnings} ->
-        collect_results(count - 1, new_warnings ++ warnings)
+        collect_results(tl(modules), new_warnings ++ warnings)
     end
   end
 
@@ -46,8 +160,8 @@ defmodule Module.ParallelChecker do
   Test cache.
   """
   def test_cache do
-    {:ok, pid} = :gen_server.start_link(__MODULE__, [[], self(), 1], [])
-    {pid, get_ets(pid)}
+    {:ok, checker} = start_link()
+    {checker, :gen_server.call(checker, :ets)}
   end
 
   @doc """
@@ -98,35 +212,26 @@ defmodule Module.ParallelChecker do
 
   ## Module checking
 
-  defp check_module(module, cache) do
-    case extract_definitions(module) do
-      {:ok, module, file, definitions, no_warn_undefined} ->
-        Module.Types.warnings(module, file, definitions, no_warn_undefined, cache)
-        |> group_warnings()
-        |> emit_warnings()
+  defp check_module(module_map, cache) do
+    %{module: module, file: file, compile_opts: compile_opts, definitions: definitions} =
+      module_map
 
-      :error ->
-        []
-    end
-  end
-
-  defp extract_definitions({module, module_map}) when is_map(module_map) do
     no_warn_undefined =
-      module_map.compile_opts
+      compile_opts
       |> extract_no_warn_undefined()
       |> merge_compiler_no_warn_undefined()
 
-    {:ok, module, module_map.file, module_map.definitions, no_warn_undefined}
-  end
+    warnings =
+      module
+      |> Module.Types.warnings(file, definitions, no_warn_undefined, cache)
+      |> group_warnings()
+      |> emit_warnings()
 
-  defp extract_definitions({module, binary}) when is_binary(binary) do
-    with {:ok, {_, [debug_info: chunk]}} <- :beam_lib.chunks(binary, [:debug_info]),
-         {:debug_info_v1, backend, data} <- chunk,
-         {:ok, module_map} <- backend.debug_info(:elixir_v1, module, data, []) do
-      extract_definitions({module, module_map})
-    else
-      _ -> :error
-    end
+    module_map
+    |> Map.get(:after_verify, [])
+    |> Enum.each(fn {verify_mod, verify_fun} -> apply(verify_mod, verify_fun, [module]) end)
+
+    warnings
   end
 
   defp extract_no_warn_undefined(compile_opts) do
@@ -139,11 +244,8 @@ defmodule Module.ParallelChecker do
 
   defp merge_compiler_no_warn_undefined(no_warn_undefined) do
     case Code.get_compiler_option(:no_warn_undefined) do
-      :all ->
-        :all
-
-      list when is_list(list) ->
-        no_warn_undefined ++ list
+      :all -> :all
+      list when is_list(list) -> no_warn_undefined ++ list
     end
   end
 
@@ -176,7 +278,7 @@ defmodule Module.ParallelChecker do
 
   defp format_locations(locations) do
     [
-      "Found at #{length(locations)} locations:\n",
+      "Invalid call found at #{length(locations)} locations:\n",
       Enum.map(locations, &format_location/1)
     ]
   end
@@ -204,106 +306,7 @@ defmodule Module.ParallelChecker do
     IO.puts(:stderr, [:elixir_errors.warning_prefix(), message])
   end
 
-  ## Server callbacks
-
-  def init([modules, send_results, schedulers]) do
-    ets = :ets.new(:checker_cache, [:set, :public, {:read_concurrency, true}])
-
-    state = %{
-      ets: ets,
-      waiting: %{},
-      send_results: send_results,
-      modules: modules,
-      spawned: 0,
-      schedulers: schedulers
-    }
-
-    {:ok, state}
-  end
-
-  def handle_call({:lock, module}, from, %{waiting: waiting} = state) do
-    case waiting do
-      %{^module => froms} ->
-        waiting = Map.put(state.waiting, module, [from | froms])
-        {:noreply, %{state | waiting: waiting}}
-
-      %{} ->
-        waiting = Map.put(state.waiting, module, [])
-        {:reply, true, %{state | waiting: waiting}}
-    end
-  end
-
-  def handle_call({:unlock, module}, _from, %{waiting: waiting} = state) do
-    froms = Map.fetch!(waiting, module)
-    Enum.each(froms, &:gen_server.reply(&1, false))
-    waiting = Map.delete(waiting, module)
-    {:reply, :ok, %{state | waiting: waiting}}
-  end
-
-  def handle_call(:get_ets, _from, %{ets: ets} = state) do
-    {:reply, ets, state}
-  end
-
-  def handle_cast(:start, state) do
-    {:noreply, spawn_checkers(state)}
-  end
-
-  def handle_info({__MODULE__, :done}, state) do
-    state = %{state | spawned: state.spawned - 1}
-
-    if state.spawned == 0 and state.modules == [] do
-      {:stop, :normal, state}
-    else
-      state = spawn_checkers(state)
-      {:noreply, state}
-    end
-  end
-
-  defp lock(server, module) do
-    :gen_server.call(server, {:lock, module}, :infinity)
-  end
-
-  defp unlock(server, module) do
-    :gen_server.call(server, {:unlock, module})
-  end
-
-  defp get_ets(server) do
-    :gen_server.call(server, :get_ets)
-  end
-
-  defp start(server) do
-    :gen_server.cast(server, :start)
-  end
-
-  defp preload_cache(ets, modules) do
-    Enum.each(modules, fn
-      {_module, map} when is_map(map) -> cache_from_module_map(ets, map)
-      {module, binary} when is_binary(binary) -> cache_from_chunk(ets, module, binary)
-    end)
-  end
-
-  defp spawn_checkers(%{modules: []} = state) do
-    state
-  end
-
-  defp spawn_checkers(%{spawned: spawned, schedulers: schedulers} = state)
-       when spawned >= schedulers do
-    state
-  end
-
-  defp spawn_checkers(%{modules: [{module, _} = verify | modules]} = state) do
-    parent = self()
-    ets = state.ets
-    send_results_pid = state.send_results
-
-    spawn_link(fn ->
-      warnings = check_module(verify, {parent, ets})
-      send(send_results_pid, {__MODULE__, module, warnings})
-      send(parent, {__MODULE__, :done})
-    end)
-
-    spawn_checkers(%{state | modules: modules, spawned: state.spawned + 1})
-  end
+  ## Cache
 
   defp cache_module({server, ets}, module) do
     if lock(server, module) do
@@ -313,14 +316,9 @@ defmodule Module.ParallelChecker do
   end
 
   defp cache_from_chunk(ets, module) do
-    case :code.get_object_code(module) do
-      {^module, binary, _filename} -> cache_from_chunk(ets, module, binary)
-      _other -> false
-    end
-  end
-
-  defp cache_from_chunk(ets, module, binary) do
-    with {:ok, {_, [{'ExCk', chunk}]}} <- :beam_lib.chunks(binary, ['ExCk']),
+    with {^module, binary, _filename} <- :code.get_object_code(module),
+         {:ok, ^module} <- binary |> :beam_lib.info() |> Keyword.fetch(:module),
+         {:ok, {_, [{'ExCk', chunk}]}} <- :beam_lib.chunks(binary, ['ExCk']),
          {:elixir_checker_v1, contents} <- :erlang.binary_to_term(chunk) do
       cache_chunk(ets, module, contents.exports)
       true
@@ -329,16 +327,6 @@ defmodule Module.ParallelChecker do
     end
   end
 
-  defp cache_from_module_map(ets, map) do
-    exports =
-      [{{:__info__, 1}, :def}] ++
-        behaviour_exports(map) ++
-        definitions_to_exports(map.definitions)
-
-    deprecated = Map.new(map.deprecated)
-    cache_info(ets, map.module, exports, deprecated, :elixir)
-  end
-
   defp cache_from_info(ets, module) do
     if Code.ensure_loaded?(module) do
       {mode, exports} = info_exports(module)
@@ -369,6 +357,23 @@ defmodule Module.ParallelChecker do
     _ -> %{}
   end
 
+  defp fetch_module_map!(binary, module) when is_binary(binary) do
+    {:ok, {_, [debug_info: chunk]}} = :beam_lib.chunks(binary, [:debug_info])
+    {:debug_info_v1, backend, data} = chunk
+    {:ok, module_map} = backend.debug_info(:elixir_v1, module, data, [])
+    module_map
+  end
+
+  defp cache_from_module_map(ets, map) do
+    exports =
+      [{{:__info__, 1}, :def}] ++
+        behaviour_exports(map) ++
+        definitions_to_exports(map.definitions)
+
+    deprecated = Map.new(map.deprecated)
+    cache_info(ets, map.module, exports, deprecated, :elixir)
+  end
+
   defp cache_info(ets, module, exports, deprecated, mode) do
     Enum.each(exports, fn {{fun, arity}, kind} ->
       reason = Map.get(deprecated, {fun, arity})
@@ -410,4 +415,96 @@ defmodule Module.ParallelChecker do
       end
     end)
   end
+
+  defp lock(server, module) do
+    :gen_server.call(server, {:lock, module}, :infinity)
+  end
+
+  defp unlock(server, module) do
+    :gen_server.call(server, {:unlock, module}, :infinity)
+  end
+
+  defp register(server, pid, ref) do
+    :gen_server.cast(server, {:register, pid, ref})
+  end
+
+  ## Server callbacks
+
+  def init(schedulers) do
+    ets = :ets.new(__MODULE__, [:set, :public, {:read_concurrency, true}])
+
+    state = %{
+      ets: ets,
+      waiting: %{},
+      modules: [],
+      spawned: 0,
+      schedulers: schedulers || max(:erlang.system_info(:schedulers_online), 2)
+    }
+
+    {:ok, state}
+  end
+
+  def handle_call(:start, _from, %{ets: ets, modules: modules} = state) do
+    for {pid, ref} <- modules do
+      send(pid, {ref, :cache, ets})
+    end
+
+    for {_pid, ref} <- modules do
+      receive do
+        {^ref, :cached} -> :ok
+      end
+    end
+
+    {:reply, modules, run_checkers(state)}
+  end
+
+  def handle_call(:ets, _from, state) do
+    {:reply, state.ets, state}
+  end
+
+  def handle_call({:lock, module}, from, %{waiting: waiting} = state) do
+    case waiting do
+      %{^module => froms} ->
+        waiting = Map.put(state.waiting, module, [from | froms])
+        {:noreply, %{state | waiting: waiting}}
+
+      %{} ->
+        waiting = Map.put(state.waiting, module, [])
+        {:reply, true, %{state | waiting: waiting}}
+    end
+  end
+
+  def handle_call({:unlock, module}, _from, %{waiting: waiting} = state) do
+    froms = Map.fetch!(waiting, module)
+    Enum.each(froms, &:gen_server.reply(&1, false))
+    waiting = Map.delete(waiting, module)
+    {:reply, :ok, %{state | waiting: waiting}}
+  end
+
+  def handle_info({__MODULE__, :done}, state) do
+    state = %{state | spawned: state.spawned - 1}
+    {:noreply, run_checkers(state)}
+  end
+
+  def handle_info({__MODULE__, :stop}, state) do
+    {:stop, :normal, state}
+  end
+
+  def handle_cast({:register, pid, ref}, %{modules: modules} = state) do
+    {:noreply, %{state | modules: [{pid, ref} | modules]}}
+  end
+
+  defp run_checkers(%{modules: []} = state) do
+    state
+  end
+
+  defp run_checkers(%{spawned: spawned, schedulers: schedulers} = state)
+       when spawned >= schedulers do
+    state
+  end
+
+  defp run_checkers(%{modules: [{pid, ref} | modules]} = state) do
+    send(pid, {ref, :check})
+    run_checkers(%{state | modules: modules, spawned: state.spawned + 1})
+  end
 end
