@@ -133,13 +133,24 @@ defmodule Module.Types do
   ## ERROR TO WARNING
 
   # Collect relevant information from context and traces to report error
+  def error_to_warning(:unable_apply, {mfa, args, expected, signature, stack}, context) do
+    {fun, arity} = context.function
+    line = get_meta(stack.last_expr)[:line]
+    location = {context.file, line, {context.module, fun, arity}}
+
+    traces = type_traces(stack, context)
+    {[signature | args], traces} = lift_all_types([signature | args], traces, context)
+    error = {:unable_apply, mfa, args, expected, signature, {location, stack.last_expr, traces}}
+    {Module.Types, error, location}
+  end
+
   def error_to_warning(:unable_unify, {left, right, stack}, context) do
     {fun, arity} = context.function
     line = get_meta(stack.last_expr)[:line]
     location = {context.file, line, {context.module, fun, arity}}
 
     traces = type_traces(stack, context)
-    {left, right, traces} = lift_all_types(left, right, traces, context)
+    {[left, right], traces} = lift_all_types([left, right], traces, context)
     error = {:unable_unify, left, right, {location, stack.last_expr, traces}}
     {Module.Types, error, location}
   end
@@ -183,58 +194,83 @@ defmodule Module.Types do
     end
   end
 
-  defp lift_all_types(left, right, traces, context) do
-    all_types = [left, right] ++ for({:type, _, type, _, _} <- traces, do: type)
-    [left, right | all_types] = Unify.lift_types(all_types, context)
+  defp lift_all_types(types, traces, context) do
+    trace_types = for({:type, _, type, _, _} <- traces, do: type)
+    {types, lift_context} = Unify.lift_types(types, context)
+    {trace_types, _lift_context} = Unify.lift_types(trace_types, lift_context)
 
     {traces, []} =
-      Enum.map_reduce(traces, all_types, fn
+      Enum.map_reduce(traces, trace_types, fn
         {:type, var, _, expr, location}, [type | acc] -> {{:type, var, type, expr, location}, acc}
         other, acc -> {other, acc}
       end)
 
-    {left, right, traces}
+    {types, traces}
   end
 
   ## FORMAT WARNINGS
 
+  def format_warning({:unable_apply, mfa, args, expected, signature, {location, expr, traces}}) do
+    {module, function, arity} = mfa
+    mfa_args = Macro.generate_arguments(arity, __MODULE__)
+    {module, function, ^arity} = call_to_mfa(erl_to_ex(module, function, mfa_args, []))
+    format_mfa = Exception.format_mfa(module, function, arity)
+    {traces, [] = _hints} = format_traces(traces, [], false)
+
+    clauses =
+      Enum.map(
+        signature,
+        &String.slice(IO.iodata_to_binary(Unify.format_type({:fun, [&1]}, false)), 1..-2)
+      )
+
+    [
+      "expected #{format_mfa} to have signature:\n\n    ",
+      Enum.map_join(args, ", ", &Unify.format_type(&1, false)),
+      " -> #{Unify.format_type(expected, false)}",
+      "\n\nbut it has signature:\n\n    ",
+      indent(Enum.join(clauses, "\n")),
+      "\n\n",
+      format_expr(expr, location),
+      traces,
+      "Conflict found at"
+    ]
+  end
+
   def format_warning({:unable_unify, left, right, {location, expr, traces}}) do
-    cond do
-      map_type?(left) and map_type?(right) and match?({:ok, _, _}, missing_field(left, right)) ->
-        {:ok, atom, known_atoms} = missing_field(left, right)
+    if map_type?(left) and map_type?(right) and match?({:ok, _, _}, missing_field(left, right)) do
+      {:ok, atom, known_atoms} = missing_field(left, right)
 
-        # Drop the last trace which is the expression map.foo
-        traces = Enum.drop(traces, 1)
-        {traces, hints} = format_traces(traces, true)
+      # Drop the last trace which is the expression map.foo
+      traces = Enum.drop(traces, 1)
+      {traces, hints} = format_traces(traces, [left, right], true)
 
-        [
-          "undefined field \"#{atom}\" ",
-          format_expr(expr, location),
-          "expected one of the following fields: ",
-          Enum.map_join(Enum.sort(known_atoms), ", ", & &1),
-          "\n\n",
-          traces,
-          format_message_hints(hints),
-          "Conflict found at"
-        ]
+      [
+        "undefined field \"#{atom}\" ",
+        format_expr(expr, location),
+        "expected one of the following fields: ",
+        Enum.map_join(Enum.sort(known_atoms), ", ", & &1),
+        "\n\n",
+        traces,
+        format_message_hints(hints),
+        "Conflict found at"
+      ]
+    else
+      simplify_left? = simplify_type?(left, right)
+      simplify_right? = simplify_type?(right, left)
 
-      true ->
-        simplify_left? = simplify_type?(left, right)
-        simplify_right? = simplify_type?(right, left)
+      {traces, hints} = format_traces(traces, [left, right], simplify_left? or simplify_right?)
 
-        {traces, hints} = format_traces(traces, simplify_left? or simplify_right?)
-
-        [
-          "incompatible types:\n\n    ",
-          Unify.format_type(left, simplify_left?),
-          " !~ ",
-          Unify.format_type(right, simplify_right?),
-          "\n\n",
-          format_expr(expr, location),
-          traces,
-          format_message_hints(hints),
-          "Conflict found at"
-        ]
+      [
+        "incompatible types:\n\n    ",
+        Unify.format_type(left, simplify_left?),
+        " !~ ",
+        Unify.format_type(right, simplify_right?),
+        "\n\n",
+        format_expr(expr, location),
+        traces,
+        format_message_hints(hints),
+        "Conflict found at"
+      ]
     end
   end
 
@@ -263,17 +299,17 @@ defmodule Module.Types do
     end
   end
 
-  defp format_traces([], _simplify?) do
+  defp format_traces([], _types, _simplify?) do
     {[], []}
   end
 
-  defp format_traces(traces, simplify?) do
+  defp format_traces(traces, types, simplify?) do
     traces
     |> Enum.uniq()
     |> Enum.reverse()
     |> Enum.map_reduce([], fn
       {:type, var, type, expr, location}, hints ->
-        {hint, hints} = format_type_hint(type, expr, hints)
+        {hint, hints} = format_type_hint(type, types, expr, hints)
 
         trace = [
           "where \"",
@@ -318,7 +354,7 @@ defmodule Module.Types do
   end
 
   defp simplify_type?(type, other) do
-    map_type?(type) and not map_type?(other)
+    map_like_type?(type) and not map_like_type?(other)
   end
 
   ## EXPRESSION FORMATTING
@@ -361,14 +397,16 @@ defmodule Module.Types do
   ## Hints
 
   defp format_message_hints(hints) do
-    hints |> Enum.uniq() |> Enum.reverse() |> Enum.map(&format_message_hint/1)
+    hints
+    |> Enum.uniq()
+    |> Enum.reverse()
+    |> Enum.map(&[format_message_hint(&1), "\n"])
   end
 
   defp format_message_hint(:inferred_dot) do
     """
     HINT: "var.field" (without parentheses) implies "var" is a map() while \
     "var.fun()" (with parentheses) implies "var" is an atom()
-
     """
   end
 
@@ -378,18 +416,24 @@ defmodule Module.Types do
     integer() unless said otherwise. For example, <<expr>> assumes "expr" \
     is an integer. Pass a modifier, such as <<expr::float>> or <<expr::binary>>, \
     to change the default behaviour.
-
     """
   end
 
-  defp format_type_hint(type, expr, hints) do
-    case format_type_hint(type, expr) do
+  defp format_message_hint({:sized_and_unsize_tuples, {size, var}}) do
+    """
+    HINT: use pattern matching or "is_tuple(#{Macro.to_string(var)}) and \
+    tuple_size(#{Macro.to_string(var)}) == #{size}" to guard a sized tuple.
+    """
+  end
+
+  defp format_type_hint(type, types, expr, hints) do
+    case format_type_hint(type, types, expr) do
       {message, hint} -> {message, [hint | hints]}
       :error -> {[], hints}
     end
   end
 
-  defp format_type_hint(type, expr) do
+  defp format_type_hint(type, types, expr) do
     cond do
       dynamic_map_dot?(type, expr) ->
         {" (due to calling var.field)", :inferred_dot}
@@ -400,6 +444,9 @@ defmodule Module.Types do
       inferred_bitstring_spec?(type, expr) ->
         {[], :inferred_bitstring_spec}
 
+      message = sized_and_unsize_tuples(expr, types) ->
+        {[], {:sized_and_unsize_tuples, message}}
+
       true ->
         :error
     end
@@ -435,15 +482,33 @@ defmodule Module.Types do
     end
   end
 
+  defp sized_and_unsize_tuples({{:., _, [:erlang, :is_tuple]}, _, [var]}, types) do
+    case Enum.find(types, &match?({:tuple, _, _}, &1)) do
+      {:tuple, size, _} ->
+        {size, var}
+
+      nil ->
+        nil
+    end
+  end
+
+  defp sized_and_unsize_tuples(_expr, _types) do
+    nil
+  end
+
   ## Formatting helpers
 
   defp indent(string) do
-    String.replace(string, "\n", "    \n")
+    String.replace(string, "\n", "\n    ")
   end
 
   defp map_type?({:map, _}), do: true
   defp map_type?(_other), do: false
 
+  defp map_like_type?({:map, _}), do: true
+  defp map_like_type?({:union, union}), do: Enum.any?(union, &map_like_type?/1)
+  defp map_like_type?(_other), do: false
+
   defp atom_type?(:atom), do: true
   defp atom_type?({:atom, _}), do: false
   defp atom_type?({:union, union}), do: Enum.all?(union, &atom_type?/1)
@@ -451,4 +516,7 @@ defmodule Module.Types do
 
   defp integer_type?(:integer), do: true
   defp integer_type?(_other), do: false
+
+  defp call_to_mfa({{:., _, [mod, fun]}, _, args}), do: {mod, fun, length(args)}
+  defp call_to_mfa({fun, _, args}) when is_atom(fun), do: {Kernel, fun, length(args)}
 end
