@@ -151,7 +151,7 @@ defmodule Module.Types.Expr do
          dynamic_value_pairs =
            Enum.map(arg_pairs, fn {:required, key, _value} -> {:required, key, :dynamic} end),
          args_type = {:map, dynamic_value_pairs ++ [{:optional, :dynamic, :dynamic}]},
-         {:ok, type, context} <- unify(args_type, map_type, stack, context) do
+         {:ok, type, context} <- unify(map_type, args_type, stack, context) do
       # Retrieve map type and overwrite with the new value types from the map update
       {:map, pairs} = resolve_var(type, context)
 
@@ -216,6 +216,30 @@ defmodule Module.Types.Expr do
     end
   end
 
+  # cond do pat -> expr end
+  def of_expr({:cond, _meta, [[{:do, clauses}]]} = expr, _expected, stack, context) do
+    stack = push_expr_stack(expr, stack)
+
+    {result, context} =
+      reduce_ok(clauses, context, fn {:->, meta, [head, body]}, context = acc ->
+        case of_expr(head, :dynamic, stack, context) do
+          {:ok, _, context} ->
+            with {:ok, _expr_type, context} <- of_expr(body, :dynamic, stack, context) do
+              {:ok, keep_warnings(acc, context)}
+            end
+
+          error ->
+            # Skip the clause if it the head has an error
+            if meta[:generated], do: {:ok, acc}, else: error
+        end
+      end)
+
+    case result do
+      :ok -> {:ok, :dynamic, context}
+      :error -> {:error, context}
+    end
+  end
+
   # case expr do pat -> expr end
   def of_expr({:case, _meta, [case_expr, [{:do, clauses}]]} = expr, _expected, stack, context) do
     stack = push_expr_stack(expr, stack)
@@ -301,7 +325,7 @@ defmodule Module.Types.Expr do
   end
 
   # for pat <- expr do expr end
-  def of_expr({:for, _meta, args} = expr, _expected, stack, context) do
+  def of_expr({:for, _meta, [_ | _] = args} = expr, _expected, stack, context) do
     stack = push_expr_stack(expr, stack)
     {clauses, [[{:do, block} | opts]]} = Enum.split(args, -1)
 
@@ -320,7 +344,7 @@ defmodule Module.Types.Expr do
   end
 
   # with pat <- expr do expr end
-  def of_expr({:with, _meta, clauses} = expr, _expected, stack, context) do
+  def of_expr({:with, _meta, [_ | _] = clauses} = expr, _expected, stack, context) do
     stack = push_expr_stack(expr, stack)
 
     case reduce_ok(clauses, context, &with_clause(&1, stack, &2)) do
@@ -334,15 +358,10 @@ defmodule Module.Types.Expr do
     # TODO: Use expected type to infer intersection return type
     stack = push_expr_stack(expr, stack)
 
-    case of_expr(fun, :dynamic, stack, context) do
-      {:ok, _fun_type, context} ->
-        case map_reduce_ok(args, context, &of_expr(&1, :dynamic, stack, &2)) do
-          {:ok, _arg_types, context} -> {:ok, :dynamic, context}
-          {:error, reason} -> {:error, reason}
-        end
-
-      {:error, reason} ->
-        {:error, reason}
+    with {:ok, _fun_type, context} <- of_expr(fun, :dynamic, stack, context),
+         {:ok, _arg_types, context} <-
+           map_reduce_ok(args, context, &of_expr(&1, :dynamic, stack, &2)) do
+      {:ok, :dynamic, context}
     end
   end
 
@@ -375,11 +394,10 @@ defmodule Module.Types.Expr do
     stack = push_expr_stack(expr2, stack)
 
     with {:ok, _expr_type, context} <- of_expr(expr1, :dynamic, stack, context),
-         {:ok, _fun_type, context} <- of_expr(fun, :dynamic, stack, context) do
-      case map_reduce_ok(args, context, &of_expr(&1, :dynamic, stack, &2)) do
-        {:ok, _arg_types, context} -> {:ok, :dynamic, context}
-        {:error, reason} -> {:error, reason}
-      end
+         {:ok, _fun_type, context} <- of_expr(fun, :dynamic, stack, context),
+         {:ok, _arg_types, context} <-
+           map_reduce_ok(args, context, &of_expr(&1, :dynamic, stack, &2)) do
+      {:ok, :dynamic, context}
     end
   end
 
@@ -475,12 +493,19 @@ defmodule Module.Types.Expr do
   end
 
   defp of_clauses(clauses, stack, context) do
-    reduce_ok(clauses, context, fn {:->, _meta, [head, body]}, context = acc ->
+    reduce_ok(clauses, context, fn {:->, meta, [head, body]}, context = acc ->
       {patterns, guards} = extract_head(head)
 
-      with {:ok, _, context} <- Pattern.of_head(patterns, guards, stack, context),
-           {:ok, _expr_type, context} <- of_expr(body, :dynamic, stack, context),
-           do: {:ok, keep_warnings(acc, context)}
+      case Pattern.of_head(patterns, guards, stack, context) do
+        {:ok, _, context} ->
+          with {:ok, _expr_type, context} <- of_expr(body, :dynamic, stack, context) do
+            {:ok, keep_warnings(acc, context)}
+          end
+
+        error ->
+          # Skip the clause if it the head has an error
+          if meta[:generated], do: {:ok, acc}, else: error
+      end
     end)
   end
 
