@@ -29,7 +29,12 @@ defmodule OptionParser do
   @type argv :: [String.t()]
   @type parsed :: keyword
   @type errors :: [{String.t(), String.t() | nil}]
-  @type options :: [switches: keyword, strict: keyword, aliases: keyword]
+  @type options :: [
+          switches: keyword,
+          strict: keyword,
+          aliases: keyword,
+          allow_nonexistent_atoms: boolean
+        ]
 
   defmodule ParseError do
     defexception [:message]
@@ -487,7 +492,7 @@ defmodule OptionParser do
   Keys must be atoms. Keys with `nil` value are discarded,
   boolean values are converted to `--key` or `--no-key`
   (if the value is `true` or `false`, respectively),
-  and all other values are converted using `Kernel.to_string/1`.
+  and all other values are converted using `to_string/1`.
 
   It is advised to pass to `to_argv/2` the same set of `options`
   given to `parse/2`. Some switches can only be reconstructed
@@ -596,9 +601,11 @@ defmodule OptionParser do
           raise ArgumentError, ":switches and :strict cannot be given together"
 
         switches = opts[:switches] ->
+          validate_switches(switches)
           {switches, false}
 
         strict = opts[:strict] ->
+          validate_switches(strict)
           {strict, true}
 
         true ->
@@ -614,6 +621,20 @@ defmodule OptionParser do
     }
   end
 
+  defp validate_switches(switches) do
+    Enum.map(switches, &validate_switch/1)
+  end
+
+  defp validate_switch({_name, type_or_type_and_modifiers}) do
+    valid = [:boolean, :count, :integer, :float, :string, :keep]
+    invalid = List.wrap(type_or_type_and_modifiers) -- valid
+
+    if invalid != [] do
+      raise ArgumentError,
+            "invalid switch types/modifiers: " <> Enum.map_join(invalid, ", ", &inspect/1)
+    end
+  end
+
   defp validate_option(value, kinds) do
     {invalid?, value} =
       cond do
@@ -812,7 +833,15 @@ defmodule OptionParser do
     if type = get_type(option, opts, types) do
       "#{option} : Missing argument of type #{type}"
     else
-      "#{option} : Unknown option"
+      msg = "#{option} : Unknown option"
+
+      case did_you_mean(option, types) do
+        {similar, score} when score > 0.8 ->
+          msg <> ". Did you mean --#{similar}?"
+
+        _ ->
+          msg
+      end
     end
   end
 
@@ -831,4 +860,17 @@ defmodule OptionParser do
       types[key]
     end
   end
+
+  defp did_you_mean(option, types) do
+    key = option |> String.trim_leading("-") |> String.replace("-", "_")
+    Enum.reduce(types, {nil, 0}, &max_similar(&1, key, &2))
+  end
+
+  defp max_similar({source, _}, target, {_, current} = best) do
+    source = Atom.to_string(source)
+
+    score = String.jaro_distance(source, target)
+    option = String.replace(source, "_", "-")
+    if score < current, do: best, else: {option, score}
+  end
 end
