@@ -24,12 +24,12 @@ defmodule Port do
   receives data from multiple inputs and concatenates them in the output.
 
   After the port was created, we sent it two commands in the form of
-  messages using `Kernel.send/2`. The first command has the binary payload
+  messages using `send/2`. The first command has the binary payload
   of "hello" and the second has "world".
 
   After sending those two messages, we invoked the IEx helper `flush()`,
   which printed all messages received from the port, in this case we got
-  "hello" and "world" back. Notice the messages are in binary because we
+  "hello" and "world" back. Note that the messages are in binary because we
   passed the `:binary` option when opening the port in `Port.open/2`. Without
   such option, it would have yielded a list of bytes.
 
@@ -180,7 +180,7 @@ defmodule Port do
   @type name ::
           {:spawn, charlist | binary}
           | {:spawn_driver, charlist | binary}
-          | {:spawn_executable, charlist | atom}
+          | {:spawn_executable, :file.name_all()}
           | {:fd, non_neg_integer, non_neg_integer}
 
   @doc """
@@ -222,7 +222,7 @@ defmodule Port do
   @doc """
   Sends `data` to the port driver `port`.
 
-  For more information, see `:erlang.port_command/2`.
+  For more information, see `:erlang.port_command/3`.
 
   Inlined by the compiler.
   """
@@ -250,7 +250,7 @@ defmodule Port do
   """
   @spec info(port) :: keyword | nil
   def info(port) do
-    nillify(:erlang.port_info(port))
+    nilify(:erlang.port_info(port))
   end
 
   @doc """
@@ -270,7 +270,7 @@ defmodule Port do
   end
 
   def info(port, item) do
-    nillify(:erlang.port_info(port, item))
+    nilify(:erlang.port_info(port, item))
   end
 
   @doc """
@@ -323,7 +323,7 @@ defmodule Port do
     :erlang.ports()
   end
 
-  @compile {:inline, nillify: 1}
-  defp nillify(:undefined), do: nil
-  defp nillify(other), do: other
+  @compile {:inline, nilify: 1}
+  defp nilify(:undefined), do: nil
+  defp nilify(other), do: other
 end
