@@ -116,7 +116,7 @@ defmodule Process do
   """
   @spec put(term, term) :: term | nil
   def put(key, value) do
-    nillify(:erlang.put(key, value))
+    nilify(:erlang.put(key, value))
   end
 
   @doc """
@@ -136,7 +136,7 @@ defmodule Process do
   """
   @spec delete(term) :: term | nil
   def delete(key) do
-    nillify(:erlang.erase(key))
+    nilify(:erlang.erase(key))
   end
 
   @doc """
@@ -383,12 +383,19 @@ defmodule Process do
   @type spawn_opt ::
           :link
           | :monitor
+          | {:monitor, monitor_option()}
           | {:priority, :low | :normal | :high}
           | {:fullsweep_after, non_neg_integer}
           | {:min_heap_size, non_neg_integer}
           | {:min_bin_vheap_size, non_neg_integer}
+          | {:max_heap_size, heap_size}
+          | {:message_queue_data, :off_heap | :on_heap}
   @type spawn_opts :: [spawn_opt]
 
+  # TODO: Use :erlang.monitor_option() on Erlang/OTP 24+
+  @typep monitor_option ::
+           [alias: :explicit_unalias | :demonitor | :reply_demonitor, tag: term()]
+
   @doc """
   Spawns the given function according to the given options.
 
@@ -643,7 +650,7 @@ defmodule Process do
   """
   @spec whereis(atom) :: pid | port | nil
   def whereis(name) do
-    nillify(:erlang.whereis(name))
+    nilify(:erlang.whereis(name))
   end
 
   @doc """
@@ -742,7 +749,7 @@ defmodule Process do
   """
   @spec info(pid) :: keyword | nil
   def info(pid) do
-    nillify(:erlang.process_info(pid))
+    nilify(:erlang.process_info(pid))
   end
 
   @doc """
@@ -763,7 +770,7 @@ defmodule Process do
   end
 
   def info(pid, spec) when is_atom(spec) or is_list(spec) do
-    nillify(:erlang.process_info(pid, spec))
+    nilify(:erlang.process_info(pid, spec))
   end
 
   @doc """
@@ -781,7 +788,7 @@ defmodule Process do
   @spec hibernate(module, atom, list) :: no_return
   defdelegate hibernate(mod, fun_name, args), to: :erlang
 
-  @compile {:inline, nillify: 1}
-  defp nillify(:undefined), do: nil
-  defp nillify(other), do: other
+  @compile {:inline, nilify: 1}
+  defp nilify(:undefined), do: nil
+  defp nilify(other), do: other
 end
