@@ -244,8 +244,16 @@ defmodule Protocol do
         ...
       end
 
-  Although doing so is not recommended as it may affect your test suite
-  performance.
+  If you are using `Mix.install/2`, you can do by passing the `consolidate_protocols`
+  option:
+
+      Mix.install(
+        deps,
+        consolidate_protocols: false
+      )
+
+  Although doing so is not recommended as it may affect the performance of
+  your code.
 
   Finally, note all protocols are compiled with `debug_info` set to `true`,
   regardless of the option set by the `elixirc` compiler. The debug info is
@@ -275,7 +283,7 @@ defmodule Protocol do
       name = unquote(name)
       arity = unquote(arity)
 
-      @functions [{name, arity} | @functions]
+      @__functions__ [{name, arity} | @__functions__]
 
       # Generate a fake definition with the user
       # signature that will be used by docs
@@ -548,6 +556,9 @@ defmodule Protocol do
           | {:error, :not_a_protocol}
           | {:error, :no_beam_info}
   def consolidate(protocol, types) when is_atom(protocol) do
+    # Ensure the types are sorted so the compiled beam is deterministic
+    types = Enum.sort(types)
+
     with {:ok, ast_info, specs, compile_info} <- beam_protocol(protocol),
          {:ok, definitions} <- change_debug_info(protocol, ast_info, types),
          do: compile(definitions, specs, compile_info)
@@ -681,6 +692,8 @@ defmodule Protocol do
   def __protocol__(name, do: block) do
     quote do
       defmodule unquote(name) do
+        @before_compile Protocol
+
         # We don't allow function definition inside protocols
         import Kernel,
           except: [
@@ -688,6 +701,7 @@ defmodule Protocol do
             def: 2,
             defp: 1,
             defp: 2,
+            defdelegate: 2,
             defguard: 1,
             defguardp: 1,
             defmacro: 1,
@@ -703,7 +717,7 @@ defmodule Protocol do
         @compile :debug_info
 
         # Set up a clear slate to store defined functions
-        @functions []
+        @__functions__ []
         @fallback_to_any false
 
         # Invoke the user given block
@@ -715,6 +729,92 @@ defmodule Protocol do
     end
   end
 
+  defp callback_ast_to_fa({kind, {:"::", meta, [{name, _, args}, _return]}, _pos})
+       when kind in [:callback, :macrocallback] do
+    [{{name, length(List.wrap(args))}, meta}]
+  end
+
+  defp callback_ast_to_fa(
+         {kind, {:when, _, [{:"::", meta, [{name, _, args}, _return]}, _vars]}, _pos}
+       )
+       when kind in [:callback, :macrocallback] do
+    [{{name, length(List.wrap(args))}, meta}]
+  end
+
+  defp callback_ast_to_fa({kind, _, _pos}) when kind in [:callback, :macrocallback] do
+    []
+  end
+
+  defp callback_metas(module, kind)
+       when kind in [:callback, :macrocallback] do
+    :lists.flatmap(&callback_ast_to_fa/1, Module.get_attribute(module, kind))
+    |> :maps.from_list()
+  end
+
+  defp get_callback_line(fa, metas),
+    do: :maps.get(fa, metas, [])[:line]
+
+  defp warn(message, env, nil) do
+    IO.warn(message, env)
+  end
+
+  defp warn(message, env, line) when is_integer(line) do
+    stacktrace = :maps.update(:line, line, env) |> Macro.Env.stacktrace()
+    IO.warn(message, stacktrace)
+  end
+
+  def __before_compile__(env) do
+    callback_metas = callback_metas(env.module, :callback)
+    callbacks = :maps.keys(callback_metas)
+    functions = Module.get_attribute(env.module, :__functions__)
+
+    if functions == [] do
+      warn(
+        "protocols must define at least one function, but none was defined",
+        env,
+        nil
+      )
+    end
+
+    # TODO: Convert the following warnings into errors in future Elixir versions
+    :lists.map(
+      fn {name, arity} = fa ->
+        warn(
+          "cannot define @callback #{name}/#{arity} inside protocol, use def/1 to outline your protocol definition",
+          env,
+          get_callback_line(fa, callback_metas)
+        )
+      end,
+      callbacks -- functions
+    )
+
+    # Macro Callbacks
+    macrocallback_metas = callback_metas(env.module, :macrocallback)
+    macrocallbacks = :maps.keys(macrocallback_metas)
+
+    :lists.map(
+      fn {name, arity} = fa ->
+        warn(
+          "cannot define @macrocallback #{name}/#{arity} inside protocol, use def/1 to outline your protocol definition",
+          env,
+          get_callback_line(fa, macrocallback_metas)
+        )
+      end,
+      macrocallbacks
+    )
+
+    # Optional Callbacks
+    optional_callbacks = Module.get_attribute(env.module, :optional_callbacks)
+
+    if optional_callbacks != [] do
+      warn(
+        "cannot define @optional_callbacks inside protocol, all of the protocol definitions are required",
+        env,
+        nil
+      )
+    end
+  end
+
   defp after_defprotocol do
     quote bind_quoted: [built_in: __built_in__()] do
       any_impl_for =
@@ -805,11 +905,11 @@ defmodule Protocol do
 
       @doc false
       @spec __protocol__(:module) :: __MODULE__
-      @spec __protocol__(:functions) :: unquote(Protocol.__functions_spec__(@functions))
+      @spec __protocol__(:functions) :: unquote(Protocol.__functions_spec__(@__functions__))
       @spec __protocol__(:consolidated?) :: boolean
       @spec __protocol__(:impls) :: :not_consolidated | {:consolidated, [module]}
       Kernel.def(__protocol__(:module), do: __MODULE__)
-      Kernel.def(__protocol__(:functions), do: unquote(:lists.sort(@functions)))
+      Kernel.def(__protocol__(:functions), do: unquote(:lists.sort(@__functions__)))
       Kernel.def(__protocol__(:consolidated?), do: false)
       Kernel.def(__protocol__(:impls), do: :not_consolidated)
     end
@@ -822,26 +922,40 @@ defmodule Protocol do
     do: [:lists.foldl(&{:|, [], [&1, &2]}, head, tail), quote(do: ...)]
 
   @doc false
-  def __impl__(protocol, opts) do
-    do_defimpl(protocol, :lists.keysort(1, opts))
+  def __impl__(protocol, opts, do_block, env) do
+    opts = Keyword.merge(opts, do_block)
+
+    {for, opts} =
+      Keyword.pop_lazy(opts, :for, fn ->
+        env.module ||
+          raise ArgumentError, "defimpl/3 expects a :for option when declared outside a module"
+      end)
+
+    for = Macro.expand_literal(for, %{env | module: Kernel, function: {:defimpl, 3}})
+
+    case opts do
+      [] -> raise ArgumentError, "defimpl expects a do-end block"
+      [do: block] -> __impl__(protocol, for, block)
+      _ -> raise ArgumentError, "unknown options given to defimpl, got: #{Macro.to_string(opts)}"
+    end
   end
 
-  defp do_defimpl(protocol, do: block, for: for) when is_list(for) do
-    for f <- for, do: do_defimpl(protocol, do: block, for: f)
+  defp __impl__(protocol, for, block) when is_list(for) do
+    for f <- for, do: __impl__(protocol, f, block)
   end
 
-  defp do_defimpl(protocol, do: block, for: for) do
+  defp __impl__(protocol, for, block) do
     # Unquote the implementation just later
     # when all variables will already be injected
     # into the module body.
     impl =
       quote unquote: false do
         @doc false
-        @spec __impl__(:for) :: unquote(for)
         @spec __impl__(:target) :: __MODULE__
+        @spec __impl__(:for) :: unquote(for)
         @spec __impl__(:protocol) :: unquote(protocol)
-        def __impl__(:for), do: unquote(for)
         def __impl__(:target), do: __MODULE__
+        def __impl__(:for), do: unquote(for)
         def __impl__(:protocol), do: unquote(protocol)
       end
 
@@ -858,12 +972,12 @@ defmodule Protocol do
         @protocol protocol
         @for for
 
-        unquote(block)
-
+        res = unquote(block)
         Module.register_attribute(__MODULE__, :__impl__, persist: true)
         @__impl__ [protocol: @protocol, for: @for]
 
         unquote(impl)
+        res
       end
     end
   end
@@ -921,14 +1035,15 @@ defmodule Protocol do
 
   @doc false
   def __ensure_defimpl__(protocol, for, env) do
-    if Protocol.consolidated?(protocol) do
+    if not Code.get_compiler_option(:ignore_already_consolidated) and
+         Protocol.consolidated?(protocol) do
       message =
         "the #{inspect(protocol)} protocol has already been consolidated, an " <>
           "implementation for #{inspect(for)} has no effect. If you want to " <>
           "implement protocols after compilation or during tests, check the " <>
           "\"Consolidation\" section in the Protocol module documentation"
 
-      IO.warn(message, Macro.Env.stacktrace(env))
+      IO.warn(message, env)
     end
 
     :ok
