@@ -47,9 +47,9 @@ defmodule Registry do
 
       {:ok, _} = Registry.start_link(keys: :unique, name: Registry.ViaTest)
       name = {:via, Registry, {Registry.ViaTest, "agent", :hello}}
-      {:ok, _} = Agent.start_link(fn -> 0 end, name: name)
+      {:ok, agent_pid} = Agent.start_link(fn -> 0 end, name: name)
       Registry.lookup(Registry.ViaTest, "agent")
-      #=> [{self(), :hello}]
+      #=> [{agent_pid, :hello}]
 
   To this point, we have been starting `Registry` using `start_link/1`.
   Typically the registry is started as part of a supervision tree though:
@@ -165,7 +165,7 @@ defmodule Registry do
   the value from the registry and sending it a message. Many parts of the standard
   library are designed to cope with that, such as `Process.monitor/1` which will
   deliver the `:DOWN` message immediately if the monitored process is already dead
-  and `Kernel.send/2` which acts as a no-op for dead processes.
+  and `send/2` which acts as a no-op for dead processes.
 
   ## ETS
 
@@ -260,6 +260,10 @@ defmodule Registry do
     end
   end
 
+  def send({registry, key, _value}, msg) do
+    Registry.send({registry, key}, msg)
+  end
+
   @doc false
   def unregister_name({registry, key}), do: unregister(registry, key)
   def unregister_name({registry, key, _value}), do: unregister(registry, key)
@@ -1061,7 +1065,9 @@ defmodule Registry do
       :ets.lookup(registry, key)
     catch
       :error, :badarg ->
-        raise ArgumentError, "unknown registry: #{inspect(registry)}"
+        raise ArgumentError,
+              "unknown registry: #{inspect(registry)}. Either the registry name is invalid or " <>
+                "the registry is not running, possibly because its application isn't started"
     else
       [{^key, value}] -> {:ok, value}
       _ -> :error
@@ -1272,7 +1278,7 @@ defmodule Registry do
 
   ## Examples
 
-  This example shows how to get everything from the registry.
+  This example shows how to get everything from the registry:
 
       iex> Registry.start_link(keys: :unique, name: Registry.SelectAllTest)
       iex> {:ok, _} = Registry.register(Registry.SelectAllTest, "hello", :value)
@@ -1280,7 +1286,7 @@ defmodule Registry do
       iex> Registry.select(Registry.SelectAllTest, [{{:"$1", :"$2", :"$3"}, [], [{{:"$1", :"$2", :"$3"}}]}])
       [{"world", self(), :value}, {"hello", self(), :value}]
 
-  Get all keys in the registry.
+  Get all keys in the registry:
 
       iex> Registry.start_link(keys: :unique, name: Registry.SelectAllTest)
       iex> {:ok, _} = Registry.register(Registry.SelectAllTest, "hello", :value)
@@ -1293,17 +1299,7 @@ defmodule Registry do
   @spec select(registry, spec) :: [term]
   def select(registry, spec)
       when is_atom(registry) and is_list(spec) do
-    spec =
-      for part <- spec do
-        case part do
-          {{key, pid, value}, guards, select} ->
-            {{key, {pid, value}}, guards, select}
-
-          _ ->
-            raise ArgumentError,
-                  "invalid match specification in Registry.select/2: #{inspect(spec)}"
-        end
-      end
+    spec = group_match_headers(spec, __ENV__.function)
 
     case key_info!(registry) do
       {_kind, partitions, nil} ->
@@ -1316,6 +1312,51 @@ defmodule Registry do
     end
   end
 
+  @doc """
+  Works like `select/2`, but only returns the number of matching records.
+
+  ## Examples
+
+  In the example below we register the current process under different
+  keys in a unique registry but with the same value:
+
+      iex> Registry.start_link(keys: :unique, name: Registry.CountSelectTest)
+      iex> {:ok, _} = Registry.register(Registry.CountSelectTest, "hello", :value)
+      iex> {:ok, _} = Registry.register(Registry.CountSelectTest, "world", :value)
+      iex> Registry.count_select(Registry.CountSelectTest, [{{:_, :_, :value}, [], [true]}])
+      2
+  """
+  @doc since: "1.14.0"
+  @spec count_select(registry, spec) :: non_neg_integer()
+  def count_select(registry, spec)
+      when is_atom(registry) and is_list(spec) do
+    spec = group_match_headers(spec, __ENV__.function)
+
+    case key_info!(registry) do
+      {_kind, partitions, nil} ->
+        Enum.reduce(0..(partitions - 1), 0, fn partition_index, acc ->
+          count = :ets.select_count(key_ets!(registry, partition_index), spec)
+          acc + count
+        end)
+
+      {_kind, 1, key_ets} ->
+        :ets.select_count(key_ets, spec)
+    end
+  end
+
+  defp group_match_headers(spec, {fun, arity}) do
+    for part <- spec do
+      case part do
+        {{key, pid, value}, guards, select} ->
+          {{key, {pid, value}}, guards, select}
+
+        _ ->
+          raise ArgumentError,
+                "invalid match specification in Registry.#{fun}/#{arity}: #{inspect(spec)}"
+      end
+    end
+  end
+
   ## Helpers
 
   @compile {:inline, hash: 2}
