@@ -43,7 +43,7 @@ defmodule System do
       system time may not match in case of time warps although the VM works towards
       aligning them. This time is not monotonic (i.e., it may decrease)
       as its behaviour is configured [by the VM time warp
-      mode](https://erlang.org/doc/apps/erts/time_correction.html#Time_Warp_Modes);
+      mode](https://www.erlang.org/doc/apps/erts/time_correction.html#Time_Warp_Modes);
 
     * `monotonic_time/0` - a monotonically increasing time provided
       by the Erlang VM.
@@ -58,10 +58,19 @@ defmodule System do
 
   For a more complete rundown on the VM support for different
   times, see the [chapter on time and time
-  correction](https://erlang.org/doc/apps/erts/time_correction.html)
+  correction](https://www.erlang.org/doc/apps/erts/time_correction.html)
   in the Erlang docs.
   """
 
+  defmodule EnvError do
+    defexception [:env]
+
+    @impl true
+    def message(%{env: env}) do
+      "could not fetch environment variable #{inspect(env)} because it is not set"
+    end
+  end
+
   @typedoc """
   The time unit to be passed to functions like `monotonic_time/1` and others.
 
@@ -463,14 +472,14 @@ defmodule System do
   @doc """
   Traps the given `signal` to execute the `fun`.
 
-  **Important**: Trapping signals may have strong implications
-  on how a system shuts down and behave in production and
-  therefore it is extremely discouraged for libraries to
-  set their own traps. Instead, they should redirect users
-  to configure them themselves. The only cases where it is
-  acceptable for libraries to set their own traps is when
-  using Elixir in script mode, such as in `.exs` files and
-  via Mix tasks.
+  > **Important**: Trapping signals may have strong implications
+  > on how a system shuts down and behave in production and
+  > therefore it is extremely discouraged for libraries to
+  > set their own traps. Instead, they should redirect users
+  > to configure them themselves. The only cases where it is
+  > acceptable for libraries to set their own traps is when
+  > using Elixir in script mode, such as in `.exs` files and
+  > via Mix tasks.
 
   An optional `id` that uniquely identifies the function
   can be given, otherwise a unique one is automatically
@@ -606,7 +615,6 @@ defmodule System do
   The returned value is a map containing name-value pairs.
   Variable names and their values are strings.
   """
-  # TODO: Remove this once we require Erlang/OTP 24+
   @spec get_env() :: %{optional(String.t()) => String.t()}
   def get_env do
     if function_exported?(:os, :env, 0) do
@@ -695,9 +703,7 @@ defmodule System do
   @doc since: "1.9.0"
   @spec fetch_env!(String.t()) :: String.t()
   def fetch_env!(varname) when is_binary(varname) do
-    get_env(varname) ||
-      raise ArgumentError,
-            "could not fetch environment variable #{inspect(varname)} because it is not set"
+    get_env(varname) || raise(EnvError, env: varname)
   end
 
   @doc """
@@ -708,8 +714,7 @@ defmodule System do
 
   For more information, see `:os.getpid/0`.
   """
-  # TODO: deprecate permanently on v1.13
-  @doc deprecated: "Use System.pid/0 instead"
+  @deprecated "Use System.pid/0 instead"
   @spec get_pid() :: binary
   def get_pid, do: IO.iodata_to_binary(:os.getpid())
 
@@ -735,11 +740,28 @@ defmodule System do
   Sets multiple environment variables.
 
   Sets a new value for each environment variable corresponding
-  to each `{key, value}` pair in `enum`.
+  to each `{key, value}` pair in `enum`. Keys are automatically
+  converted to strings, values are sent as is. `nil` values erase
+  the given keys.
   """
   @spec put_env(Enumerable.t()) :: :ok
   def put_env(enum) do
-    Enum.each(enum, fn {key, val} -> put_env(key, val) end)
+    Enum.each(enum, fn
+      {key, nil} ->
+        :os.unsetenv(to_charlist(key))
+
+      {key, val} ->
+        key = to_charlist(key)
+
+        case :string.find(key, "=") do
+          :nomatch ->
+            :os.putenv(key, to_charlist(val))
+
+          _ ->
+            raise ArgumentError,
+                  "cannot execute System.put_env/1 for key with \"=\", got: #{inspect(key)}"
+        end
+    end)
   end
 
   @doc """
@@ -756,23 +778,11 @@ defmodule System do
   @doc """
   Deprecated mechanism to retrieve the last exception stacktrace.
 
-  Starting from Erlang/OTP 23, this function will always return an
-  empty list.
+  It always return an empty list.
   """
-  # TODO: Remove conditional on Erlang/OTP 23+.
-  # Note Elixir may be compiled in an earlier Erlang version but runs on a
-  # newer one, so we need the check at compilation time and runtime.
   @deprecated "Use __STACKTRACE__ instead"
-  if function_exported?(:erlang, :get_stacktrace, 0) do
-    def stacktrace do
-      if function_exported?(:erlang, :get_stacktrace, 0) do
-        apply(:erlang, :get_stacktrace, [])
-      else
-        []
-      end
-    end
-  else
-    def stacktrace, do: []
+  def stacktrace do
+    []
   end
 
   @doc """
@@ -857,8 +867,13 @@ defmodule System do
   All applications are taken down smoothly, all code is unloaded, and all ports
   are closed before the system terminates by calling `halt/1`.
 
-  `status` must be a non-negative integer value which is returned by the
-  runtime system to the operating system.
+  `status` must be a non-negative integer or a binary.
+
+    * If an integer, the runtime system exits with the integer value which is
+      returned to the operating system.
+
+    * If a binary, an Erlang crash dump is produced with status as slogan, and
+      then the runtime system exits with status code 1.
 
   Note that on many platforms, only the status codes 0-255 are supported
   by the operating system.
@@ -886,18 +901,18 @@ defmodule System do
 
   It uses `sh` for Unix-like systems and `cmd` for Windows.
 
-  **Important**: Use this function with care. In particular, **never
-  pass untrusted user input to this function**, as the user would be
-  able to perform "command injection attacks" by executing any code
-  directly on the machine. Generally speaking, prefer to use `cmd/3`
-  over this function.
+  > **Important:**: Use this function with care. In particular, **never
+  > pass untrusted user input to this function**, as the user would be
+  > able to perform "command injection attacks" by executing any code
+  > directly on the machine. Generally speaking, prefer to use `cmd/3`
+  > over this function.
 
   ## Examples
 
       iex> System.shell("echo hello")
       {"hello\n", 0}
 
-  If you want to stream the devices to IO as they come:
+  If you want to stream the output to Standard IO as it arrives:
 
       iex> System.shell("echo hello", into: IO.stream())
       hello
@@ -914,27 +929,24 @@ defmodule System do
 
     # Finding shell command logic from :os.cmd in OTP
     # https://github.com/erlang/otp/blob/8deb96fb1d017307e22d2ab88968b9ef9f1b71d0/lib/kernel/src/os.erl#L184
-    command =
-      case :os.type() do
-        {:unix, _} ->
-          command =
-            command
-            |> String.replace("\"", "\\\"")
-            |> String.to_charlist()
+    case :os.type() do
+      {:unix, _} ->
+        shell_path = :os.find_executable('sh') || :erlang.error(:enoent, [command, opts])
+        command = "(#{command}\n) </dev/null"
+        do_cmd({:spawn_executable, shell_path}, [args: ["-c", command]], opts)
 
-          'sh -c "' ++ command ++ '"'
-
-        {:win32, osname} ->
-          command = String.to_charlist(command)
+      {:win32, osname} ->
+        command = String.to_charlist(command)
 
+        command =
           case {System.get_env("COMSPEC"), osname} do
             {nil, :windows} -> 'command.com /s /c ' ++ command
             {nil, _} -> 'cmd /s /c ' ++ command
             {cmd, _} -> '#{cmd} /s /c ' ++ command
           end
-      end
 
-    do_cmd({:spawn, command}, [], opts)
+        do_cmd({:spawn, command}, [], opts)
+    end
   end
 
   @doc ~S"""
@@ -969,7 +981,7 @@ defmodule System do
       iex> System.cmd("echo", ["hello"], env: [{"MIX_ENV", "test"}])
       {"hello\n", 0}
 
-  If you want to stream the devices to IO as they come:
+  If you want to stream the output to Standard IO as it arrives:
 
       iex> System.cmd("echo", ["hello"], into: IO.stream())
       hello
