@@ -35,7 +35,7 @@ defmodule Task do
 
     1. If you are using async tasks, you **must await** a reply
        as they are *always* sent. If you are not expecting a reply,
-       consider using `Task.start_link/1` detailed below.
+       consider using `Task.start_link/1` as detailed below.
 
     2. async tasks link the caller and the spawned process. This
        means that, if the caller crashes, the task will crash
@@ -69,41 +69,42 @@ defmodule Task do
         {Task.Supervisor, name: MyApp.TaskSupervisor}
       ], strategy: :one_for_one)
 
-  And now you can use async/await once again passig the name of
-  the supervisor isntead of the pid:
+  And now you can use async/await by passing the name of
+  the supervisor instead of the pid:
 
       Task.Supervisor.async(MyApp.TaskSupervisor, fn ->
         # Do something
       end)
       |> Task.await()
 
-  We encourage developers to rely on supervised tasks as much as
-  possible. Supervised tasks enable a huge variety of patterns
-  which allows you explicit control on how to handle the results,
-  errors, and timeouts. Here is a summary:
+  We encourage developers to rely on supervised tasks as much as possible.
+  Supervised tasks improves the visibility of how many tasks are running
+  at a given moment and enable a huge variety of patterns that gives you
+  explicit control on how to handle the results, errors, and timeouts.
+  Here is a summary:
 
-    * Use `Task.Supervisor.start_child/2` to start a fire-and-forget
-      task and you don't care about its results nor about if it completes
-      successfully
+    * Using `Task.Supervisor.start_child/2` allows you to start a fire-and-forget
+      task that you don't care about its results or if it completes successfully or not.
 
-    * Use `Task.Supervisor.async/2` + `Task.await/2` allows you to execute
+    * Using `Task.Supervisor.async/2` + `Task.await/2` allows you to execute
       tasks concurrently and retrieve its result. If the task fails,
-      the caller will also fail
+      the caller will also fail.
 
-    * Use `Task.Supervisor.async_nolink/2` + `Task.yield/2` + `Task.shutdown/2`
+    * Using `Task.Supervisor.async_nolink/2` + `Task.yield/2` + `Task.shutdown/2`
       allows you to execute tasks concurrently and retrieve their results
       or the reason they failed within a given time frame. If the task fails,
-      the caller won't fail: you will receive the error reason either on
-      `yield` or `shutdown`
+      the caller won't fail. You will receive the error reason either on
+      `yield` or `shutdown`.
 
-  See the `Task.Supervisor` module for details on the supported operations.
+  Furthermore, the supervisor guarantee all tasks first terminate, within a
+  configurable shutdown period, when your application shuts down. See the
+  `Task.Supervisor` module for details on the supported operations.
 
   ### Distributed tasks
 
-  Since Elixir provides a `Task.Supervisor`, it is easy to use one
-  to dynamically start tasks across nodes:
+  With `Task.Supervisor`, it is easy to dynamically start tasks across nodes:
 
-      # On the remote node
+      # On the remote node named :remote@local
       Task.Supervisor.start_link(name: MyApp.DistSupervisor)
 
       # On the client
@@ -111,8 +112,8 @@ defmodule Task do
       Task.Supervisor.async(supervisor, MyMod, :my_fun, [arg1, arg2, arg3])
 
   Note that, when working with distributed tasks, one should use the
-  `Task.Supervisor.async/4` function that expects explicit module, function,
-  and arguments, instead of `Task.Supervisor.async/2` that works with anonymous
+  `Task.Supervisor.async/5` function that expects explicit module, function,
+  and arguments, instead of `Task.Supervisor.async/3` that works with anonymous
   functions. That's because anonymous functions expect the same module version
   to exist on all involved nodes. Check the `Agent` module documentation for
   more information on distributed processes as the limitations described there
@@ -131,7 +132,7 @@ defmodule Task do
 
   This is often useful when you need to execute some steps while
   setting up your supervision tree. For example: to warm up caches,
-  log the initialization status, etc.
+  log the initialization status, and such.
 
   If you don't want to put the Task code directly under the `Supervisor`,
   you can wrap the `Task` in its own module, similar to how you would
@@ -156,8 +157,8 @@ defmodule Task do
       ], strategy: :one_for_one)
 
   Since these tasks are supervised and not directly linked to the caller,
-  they cannot be awaited on. By default, the functions `Task.start`
-  and `Task.start_link` are for fire-and-forget tasks, where you don't
+  they cannot be awaited on. By default, the functions `Task.start/1`
+  and `Task.start_link/1` are for fire-and-forget tasks, where you don't
   care about the results or if it completes successfully or not.
 
   `use Task` defines a `child_spec/1` function, allowing the
@@ -193,8 +194,8 @@ defmodule Task do
   For example, we recommend developers to always start tasks under a supervisor.
   This provides more visibility and allows you to control how those tasks are
   terminated when a node shuts down. That might look something like
-  `Task.Supervisor.start_child(MySupervisor, task_specification)`. This means
-  that, although your code is the one who invokes the task, the actual ancestor of
+  `Task.Supervisor.start_child(MySupervisor, task_function)`. This means
+  that, although your code is the one invoking the task, the actual ancestor of
   the task is the supervisor, as the supervisor is the one effectively starting it.
 
   To track the relationship between your code and the task, we use the `$callers`
@@ -224,26 +225,30 @@ defmodule Task do
 
   It contains these fields:
 
+    * `:mfa` - a three-element tuple containing the module, function name,
+      and arity invoked to start the task in `async/1` and `async/3`
+
+    * `:owner` - the PID of the process that started the task
+
     * `:pid` - the PID of the task process; `nil` if the task does
       not use a task process
 
     * `:ref` - the task monitor reference
 
-    * `:owner` - the PID of the process that started the task
-
   """
-  @enforce_keys [:pid, :ref, :owner]
-  defstruct pid: nil, ref: nil, owner: nil
+  @enforce_keys [:mfa, :owner, :pid, :ref]
+  defstruct @enforce_keys
 
   @typedoc """
   The Task type.
 
-  See `%Task{}` for information about each field of the structure.
+  See [`%Task{}`](`__struct__/0`) for information about each field of the structure.
   """
   @type t :: %__MODULE__{
+          mfa: mfa(),
+          owner: pid(),
           pid: pid() | nil,
-          ref: reference() | nil,
-          owner: pid() | nil
+          ref: reference()
         }
 
   defguardp is_timeout(timeout)
@@ -391,13 +396,13 @@ defmodule Task do
       Task.await(x) + y
 
   As before, if `heavy_fun/0` fails, the whole computation will
-  fail, including the parent process. If you don't want the task
+  fail, including the caller process. If you don't want the task
   to fail then you must change the `heavy_fun/0` code in the
   same way you would achieve it if you didn't have the async call.
   For example, to either return `{:ok, val} | :error` results or,
   in more extreme cases, by using `try/rescue`. In other words,
-  an asynchronous task should be thought of as an extension of a
-  process rather than a mechanism to isolate it from all errors.
+  an asynchronous task should be thought of as an extension of the
+  caller process rather than a mechanism to isolate it from all errors.
 
   If you don't want to link the caller to the task, then you
   must use a supervised task with `Task.Supervisor` and call
@@ -417,29 +422,91 @@ defmodule Task do
     * Unlinking the task process started with `async`/`await`.
       If you unlink the processes and the task does not belong
       to any supervisor, you may leave dangling tasks in case
-      the parent dies.
+      the caller process dies.
 
+  ## Metadata
+
+  The task created with this function stores `:erlang.apply/2` in
+  its `:mfa` metadata field, which is used internally to apply
+  the anonymous function. Use `async/3` if you want another function
+  to be used as metadata.
   """
   @spec async((() -> any)) :: t
   def async(fun) when is_function(fun, 0) do
     async(:erlang, :apply, [fun, []])
   end
 
+  # TODO: Remove conditional on Erlang/OTP 24
+  @compile {:no_warn_undefined, {:erlang, :monitor, 3}}
+
   @doc """
   Starts a task that must be awaited on.
 
   Similar to `async/1` except the function to be started is
   specified by the given `module`, `function_name`, and `args`.
+  The `module`, `function_name`, and its arity are stored as
+  a tuple in the `:mfa` field for reflection purposes.
   """
   @spec async(module, atom, [term]) :: t
   def async(module, function_name, args)
       when is_atom(module) and is_atom(function_name) and is_list(args) do
-    mfa = {module, function_name, args}
+    mfargs = {module, function_name, args}
     owner = self()
-    {:ok, pid} = Task.Supervised.start_link(get_owner(owner), get_callers(owner), :nomonitor, mfa)
-    ref = Process.monitor(pid)
-    send(pid, {owner, ref})
-    %Task{pid: pid, ref: ref, owner: owner}
+    {:ok, pid} = Task.Supervised.start_link(get_owner(owner), :nomonitor)
+
+    {reply_to, ref} =
+      if function_exported?(:erlang, :monitor, 3) do
+        ref = :erlang.monitor(:process, pid, alias: :demonitor)
+        {ref, ref}
+      else
+        {owner, Process.monitor(pid)}
+      end
+
+    send(pid, {owner, ref, reply_to, get_callers(owner), mfargs})
+    %Task{pid: pid, ref: ref, owner: owner, mfa: {module, function_name, length(args)}}
+  end
+
+  @doc """
+  Starts a task that immediately completes with the given `result`.
+
+  Unlike `async/1`, this task does not spawn a linked process. It can
+  be awaited or yielded like any other task.
+
+  ## Usage
+
+  In some cases, it is useful to create a "completed" task that represents
+  a task that has already run and generated a result. For example, when
+  processing data you may be able to determine that certain inputs are
+  invalid before dispatching them for further processing:
+
+      def process(data) do
+        tasks =
+          for entry <- data do
+            if invalid_input?(entry) do
+              Task.completed({:error, :invalid_input})
+            else
+              Task.async(fn -> further_process(entry) end)
+            end
+          end
+
+        Task.await_many(tasks)
+      end
+
+  In many cases, `Task.completed/1` may be avoided in favor of returning the
+  result directly.  You should generally only require this variant when working
+  with mixed asynchrony, when a group of inputs will be handled partially
+  synchronously and partially asynchronously.
+  """
+  @doc since: "1.13.0"
+  @spec completed(any) :: t
+  def completed(result) do
+    ref = make_ref()
+    owner = self()
+
+    # "complete" the task immediately
+    send(owner, {ref, result})
+
+    %Task{pid: nil, ref: ref, owner: owner, mfa: {Task, :completed, 1}}
   end
 
   @doc """
@@ -447,22 +514,25 @@ defmodule Task do
   is mapped concurrently on each element in `enumerable`.
 
   Each element of `enumerable` will be prepended to the given `args` and
-  processed by its own task. The tasks will be linked to an intermediate
-  process that is then linked to the current process. This means a failure
-  in a task terminates the current process and a failure in the current process
-  terminates all tasks.
+  processed by its own task. Those tasks will be linked to an intermediate
+  process that is then linked to the caller process. This means a failure
+  in a task terminates the caller process and a failure in the caller
+  process terminates all tasks.
 
   When streamed, each task will emit `{:ok, value}` upon successful
   completion or `{:exit, reason}` if the caller is trapping exits.
-  The order of results depends on the value of the `:ordered` option.
+  It's possible to have `{:exit, {element, reason}}` for exits
+  using the `:zip_input_on_exit` option. The order of results depends
+  on the value of the `:ordered` option.
 
   The level of concurrency and the time tasks are allowed to run can
   be controlled via options (see the "Options" section below).
 
   Consider using `Task.Supervisor.async_stream/6` to start tasks
-  under a supervisor. If you find yourself trapping exits to handle exits
-  inside the async stream, consider using `Task.Supervisor.async_stream_nolink/6`
-  to start tasks that are not linked to the calling process.
+  under a supervisor. If you find yourself trapping exits to ensure
+  errors in the tasks do not terminate the caller process, consider
+  using `Task.Supervisor.async_stream_nolink/6` to start tasks that
+  are not linked to the caller process.
 
   ## Options
 
@@ -483,10 +553,15 @@ defmodule Task do
 
     * `:on_timeout` - what to do when a task times out. The possible
       values are:
-      * `:exit` (default) - the process that spawned the tasks exits.
+      * `:exit` (default) - the caller (the process that spawned the tasks) exits.
       * `:kill_task` - the task that timed out is killed. The value
         emitted for that task is `{:exit, :timeout}`.
 
+    * `:zip_input_on_exit` - (since v1.14.0) adds the original
+      input to `:exit` tuples. The value emitted for that task is
+      `{:exit, {input, reason}}`, where `input` is the collection element
+      that caused an exited during processing. Defaults to `false`.
+
   ## Example
 
   Let's build a stream and then enumerate it:
@@ -508,11 +583,32 @@ defmodule Task do
       stream = Task.async_stream(collection, Mod, :expensive_fun, [], ordered: false)
       Stream.run(stream)
 
-  ## Attention: async + take
+  ## First async tasks to complete
 
-  Given items in an async stream are processed concurrently, doing
-  `async_stream` followed by `Enum.take/2` may cause more items than
-  requested to be processed. Let's see an example:
+  You can also use `async_stream/3` to execute M tasks and find the N tasks
+  to complete. For example:
+
+      [
+        &heavy_call_1/0,
+        &heavy_call_2/0,
+        &heavy_call_3/0
+      ]
+      |> Task.async_stream(fn fun -> fun.() end, ordered: false, max_concurrency: 3)
+      |> Stream.filter(&match?({:ok, _}, &1))
+      |> Enum.take(2)
+
+  In the example above, we are executing three tasks and waiting for the
+  first 2 to complete. We use `Stream.filter/2` to restrict ourselves only
+  to successfully completed tasks, and then use `Enum.take/2` to retrieve
+  N items. Note it is important to set both `ordered: false` and
+  `max_concurrency: M`, where M is the number of tasks, to make sure all
+  calls execute concurrently.
+
+  ### Attention: unbound async + take
+
+  If you want to potentially process a high number of items and keep only
+  part of the results, you may end-up processing more items than desired.
+  Let's see an example:
 
       1..100
       |> Task.async_stream(fn i ->
@@ -521,13 +617,15 @@ defmodule Task do
       end)
       |> Enum.take(10)
 
-  For a machine with 8 cores, the above will process 16 items instead
-  of 10. The reason is that `async_stream/5` always have 8 elements
-  processing at once. So by the time `Enum` says it got all elements
-  it needed, there are still 6 elements left to be processed.
+  Running the example above in a machine with 8 cores will process 16 items,
+  even though you want only 10 elements, since `async_stream/3` process items
+  concurrently. That's because it will process 8 elements at once. Then all 8
+  elements complete at roughly the same time, causing 8 elements to be kicked
+  off for processing. Out of these extra 8, only 2 will be used, and the rest
+  will be terminated.
 
-  The solution here is to use `Stream.take/2` instead of `Enum.take/2`
-  to filter elements before-hand:
+  Depending on the problem, you can filter or limit the number of elements
+  upfront:
 
       1..100
       |> Stream.take(10)
@@ -537,9 +635,11 @@ defmodule Task do
       end)
       |> Enum.to_list()
 
-  If for some reason you cannot take the elements before hand,
-  you can use `:max_concurrency` to limit how many elements
-  may be over processed at the cost of reducing concurrency.
+  In other cases, you likely want to tweak `:max_concurrency` to limit how
+  many elements may be over processed at the cost of reducing concurrency.
+  You can also set the number of elements to take to be a multiple of
+  `:max_concurrency`. For instance, setting `max_concurrency: 5` in the
+  example above.
   """
   @doc since: "1.4.0"
   @spec async_stream(Enumerable.t(), module, atom, [term], keyword) :: Enumerable.t()
@@ -556,8 +656,8 @@ defmodule Task do
   module-function-arguments tuple. `fun` must be a one-arity anonymous function.
 
   Each `enumerable` element is passed as argument to the given function `fun` and
-  processed by its own task. The tasks will be linked to the current process,
-  similarly to `async/1`.
+  processed by its own task. The tasks will be linked to the caller process, similarly
+  to `async/1`.
 
   ## Example
 
@@ -578,10 +678,14 @@ defmodule Task do
   end
 
   defp build_stream(enumerable, fun, options) do
-    &Task.Supervised.stream(enumerable, &1, &2, fun, options, fn [owner | _] = callers, mfa ->
-      {:ok, pid} = Task.Supervised.start_link(get_owner(owner), callers, :nomonitor, mfa)
-      {:ok, :link, pid}
-    end)
+    fn acc, acc_fun ->
+      owner = get_owner(self())
+
+      Task.Supervised.stream(enumerable, acc, acc_fun, get_callers(self()), fun, options, fn ->
+        {:ok, pid} = Task.Supervised.start_link(owner, :nomonitor)
+        {:ok, :link, pid}
+      end)
+    end
   end
 
   # Returns a tuple with the node where this is executed and either the
@@ -607,14 +711,14 @@ defmodule Task do
   @doc ~S"""
   Awaits a task reply and returns it.
 
-  In case the task process dies, the current process will exit with the same
+  In case the task process dies, the caller process will exit with the same
   reason as the task.
 
   A timeout, in milliseconds or `:infinity`, can be given with a default value
-  of `5000`. If the timeout is exceeded, then the current process will exit. If
-  the task process is linked to the current process which is the case when a
-  task is started with `async`, then the task process will also exit. If the
-  task process is trapping exits or not linked to the current process, then it
+  of `5000`. If the timeout is exceeded, then the caller process will exit.
+  If the task process is linked to the caller process which is the case when
+  a task is started with `async`, then the task process will also exit. If the
+  task process is trapping exits or not linked to the caller process, then it
   will continue to run.
 
   This function assumes the task's monitor is still active or the monitor's
@@ -727,6 +831,50 @@ defmodule Task do
     end
   end
 
+  @doc """
+  Ignores an existing task.
+
+  This means the task will continue running, but it will be unlinked
+  and you can no longer yield, await or shut it down.
+
+  Returns `{:ok, reply}` if the reply is received before ignoring the task,
+  `{:exit, reason}` if the task died before ignoring it, otherwise `nil`.
+
+  Important: avoid using [`Task.async/1,3`](`async/1`) and then immediately ignoring
+  the task. If you want to start tasks you don't care about their
+  results, use `Task.Supervisor.start_child/2` instead.
+
+  Requires Erlang/OTP 24+.
+  """
+  @doc since: "1.13.0"
+  def ignore(%Task{ref: ref, pid: pid, owner: owner} = task) do
+    unless function_exported?(:erlang, :monitor, 3) do
+      raise "Task.ignore/1 requires Erlang/OTP 24+"
+    end
+
+    if owner != self() do
+      raise ArgumentError, invalid_owner_error(task)
+    end
+
+    receive do
+      {^ref, reply} ->
+        Process.unlink(pid)
+        Process.demonitor(ref, [:flush])
+        {:ok, reply}
+
+      {:DOWN, ^ref, _, proc, :noconnection} ->
+        exit({reason(:noconnection, proc), {__MODULE__, :ignore, [task]}})
+
+      {:DOWN, ^ref, _, _, reason} ->
+        {:exit, reason}
+    after
+      0 ->
+        Process.unlink(pid)
+        Process.demonitor(ref, [:flush])
+        nil
+    end
+  end
+
   @doc """
   Awaits replies from multiple tasks and returns them.
 
@@ -734,16 +882,16 @@ defmodule Task do
   given time interval. It returns a list of the results, in the same order as
   the tasks supplied in the `tasks` input argument.
 
-  If any of the task processes dies, the current process will exit with the
-  same reason as that task.
+  If any of the task processes dies, the caller process will exit with the same
+  reason as that task.
 
   A timeout, in milliseconds or `:infinity`, can be given with a default value
-  of `5000`. If the timeout is exceeded, then the current process will exit.
-  Any task processes that are linked to the current process (which is the case
+  of `5000`. If the timeout is exceeded, then the caller process will exit.
+  Any task processes that are linked to the caller process (which is the case
   when a task is started with `async`) will also exit. Any task processes that
-  are trapping exits or not linked to the current process will continue to run.
+  are trapping exits or not linked to the caller process will continue to run.
 
-  This function assumes the tasks' monitors are still active or the monitors'
+  This function assumes the tasks' monitors are still active or the monitor's
   `:DOWN` message is in the message queue. If any tasks have been demonitored,
   or the message already received, this function will wait for the duration of
   the timeout.
@@ -856,16 +1004,17 @@ defmodule Task do
   end
 
   @doc ~S"""
-  Temporarily blocks the current process waiting for a task reply.
+  Temporarily blocks the caller process waiting for a task reply.
 
   Returns `{:ok, reply}` if the reply is received, `nil` if
   no reply has arrived, or `{:exit, reason}` if the task has already
   exited. Keep in mind that normally a task failure also causes
   the process owning the task to exit. Therefore this function can
-  return `{:exit, reason}` only if
+  return `{:exit, reason}` if at least one of the conditions below apply:
 
     * the task process exited with the reason `:normal`
-    * it isn't linked to the caller
+    * the task isn't linked to the caller (the task was started
+      with `Task.Supervisor.async_nolink/2` or `Task.Supervisor.async_nolink/4`)
     * the caller is trapping exits
 
   A timeout, in milliseconds or `:infinity`, can be given with a default value
@@ -890,6 +1039,18 @@ defmodule Task do
           nil
       end
 
+  If you intend to check on the task but leave it running after the timeout,
+  you can chain this together with `ignore/1`, like so:
+
+      case Task.yield(task, timeout) || Task.ignore(task) do
+        {:ok, result} ->
+          result
+
+        nil ->
+          Logger.warn("Failed to get a result in #{timeout}ms")
+          nil
+      end
+
   That ensures that if the task completes after the `timeout` but before `shutdown/1`
   has been called, you will still get the result, since `shutdown/1` is designed to
   handle this case and return the result.
@@ -942,8 +1103,9 @@ defmodule Task do
 
   `Task.yield_many/2` allows developers to spawn multiple tasks
   and retrieve the results received in a given timeframe.
-  If we combine it with `Task.shutdown/2`, it allows us to gather
-  those results and cancel the tasks that have not replied in time.
+  If we combine it with `Task.shutdown/2` (or `Task.ignore/1`),
+  it allows us to gather those results and cancel (or ignore)
+  the tasks that have not replied in time.
 
   Let's see an example.
 
@@ -1028,7 +1190,8 @@ defmodule Task do
   Unlinks and shuts down the task, and then checks for a reply.
 
   Returns `{:ok, reply}` if the reply is received while shutting down the task,
-  `{:exit, reason}` if the task died, otherwise `nil`.
+  `{:exit, reason}` if the task died, otherwise `nil`. Once shut down,
+  you can no longer await or yield it.
 
   The second argument is either a timeout or `:brutal_kill`. In case
   of a timeout, a `:shutdown` exit signal is sent to the task process
