@@ -1,6 +1,6 @@
 % Holds the logic responsible for function definitions (def(p) and defmacro(p)).
 -module(elixir_def).
--export([setup/1, reset_last/1, local_for/4,
+-export([setup/1, reset_last/1, local_for/5, external_for/5,
   take_definition/2, store_definition/5, store_definition/9,
   fetch_definitions/2, format_error/1]).
 -include("elixir.hrl").
@@ -16,18 +16,32 @@ reset_last({DataSet, _DataBag}) ->
 reset_last(Module) when is_atom(Module) ->
   reset_last(elixir_module:data_tables(Module)).
 
-local_for(Module, Name, Arity, Kinds) ->
+%% Finds the local definition for the current module.
+local_for(Meta, Name, Arity, Kinds, E) ->
+  External = fun({Mod, Fun}, Args) ->
+    invoke_external([{from_macro, true} | Meta], Mod, Fun, Args, E)
+  end,
+  fun_for(Meta, ?key(E, module), Name, Arity, Kinds, {value, External}).
+
+%% Finds the local definition for an external module.
+external_for(Meta, Module, Name, Arity, Kinds) ->
+  fun_for(Meta, Module, Name, Arity, Kinds, none).
+
+fun_for(Meta, Module, Name, Arity, Kinds, External) ->
   Tuple = {Name, Arity},
 
   try
     {Set, Bag} = elixir_module:data_tables(Module),
     {ets:lookup(Set, {def, Tuple}), ets:lookup(Bag, {clauses, Tuple})}
   of
-    {[{_, Kind, Meta, _, _, _}], Clauses} ->
+    {[{_, Kind, LocalMeta, _, _, _}], ClausesPairs} ->
       case (Kinds == all) orelse (lists:member(Kind, Kinds)) of
-        true -> elixir_erl:definition_to_anonymous(Module, Kind, Meta,
-                                                   [Clause || {_, Clause} <- Clauses]);
-        false -> false
+        true ->
+          Local = {value, fun(Fun, Args) -> invoke_local(Meta, Module, Fun, Args, External) end},
+          Clauses = [Clause || {_, Clause} <- ClausesPairs],
+          elixir_erl:definition_to_anonymous(Kind, LocalMeta, Clauses, Local, External);
+        false ->
+          false
       end;
     {[], _} ->
       false
@@ -35,6 +49,21 @@ local_for(Module, Name, Arity, Kinds) ->
     _:_ -> false
   end.
 
+invoke_local(Meta, Module, ErlName, Args, External) ->
+  {Name, Arity} = elixir_utils:erl_fa_to_elixir_fa(ErlName, length(Args)),
+
+  case fun_for(Meta, Module, Name, Arity, all, External) of
+    false ->
+      {current_stacktrace, [_ | T]} = erlang:process_info(self(), current_stacktrace),
+      erlang:raise(error, undef, [{Module, Name, Arity, []} | T]);
+    Fun ->
+      apply(Fun, Args)
+  end.
+
+invoke_external(Meta, Mod, Name, Args, E) ->
+  is_map(E) andalso elixir_env:trace({require, Meta, Mod, []}, E),
+  apply(Mod, Name, Args).
+
 %% Take a definition out of the table
 
 take_definition(Module, {Name, Arity} = Tuple) ->
@@ -146,8 +175,8 @@ store_definition(Kind, CheckClauses, Call, Body, Pos) ->
 
   {File, DefMeta} =
     case retrieve_location(Location, ?key(E, module)) of
-      {F, L} ->
-        {F, [{line, Line}, {file, {F, L}} | Generated]};
+      {AF, RF, L} ->
+        {AF, [{line, Line}, {file, {RF, L}} | Generated]};
       nil ->
         {nil, [{line, Line} | Generated]}
     end,
@@ -162,10 +191,10 @@ store_definition(Kind, CheckClauses, Call, Body, Pos) ->
 store_definition(Meta, Kind, CheckClauses, Name, Arity, DefaultsArgs, Guards, Body, File, ER) ->
   Module = ?key(ER, module),
   Tuple = {Name, Arity},
-  E = env_for_expansion(Kind, Tuple, ER),
+  {S, E} = env_for_expansion(Kind, Tuple, ER),
 
-  {Args, Defaults} = unpack_defaults(Kind, Meta, Name, DefaultsArgs, E),
-  Clauses = [elixir_clauses:def(Clause, E) ||
+  {Args, Defaults} = unpack_defaults(Kind, Meta, Name, DefaultsArgs, S, E),
+  Clauses = [elixir_clauses:def(Clause, S, E) ||
              Clause <- def_to_clauses(Kind, Meta, Args, Guards, Body, E)],
 
   DefaultsLength = length(Defaults),
@@ -181,24 +210,25 @@ store_definition(Meta, Kind, CheckClauses, Name, Arity, DefaultsArgs, Guards, Bo
   Tuple.
 
 env_for_expansion(Kind, Tuple, E) when Kind =:= defmacro; Kind =:= defmacrop ->
-  E#{function := Tuple, contextual_vars := ['__CALLER__']};
+  S = elixir_env:env_to_ex(E),
+  {S#elixir_ex{caller=true}, E#{function := Tuple}};
 env_for_expansion(_Kind, Tuple, E) ->
-  E#{function := Tuple, contextual_vars := []}.
+  {elixir_env:env_to_ex(E), E#{function := Tuple}}.
 
 retrieve_location(Location, Module) ->
   {Set, _} = elixir_module:data_tables(Module),
   case ets:take(Set, file) of
     [] when is_tuple(Location) ->
       {File, Line} = Location,
-      {elixir_utils:relative_to_cwd(File), Line};
+      {filename:absname(File), elixir_utils:relative_to_cwd(File), Line};
     [] ->
       nil;
-    [{file, File, _}] when is_binary(File) ->
+    [{file, File, _, _}] when is_binary(File) ->
       'Elixir.Module':delete_attribute(Module, file),
-      {elixir_utils:relative_to_cwd(File), 0};
-    [{file, {File, Line}, _}] when is_binary(File) andalso is_integer(Line) ->
+      {filename:absname(File), elixir_utils:relative_to_cwd(File), 0};
+    [{file, {File, Line}, _, _}] when is_binary(File) andalso is_integer(Line) ->
       'Elixir.Module':delete_attribute(Module, file),
-      {elixir_utils:relative_to_cwd(File), Line}
+      {filename:absname(File), elixir_utils:relative_to_cwd(File), Line}
   end.
 
 run_with_location_change(nil, E, Callback) ->
@@ -259,27 +289,27 @@ store_definition(Check, Kind, Meta, Name, Arity, File, Module, Defaults, Clauses
 
 %% Handling of defaults
 
-unpack_defaults(Kind, Meta, Name, Args, E) ->
-  Expanded = expand_defaults(Args, E#{context := nil}),
-  unpack_defaults(Kind, Meta, Name, Expanded, [], []).
+unpack_defaults(Kind, Meta, Name, Args, S, E) ->
+  {Expanded, #elixir_ex{unused={_, VersionOffset}}} = expand_defaults(Args, S, E#{context := nil}, []),
+  unpack_expanded(Kind, Meta, Name, Expanded, VersionOffset, [], []).
 
-unpack_defaults(Kind, Meta, Name, [{'\\\\', DefaultMeta, [Expr, _]} | T] = List, Acc, Clauses) ->
-  Base = match_defaults(Acc, length(Acc), []),
-  {Args, Invoke} = extract_defaults(List, length(Base), [], []),
+unpack_expanded(Kind, Meta, Name, [{'\\\\', DefaultMeta, [Expr, _]} | T] = List, VersionOffset, Acc, Clauses) ->
+  Base = match_defaults(Acc, length(Acc) + VersionOffset, []),
+  {Args, Invoke} = extract_defaults(List, length(Base) + VersionOffset, [], []),
   Clause = {Meta, Base ++ Args, [], {super, [{super, {Kind, Name}} | DefaultMeta], Base ++ Invoke}},
-  unpack_defaults(Kind, Meta, Name, T, [Expr | Acc], [Clause | Clauses]);
-unpack_defaults(Kind, Meta, Name, [H | T], Acc, Clauses) ->
-  unpack_defaults(Kind, Meta, Name, T, [H | Acc], Clauses);
-unpack_defaults(_Kind, _Meta, _Name, [], Acc, Clauses) ->
+  unpack_expanded(Kind, Meta, Name, T, VersionOffset, [Expr | Acc], [Clause | Clauses]);
+unpack_expanded(Kind, Meta, Name, [H | T], VersionOffset, Acc, Clauses) ->
+  unpack_expanded(Kind, Meta, Name, T, VersionOffset, [H | Acc], Clauses);
+unpack_expanded(_Kind, _Meta, _Name, [], _VersionOffset, Acc, Clauses) ->
   {lists:reverse(Acc), lists:reverse(Clauses)}.
 
-expand_defaults([{'\\\\', Meta, [Expr, Default]} | Args], E) ->
-  {ExpandedDefault, _} = elixir_expand:expand(Default, E),
-  [{'\\\\', Meta, [Expr, ExpandedDefault]} | expand_defaults(Args, E)];
-expand_defaults([Arg | Args], E) ->
-  [Arg | expand_defaults(Args, E)];
-expand_defaults([], _E) ->
-  [].
+expand_defaults([{'\\\\', Meta, [Expr, Default]} | Args], S, E, Acc) ->
+  {ExpandedDefault, SE, _} = elixir_expand:expand(Default, S, E),
+  expand_defaults(Args, SE, E, [{'\\\\', Meta, [Expr, ExpandedDefault]} | Acc]);
+expand_defaults([Arg | Args], S, E, Acc) ->
+   expand_defaults(Args, S, E, [Arg | Acc]);
+expand_defaults([], S, _E, Acc) ->
+  {lists:reverse(Acc), S}.
 
 extract_defaults([{'\\\\', _, [_Expr, Default]} | T], Counter, NewArgs, NewInvoke) ->
   extract_defaults(T, Counter, NewArgs, [Default | NewInvoke]);
@@ -289,7 +319,7 @@ extract_defaults([_ | T], Counter, NewArgs, NewInvoke) ->
 extract_defaults([], _Counter, NewArgs, NewInvoke) ->
   {lists:reverse(NewArgs), lists:reverse(NewInvoke)}.
 
-match_defaults([], 0, Acc) ->
+match_defaults([], _Counter, Acc) ->
   Acc;
 match_defaults([_ | T], Counter, Acc) ->
   NewCounter = Counter - 1,
