@@ -1,6 +1,7 @@
 %% Compiler backend to Erlang.
+
 -module(elixir_erl).
--export([elixir_to_erl/1, definition_to_anonymous/4, compile/1, consolidate/3,
+-export([elixir_to_erl/1, elixir_to_erl/2, definition_to_anonymous/5, compile/1, consolidate/3,
          get_ann/1, debug_info/4, scope/2, format_error/1]).
 -include("elixir.hrl").
 -define(typespecs, 'Elixir.Kernel.Typespec').
@@ -12,18 +13,18 @@ debug_info(elixir_v1, _Module, none, _Opts) ->
 debug_info(elixir_v1, _Module, {elixir_v1, Map, _Specs}, _Opts) ->
   {ok, Map};
 debug_info(erlang_v1, _Module, {elixir_v1, Map, Specs}, _Opts) ->
-  {Prefix, Forms, _, _, _, _} = dynamic_form(Map),
+  {Prefix, Forms, _, _, _} = dynamic_form(Map),
   {ok, Prefix ++ Specs ++ Forms};
 debug_info(core_v1, _Module, {elixir_v1, Map, Specs}, Opts) ->
-  {Prefix, Forms, _, _, _, _} = dynamic_form(Map),
+  {Prefix, Forms, _, _, _} = dynamic_form(Map),
   #{compile_opts := CompileOpts} = Map,
   AllOpts = CompileOpts ++ Opts,
 
-  %% Do not rely on elixir_erl_compiler because we don't
+  %% Do not rely on elixir_erl_compiler because we don't warn
   %% warnings nor the other functionality provided there.
   case elixir_erl_compiler:erl_to_core(Prefix ++ Specs ++ Forms, AllOpts) of
     {ok, CoreForms, _} ->
-      try compile:noenv_forms(CoreForms, [?NO_SPAWN_COMPILER_PROCESS, from_core, core, return | AllOpts]) of
+      try compile:noenv_forms(CoreForms, [no_spawn_compiler_process, from_core, core, return | AllOpts]) of
         {ok, _, Core, _} -> {ok, Core};
         _What -> {error, failed_conversion}
       catch
@@ -43,72 +44,50 @@ get_ann(Opts) when is_list(Opts) ->
 get_ann([{generated, true} | T], _, Line) -> get_ann(T, true, Line);
 get_ann([{line, Line} | T], Gen, _) when is_integer(Line) -> get_ann(T, Gen, Line);
 get_ann([_ | T], Gen, Line) -> get_ann(T, Gen, Line);
-get_ann([], Gen, Line) -> erl_anno:set_generated(Gen, Line).
+get_ann([], Gen, Line) -> erl_anno:set_generated(Gen, erl_anno:new(Line)).
 
 %% Converts an Elixir definition to an anonymous function.
 
-definition_to_anonymous(Module, Kind, Meta, Clauses) ->
-  ErlClauses = [translate_clause(Kind, Clause, true) || Clause <- Clauses],
+definition_to_anonymous(Kind, Meta, Clauses, LocalHandler, ExternalHandler) ->
+  ErlClauses = [translate_clause(Kind, 0, Clause, true) || Clause <- Clauses],
   Fun = {'fun', ?ann(Meta), {clauses, ErlClauses}},
-  LocalHandler = fun(LocalName, LocalArgs) -> invoke_local(Module, LocalName, LocalArgs) end,
-  {value, Result, _Binding} = erl_eval:expr(Fun, [], {value, LocalHandler}),
+  {value, Result, _Binding} = erl_eval:expr(Fun, [], LocalHandler, ExternalHandler),
   Result.
 
-invoke_local(Module, ErlName, Args) ->
-  {Name, Arity} = elixir_utils:erl_fa_to_elixir_fa(ErlName, length(Args)),
-
-  case elixir_def:local_for(Module, Name, Arity, all) of
-    false ->
-      {current_stacktrace, [_ | T]} = erlang:process_info(self(), current_stacktrace),
-      erlang:raise(error, undef, [{Module, Name, Arity, []} | T]);
-    Fun ->
-      apply(Fun, Args)
-  end.
-
 %% Converts Elixir quoted literals to Erlang AST.
+elixir_to_erl(Tree) ->
+  elixir_to_erl(Tree, erl_anno:new(0)).
 
-elixir_to_erl(Tree) when is_tuple(Tree) ->
-  {tuple, 0, [elixir_to_erl(X) || X <- tuple_to_list(Tree)]};
-elixir_to_erl([]) ->
-  {nil, 0};
-elixir_to_erl(<<>>) ->
-  {bin, 0, []};
-elixir_to_erl(Tree) when is_list(Tree) ->
-  elixir_to_erl_cons(Tree);
-elixir_to_erl(Tree) when is_atom(Tree) ->
-  {atom, 0, Tree};
-elixir_to_erl(Tree) when is_integer(Tree) ->
-  {integer, 0, Tree};
-elixir_to_erl(Tree) when is_float(Tree) ->
-  {float, 0, Tree};
-elixir_to_erl(Tree) when is_binary(Tree) ->
+elixir_to_erl(Tree, Ann) when is_tuple(Tree) ->
+  {tuple, Ann, [elixir_to_erl(X, Ann) || X <- tuple_to_list(Tree)]};
+elixir_to_erl([], Ann) ->
+  {nil, Ann};
+elixir_to_erl(<<>>, Ann) ->
+  {bin, Ann, []};
+elixir_to_erl(#{} = Map, Ann) ->
+  Assocs = [{map_field_assoc, Ann, elixir_to_erl(K, Ann), elixir_to_erl(V, Ann)} || {K, V} <- maps:to_list(Map)],
+  {map, Ann, Assocs};
+elixir_to_erl(Tree, Ann) when is_list(Tree) ->
+  elixir_to_erl_cons(Tree, Ann);
+elixir_to_erl(Tree, Ann) when is_atom(Tree) ->
+  {atom, Ann, Tree};
+elixir_to_erl(Tree, Ann) when is_integer(Tree) ->
+  {integer, Ann, Tree};
+elixir_to_erl(Tree, Ann) when is_float(Tree) ->
+  {float, Ann, Tree};
+elixir_to_erl(Tree, Ann) when is_binary(Tree) ->
   %% Note that our binaries are UTF-8 encoded and we are converting
   %% to a list using binary_to_list. The reason for this is that Erlang
   %% considers a string in a binary to be encoded in latin1, so the bytes
   %% are not changed in any fashion.
-  {bin, 0, [{bin_element, 0, {string, 0, binary_to_list(Tree)}, default, default}]};
-elixir_to_erl(Function) when is_function(Function) ->
-  case (erlang:fun_info(Function, type) == {type, external}) andalso
-       (erlang:fun_info(Function, env) == {env, []}) of
-    true ->
-      {module, Module} = erlang:fun_info(Function, module),
-      {name, Name}     = erlang:fun_info(Function, name),
-      {arity, Arity}   = erlang:fun_info(Function, arity),
-
-      {'fun', 0, {function,
-        {atom, 0, Module},
-        {atom, 0, Name},
-        {integer, 0, Arity}}};
-    false ->
-      error(badarg)
-  end;
-elixir_to_erl(Pid) when is_pid(Pid) ->
-  ?remote(0, erlang, binary_to_term, [elixir_erl:elixir_to_erl(term_to_binary(Pid))]);
-elixir_to_erl(_Other) ->
+  {bin, Ann, [{bin_element, Ann, {string, Ann, binary_to_list(Tree)}, default, default}]};
+elixir_to_erl(Pid, Ann) when is_pid(Pid) ->
+  ?remote(Ann, erlang, binary_to_term, [elixir_to_erl(term_to_binary(Pid), Ann)]);
+elixir_to_erl(_Other, _Ann) ->
   error(badarg).
 
-elixir_to_erl_cons([H | T]) -> {cons, 0, elixir_to_erl(H), elixir_to_erl_cons(T)};
-elixir_to_erl_cons(T) -> elixir_to_erl(T).
+elixir_to_erl_cons([H | T], Ann) -> {cons, Ann, elixir_to_erl(H, Ann), elixir_to_erl_cons(T, Ann)};
+elixir_to_erl_cons(T, Ann) -> elixir_to_erl(T, Ann).
 
 %% Returns a scope for translation.
 
@@ -118,83 +97,82 @@ scope(_Meta, ExpandCaptures) ->
 %% Static compilation hook, used in protocol consolidation
 
 consolidate(Map, TypeSpecs, Chunks) ->
-  {Prefix, Forms, _Def, _Defmacro, _Macros, _NoWarnUndefined} = dynamic_form(Map),
+  {Prefix, Forms, _Def, _Defmacro, _Macros} = dynamic_form(Map),
   load_form(Map, Prefix, Forms, TypeSpecs, Chunks).
 
 %% Dynamic compilation hook, used in regular compiler
 
-compile(Map) ->
-  elixir_erl_compiler:spawn(fun spawned_compile/1, [Map]).
-
-spawned_compile(#{module := Module, line := Line} = Map) ->
+compile(#{module := Module, line := Line} = Map) ->
   {Set, Bag} = elixir_module:data_tables(Module),
 
   TranslatedTypespecs =
-    case elixir_config:get(bootstrap) andalso
+    case elixir_config:is_bootstrap() andalso
           (code:ensure_loaded(?typespecs) /= {module, ?typespecs}) of
       true -> {[], [], [], [], []};
       false -> ?typespecs:translate_typespecs_for_module(Set, Bag)
     end,
 
-  {Prefix, Forms, Def, Defmacro, Macros, NoWarnUndefined} = dynamic_form(Map),
+  {Prefix, Forms, Def, Defmacro, Macros} = dynamic_form(Map),
   {Types, Callbacks, TypeSpecs} = typespecs_form(Map, TranslatedTypespecs, Macros),
 
   DocsChunk = docs_chunk(Set, Module, Line, Def, Defmacro, Types, Callbacks),
-  CheckerChunk = checker_chunk(Map, NoWarnUndefined),
+  CheckerChunk = checker_chunk(Map),
   load_form(Map, Prefix, Forms, TypeSpecs, DocsChunk ++ CheckerChunk).
 
 dynamic_form(#{module := Module, line := Line, relative_file := RelativeFile,
                attributes := Attributes, definitions := Definitions, unreachable := Unreachable,
-               deprecated := Deprecated, compile_opts := Opts}) ->
+               deprecated := Deprecated, compile_opts := Opts} = Map) ->
   {Def, Defmacro, Macros, Exports, Functions} =
-    split_definition(Definitions, Unreachable, [], [], [], [], {[], []}),
+    split_definition(Definitions, Unreachable, Line, [], [], [], [], {[], []}),
 
-  {NoWarnUndefined, FilteredOpts} = split_no_warn_undefined(Opts, [], []),
+  FilteredOpts = lists:filter(fun({no_warn_undefined, _}) -> false; (_) -> true end, Opts),
   Location = {elixir_utils:characters_to_list(RelativeFile), Line},
+
   Prefix = [{attribute, Line, file, Location},
             {attribute, Line, module, Module},
             {attribute, Line, compile, [no_auto_import | FilteredOpts]}],
 
-  Forms0 = functions_form(Line, Module, Def, Defmacro, Exports, Functions, Deprecated),
+  Struct = maps:get(struct, Map, nil),
+  Forms0 = functions_form(Line, Module, Def, Defmacro, Exports, Functions, Deprecated, Struct),
   Forms1 = attributes_form(Line, Attributes, Forms0),
-  {Prefix, Forms1, Def, Defmacro, Macros, NoWarnUndefined}.
+  {Prefix, Forms1, Def, Defmacro, Macros}.
 
 % Definitions
 
-split_definition([{Tuple, Kind, Meta, Clauses} | T], Unreachable,
+split_definition([{Tuple, Kind, Meta, Clauses} | T], Unreachable, Line,
                  Def, Defmacro, Macros, Exports, Functions) ->
   case lists:member(Tuple, Unreachable) of
     false ->
-      split_definition(Tuple, Kind, Meta, Clauses, T, Unreachable,
+      split_definition(Tuple, Kind, Meta, Clauses, T, Unreachable, Line,
                        Def, Defmacro, Macros, Exports, Functions);
     true ->
-      split_definition(T, Unreachable, Def, Defmacro, Macros, Exports, Functions)
+      split_definition(T, Unreachable, Line, Def, Defmacro, Macros, Exports, Functions)
   end;
-split_definition([], _Unreachable, Def, Defmacro, Macros, Exports, {Head, Tail}) ->
+split_definition([], _Unreachable, _Line, Def, Defmacro, Macros, Exports, {Head, Tail}) ->
   {Def, Defmacro, Macros, Exports, Head ++ Tail}.
 
-split_definition(Tuple, def, Meta, Clauses, T, Unreachable,
+split_definition(Tuple, def, Meta, Clauses, T, Unreachable, Line,
                  Def, Defmacro, Macros, Exports, Functions) ->
-  {_, _, N, A, _} = Entry = translate_definition(def, Meta, Tuple, Clauses),
-  split_definition(T, Unreachable, [Tuple | Def], Defmacro, Macros, [{N, A} | Exports],
+  {_, _, N, A, _} = Entry = translate_definition(def, Line, Meta, Tuple, Clauses),
+  split_definition(T, Unreachable, Line, [Tuple | Def], Defmacro, Macros, [{N, A} | Exports],
                    add_definition(Meta, Entry, Functions));
 
-split_definition(Tuple, defp, Meta, Clauses, T, Unreachable,
+split_definition(Tuple, defp, Meta, Clauses, T, Unreachable, Line,
                  Def, Defmacro, Macros, Exports, Functions) ->
-  Entry = translate_definition(defp, Meta, Tuple, Clauses),
-  split_definition(T, Unreachable, Def, Defmacro, Macros, Exports,
+  Entry = translate_definition(defp, Line, Meta, Tuple, Clauses),
+  split_definition(T, Unreachable, Line, Def, Defmacro, Macros, Exports,
                    add_definition(Meta, Entry, Functions));
 
-split_definition(Tuple, defmacro, Meta, Clauses, T, Unreachable,
+split_definition(Tuple, defmacro, Meta, Clauses, T, Unreachable, Line,
                  Def, Defmacro, Macros, Exports, Functions) ->
-  {_, _, N, A, _} = Entry = translate_definition(defmacro, Meta, Tuple, Clauses),
-  split_definition(T, Unreachable, Def, [Tuple | Defmacro], [Tuple | Macros], [{N, A} | Exports],
+  {_, _, N, A, _} = Entry = translate_definition(defmacro, Line, Meta, Tuple, Clauses),
+  split_definition(T, Unreachable, Line, Def, [Tuple | Defmacro], [Tuple | Macros], [{N, A} | Exports],
                    add_definition(Meta, Entry, Functions));
 
-split_definition(Tuple, defmacrop, Meta, Clauses, T, Unreachable,
+split_definition(Tuple, defmacrop, Meta, Clauses, T, Unreachable, Line,
                  Def, Defmacro, Macros, Exports, Functions) ->
-  Entry = translate_definition(defmacro, Meta, Tuple, Clauses),
-  split_definition(T, Unreachable, Def, Defmacro, [Tuple | Macros], Exports,
+  Entry = translate_definition(defmacro, Line, Meta, Tuple, Clauses),
+  split_definition(T, Unreachable, Line, Def, Defmacro, [Tuple | Macros], Exports,
                    add_definition(Meta, Entry, Functions)).
 
 add_definition(Meta, Body, {Head, Tail}) ->
@@ -211,23 +189,30 @@ add_definition(Meta, Body, {Head, Tail}) ->
       {[Body | Head], Tail}
   end.
 
-translate_definition(Kind, Meta, {Name, Arity}, Clauses) ->
-  ErlClauses = [translate_clause(Kind, Clause, false) || Clause <- Clauses],
+translate_definition(Kind, Line, Meta, {Name, Arity}, Clauses) ->
+  ErlClauses = [translate_clause(Kind, Line, Clause, false) || Clause <- Clauses],
 
   case is_macro(Kind) of
     true -> {function, ?ann(Meta), elixir_utils:macro_name(Name), Arity + 1, ErlClauses};
     false -> {function, ?ann(Meta), Name, Arity, ErlClauses}
   end.
 
-translate_clause(Kind, {Meta, Args, Guards, Body}, ExpandCaptures) ->
+translate_clause(Kind, Line, {Meta, Args, Guards, Body}, ExpandCaptures) ->
   S = scope(Meta, ExpandCaptures),
 
-  {TClause, TS} = elixir_erl_clauses:clause(Meta,
-                    fun elixir_erl_pass:translate_args/2, Args, Body, Guards, S),
+  %% If the line matches the module line, then it is most likely an
+  %% auto-generated function and we don't want to track its contents.
+  Ann =
+    case ?line(Meta) of
+      Line -> erl_anno:set_generated(true, erl_anno:new(0));
+      _ -> ?ann(Meta)
+    end,
+
+  {TClause, TS} =
+    elixir_erl_clauses:clause(Ann, fun elixir_erl_pass:translate_args/3, Args, Body, Guards, S),
 
   case is_macro(Kind) of
     true ->
-      Ann = ?ann(Meta),
       FArgs = {var, Ann, '_@CALLER'},
       MClause = setelement(3, TClause, [FArgs | element(3, TClause)]),
 
@@ -235,7 +220,7 @@ translate_clause(Kind, {Meta, Args, Guards, Body}, ExpandCaptures) ->
         true  ->
           FBody = {'match', Ann,
             {'var', Ann, '__CALLER__'},
-            ?remote(Ann, elixir_env, linify, [{var, Ann, '_@CALLER'}])
+            ?remote(Ann, elixir_env, to_caller, [{var, Ann, '_@CALLER'}])
           },
           setelement(5, MClause, [FBody | element(5, TClause)]);
         false ->
@@ -251,13 +236,15 @@ is_macro(_)         -> false.
 
 % Functions
 
-functions_form(Line, Module, Def, Defmacro, Exports, Body, Deprecated) ->
-  {Spec, Info} = add_info_function(Line, Module, Def, Defmacro, Deprecated),
+functions_form(Line, Module, Def, Defmacro, Exports, Body, Deprecated, Struct) ->
+  {Spec, Info} = add_info_function(Line, Module, Def, Defmacro, Deprecated, Struct),
   [{attribute, Line, export, lists:sort([{'__info__', 1} | Exports])}, Spec, Info | Body].
 
-add_info_function(Line, Module, Def, Defmacro, Deprecated) ->
-  AllowedAttrs = [attributes, compile, functions, macros, md5, module, deprecated],
+add_info_function(Line, Module, Def, Defmacro, Deprecated, Struct) ->
+  AllowedAttrs = [attributes, compile, functions, macros, md5, exports_md5, module, deprecated, struct],
   AllowedArgs = lists:map(fun(Atom) -> {atom, Line, Atom} end, AllowedAttrs),
+  SortedDef = lists:sort(Def),
+  SortedDefmacro = lists:sort(Defmacro),
 
   Spec =
     {attribute, Line, spec, {{'__info__', 1},
@@ -271,9 +258,11 @@ add_info_function(Line, Module, Def, Defmacro, Deprecated) ->
 
   Info =
     {function, 0, '__info__', 1, [
-      direct_module_info(Module),
-      functions_info(Def),
-      macros_info(Defmacro),
+      get_module_info(Module),
+      functions_info(SortedDef),
+      macros_info(SortedDefmacro),
+      struct_info(Struct),
+      exports_md5_info(Struct, SortedDef, SortedDefmacro),
       get_module_info(Module, attributes),
       get_module_info(Module, compile),
       get_module_info(Module, md5),
@@ -282,21 +271,33 @@ add_info_function(Line, Module, Def, Defmacro, Deprecated) ->
 
   {Spec, Info}.
 
-direct_module_info(Module) ->
+get_module_info(Module) ->
   {clause, 0, [{atom, 0, module}], [], [{atom, 0, Module}]}.
 
+exports_md5_info(Struct, Def, Defmacro) ->
+  %% Deprecations do not need to be part of exports_md5 because it is always
+  %% checked by the runtime pass, so it is not really part of compilation.
+  Md5 = erlang:md5(erlang:term_to_binary({Def, Defmacro, Struct})),
+  {clause, 0, [{atom, 0, exports_md5}], [], [elixir_to_erl(Md5)]}.
+
 functions_info(Def) ->
-  {clause, 0, [{atom, 0, functions}], [], [elixir_erl:elixir_to_erl(lists:sort(Def))]}.
+  {clause, 0, [{atom, 0, functions}], [], [elixir_to_erl(Def)]}.
 
 macros_info(Defmacro) ->
-  {clause, 0, [{atom, 0, macros}], [], [elixir_erl:elixir_to_erl(lists:sort(Defmacro))]}.
+  {clause, 0, [{atom, 0, macros}], [], [elixir_to_erl(Defmacro)]}.
+
+struct_info(nil) ->
+  {clause, 0, [{atom, 0, struct}], [], [{atom, 0, nil}]};
+struct_info(Fields) ->
+  FieldsWithoutDefault = [maps:remove(default, FieldInfo) || FieldInfo <- Fields],
+  {clause, 0, [{atom, 0, struct}], [], [elixir_to_erl(FieldsWithoutDefault)]}.
 
 get_module_info(Module, Key) ->
   Call = ?remote(0, erlang, get_module_info, [{atom, 0, Module}, {var, 0, 'Key'}]),
   {clause, 0, [{match, 0, {var, 0, 'Key'}, {atom, 0, Key}}], [], [Call]}.
 
 deprecated_info(Deprecated) ->
-  {clause, 0, [{atom, 0, deprecated}], [], [elixir_erl:elixir_to_erl(Deprecated)]}.
+  {clause, 0, [{atom, 0, deprecated}], [], [elixir_to_erl(Deprecated)]}.
 
 % Typespecs
 
@@ -479,61 +480,69 @@ docs_chunk(Set, Module, Line, Def, Defmacro, Types, Callbacks) ->
       []
   end.
 
-doc_value(Doc) ->
+doc_value(Doc, Name) ->
   case Doc of
-    nil -> none;
-    false -> hidden;
-    Doc -> #{<<"en">> => Doc}
+    false ->
+      hidden;
+    nil ->
+      case erlang:atom_to_list(Name) of
+        [$_ | _] -> hidden;
+        _ -> none
+      end;
+    Doc ->
+      #{<<"en">> => Doc}
   end.
 
 get_moduledoc(Line, Set) ->
   case ets:lookup_element(Set, moduledoc, 2) of
     nil -> {Line, none};
-    {DocLine, Doc} -> {DocLine, doc_value(Doc)}
+    {DocLine, false} -> {DocLine, hidden};
+    {DocLine, nil} -> {DocLine, none};
+    {DocLine, Doc} -> {DocLine, #{<<"en">> => Doc}}
   end.
 
 get_moduledoc_meta(Set) ->
   case ets:lookup(Set, {moduledoc, meta}) of
     [] -> #{};
-    [{{moduledoc, meta}, Map, _}] when is_map(Map) -> Map
+    [{{moduledoc, meta}, Map}] when is_map(Map) -> Map
   end.
 
 get_docs(Set, Module, Definitions, Kind) ->
   [{Key,
     erl_anno:new(Line),
     [signature_to_binary(Module, Name, Signature)],
-    doc_value(Doc),
+    doc_value(Doc, Name),
     Meta
    } || {Name, Arity} <- Definitions,
-        {Key, Line, Signature, Doc, Meta} <- ets:lookup(Set, {Kind, Name, Arity})].
+        {Key, _Ctx, Line, Signature, Doc, Meta} <- ets:lookup(Set, {Kind, Name, Arity})].
 
 get_callback_docs(Set, Callbacks) ->
   [{Key,
     erl_anno:new(Line),
     [],
-    doc_value(Doc),
+    doc_value(Doc, Name),
     Meta
-   } || Callback <- Callbacks, {Key, Line, Doc, Meta} <- ets:lookup(Set, Callback)].
+   } || Callback <- Callbacks, {{_, Name, _} = Key, Line, Doc, Meta} <- ets:lookup(Set, Callback)].
 
 get_type_docs(Set, Types) ->
   [{Key,
     erl_anno:new(Line),
     [],
-    doc_value(Doc),
+    doc_value(Doc, Name),
     Meta
    } || {_Kind, {Name, Arity}, _, _, true} <- Types,
         {Key, Line, Doc, Meta} <- ets:lookup(Set, {type, Name, Arity})].
 
 signature_to_binary(_Module, Name, _Signature) when Name == '__aliases__'; Name == '__block__' ->
-  <<(atom_to_binary(Name, utf8))/binary, "(args)">>;
+  <<(atom_to_binary(Name))/binary, "(args)">>;
 
 signature_to_binary(_Module, fn, _Signature) ->
-  <<"fn">>;
+  <<"fn(clauses)">>;
 
 signature_to_binary(_Module, Name, _Signature)
     when Name == '__CALLER__'; Name == '__DIR__'; Name == '__ENV__';
          Name == '__MODULE__'; Name == '__STACKTRACE__'; Name == '%{}' ->
-  atom_to_binary(Name, utf8);
+  atom_to_binary(Name);
 
 signature_to_binary(_Module, '%', _) ->
   <<"%struct{}">>;
@@ -542,9 +551,11 @@ signature_to_binary(Module, '__struct__', []) ->
   <<"%", ('Elixir.Kernel':inspect(Module))/binary, "{}">>;
 
 signature_to_binary(_, Name, Signature) ->
-  'Elixir.Macro':to_string({Name, [], Signature}).
+  Quoted = {Name, [{closing, []}], Signature},
+  Doc = 'Elixir.Inspect.Algebra':format('Elixir.Code':quoted_to_algebra(Quoted), infinity),
+  'Elixir.IO':iodata_to_binary(Doc).
 
-checker_chunk(#{definitions := Definitions, deprecated := Deprecated, is_behaviour := IsBehaviour}, NoWarnUndefined) ->
+checker_chunk(#{definitions := Definitions, deprecated := Deprecated, is_behaviour := IsBehaviour}) ->
   DeprecatedMap = maps:from_list(Deprecated),
 
   Exports =
@@ -559,8 +570,7 @@ checker_chunk(#{definitions := Definitions, deprecated := Deprecated, is_behavio
     end, [], Definitions),
 
   Contents = #{
-    exports => lists:sort(behaviour_info_exports(IsBehaviour) ++ Exports),
-    no_warn_undefined => NoWarnUndefined
+    exports => lists:sort(behaviour_info_exports(IsBehaviour) ++ Exports)
   },
 
   [{<<"ExCk">>, erlang:term_to_binary({elixir_checker_v1, Contents})}].
@@ -568,16 +578,6 @@ checker_chunk(#{definitions := Definitions, deprecated := Deprecated, is_behavio
 behaviour_info_exports(true) -> [{{behaviour_info, 1}, #{kind => def, deprecated_reason => nil}}];
 behaviour_info_exports(false) -> [].
 
-split_no_warn_undefined([{no_warn_undefined, NoWarnUndefined} | CompileOpts], AccNWU, AccCO) ->
-  split_no_warn_undefined(CompileOpts, list_wrap(NoWarnUndefined) ++ AccNWU, AccCO);
-split_no_warn_undefined([Opt | CompileOpts], AccNWU, AccCO) ->
-  split_no_warn_undefined(CompileOpts, AccNWU, [Opt | AccCO]);
-split_no_warn_undefined([], AccNWU, AccCO) ->
-  {AccNWU, lists:reverse(AccCO)}.
-
-list_wrap(List) when is_list(List) -> List;
-list_wrap(Other) -> [Other].
-
 %% Errors
 
 form_error(#{line := Line, file := File}, Error) ->
