@@ -64,20 +64,18 @@ next_counter(Module) ->
 
 %% Compilation hook
 
-compile(Module, _Block, _Vars, #{line := Line, file := File}) when Module == nil; is_boolean(Module) ->
-  elixir_errors:form_error([{line, Line}], File, ?MODULE, {invalid_module, Module});
 compile(Module, Block, Vars, Env) when is_atom(Module) ->
-  #{line := Line, current_vars := {Read, _}, function := Function} = Env,
+  #{line := Line, function := Function, versioned_vars := OldVerVars} = Env,
 
-  %% In case we are generating a module from inside a function,
-  %% we get rid of the lexical tracker information as, at this
-  %% point, the lexical tracker process is long gone.
-  ResetEnv = elixir_env:reset_unused_vars(Env),
+  {VerVars, _} =
+    lists:mapfoldl(fun({Var, _}, I) -> {{Var, I}, I + 1} end, 0, maps:to_list(OldVerVars)),
+
+  BaseEnv = Env#{module := Module, versioned_vars := maps:from_list(VerVars)},
 
   MaybeLexEnv =
     case Function of
-      nil -> ResetEnv#{module := Module, current_vars := {Read, false}};
-      _   -> ResetEnv#{lexical_tracker := nil, tracers := [], function := nil, module := Module, current_vars := {Read, false}}
+      nil -> BaseEnv;
+      _   -> BaseEnv#{lexical_tracker := nil, tracers := [], function := nil}
     end,
 
   case MaybeLexEnv of
@@ -96,6 +94,7 @@ compile(Module, _Block, _Vars, #{line := Line, file := File}) ->
 compile(Line, Module, Block, Vars, E) ->
   File = ?key(E, file),
   check_module_availability(Line, File, Module),
+  ModuleAsCharlist = validate_module_name(Line, File, Module),
 
   CompilerModules = compiler_modules(),
   {Tables, Ref} = build(Line, File, Module),
@@ -103,53 +102,69 @@ compile(Line, Module, Block, Vars, E) ->
 
   try
     put_compiler_modules([Module | CompilerModules]),
-    elixir_env:trace({defmodule, [{line, Line}]}, E),
     {Result, NE} = eval_form(Line, Module, DataBag, Block, Vars, E),
+    CheckerInfo = checker_info(),
 
-    PersistedAttributes = ets:lookup_element(DataBag, persisted_attributes, 2),
-    Attributes = attributes(DataSet, DataBag, PersistedAttributes),
-    {AllDefinitions, Private} = elixir_def:fetch_definitions(File, Module),
+    {Binary, PersistedAttributes, Autoload} =
+      elixir_erl_compiler:spawn(fun() ->
+        PersistedAttributes = ets:lookup_element(DataBag, persisted_attributes, 2),
+        Attributes = attributes(DataSet, DataBag, PersistedAttributes),
+        {AllDefinitions, Private} = elixir_def:fetch_definitions(File, Module),
 
-    OnLoadAttribute = lists:keyfind(on_load, 1, Attributes),
-    NewPrivate = validate_on_load_attribute(OnLoadAttribute, AllDefinitions, Private, File, Line),
+        OnLoadAttribute = lists:keyfind(on_load, 1, Attributes),
+        NewPrivate = validate_on_load_attribute(OnLoadAttribute, AllDefinitions, Private, File, Line),
 
-    DialyzerAttribute = lists:keyfind(dialyzer, 1, Attributes),
-    validate_dialyzer_attribute(DialyzerAttribute, AllDefinitions, File, Line),
+        DialyzerAttribute = lists:keyfind(dialyzer, 1, Attributes),
+        validate_dialyzer_attribute(DialyzerAttribute, AllDefinitions, File, Line),
 
-    Unreachable = elixir_locals:warn_unused_local(File, Module, AllDefinitions, NewPrivate),
-    elixir_locals:ensure_no_undefined_local(File, Module, AllDefinitions),
-    elixir_locals:ensure_no_import_conflict(File, Module, AllDefinitions),
+        Unreachable = elixir_locals:warn_unused_local(File, Module, AllDefinitions, NewPrivate),
+        elixir_locals:ensure_no_undefined_local(File, Module, AllDefinitions),
+        elixir_locals:ensure_no_import_conflict(File, Module, AllDefinitions),
 
-    %% We stop tracking locals here to avoid race conditions in case after_load
-    %% evaluates code in a separate process that may write to locals table.
-    elixir_locals:stop({DataSet, DataBag}),
-    make_readonly(Module),
+        %% We stop tracking locals here to avoid race conditions in case after_load
+        %% evaluates code in a separate process that may write to locals table.
+        elixir_locals:stop({DataSet, DataBag}),
+        make_readonly(Module),
 
-    (not elixir_config:get(bootstrap)) andalso
-     'Elixir.Module':check_behaviours_and_impls(E, DataSet, DataBag, AllDefinitions),
+        (not elixir_config:is_bootstrap()) andalso
+         'Elixir.Module':check_derive_behaviours_and_impls(E, DataSet, DataBag, AllDefinitions),
 
-    RawCompileOpts = bag_lookup_element(DataBag, {accumulate, compile}, 2),
-    CompileOpts = validate_compile_opts(RawCompileOpts, AllDefinitions, Unreachable, File, Line),
+        RawCompileOpts = bag_lookup_element(DataBag, {accumulate, compile}, 2),
+        CompileOpts = validate_compile_opts(RawCompileOpts, AllDefinitions, Unreachable, File, Line),
 
-    ModuleMap = #{
-      struct => get_struct(DataSet),
-      module => Module,
-      line => Line,
-      file => File,
-      relative_file => elixir_utils:relative_to_cwd(File),
-      attributes => Attributes,
-      definitions => AllDefinitions,
-      unreachable => Unreachable,
-      compile_opts => CompileOpts,
-      deprecated => get_deprecated(DataBag),
-      is_behaviour => is_behaviour(DataBag)
-    },
+        AfterVerify = bag_lookup_element(DataBag, {accumulate, after_verify}, 2),
+        [elixir_env:trace({remote_function, [], VerifyMod, VerifyFun, 1}, E) ||
+         {VerifyMod, VerifyFun} <- AfterVerify],
 
-    Binary = elixir_erl:compile(ModuleMap),
-    autoload_module(Module, Binary, CompileOpts, NE),
+        ModuleMap = #{
+          struct => get_struct(DataSet),
+          module => Module,
+          line => Line,
+          file => File,
+          relative_file => elixir_utils:relative_to_cwd(File),
+          attributes => Attributes,
+          definitions => AllDefinitions,
+          unreachable => Unreachable,
+          after_verify => AfterVerify,
+          compile_opts => CompileOpts,
+          deprecated => get_deprecated(DataBag),
+          is_behaviour => is_behaviour(DataBag)
+        },
+
+        Binary = elixir_erl:compile(ModuleMap),
+        Autoload = proplists:get_value(autoload, CompileOpts, true),
+        spawn_parallel_checker(CheckerInfo, Module, ModuleMap),
+        {Binary, PersistedAttributes, Autoload}
+      end),
+
+    Autoload andalso code:load_binary(Module, beam_location(ModuleAsCharlist), Binary),
     eval_callbacks(Line, DataBag, after_compile, [NE, Binary], NE),
+    elixir_env:trace({on_module, Binary, none}, E),
     warn_unused_attributes(File, DataSet, DataBag, PersistedAttributes),
-    make_module_available(Module, Binary, ModuleMap),
+    make_module_available(Module, Binary),
+    (CheckerInfo == undefined) andalso
+      [VerifyMod:VerifyFun(Module) ||
+       {VerifyMod, VerifyFun} <- bag_lookup_element(DataBag, {accumulate, after_verify}, 2)],
     {module, Module, Binary, Result}
   catch
     error:undef:Stacktrace ->
@@ -225,7 +240,7 @@ is_behaviour(DataBag) ->
 %% exception message suggesting the current module is not loaded. This is
 %% misleading so use a custom reason.
 compile_undef(Module, Fun, Arity, Stack) ->
-  case elixir_config:get(bootstrap) of
+  case elixir_config:is_bootstrap() of
     false ->
       Opts = [{module, Module}, {function, Fun}, {arity, Arity},
               {reason, 'function not available'}],
@@ -238,7 +253,8 @@ compile_undef(Module, Fun, Arity, Stack) ->
 %% Handle reserved modules and duplicates.
 
 check_module_availability(Line, File, Module) ->
-  Reserved = ['Elixir.Any', 'Elixir.BitString', 'Elixir.PID',
+  Reserved = ['Elixir.True', 'Elixir.False', 'Elixir.Nil',
+              'Elixir.Any', 'Elixir.BitString', 'Elixir.PID',
               'Elixir.Reference', 'Elixir.Elixir', 'Elixir'],
 
   case lists:member(Module, Reserved) of
@@ -258,12 +274,23 @@ check_module_availability(Line, File, Module) ->
       ok
   end.
 
+validate_module_name(Line, File, Module) when Module == nil; is_boolean(Module) ->
+  elixir_errors:form_error([{line, Line}], File, ?MODULE, {invalid_module, Module});
+validate_module_name(Line, File, Module) ->
+  Charlist = atom_to_list(Module),
+  case lists:any(fun(Char) -> (Char =:= $/) or (Char =:= $\\) end, Charlist) of
+    true ->
+      elixir_errors:form_error([{line, Line}], File, ?MODULE, {invalid_module, Module});
+    false ->
+      Charlist
+  end.
+
 %% Hook that builds both attribute and functions and set up common hooks.
 
 build(Line, File, Module) ->
   %% In the set table we store:
   %%
-  %% * {Attribute, Value, AccumulateOrReadOrUnreadline}
+  %% * {Attribute, Value, AccumulateOrUnsetOrReadOrUnreadline, TraceLineOrNil}
   %% * {{elixir, ...}, ...}
   %% * {{cache, ...}, ...}
   %% * {{function, Tuple}, ...}, {{macro, Tuple}, ...}
@@ -292,25 +319,26 @@ build(Line, File, Module) ->
   DataBag = ets:new(Module, [duplicate_bag, public]),
 
   ets:insert(DataSet, [
-    % {Key, Value, ReadOrUnreadLine}
-    {moduledoc, nil, nil},
+    % {Key, Value, ReadOrUnreadLine, TraceLine}
+    {moduledoc, nil, nil, []},
 
-    % {Key, Value, accumulate}
-    {after_compile, [], accumulate},
-    {before_compile, [], accumulate},
-    {behaviour, [], accumulate},
-    {compile, [], accumulate},
-    {derive, [], accumulate},
-    {dialyzer, [], accumulate},
-    {external_resource, [], accumulate},
-    {on_definition, [], accumulate},
-    {type, [], accumulate},
-    {opaque, [], accumulate},
-    {typep, [], accumulate},
-    {spec, [], accumulate},
-    {callback, [], accumulate},
-    {macrocallback, [], accumulate},
-    {optional_callbacks, [], accumulate},
+    % {Key, Value, accumulate, TraceLine}
+    {after_compile, [], accumulate, []},
+    {after_verify, [], accumulate, []},
+    {before_compile, [], accumulate, []},
+    {behaviour, [], accumulate, []},
+    {compile, [], accumulate, []},
+    {derive, [], accumulate, []},
+    {dialyzer, [], accumulate, []},
+    {external_resource, [], accumulate, []},
+    {on_definition, [], accumulate, []},
+    {type, [], accumulate, []},
+    {opaque, [], accumulate, []},
+    {typep, [], accumulate, []},
+    {spec, [], accumulate, []},
+    {callback, [], accumulate, []},
+    {macrocallback, [], accumulate, []},
+    {optional_callbacks, [], accumulate, []},
 
     % Others
     {?counter_attr, 0}
@@ -320,7 +348,7 @@ build(Line, File, Module) ->
   ets:insert(DataBag, [{persisted_attributes, Attr} || Attr <- Persisted]),
 
   OnDefinition =
-    case elixir_config:get(bootstrap) of
+    case elixir_config:is_bootstrap() of
       false -> {'Elixir.Module', compile_definition_attributes};
       _ -> {elixir_module, delete_definition_attributes}
     end,
@@ -348,9 +376,9 @@ build(Line, File, Module) ->
 %% Handles module and callback evaluations.
 
 eval_form(Line, Module, DataBag, Block, Vars, E) ->
-  {Value, EE} = elixir_compiler:eval_forms(Block, Vars, E),
+  {Value, EE} = elixir_compiler:compile(Block, Vars, E),
   elixir_overridable:store_not_overridden(Module),
-  EV = elixir_env:linify({Line, elixir_env:reset_vars(EE)}),
+  EV = (elixir_env:reset_vars(EE))#{line := Line},
   EC = eval_callbacks(Line, DataBag, before_compile, [EV], EV),
   elixir_overridable:store_not_overridden(Module),
   {Value, EC}.
@@ -358,17 +386,19 @@ eval_form(Line, Module, DataBag, Block, Vars, E) ->
 eval_callbacks(Line, DataBag, Name, Args, E) ->
   Callbacks = bag_lookup_element(DataBag, {accumulate, Name}, 2),
   lists:foldl(fun({M, F}, Acc) ->
-    expand_callback(Line, M, F, Args, elixir_env:reset_vars(Acc),
-                    fun(AM, AF, AA) -> apply(AM, AF, AA) end)
+    expand_callback(Line, M, F, Args, Acc, fun(AM, AF, AA) -> apply(AM, AF, AA) end)
   end, E, Callbacks).
 
-expand_callback(Line, M, F, Args, E, Fun) ->
+expand_callback(Line, M, F, Args, Acc, Fun) ->
+  E = elixir_env:reset_vars(Acc),
+  S = elixir_env:env_to_ex(E),
   Meta = [{line, Line}, {required, true}],
 
-  {EE, ET} = elixir_dispatch:dispatch_require(Meta, M, F, Args, E, fun(AM, AF, AA) ->
-    Fun(AM, AF, AA),
-    {ok, E}
-  end),
+  {EE, _S, ET} =
+    elixir_dispatch:dispatch_require(Meta, M, F, Args, S, E, fun(AM, AF, AA) ->
+      Fun(AM, AF, AA),
+      {ok, S, E}
+    end),
 
   if
     is_atom(EE) ->
@@ -391,9 +421,9 @@ attributes(DataSet, DataBag, PersistedAttributes) ->
 
 lookup_attribute(DataSet, DataBag, Key) when is_atom(Key) ->
   case ets:lookup(DataSet, Key) of
-    [{_, _, accumulate}] -> bag_lookup_element(DataBag, {accumulate, Key}, 2);
-    [{_, _, unset}] -> [];
-    [{_, Value, _}] -> [Value];
+    [{_, _, accumulate, _}] -> bag_lookup_element(DataBag, {accumulate, Key}, 2);
+    [{_, _, unset, _}] -> [];
+    [{_, Value, _, _}] -> [Value];
     [] -> []
   end.
 
@@ -402,19 +432,14 @@ warn_unused_attributes(File, DataSet, DataBag, PersistedAttrs) ->
   %% This is the same list as in Module.put_attribute
   %% without moduledoc which are never warned on.
   Attrs = [doc, typedoc, impl, deprecated | StoredAttrs -- PersistedAttrs],
-  Query = [{{Attr, '_', '$1'}, [{is_integer, '$1'}], [[Attr, '$1']]} || Attr <- Attrs],
+  Query = [{{Attr, '_', '$1', '_'}, [{is_integer, '$1'}], [[Attr, '$1']]} || Attr <- Attrs],
   [elixir_errors:form_warn([{line, Line}], File, ?MODULE, {unused_attribute, Key})
    || [Key, Line] <- ets:select(DataSet, Query)].
 
 get_struct(Set) ->
-  case ets:lookup(Set, '__struct__') of
+  case ets:lookup(Set, {elixir, struct}) of
     [] -> nil;
-    [{_, Struct, _}] ->
-      case ets:lookup(Set, enforce_keys) of
-        [] -> {Struct, []};
-        [{_, EnforceKeys, _}] when is_list(EnforceKeys) -> {Struct, EnforceKeys};
-        [{_, EnforceKeys, _}] -> {Struct, [EnforceKeys]}
-      end
+    [{_, Struct}] -> Struct
   end.
 
 get_deprecated(Bag) ->
@@ -427,38 +452,41 @@ bag_lookup_element(Table, Name, Pos) ->
     error:badarg -> []
   end.
 
-%% Takes care of autoloading the module if configured.
-
-autoload_module(Module, Binary, Opts, E) ->
-  case proplists:get_value(autoload, Opts, true) of
-    true  -> code:load_binary(Module, beam_location(E), Binary);
-    false -> ok
-  end.
-
-beam_location(#{module := Module}) ->
+beam_location(ModuleAsCharlist) ->
   case get(elixir_compiler_dest) of
     Dest when is_binary(Dest) ->
-      filename:join(elixir_utils:characters_to_list(Dest), atom_to_list(Module) ++ ".beam");
+      filename:join(elixir_utils:characters_to_list(Dest), ModuleAsCharlist ++ ".beam");
     _ ->
       ""
   end.
 
 %% Integration with elixir_compiler that makes the module available
 
-make_module_available(Module, Binary, ModuleMap) ->
+checker_info() ->
+  case get(elixir_checker_info) of
+    undefined -> undefined;
+    _ -> 'Elixir.Module.ParallelChecker':get()
+  end.
+
+spawn_parallel_checker(undefined, _Module, _ModuleMap) ->
+  nil;
+spawn_parallel_checker(CheckerInfo, Module, ModuleMap) ->
+  'Elixir.Module.ParallelChecker':spawn(CheckerInfo, Module, ModuleMap).
+
+make_module_available(Module, Binary) ->
   case get(elixir_module_binaries) of
     Current when is_list(Current) ->
-      put(elixir_module_binaries, [{Module, ModuleMap, Binary} | Current]);
+      put(elixir_module_binaries, [{Module, Binary} | Current]);
     _ ->
       ok
   end,
 
-  case get(elixir_compiler_pid) of
+  case get(elixir_compiler_info) of
     undefined ->
       ok;
-    PID ->
+    {PID, _} ->
       Ref = make_ref(),
-      PID ! {module_available, self(), Ref, get(elixir_compiler_file), Module, Binary, ModuleMap},
+      PID ! {module_available, self(), Ref, get(elixir_compiler_file), Module, Binary},
       receive {Ref, ack} -> ok end
   end.
 
@@ -510,7 +538,7 @@ format_error({bad_inline, {Name, Arity}}) ->
 format_error({bad_dialyzer, Key, {Name, Arity}}) ->
   io_lib:format("undefined function ~ts/~B given to @dialyzer :~ts", [Name, Arity, Key]);
 format_error({undefined_on_load, {Name, Arity}}) ->
-  io_lib:format("@on_load function ~ts/~B is undefined", [Name, Arity]);
+  io_lib:format("undefined function ~ts/~B given to @on_load", [Name, Arity]);
 format_error({wrong_kind_on_load, {Name, Arity}, WrongKind}) ->
   io_lib:format("expected @on_load function ~ts/~B to be a function, got \"~ts\"",
                 [Name, Arity, WrongKind]);
