@@ -1,6 +1,6 @@
 -module(elixir_rewrite).
 -compile({inline, [inner_inline/4, inner_rewrite/5]}).
--export([erl_to_ex/3, inline/3, rewrite/5, match_rewrite/5, guard_rewrite/5, format_error/1]).
+-export([erl_to_ex/3, inline/3, rewrite/5, match_rewrite/5, guard_rewrite/6, format_error/1]).
 -include("elixir.hrl").
 
 %% Convenience variables
@@ -51,6 +51,7 @@ erl_to_ex(Mod, Fun, Args) ->
 inline(Mod, Fun, Arity) -> inner_inline(ex_to_erl, Mod, Fun, Arity).
 
 ?inline(?atom, to_charlist, 1, erlang, atom_to_list);
+?inline(?atom, to_string, 1, erlang, atom_to_binary);
 
 ?inline(?bitwise, 'bnot', 1, erlang, 'bnot');
 ?inline(?bitwise, 'band', 2, erlang, 'band');
@@ -157,6 +158,8 @@ inline(Mod, Fun, Arity) -> inner_inline(ex_to_erl, Mod, Fun, Arity).
 ?inline(?node, spawn, 5, erlang, spawn_opt);
 ?inline(?node, spawn_link, 2, erlang, spawn_link);
 ?inline(?node, spawn_link, 4, erlang, spawn_link);
+?inline(?node, spawn_monitor, 2, erlang, spawn_monitor);
+?inline(?node, spawn_monitor, 4, erlang, spawn_monitor);
 
 ?inline(?port, close, 1, erlang, port_close);
 ?inline(?port, command, 2, erlang, port_command);
@@ -232,7 +235,6 @@ rewrite(Receiver, DotMeta, Right, Meta, Args) ->
   {EReceiver, ERight, EArgs} = inner_rewrite(ex_to_erl, DotMeta, Receiver, Right, Args),
   {{'.', DotMeta, [EReceiver, ERight]}, Meta, EArgs}.
 
-?rewrite(?atom, to_string, [Arg], erlang, atom_to_binary, [Arg, utf8]);
 ?rewrite(?kernel, is_map_key, [Map, Key], erlang, is_map_key, [Key, Map]);
 ?rewrite(?map, delete, [Map, Key], maps, remove, [Key, Map]);
 ?rewrite(?map, fetch, [Map, Key], maps, find, [Key, Map]);
@@ -276,6 +278,11 @@ inner_rewrite(erl_to_ex, _Meta, erlang, element, [{{'.', _, [erlang, '+']}, _, [
 inner_rewrite(erl_to_ex, _Meta, erlang, setelement, [{{'.', _, [erlang, '+']}, _, [Index, 1]}, Tuple, Value]) ->
   {?kernel, put_elem, [Tuple, Index, Value]};
 
+inner_rewrite(erl_to_ex, _Meta, erlang, 'orelse', [_, _] = Args) ->
+  {?kernel, 'or', Args};
+inner_rewrite(erl_to_ex, _Meta, erlang, 'andalso', [_, _] = Args) ->
+  {?kernel, 'and', Args};
+
 inner_rewrite(_To, _Meta, Mod, Fun, Args) -> {Mod, Fun, Args}.
 
 increment(_Meta, Number) when is_number(Number) ->
@@ -312,22 +319,26 @@ static_append(_, _, _) -> throw(impossible).
 %% Guard rewrite is similar to regular rewrite, except
 %% it also verifies the resulting function is supported in
 %% guard context - only certain BIFs and operators are.
-guard_rewrite(Receiver, DotMeta, Right, Meta, Args) ->
+guard_rewrite(Receiver, DotMeta, Right, Meta, Args, Context) ->
   case inner_rewrite(ex_to_erl, DotMeta, Receiver, Right, Args) of
     {erlang, RRight, RArgs} ->
       case allowed_guard(RRight, length(RArgs)) of
         true -> {ok, {{'.', DotMeta, [erlang, RRight]}, Meta, RArgs}};
-        false -> {error, {invalid_guard, Receiver, Right, length(Args)}}
+        false -> {error, {invalid_guard, Receiver, Right, length(Args), Context}}
       end;
-    _ -> {error, {invalid_guard, Receiver, Right, length(Args)}}
+    _ -> {error, {invalid_guard, Receiver, Right, length(Args), Context}}
   end.
 
+%% erlang:is_record/2-3 are compiler guards in Erlang which we
+%% need to explicitly forbid as they are allowed in erl_internal.
+allowed_guard(is_record, 2) -> false;
+allowed_guard(is_record, 3) -> false;
 allowed_guard(Right, Arity) ->
   erl_internal:guard_bif(Right, Arity) orelse elixir_utils:guard_op(Right, Arity).
 
-format_error({invalid_guard, Receiver, Right, Arity}) ->
-  io_lib:format("cannot invoke remote function ~ts.~ts/~B inside guards",
-                ['Elixir.Macro':to_string(Receiver), Right, Arity]);
+format_error({invalid_guard, Receiver, Right, Arity, Context}) ->
+  io_lib:format("cannot invoke remote function ~ts.~ts/~B inside ~ts",
+                ['Elixir.Macro':to_string(Receiver), Right, Arity, Context]);
 format_error({invalid_match, Receiver, Right, Arity}) ->
   io_lib:format("cannot invoke remote function ~ts.~ts/~B inside a match",
                 ['Elixir.Macro':to_string(Receiver), Right, Arity]);
