@@ -6,6 +6,8 @@ defmodule ApplicationTest do
   import PathHelpers
   import ExUnit.CaptureIO
 
+  @app :elixir
+
   test "application environment" do
     assert_raise ArgumentError, ~r/because the application was not loaded nor configured/, fn ->
       Application.fetch_env!(:unknown, :unknown)
@@ -75,12 +77,14 @@ defmodule ApplicationTest do
 
       assert Application.put_env(:elixir, :unknown, nested: [key: :value]) == :ok
 
+      assert compile_env(@app, :unknown, :default) == [nested: [key: :value]]
       assert compile_env(:elixir, :unknown, :default) == [nested: [key: :value]]
       assert_received {:compile_env, :elixir, [:unknown], {:ok, [nested: [key: :value]]}}
 
       assert compile_env(:elixir, :unknown) == [nested: [key: :value]]
       assert_received {:compile_env, :elixir, [:unknown], {:ok, [nested: [key: :value]]}}
 
+      assert compile_env!(@app, :unknown) == [nested: [key: :value]]
       assert compile_env!(:elixir, :unknown) == [nested: [key: :value]]
       assert_received {:compile_env, :elixir, [:unknown], {:ok, [nested: [key: :value]]}}
 
@@ -113,13 +117,23 @@ defmodule ApplicationTest do
     def trace(_, _), do: :ok
 
     defp compile_env(app, key, default \\ nil) do
-      code = quote do: Application.compile_env(unquote(app), unquote(key), unquote(default))
+      code =
+        quote do
+          require Application
+          Application.compile_env(unquote(app), unquote(key), unquote(default))
+        end
+
       {result, _binding} = Code.eval_quoted(code, [], tracers: [__MODULE__])
       result
     end
 
     defp compile_env!(app, key) do
-      code = quote do: Application.compile_env!(unquote(app), unquote(key))
+      code =
+        quote do
+          require Application
+          Application.compile_env!(unquote(app), unquote(key))
+        end
+
       {result, _binding} = Code.eval_quoted(code, [], tracers: [__MODULE__])
       result
     end
