@@ -7,6 +7,20 @@ defmodule Code.Formatter.GeneralTest do
 
   @short_length [line_length: 10]
 
+  test "does not emit warnings" do
+    assert_format "fn -> end", "fn -> nil end"
+  end
+
+  describe "unicode normalization" do
+    test "with nfc normalizations" do
+      assert_format "ç", "ç"
+    end
+
+    test "with custom normalizations" do
+      assert_format "µs", "μs"
+    end
+  end
+
   describe "aliases" do
     test "with atom-only parts" do
       assert_same "Elixir"
@@ -115,6 +129,93 @@ defmodule Code.Formatter.GeneralTest do
                   """,
                   @short_length
     end
+
+    test "with custom formatting" do
+      bad = """
+      ~W/foo  bar  baz/
+      """
+
+      good = """
+      ~W/foo bar baz/
+      """
+
+      formatter = fn content, opts ->
+        assert opts == [file: nil, line: 1, sigil: :W, modifiers: [], opening_delimiter: "/"]
+        content |> String.split(~r/ +/) |> Enum.join(" ")
+      end
+
+      assert_format bad, good, sigils: [W: formatter]
+
+      bad = """
+      var = ~W<foo  bar  baz>abc
+      """
+
+      good = """
+      var = ~W<foo bar baz>abc
+      """
+
+      formatter = fn content, opts ->
+        assert opts == [file: nil, line: 1, sigil: :W, modifiers: 'abc', opening_delimiter: "<"]
+        content |> String.split(~r/ +/) |> Enum.intersperse(" ")
+      end
+
+      assert_format bad, good, sigils: [W: formatter]
+    end
+
+    test "with custom formatting on heredocs" do
+      bad = """
+      ~W'''
+      foo  bar  baz
+      '''
+      """
+
+      good = """
+      ~W'''
+      foo bar baz
+      '''
+      """
+
+      formatter = fn content, opts ->
+        assert opts == [file: nil, line: 1, sigil: :W, modifiers: [], opening_delimiter: "'''"]
+        content |> String.split(~r/ +/) |> Enum.join(" ")
+      end
+
+      assert_format bad, good, sigils: [W: formatter]
+
+      bad = ~S'''
+      if true do
+        ~W"""
+        foo
+        bar
+        baz
+        """abc
+      end
+      '''
+
+      good = ~S'''
+      if true do
+        ~W"""
+        foo
+        bar
+        baz
+        """abc
+      end
+      '''
+
+      formatter = fn content, opts ->
+        assert opts == [
+                 file: nil,
+                 line: 2,
+                 sigil: :W,
+                 modifiers: 'abc',
+                 opening_delimiter: ~S/"""/
+               ]
+
+        content |> String.split(~r/ +/) |> Enum.join("\n")
+      end
+
+      assert_format bad, good, sigils: [W: formatter]
+    end
   end
 
   describe "anonymous functions" do
