@@ -265,6 +265,20 @@ defmodule EnumTest do
     end
   end
 
+  test "drop/2 with streams" do
+    drop_stream = fn list, count -> list |> Stream.map(& &1) |> Enum.drop(count) end
+
+    assert drop_stream.([1, 2, 3], 0) == [1, 2, 3]
+    assert drop_stream.([1, 2, 3], 1) == [2, 3]
+    assert drop_stream.([1, 2, 3], 2) == [3]
+    assert drop_stream.([1, 2, 3], 3) == []
+    assert drop_stream.([1, 2, 3], 4) == []
+    assert drop_stream.([1, 2, 3], -1) == [1, 2]
+    assert drop_stream.([1, 2, 3], -2) == [1]
+    assert drop_stream.([1, 2, 3], -4) == []
+    assert drop_stream.([], 3) == []
+  end
+
   test "drop_every/2" do
     assert Enum.drop_every([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], 2) == [2, 4, 6, 8, 10]
     assert Enum.drop_every([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], 3) == [2, 3, 5, 6, 8, 9]
@@ -737,19 +751,19 @@ defmodule EnumTest do
     # please note the order of following assertions is important
     seed1 = {1406, 407_414, 139_258}
     seed2 = {1306, 421_106, 567_597}
-    :rand.seed(:exrop, seed1)
-    assert Enum.random([1, 2]) == 2
-    assert Enum.random([1, 2]) == 2
-    :rand.seed(:exrop, seed1)
+    :rand.seed(:exsss, seed1)
+    assert Enum.random([1, 2]) == 1
     assert Enum.random([1, 2]) == 2
+    :rand.seed(:exsss, seed1)
+    assert Enum.random([1, 2]) == 1
     assert Enum.random([1, 2, 3]) == 1
-    assert Enum.random([1, 2, 3, 4]) == 1
-    assert Enum.random([1, 2, 3, 4, 5]) == 2
-    :rand.seed(:exrop, seed2)
-    assert Enum.random([1, 2]) == 2
-    assert Enum.random([1, 2, 3]) == 1
-    assert Enum.random([1, 2, 3, 4]) == 1
-    assert Enum.random([1, 2, 3, 4, 5]) == 1
+    assert Enum.random([1, 2, 3, 4]) == 2
+    assert Enum.random([1, 2, 3, 4, 5]) == 3
+    :rand.seed(:exsss, seed2)
+    assert Enum.random([1, 2]) == 1
+    assert Enum.random([1, 2, 3]) == 2
+    assert Enum.random([1, 2, 3, 4]) == 4
+    assert Enum.random([1, 2, 3, 4, 5]) == 3
   end
 
   test "reduce/2" do
@@ -803,6 +817,184 @@ defmodule EnumTest do
     assert Enum.reverse_slice([1, 2, 3], 10, 10) == [1, 2, 3]
   end
 
+  describe "slide/3" do
+    test "on an empty enum produces an empty list" do
+      for enum <- [[], %{}, 0..-1//1, MapSet.new()] do
+        assert Enum.slide(enum, 0..0, 0) == []
+        assert Enum.slide(enum, 1..1, 2) == []
+      end
+    end
+
+    test "on a single-element enumerable is the same as transforming to list" do
+      for enum <- [["foo"], [1], [%{foo: "bar"}], %{foo: :bar}, MapSet.new(["foo"]), 1..1] do
+        assert Enum.slide(enum, 0..0, 0) == Enum.to_list(enum)
+      end
+    end
+
+    test "moves a single element" do
+      for zero_to_20 <- [0..20, Enum.to_list(0..20)] do
+        expected_numbers = Enum.flat_map([0..7, [14], 8..13, 15..20], &Enum.to_list/1)
+        assert Enum.slide(zero_to_20, 14..14, 8) == expected_numbers
+      end
+
+      assert Enum.slide([:a, :b, :c, :d, :e, :f], 3..3, 2) == [:a, :b, :d, :c, :e, :f]
+      assert Enum.slide([:a, :b, :c, :d, :e, :f], 3, 3) == [:a, :b, :c, :d, :e, :f]
+    end
+
+    test "on a subsection of a list reorders the range correctly" do
+      for zero_to_20 <- [0..20, Enum.to_list(0..20)] do
+        expected_numbers = Enum.flat_map([0..7, 14..18, 8..13, 19..20], &Enum.to_list/1)
+        assert Enum.slide(zero_to_20, 14..18, 8) == expected_numbers
+      end
+
+      assert Enum.slide([:a, :b, :c, :d, :e, :f], 3..4, 2) == [:a, :b, :d, :e, :c, :f]
+    end
+
+    test "handles negative indices" do
+      make_negative_range = fn first..last, length ->
+        (first - length)..(last - length)//1
+      end
+
+      test_specs = [
+        {[], 0..0, 0},
+        {[1], 0..0, 0},
+        {[-2, 1], 1..1, 1},
+        {[4, -3, 2, -1], 3..3, 2},
+        {[-5, -3, 4, 4, 5], 0..2, 3},
+        {[0, 1, 2, 3, 4, 5, 6, 7, 8, 9], 4..7, 9},
+        {[0, 1, 2, 3, 4, 5, 6, 7, 8, 9], 4..7, 0}
+      ]
+
+      for {list, range, insertion_point} <- test_specs do
+        negative_range = make_negative_range.(range, length(list))
+
+        assert Enum.slide(list, negative_range, insertion_point) ==
+                 Enum.slide(list, range, insertion_point)
+      end
+    end
+
+    test "handles mixed positive and negative indices" do
+      for zero_to_20 <- [0..20, Enum.to_list(0..20)] do
+        assert Enum.slide(zero_to_20, -6..-1, 8) ==
+                 Enum.slide(zero_to_20, 15..20, 8)
+
+        assert Enum.slide(zero_to_20, 15..-1//1, 8) ==
+                 Enum.slide(zero_to_20, 15..20, 8)
+
+        assert Enum.slide(zero_to_20, -6..20, 8) ==
+                 Enum.slide(zero_to_20, 15..20, 8)
+
+        assert Enum.slide(zero_to_20, -100..5, 8) ==
+                 Enum.slide(zero_to_20, 0..5, 8)
+      end
+    end
+
+    test "raises an error when the step is not exactly 1" do
+      slide_ranges_that_should_fail = [2..10//2, 8..-1, 10..2//-1, 10..4//-2, -1..-8//-1]
+
+      for zero_to_20 <- [0..20, Enum.to_list(0..20)],
+          range_that_should_fail <- slide_ranges_that_should_fail do
+        assert_raise(ArgumentError, fn ->
+          Enum.slide(zero_to_20, range_that_should_fail, 1)
+        end)
+      end
+    end
+
+    test "doesn't change the order when the first and middle indices match" do
+      for zero_to_20 <- [0..20, Enum.to_list(0..20)] do
+        assert Enum.slide(zero_to_20, 8..18, 8) == Enum.to_list(0..20)
+      end
+
+      assert Enum.slide([:a, :b, :c, :d, :e, :f], 1..3, 1) == [:a, :b, :c, :d, :e, :f]
+    end
+
+    test "on the whole of an enumerable reorders it correctly" do
+      for zero_to_20 <- [0..20, Enum.to_list(0..20)] do
+        expected_numbers = Enum.flat_map([10..20, 0..9], &Enum.to_list/1)
+        assert Enum.slide(zero_to_20, 10..20, 0) == expected_numbers
+      end
+
+      assert Enum.slide([:a, :b, :c, :d, :e, :f], 4..5, 0) == [:e, :f, :a, :b, :c, :d]
+    end
+
+    test "raises when the insertion point is inside the range" do
+      for zero_to_20 <- [0..20, Enum.to_list(0..20)] do
+        assert_raise ArgumentError, fn ->
+          Enum.slide(zero_to_20, 10..18, 14)
+        end
+      end
+    end
+
+    test "accepts range starts that are off the end of the enum, returning the input list" do
+      assert Enum.slide([], 1..5, 0) == []
+
+      for zero_to_20 <- [0..20, Enum.to_list(0..20)] do
+        assert Enum.slide(zero_to_20, 21..25, 3) == Enum.to_list(0..20)
+      end
+    end
+
+    test "accepts range ends that are off the end of the enum, truncating the moved range" do
+      for zero_to_10 <- [0..10, Enum.to_list(0..10)] do
+        assert Enum.slide(zero_to_10, 8..15, 4) == Enum.slide(zero_to_10, 8..10, 4)
+      end
+    end
+
+    test "matches behavior for lists vs. ranges" do
+      range = 0..20
+      list = Enum.to_list(range)
+      # Below 32 elements, the map implementation currently sticks values in order.
+      # If ever the MapSet implementation changes, this will fail (not affecting the correctness
+      # of slide). I figured it'd be worth testing this for the time being just to have
+      # another enumerable (aside from range) testing the generic implementation.
+      set = MapSet.new(list)
+
+      test_specs = [
+        {0..0, 0},
+        {0..0, 20},
+        {11..11, 14},
+        {11..11, 3},
+        {4..8, 19},
+        {4..8, 0},
+        {4..8, 2},
+        {10..20, 0},
+        {2..1//1, -20}
+      ]
+
+      for {slide_range, insertion_point} <- test_specs do
+        slide = &Enum.slide(&1, slide_range, insertion_point)
+        assert slide.(list) == slide.(set)
+        assert slide.(list) == slide.(range)
+      end
+    end
+
+    test "inserts at negative indices" do
+      for zero_to_5 <- [0..5, Enum.to_list(0..5)] do
+        assert Enum.slide(zero_to_5, 0, -1) == [1, 2, 3, 4, 5, 0]
+        assert Enum.slide(zero_to_5, 1, -1) == [0, 2, 3, 4, 5, 1]
+        assert Enum.slide(zero_to_5, 1..2, -2) == [0, 3, 4, 1, 2, 5]
+        assert Enum.slide(zero_to_5, -5..-4//1, -2) == [0, 3, 4, 1, 2, 5]
+      end
+
+      assert Enum.slide([:a, :b, :c, :d, :e, :f], -5..-3//1, -2) ==
+               Enum.slide([:a, :b, :c, :d, :e, :f], 1..3, 4)
+    end
+
+    test "raises when insertion index would fall inside the range" do
+      for zero_to_5 <- [0..5, Enum.to_list(0..5)] do
+        assert_raise ArgumentError, fn ->
+          Enum.slide(zero_to_5, 2..3, -3)
+        end
+      end
+
+      for zero_to_10 <- [0..10, Enum.to_list(0..10)],
+          insertion_idx <- 3..5 do
+        assert_raise ArgumentError, fn ->
+          assert Enum.slide(zero_to_10, 2..5, insertion_idx)
+        end
+      end
+    end
+  end
+
   test "scan/2" do
     assert Enum.scan([1, 2, 3, 4, 5], &(&1 + &2)) == [1, 3, 6, 10, 15]
     assert Enum.scan([], &(&1 + &2)) == []
@@ -815,8 +1007,8 @@ defmodule EnumTest do
 
   test "shuffle/1" do
     # set a fixed seed so the test can be deterministic
-    :rand.seed(:exrop, {1374, 347_975, 449_264})
-    assert Enum.shuffle([1, 2, 3, 4, 5]) == [3, 1, 4, 2, 5]
+    :rand.seed(:exsss, {1374, 347_975, 449_264})
+    assert Enum.shuffle([1, 2, 3, 4, 5]) == [1, 3, 4, 5, 2]
   end
 
   test "slice/2" do
@@ -824,29 +1016,31 @@ defmodule EnumTest do
     assert Enum.slice(list, 0..0) == [1]
     assert Enum.slice(list, 0..1) == [1, 2]
     assert Enum.slice(list, 0..2) == [1, 2, 3]
-    assert Enum.slice(list, 1, 2) == [2, 3]
-    assert Enum.slice(list, 1, 0) == []
-    assert Enum.slice(list, 2, 5) == [3, 4, 5]
-    assert Enum.slice(list, 2, 6) == [3, 4, 5]
-    assert Enum.slice(list, 5, 5) == []
-    assert Enum.slice(list, 6, 5) == []
-    assert Enum.slice(list, 6, 0) == []
-    assert Enum.slice(list, -6, 0) == []
-    assert Enum.slice(list, -6, 5) == []
-    assert Enum.slice(list, -2, 5) == [4, 5]
-    assert Enum.slice(list, -3, 1) == [3]
 
-    assert_raise FunctionClauseError, fn ->
-      Enum.slice(list, 0, -1)
-    end
+    assert Enum.slice(list, 0..10//2) == [1, 3, 5]
+    assert Enum.slice(list, 0..10//3) == [1, 4]
+    assert Enum.slice(list, 0..10//4) == [1, 5]
+    assert Enum.slice(list, 0..10//5) == [1]
+    assert Enum.slice(list, 0..10//6) == [1]
 
-    assert_raise FunctionClauseError, fn ->
-      Enum.slice(list, 0.99, 0)
-    end
+    assert Enum.slice(list, 0..2//2) == [1, 3]
+    assert Enum.slice(list, 0..2//3) == [1]
 
-    assert_raise FunctionClauseError, fn ->
-      Enum.slice(list, 0, 0.99)
-    end
+    assert Enum.slice(list, 0..-1//2) == [1, 3, 5]
+    assert Enum.slice(list, 0..-1//3) == [1, 4]
+    assert Enum.slice(list, 0..-1//4) == [1, 5]
+    assert Enum.slice(list, 0..-1//5) == [1]
+    assert Enum.slice(list, 0..-1//6) == [1]
+
+    assert Enum.slice(list, 1..-1//2) == [2, 4]
+    assert Enum.slice(list, 1..-1//3) == [2, 5]
+    assert Enum.slice(list, 1..-1//4) == [2]
+    assert Enum.slice(list, 1..-1//5) == [2]
+
+    assert Enum.slice(list, -4..-1//2) == [2, 4]
+    assert Enum.slice(list, -4..-1//3) == [2, 5]
+    assert Enum.slice(list, -4..-1//4) == [2]
+    assert Enum.slice(list, -4..-1//5) == [2]
   end
 
   test "slice/3" do
@@ -862,7 +1056,7 @@ defmodule EnumTest do
     assert Enum.slice(list, 6, 5) == []
     assert Enum.slice(list, 6, 0) == []
     assert Enum.slice(list, -6, 0) == []
-    assert Enum.slice(list, -6, 5) == []
+    assert Enum.slice(list, -6, 5) == [1, 2, 3, 4, 5]
     assert Enum.slice(list, -2, 5) == [4, 5]
     assert Enum.slice(list, -3, 1) == [3]
 
@@ -884,6 +1078,28 @@ defmodule EnumTest do
     assert [1, 2, 3] |> Stream.cycle() |> Enum.slice(0, 5) == [1, 2, 3, 1, 2]
     assert [1, 2, 3] |> Stream.cycle() |> Enum.slice(0..1) == [1, 2]
     assert [1, 2, 3] |> Stream.cycle() |> Enum.slice(0..4) == [1, 2, 3, 1, 2]
+    assert [1, 2, 3] |> Stream.cycle() |> Enum.slice(0..4//2) == [1, 3, 2]
+    assert [1, 2, 3] |> Stream.cycle() |> Enum.slice(0..5//2) == [1, 3, 2]
+  end
+
+  test "slice on pruned infinite streams" do
+    assert [1, 2, 3] |> Stream.cycle() |> Stream.take(10) |> Enum.slice(0, 2) == [1, 2]
+    assert [1, 2, 3] |> Stream.cycle() |> Stream.take(10) |> Enum.slice(0, 5) == [1, 2, 3, 1, 2]
+    assert [1, 2, 3] |> Stream.cycle() |> Stream.take(10) |> Enum.slice(0..1) == [1, 2]
+    assert [1, 2, 3] |> Stream.cycle() |> Stream.take(10) |> Enum.slice(0..4) == [1, 2, 3, 1, 2]
+    assert [1, 2, 3] |> Stream.cycle() |> Stream.take(10) |> Enum.slice(0..4//2) == [1, 3, 2]
+
+    assert [1, 2, 3] |> Stream.cycle() |> Stream.take(10) |> Enum.slice(-10..-9//1) ==
+             [1, 2]
+
+    assert [1, 2, 3] |> Stream.cycle() |> Stream.take(10) |> Enum.slice(-10..-6//1) ==
+             [1, 2, 3, 1, 2]
+
+    assert [1, 2, 3] |> Stream.cycle() |> Stream.take(10) |> Enum.slice(-10..-6//2) ==
+             [1, 3, 2]
+
+    assert [1, 2, 3] |> Stream.cycle() |> Stream.take(10) |> Enum.slice(-9..-5//2) ==
+             [2, 1, 3]
   end
 
   test "sort/1" do
@@ -1150,17 +1366,17 @@ defmodule EnumTest do
     # please note the order of following assertions is important
     seed1 = {1406, 407_414, 139_258}
     seed2 = {1406, 421_106, 567_597}
-    :rand.seed(:exrop, seed1)
+    :rand.seed(:exsss, seed1)
     assert Enum.take_random([1, 2, 3], 1) == [3]
-    assert Enum.take_random([1, 2, 3], 2) == [2, 1]
-    assert Enum.take_random([1, 2, 3], 3) == [1, 2, 3]
-    assert Enum.take_random([1, 2, 3], 4) == [3, 1, 2]
-    :rand.seed(:exrop, seed2)
-    assert Enum.take_random([1, 2, 3], 1) == [1]
-    assert Enum.take_random([1, 2, 3], 2) == [1, 2]
-    assert Enum.take_random([1, 2, 3], 3) == [2, 3, 1]
+    assert Enum.take_random([1, 2, 3], 2) == [3, 2]
+    assert Enum.take_random([1, 2, 3], 3) == [3, 1, 2]
     assert Enum.take_random([1, 2, 3], 4) == [1, 3, 2]
-    assert Enum.take_random([1, 2, 3], 129) == [2, 1, 3]
+    :rand.seed(:exsss, seed2)
+    assert Enum.take_random([1, 2, 3], 1) == [1]
+    assert Enum.take_random([1, 2, 3], 2) == [3, 1]
+    assert Enum.take_random([1, 2, 3], 3) == [3, 1, 2]
+    assert Enum.take_random([1, 2, 3], 4) == [2, 1, 3]
+    assert Enum.take_random([1, 2, 3], 129) == [2, 3, 1]
 
     # assert that every item in the sample comes from the input list
     list = for _ <- 1..100, do: make_ref()
@@ -1219,6 +1435,13 @@ defmodule EnumTest do
 
     assert Enum.with_index([1, 2, 3], fn element, index -> {index, element} end) ==
              [{0, 1}, {1, 2}, {2, 3}]
+
+    assert Enum.with_index(1..0//1) == []
+    assert Enum.with_index(1..3) == [{1, 0}, {2, 1}, {3, 2}]
+    assert Enum.with_index(1..3, 10) == [{1, 10}, {2, 11}, {3, 12}]
+
+    assert Enum.with_index(1..3, fn element, index -> {index, element} end) ==
+             [{0, 1}, {1, 2}, {2, 3}]
   end
 
   test "zip/2" do
@@ -1771,17 +1994,17 @@ defmodule EnumTest.Range do
     # please note the order of following assertions is important
     seed1 = {1406, 407_414, 139_258}
     seed2 = {1306, 421_106, 567_597}
-    :rand.seed(:exrop, seed1)
-    assert Enum.random(1..2) == 2
+    :rand.seed(:exsss, seed1)
+    assert Enum.random(1..2) == 1
     assert Enum.random(1..3) == 1
     assert Enum.random(3..1) == 2
 
-    :rand.seed(:exrop, seed2)
-    assert Enum.random(1..2) == 2
-    assert Enum.random(1..3) == 1
+    :rand.seed(:exsss, seed2)
+    assert Enum.random(1..2) == 1
+    assert Enum.random(1..3) == 2
 
-    assert Enum.random(1..10//2) == 9
-    assert Enum.random(1..10//2) == 1
+    assert Enum.random(1..10//2) == 7
+    assert Enum.random(1..10//2) == 5
   end
 
   test "reduce/2" do
@@ -1845,9 +2068,9 @@ defmodule EnumTest.Range do
 
   test "shuffle/1" do
     # set a fixed seed so the test can be deterministic
-    :rand.seed(:exrop, {1374, 347_975, 449_264})
-    assert Enum.shuffle(1..5) == [3, 1, 4, 2, 5]
-    assert Enum.shuffle(1..10//2) == [3, 9, 7, 5, 1]
+    :rand.seed(:exsss, {1374, 347_975, 449_264})
+    assert Enum.shuffle(1..5) == [1, 3, 4, 5, 2]
+    assert Enum.shuffle(1..10//2) == [3, 9, 7, 1, 5]
   end
 
   test "slice/2" do
@@ -1864,11 +2087,36 @@ defmodule EnumTest.Range do
     assert Enum.slice(1..5, 6..0) == []
     assert Enum.slice(1..5, -3..0) == []
     assert Enum.slice(1..5, -3..1) == []
-    assert Enum.slice(1..5, -6..0) == []
-    assert Enum.slice(1..5, -6..5) == []
+    assert Enum.slice(1..5, -6..0) == [1]
+    assert Enum.slice(1..5, -6..5) == [1, 2, 3, 4, 5]
+    assert Enum.slice(1..5, -6..-1) == [1, 2, 3, 4, 5]
     assert Enum.slice(1..5, -5..-1) == [1, 2, 3, 4, 5]
     assert Enum.slice(1..5, -5..-3) == [1, 2, 3]
-    assert Enum.slice(1..5, -6..-1) == []
+
+    assert Enum.slice(1..5, 0..10//2) == [1, 3, 5]
+    assert Enum.slice(1..5, 0..10//3) == [1, 4]
+    assert Enum.slice(1..5, 0..10//4) == [1, 5]
+    assert Enum.slice(1..5, 0..10//5) == [1]
+    assert Enum.slice(1..5, 0..10//6) == [1]
+
+    assert Enum.slice(1..5, 0..2//2) == [1, 3]
+    assert Enum.slice(1..5, 0..2//3) == [1]
+
+    assert Enum.slice(1..5, 0..-1//2) == [1, 3, 5]
+    assert Enum.slice(1..5, 0..-1//3) == [1, 4]
+    assert Enum.slice(1..5, 0..-1//4) == [1, 5]
+    assert Enum.slice(1..5, 0..-1//5) == [1]
+    assert Enum.slice(1..5, 0..-1//6) == [1]
+
+    assert Enum.slice(1..5, 1..-1//2) == [2, 4]
+    assert Enum.slice(1..5, 1..-1//3) == [2, 5]
+    assert Enum.slice(1..5, 1..-1//4) == [2]
+    assert Enum.slice(1..5, 1..-1//5) == [2]
+
+    assert Enum.slice(1..5, -4..-1//2) == [2, 4]
+    assert Enum.slice(1..5, -4..-1//3) == [2, 5]
+    assert Enum.slice(1..5, -4..-1//4) == [2]
+    assert Enum.slice(1..5, -4..-1//5) == [2]
 
     assert Enum.slice(5..1, 0..0) == [5]
     assert Enum.slice(5..1, 0..1) == [5, 4]
@@ -1881,11 +2129,11 @@ defmodule EnumTest.Range do
     assert Enum.slice(5..1, 5..5) == []
     assert Enum.slice(5..1, 6..5) == []
     assert Enum.slice(5..1, 6..0) == []
-    assert Enum.slice(5..1, -6..0) == []
-    assert Enum.slice(5..1, -6..5) == []
+    assert Enum.slice(5..1, -6..0) == [5]
+    assert Enum.slice(5..1, -6..5) == [5, 4, 3, 2, 1]
+    assert Enum.slice(5..1, -6..-1) == [5, 4, 3, 2, 1]
     assert Enum.slice(5..1, -5..-1) == [5, 4, 3, 2, 1]
     assert Enum.slice(5..1, -5..-3) == [5, 4, 3]
-    assert Enum.slice(5..1, -6..-1) == []
 
     assert Enum.slice(1..10//2, 0..0) == [1]
     assert Enum.slice(1..10//2, 0..1) == [1, 3]
@@ -1900,15 +2148,15 @@ defmodule EnumTest.Range do
     assert Enum.slice(1..10//2, 6..0) == []
     assert Enum.slice(1..10//2, -3..0) == []
     assert Enum.slice(1..10//2, -3..1) == []
-    assert Enum.slice(1..10//2, -6..0) == []
-    assert Enum.slice(1..10//2, -6..5) == []
+    assert Enum.slice(1..10//2, -6..0) == [1]
+    assert Enum.slice(1..10//2, -6..5) == [1, 3, 5, 7, 9]
+    assert Enum.slice(1..10//2, -6..-1) == [1, 3, 5, 7, 9]
     assert Enum.slice(1..10//2, -5..-1) == [1, 3, 5, 7, 9]
     assert Enum.slice(1..10//2, -5..-3) == [1, 3, 5]
-    assert Enum.slice(1..10//2, -6..-1) == []
 
     assert_raise ArgumentError,
-                 "Enum.slice/2 does not accept ranges with custom steps, got: 1..3//2",
-                 fn -> Enum.slice(1..5, 1..3//2) end
+                 "Enum.slice/2 does not accept ranges with negative steps, got: 1..3//-2",
+                 fn -> Enum.slice(1..5, 1..3//-2) end
   end
 
   test "slice/3" do
@@ -1923,7 +2171,7 @@ defmodule EnumTest.Range do
     assert Enum.slice(1..5, 6, 5) == []
     assert Enum.slice(1..5, 6, 0) == []
     assert Enum.slice(1..5, -6, 0) == []
-    assert Enum.slice(1..5, -6, 5) == []
+    assert Enum.slice(1..5, -6, 5) == [1, 2, 3, 4, 5]
     assert Enum.slice(1..5, -2, 5) == [4, 5]
     assert Enum.slice(1..5, -3, 1) == [3]
 
@@ -1951,7 +2199,7 @@ defmodule EnumTest.Range do
     assert Enum.slice(5..1, 6, 5) == []
     assert Enum.slice(5..1, 6, 0) == []
     assert Enum.slice(5..1, -6, 0) == []
-    assert Enum.slice(5..1, -6, 5) == []
+    assert Enum.slice(5..1, -6, 5) == [5, 4, 3, 2, 1]
 
     assert Enum.slice(1..10//2, 0, 0) == []
     assert Enum.slice(1..10//2, 0, 1) == [1]
@@ -1964,7 +2212,7 @@ defmodule EnumTest.Range do
     assert Enum.slice(1..10//2, 6, 5) == []
     assert Enum.slice(1..10//2, 6, 0) == []
     assert Enum.slice(1..10//2, -6, 0) == []
-    assert Enum.slice(1..10//2, -6, 5) == []
+    assert Enum.slice(1..10//2, -6, 5) == [1, 3, 5, 7, 9]
     assert Enum.slice(1..10//2, -2, 5) == [7, 9]
     assert Enum.slice(1..10//2, -3, 1) == [5]
   end
@@ -2040,6 +2288,7 @@ defmodule EnumTest.Range do
     assert Enum.take(1..3, -2) == [2, 3]
     assert Enum.take(1..3, -4) == [1, 2, 3]
     assert Enum.take(1..0, 3) == [1, 0]
+    assert Enum.take(1..0//1, -3) == []
   end
 
   test "take_every/2" do
@@ -2064,29 +2313,29 @@ defmodule EnumTest.Range do
     # please note the order of following assertions is important
     seed1 = {1406, 407_414, 139_258}
     seed2 = {1406, 421_106, 567_597}
-    :rand.seed(:exrop, seed1)
+    :rand.seed(:exsss, seed1)
     assert Enum.take_random(1..3, 1) == [3]
-    :rand.seed(:exrop, seed1)
-    assert Enum.take_random(1..3, 2) == [3, 2]
-    :rand.seed(:exrop, seed1)
-    assert Enum.take_random(1..3, 3) == [3, 2, 1]
-    :rand.seed(:exrop, seed1)
-    assert Enum.take_random(1..3, 4) == [3, 2, 1]
-    :rand.seed(:exrop, seed1)
+    :rand.seed(:exsss, seed1)
+    assert Enum.take_random(1..3, 2) == [3, 1]
+    :rand.seed(:exsss, seed1)
+    assert Enum.take_random(1..3, 3) == [3, 1, 2]
+    :rand.seed(:exsss, seed1)
+    assert Enum.take_random(1..3, 4) == [3, 1, 2]
+    :rand.seed(:exsss, seed1)
     assert Enum.take_random(3..1, 1) == [1]
-    :rand.seed(:exrop, seed2)
+    :rand.seed(:exsss, seed2)
     assert Enum.take_random(1..3, 1) == [1]
-    :rand.seed(:exrop, seed2)
+    :rand.seed(:exsss, seed2)
     assert Enum.take_random(1..3, 2) == [1, 3]
-    :rand.seed(:exrop, seed2)
+    :rand.seed(:exsss, seed2)
     assert Enum.take_random(1..3, 3) == [1, 3, 2]
-    :rand.seed(:exrop, seed2)
+    :rand.seed(:exsss, seed2)
     assert Enum.take_random(1..3, 4) == [1, 3, 2]
 
     # make sure optimizations don't change fixed seeded tests
-    :rand.seed(:exrop, {101, 102, 103})
+    :rand.seed(:exsss, {101, 102, 103})
     one = Enum.take_random(1..100, 1)
-    :rand.seed(:exrop, {101, 102, 103})
+    :rand.seed(:exsss, {101, 102, 103})
     two = Enum.take_random(1..100, 2)
     assert hd(one) == hd(two)
   end
@@ -2149,12 +2398,12 @@ defmodule EnumTest.Map do
     map = %{a: 1, b: 2, c: 3}
     seed1 = {1406, 407_414, 139_258}
     seed2 = {1406, 421_106, 567_597}
-    :rand.seed(:exrop, seed1)
-    assert Enum.random(map) == {:a, 1}
+    :rand.seed(:exsss, seed1)
+    assert Enum.random(map) == {:c, 3}
     assert Enum.random(map) == {:a, 1}
     assert Enum.random(map) == {:b, 2}
 
-    :rand.seed(:exrop, seed2)
+    :rand.seed(:exsss, seed2)
     assert Enum.random(map) == {:c, 3}
     assert Enum.random(map) == {:b, 2}
   end
@@ -2171,21 +2420,21 @@ defmodule EnumTest.Map do
     map = %{a: 1, b: 2, c: 3}
     seed1 = {1406, 407_414, 139_258}
     seed2 = {1406, 421_106, 567_597}
-    :rand.seed(:exrop, seed1)
+    :rand.seed(:exsss, seed1)
     assert Enum.take_random(map, 1) == [c: 3]
-    :rand.seed(:exrop, seed1)
-    assert Enum.take_random(map, 2) == [c: 3, b: 2]
-    :rand.seed(:exrop, seed1)
-    assert Enum.take_random(map, 3) == [c: 3, b: 2, a: 1]
-    :rand.seed(:exrop, seed1)
-    assert Enum.take_random(map, 4) == [c: 3, b: 2, a: 1]
-    :rand.seed(:exrop, seed2)
+    :rand.seed(:exsss, seed1)
+    assert Enum.take_random(map, 2) == [c: 3, a: 1]
+    :rand.seed(:exsss, seed1)
+    assert Enum.take_random(map, 3) == [c: 3, a: 1, b: 2]
+    :rand.seed(:exsss, seed1)
+    assert Enum.take_random(map, 4) == [c: 3, a: 1, b: 2]
+    :rand.seed(:exsss, seed2)
     assert Enum.take_random(map, 1) == [a: 1]
-    :rand.seed(:exrop, seed2)
+    :rand.seed(:exsss, seed2)
     assert Enum.take_random(map, 2) == [a: 1, c: 3]
-    :rand.seed(:exrop, seed2)
+    :rand.seed(:exsss, seed2)
     assert Enum.take_random(map, 3) == [a: 1, c: 3, b: 2]
-    :rand.seed(:exrop, seed2)
+    :rand.seed(:exsss, seed2)
     assert Enum.take_random(map, 4) == [a: 1, c: 3, b: 2]
   end
 
@@ -2242,7 +2491,7 @@ defmodule EnumTest.Map do
     assert Enum.slice(map, 6, 5) == []
     assert Enum.slice(map, 6, 0) == []
     assert Enum.slice(map, -6, 0) == []
-    assert Enum.slice(map, -6, 5) == []
+    assert Enum.slice(map, -6, 5) == [a: 1, b: 2, c: 3, d: 4, e: 5]
     assert Enum.slice(map, -2, 5) == [d: 4, e: 5]
     assert Enum.slice(map, -3, 1) == [c: 3]
 
@@ -2269,7 +2518,7 @@ defmodule EnumTest.Map do
     assert Enum.slice(map, 6, 5) == []
     assert Enum.slice(map, 6, 0) == []
     assert Enum.slice(map, -6, 0) == []
-    assert Enum.slice(map, -6, 5) == []
+    assert Enum.slice(map, -6, 5) == [a: 1, b: 2, c: 3, d: 4, e: 5]
     assert Enum.slice(map, -2, 5) == [d: 4, e: 5]
     assert Enum.slice(map, -3, 1) == [c: 3]
 
