@@ -3,22 +3,12 @@ Code.require_file("test_helper.exs", __DIR__)
 defmodule ExceptionTest do
   use ExUnit.Case, async: true
 
-  doctest Exception
-
-  test "raising preserves the stacktrace" do
-    stacktrace =
-      try do
-        raise "a"
-      rescue
-        _ -> hd(__STACKTRACE__)
-      end
-
-    file = __ENV__.file |> Path.relative_to_cwd() |> String.to_charlist()
-
-    assert {__MODULE__, :"test raising preserves the stacktrace", _, [file: ^file, line: 11]} =
-             stacktrace
+  defp capture_err(fun) do
+    ExUnit.CaptureIO.capture_io(:stderr, fun)
   end
 
+  doctest Exception
+
   test "message/1" do
     defmodule BadException do
       def message(exception) do
@@ -43,6 +33,12 @@ defmodule ExceptionTest do
     assert Exception.normalize({:EXIT, self()}, :badarg, []) == :badarg
     assert Exception.normalize(:error, :badarg, []).__struct__ == ArgumentError
     assert Exception.normalize(:error, %ArgumentError{}, []).__struct__ == ArgumentError
+
+    assert %ErlangError{original: :no_translation, reason: ": foo"} =
+             Exception.normalize(:error, :no_translation, [
+               {:io, :put_chars, [self(), <<222>>],
+                [error_info: %{module: __MODULE__, function: :dummy_error_extras}]}
+             ])
   end
 
   test "format/2 without stacktrace" do
@@ -416,23 +412,93 @@ defmodule ExceptionTest do
       assert Exception.blame(:exit, :function_clause, stack) == {:function_clause, stack}
     end
 
+    test "reverts is_struct macro on guards for blaming" do
+      import PathHelpers
+
+      write_beam(
+        defmodule Req do
+          def get!(url)
+              when is_binary(url) or (is_struct(url) and is_struct(url, URI) and false) do
+            url
+          end
+
+          def get!(url, url_module)
+              when is_binary(url) or (is_struct(url) and is_struct(url, url_module) and false) do
+            url
+          end
+        end
+      )
+
+      :code.delete(Req)
+      :code.purge(Req)
+
+      assert blame_message(Req, & &1.get!(url: "https://elixir-lang.org")) =~ """
+             no function clause matching in ExceptionTest.Req.get!/1
+
+             The following arguments were given to ExceptionTest.Req.get!/1:
+
+                 # 1
+                 [url: "https://elixir-lang.org"]
+
+             Attempted function clauses (showing 1 out of 1):
+
+                 def get!(url) when -is_binary(url)- or -is_struct(url)- and -is_struct(url, URI)- and -false-
+             """
+
+      elixir_uri = %URI{} = URI.parse("https://elixir-lang.org")
+
+      assert blame_message(Req, & &1.get!(elixir_uri, URI)) =~ """
+             no function clause matching in ExceptionTest.Req.get!/2
+
+             The following arguments were given to ExceptionTest.Req.get!/2:
+
+                 # 1
+                 %URI{scheme: \"https\", authority: \"elixir-lang.org\", userinfo: nil, host: \"elixir-lang.org\", port: 443, path: nil, query: nil, fragment: nil}
+
+                 # 2
+                 URI
+
+             Attempted function clauses (showing 1 out of 1):
+
+                 def get!(url, url_module) when -is_binary(url)- or is_struct(url) and is_struct(url, url_module) and -false-
+             """
+
+      assert blame_message(Req, & &1.get!(elixir_uri)) =~ """
+             no function clause matching in ExceptionTest.Req.get!/1
+
+             The following arguments were given to ExceptionTest.Req.get!/1:
+
+                 # 1
+                 %URI{scheme: \"https\", authority: \"elixir-lang.org\", userinfo: nil, host: \"elixir-lang.org\", port: 443, path: nil, query: nil, fragment: nil}
+
+             Attempted function clauses (showing 1 out of 1):
+
+                 def get!(url) when -is_binary(url)- or is_struct(url) and is_struct(url, URI) and -false-
+             """
+    end
+
     test "annotates badarg on apply" do
-      assert blame_message([], & &1.foo) ==
-               "you attempted to apply :foo on []. If you are using apply/3, make sure " <>
+      assert blame_message([], & &1.foo()) ==
+               "you attempted to apply a function named :foo on []. If you are using Kernel.apply/3, make sure " <>
                  "the module is an atom. If you are using the dot syntax, such as " <>
-                 "map.field or module.function(), make sure the left side of the dot is an atom or a map"
+                 "module.function(), make sure the left-hand side of the dot is a module atom"
 
       assert blame_message([], &apply(&1, :foo, [])) ==
-               "you attempted to apply :foo on []. If you are using apply/3, make sure " <>
+               "you attempted to apply a function named :foo on []. If you are using Kernel.apply/3, make sure " <>
                  "the module is an atom. If you are using the dot syntax, such as " <>
-                 "map.field or module.function(), make sure the left side of the dot is an atom or a map"
+                 "module.function(), make sure the left-hand side of the dot is a module atom"
 
       assert blame_message([], &apply(Kernel, &1, [1, 2])) ==
-               "you attempted to apply [] on module Kernel. Functions (the second argument of apply) must always be an atom"
+               "you attempted to apply a function named [] on module Kernel. However [] is not a valid function name. " <>
+                 "Function names (the second argument of apply) must always be an atom"
 
       assert blame_message(123, &apply(Kernel, :+, &1)) ==
-               "you attempted to apply :+ on module Kernel with arguments 123. " <>
-                 "Arguments (the third argument of apply) must always be a list"
+               "you attempted to apply a function named :+ on module Kernel with arguments 123. " <>
+                 "Arguments (the third argument of apply) must always be a proper list"
+
+      assert blame_message(123, &apply(Kernel, :+, [&1 | 456])) ==
+               "you attempted to apply a function named :+ on module Kernel with arguments [123 | 456]. " <>
+                 "Arguments (the third argument of apply) must always be a proper list"
     end
 
     test "annotates function clause errors" do
@@ -455,13 +521,13 @@ defmodule ExceptionTest do
 
     test "annotates undefined function error with suggestions" do
       assert blame_message(Enum, & &1.map(:ok)) == """
-             function Enum.map/1 is undefined or private. Did you mean one of:
+             function Enum.map/1 is undefined or private. Did you mean:
 
                    * map/2
              """
 
       assert blame_message(Enum, & &1.man(:ok)) == """
-             function Enum.man/1 is undefined or private. Did you mean one of:
+             function Enum.man/1 is undefined or private. Did you mean:
 
                    * map/2
                    * max/1
@@ -470,12 +536,10 @@ defmodule ExceptionTest do
                    * min/1
              """
 
-      assert blame_message(:erlang, & &1.gt_cookie()) == """
-             function :erlang.gt_cookie/0 is undefined or private. Did you mean one of:
-
-                   * get_cookie/0
-                   * set_cookie/2
-             """
+      message = blame_message(:erlang, & &1.gt_cookie())
+      assert message =~ "function :erlang.gt_cookie/0 is undefined or private. Did you mean:"
+      assert message =~ "* get_cookie/0"
+      assert message =~ "* set_cookie/2"
     end
 
     test "annotates undefined function clause error with macro hints" do
@@ -485,39 +549,73 @@ defmodule ExceptionTest do
                  "you intend to invoke this macro"
     end
 
-    test "annotates undefined function clause error with callback hints hints" do
-      defmodule Behaviour do
-        @callback callback() :: :ok
-        @optional_callbacks callback: 0
-      end
+    test "annotates undefined function clause error with callback hints" do
+      capture_err(fn ->
+        Code.eval_string("""
+          defmodule Behaviour do
+            @callback callback() :: :ok
+          end
 
-      defmodule Implementation do
-        @behaviour Behaviour
-      end
+          defmodule Implementation do
+            @behaviour Behaviour
+          end
+        """)
+      end)
 
       assert blame_message(Implementation, & &1.callback()) ==
-               "function ExceptionTest.Implementation.callback/0 is undefined or private" <>
-                 ", but the behaviour ExceptionTest.Behaviour expects it to be present"
+               "function Implementation.callback/0 is undefined or private" <>
+                 ", but the behaviour Behaviour expects it to be present"
     end
 
-    if :erlang.system_info(:otp_release) >= '23' do
-      test "annotates undefined function clause error with otp obsolete hints" do
-        assert blame_message(:erlang, & &1.hash(1, 2)) ==
-                 "function :erlang.hash/2 is undefined or private, use erlang:phash2/2 instead"
+    test "does not annotate undefined function clause error with callback hints when callback is optional" do
+      defmodule BehaviourWithOptional do
+        @callback callback() :: :ok
+        @callback optional() :: :ok
+        @optional_callbacks callback: 0, optional: 0
       end
+
+      defmodule ImplementationWithOptional do
+        @behaviour BehaviourWithOptional
+        def callback(), do: :ok
+      end
+
+      assert blame_message(ImplementationWithOptional, & &1.optional()) ==
+               "function ExceptionTest.ImplementationWithOptional.optional/0 is undefined or private"
+    end
+
+    test "annotates undefined function clause error with otp obsolete hints" do
+      assert blame_message(:erlang, & &1.hash(1, 2)) ==
+               "function :erlang.hash/2 is undefined or private, use erlang:phash2/2 instead"
+    end
+
+    test "annotates undefined key error with nil hints" do
+      assert blame_message(nil, & &1.foo) ==
+               "key :foo not found in: nil. If you are using the dot syntax, " <>
+                 "such as map.field, make sure the left-hand side of the dot is a map"
+
+      # we use `Code.eval_string/1` to escape the formatter and warnings
+      assert blame_message("nil.foo", &Code.eval_string/1) ==
+               "key :foo not found in: nil. If you are using the dot syntax, " <>
+                 "such as map.field, make sure the left-hand side of the dot is a map"
     end
 
     test "annotates undefined function clause error with nil hints" do
-      assert blame_message(nil, & &1.foo) ==
+      assert blame_message(nil, & &1.foo()) ==
                "function nil.foo/0 is undefined. If you are using the dot syntax, " <>
-                 "such as map.field or module.function(), make sure the left side of the dot is an atom or a map"
+                 "such as module.function(), make sure the left-hand side of " <>
+                 "the dot is a module atom"
+
+      assert blame_message("nil.foo()", &Code.eval_string/1) ==
+               "function nil.foo/0 is undefined. If you are using the dot syntax, " <>
+                 "such as module.function(), make sure the left-hand side of " <>
+                 "the dot is a module atom"
     end
 
     test "annotates key error with suggestions if keys are atoms" do
       message = blame_message(%{first: nil, second: nil}, fn map -> map.firts end)
 
       assert message == """
-             key :firts not found in: %{first: nil, second: nil}. Did you mean one of:
+             key :firts not found in: %{first: nil, second: nil}. Did you mean:
 
                    * :first
              """
@@ -535,7 +633,7 @@ defmodule ExceptionTest do
         blame_message([first: nil, second: nil], fn kwlist -> Keyword.fetch!(kwlist, :firts) end)
 
       assert message == """
-             key :firts not found in: [first: nil, second: nil]. Did you mean one of:
+             key :firts not found in: [first: nil, second: nil]. Did you mean:
 
                    * :first
              """
@@ -544,7 +642,7 @@ defmodule ExceptionTest do
     test "annotates key error with suggestions for structs" do
       message = blame_message(%URI{}, fn map -> map.schema end)
       assert message =~ "key :schema not found in: %URI{"
-      assert message =~ "Did you mean one of:"
+      assert message =~ "Did you mean:"
       assert message =~ "* :scheme"
     end
 
@@ -567,7 +665,7 @@ defmodule ExceptionTest do
     end
 
     test "annotates band arithmetic errors" do
-      use Bitwise
+      import Bitwise
 
       assert blame_message(:foo, &band(&1, 10)) ==
                "bad argument in arithmetic expression: Bitwise.band(:foo, 10)"
@@ -577,7 +675,7 @@ defmodule ExceptionTest do
     end
 
     test "annotates bor arithmetic errors" do
-      use Bitwise
+      import Bitwise
 
       assert blame_message(:foo, &bor(&1, 10)) ==
                "bad argument in arithmetic expression: Bitwise.bor(:foo, 10)"
@@ -587,14 +685,14 @@ defmodule ExceptionTest do
     end
 
     test "annotates bxor arithmetic errors" do
-      use Bitwise
+      import Bitwise
 
       assert blame_message(:foo, &bxor(&1, 10)) ==
                "bad argument in arithmetic expression: Bitwise.bxor(:foo, 10)"
     end
 
     test "annotates bsl arithmetic errors" do
-      use Bitwise
+      import Bitwise
 
       assert blame_message(:foo, &bsl(10, &1)) ==
                "bad argument in arithmetic expression: Bitwise.bsl(10, :foo)"
@@ -604,7 +702,7 @@ defmodule ExceptionTest do
     end
 
     test "annotates bsr arithmetic errors" do
-      use Bitwise
+      import Bitwise
 
       assert blame_message(:foo, &bsr(10, &1)) ==
                "bad argument in arithmetic expression: Bitwise.bsr(10, :foo)"
@@ -614,13 +712,10 @@ defmodule ExceptionTest do
     end
 
     test "annotates bnot arithmetic errors" do
-      use Bitwise
+      import Bitwise
 
       assert blame_message(:foo, &bnot(&1)) ==
                "bad argument in arithmetic expression: Bitwise.bnot(:foo)"
-
-      assert blame_message(:foo, &(~~~&1)) ==
-               "bad argument in arithmetic expression: Bitwise.bnot(:foo)"
     end
 
     defp blame_message(arg, fun) do
@@ -684,7 +779,7 @@ defmodule ExceptionTest do
       {:ok, :def, clauses} = Exception.blame_mfa(Blaming, :even_and_odd, [1, 1])
 
       assert annotated_clauses_to_string(clauses) == [
-               "{[+foo+, +bar+], [+is_integer(foo)+ and -Bitwise.band(foo, 1) == 0- and (+is_integer(bar)+ and +Bitwise.band(bar, 1) == 1+)]}"
+               "{[+foo+, +bar+], [+is_integer(foo)+ and -Bitwise.band(foo, 1) == 0- and +is_integer(bar)+ and +Bitwise.band(bar, 1) == 1+]}"
              ]
 
       {:ok, :defmacro, clauses} = Exception.blame_mfa(Kernel, :!, [true])
@@ -696,19 +791,21 @@ defmodule ExceptionTest do
     end
 
     defp annotated_clauses_to_string(clauses) do
-      Enum.map(clauses, fn args_and_clauses ->
-        Macro.to_string(args_and_clauses, fn
-          %{match?: true, node: node}, _string ->
-            "+" <> Macro.to_string(node) <> "+"
-
-          %{match?: false, node: node}, _string ->
-            "-" <> Macro.to_string(node) <> "-"
-
-          _node, string ->
-            string
-        end)
+      Enum.map(clauses, fn {args, clauses} ->
+        args = Enum.map_join(args, ", ", &arg_to_string/1)
+        clauses = Enum.map_join(clauses, ", ", &clause_to_string/1)
+        "{[#{args}], [#{clauses}]}"
       end)
     end
+
+    defp arg_to_string(%{match?: true, node: node}), do: "+" <> Macro.to_string(node) <> "+"
+    defp arg_to_string(%{match?: false, node: node}), do: "-" <> Macro.to_string(node) <> "-"
+
+    defp clause_to_string({op, _, [left, right]}),
+      do: clause_to_string(left) <> " #{op} " <> clause_to_string(right)
+
+    defp clause_to_string(other),
+      do: arg_to_string(other)
   end
 
   describe "exception messages" do
@@ -778,7 +875,7 @@ defmodule ExceptionTest do
     end
   end
 
-  if :erlang.system_info(:otp_release) >= '24' do
+  if System.otp_release() >= "24" do
     describe "error_info" do
       test "badarg on erlang" do
         assert message(:erlang, & &1.element("foo", "bar")) == """
@@ -808,14 +905,30 @@ defmodule ExceptionTest do
                  * 1st argument: counters array size reached a system limit
                """
       end
+    end
+  end
 
-      defp message(arg, fun) do
-        try do
-          fun.(arg)
-        rescue
-          e -> Exception.message(e)
-        end
+  if System.otp_release() >= "25" do
+    describe "binary constructor error info" do
+      defp concat(a, b), do: a <> b
+
+      test "on binary concatenation" do
+        assert message(123, &concat(&1, "bar")) ==
+                 "construction of binary failed: segment 1 of type 'binary': expected a binary but got: 123"
+
+        assert message(~D[0001-02-03], &concat(&1, "bar")) ==
+                 "construction of binary failed: segment 1 of type 'binary': expected a binary but got: ~D[0001-02-03]"
       end
     end
   end
+
+  defp message(arg, fun) do
+    try do
+      fun.(arg)
+    rescue
+      e -> Exception.message(e)
+    end
+  end
+
+  def dummy_error_extras(_exception, _stacktrace), do: %{general: "foo"}
 end
