@@ -62,12 +62,23 @@ defmodule IOTest do
     assert 'F' == IO.getn(file, "λ")
     assert 'OO' == IO.getn(file, "", 2)
     assert '\n' == IO.getn(file, "λ", 99)
+    assert :eof == IO.getn(file, "λ", 1)
+    assert File.close(file) == :ok
+  end
+
+  test "getn with eof" do
+    {:ok, file} = File.open(Path.expand('fixtures/file.txt', __DIR__), [:charlist])
+    assert 'F' == IO.getn(file, "λ")
+    assert 'OO\n' == IO.getn(file, "", :eof)
+    assert :eof == IO.getn(file, "", :eof)
     assert File.close(file) == :ok
   end
 
   test "getn with UTF-8 and binary" do
     {:ok, file} = File.open(Path.expand('fixtures/utf8.txt', __DIR__), [:utf8])
     assert "Русский" == IO.getn(file, "", 7)
+    assert "\n日\n" == IO.getn(file, "", :eof)
+    assert :eof == IO.getn(file, "", :eof)
     assert File.close(file) == :ok
   end
 
@@ -85,17 +96,24 @@ defmodule IOTest do
     assert File.close(file) == :ok
   end
 
-  test "readall" do
+  test "read with all" do
     {:ok, file} = File.open(Path.expand('fixtures/file.txt', __DIR__))
     assert "FOO\n" == IO.read(file, :all)
     assert "" == IO.read(file, :all)
     assert File.close(file) == :ok
   end
 
-  test "readall with UTF-8 and binary" do
+  test "read with eof" do
+    {:ok, file} = File.open(Path.expand('fixtures/file.txt', __DIR__))
+    assert "FOO\n" == IO.read(file, :eof)
+    assert :eof == IO.read(file, :eof)
+    assert File.close(file) == :ok
+  end
+
+  test "read with eof and UTF-8 and binary" do
     {:ok, file} = File.open(Path.expand('fixtures/utf8.txt', __DIR__), [:utf8])
-    assert "Русский\n日\n" == IO.read(file, :all)
-    assert "" == IO.read(file, :all)
+    assert "Русский\n日\n" == IO.read(file, :eof)
+    assert :eof == IO.read(file, :eof)
     assert File.close(file) == :ok
   end
 
@@ -113,14 +131,21 @@ defmodule IOTest do
     assert File.close(file) == :ok
   end
 
-  test "binreadall" do
+  test "binread with all" do
     {:ok, file} = File.open(Path.expand('fixtures/utf8.txt', __DIR__))
     assert "Русский\n日\n" == IO.binread(file, :all)
     assert "" == IO.binread(file, :all)
     assert File.close(file) == :ok
   end
 
-  test "binreadline" do
+  test "binread with eof" do
+    {:ok, file} = File.open(Path.expand('fixtures/utf8.txt', __DIR__))
+    assert "Русский\n日\n" == IO.binread(file, :eof)
+    assert :eof == IO.binread(file, :eof)
+    assert File.close(file) == :ok
+  end
+
+  test "binread with line" do
     {:ok, file} = File.open(Path.expand('fixtures/utf8.txt', __DIR__))
     assert "Русский\n" == IO.binread(file, :line)
     assert "日\n" == IO.binread(file, :line)
@@ -134,27 +159,81 @@ defmodule IOTest do
     assert capture_io(fn -> IO.puts(13) end) == "13\n"
   end
 
-  test "warn with chardata" do
-    assert capture_io(:stderr, fn -> IO.warn("hello") end) =~
-             "hello\n  (ex_unit #{System.version()}) lib/ex_unit"
+  describe "warn" do
+    test "with chardata" do
+      assert capture_io(:stderr, fn -> IO.warn("hello") end) =~
+               "hello\n  (ex_unit #{System.version()}) lib/ex_unit"
 
-    assert capture_io(:stderr, fn -> IO.warn('hello') end) =~
-             "hello\n  (ex_unit #{System.version()}) lib/ex_unit"
+      assert capture_io(:stderr, fn -> IO.warn('hello') end) =~
+               "hello\n  (ex_unit #{System.version()}) lib/ex_unit"
 
-    assert capture_io(:stderr, fn -> IO.warn(:hello) end) =~
-             "hello\n  (ex_unit #{System.version()}) lib/ex_unit"
+      assert capture_io(:stderr, fn -> IO.warn(:hello) end) =~
+               "hello\n  (ex_unit #{System.version()}) lib/ex_unit"
 
-    assert capture_io(:stderr, fn -> IO.warn(13) end) =~
-             "13\n  (ex_unit #{System.version()}) lib/ex_unit"
+      assert capture_io(:stderr, fn -> IO.warn(13) end) =~
+               "13\n  (ex_unit #{System.version()}) lib/ex_unit"
+    end
 
-    assert capture_io(:stderr, fn -> IO.warn("hello", []) end) =~ "hello\n"
+    test "no stacktrace" do
+      assert capture_io(:stderr, fn -> IO.warn("hello", []) end) =~ "hello\n\n"
+    end
 
-    stacktrace = [{IEx.Evaluator, :eval, 4, [file: 'lib/iex/evaluator.ex', line: 108]}]
+    test "with stacktrace" do
+      stacktrace = [{IEx.Evaluator, :eval, 4, [file: 'lib/iex/evaluator.ex', line: 108]}]
 
-    assert capture_io(:stderr, fn -> IO.warn("hello", stacktrace) end) =~ """
-           hello
-             lib/iex/evaluator.ex:108: IEx.Evaluator.eval/4
-           """
+      assert capture_io(:stderr, fn -> IO.warn("hello", stacktrace) end) =~ """
+             hello
+               lib/iex/evaluator.ex:108: IEx.Evaluator.eval/4
+             """
+    end
+
+    test "with env" do
+      assert capture_io(:stderr, fn -> IO.warn("hello", __ENV__) end) =~ ~r"""
+             hello
+               (lib/elixir/)?test/elixir/io_test.exs:#{__ENV__.line - 2}: IOTest.\"test warn with env\"/1
+             """
+    end
+
+    test "with options" do
+      assert capture_io(:stderr, fn ->
+               IO.warn("hello", line: 13, file: "lib/foo.ex", module: Foo, function: {:bar, 1})
+             end) =~ """
+             hello
+               lib/foo.ex:13: Foo.bar/1
+             """
+
+      assert capture_io(:stderr, fn ->
+               IO.warn("hello", file: "lib/foo.ex", module: Foo, function: {:bar, 1})
+             end) =~ """
+             hello
+               lib/foo.ex: Foo.bar/1
+             """
+
+      assert capture_io(:stderr, fn -> IO.warn("hello", file: "lib/foo.ex", module: Foo) end) =~
+               """
+               hello
+                 lib/foo.ex: Foo (module)
+               """
+
+      assert capture_io(:stderr, fn -> IO.warn("hello", file: "lib/foo.ex") end) =~ """
+             hello
+               lib/foo.ex: (file)
+             """
+
+      assert capture_io(:stderr, fn ->
+               IO.warn("hello", file: "lib/foo.ex", function: {:bar, 1})
+             end) =~ """
+             hello
+               lib/foo.ex: (file)
+             """
+
+      assert capture_io(:stderr, fn ->
+               IO.warn("hello", line: 13, module: Foo, function: {:bar, 1})
+             end) =~ """
+             hello
+
+             """
+    end
   end
 
   test "write with chardata" do
