@@ -98,6 +98,19 @@ defmodule Kernel.ComprehensionTest do
     assert Process.get(:into_halt)
   end
 
+  test "nested for comprehensions with unique values" do
+    assert for(x <- [1, 1, 2], uniq: true, do: for(y <- [3, 3], uniq: true, do: x * y)) == [
+             [3],
+             [6]
+           ]
+
+    assert for(<<x <- "abcabc">>,
+             uniq: true,
+             into: "",
+             do: for(<<y <- "zz">>, uniq: true, into: "", do: to_bin(x) <> to_bin(y))
+           ) == "azbzcz"
+  end
+
   test "for comprehensions with nilly filters" do
     assert for(x <- 1..3, nilly(), do: x * 2) == []
   end
@@ -133,6 +146,19 @@ defmodule Kernel.ComprehensionTest do
     assert for({_, _} = x <- [foo: :bar], do: x) == [foo: :bar]
   end
 
+  test "for comprehensions with shadowing" do
+    assert for(
+             a <-
+               (
+                 b = 1
+                 _ = b
+                 [1]
+               ),
+             b <- [2],
+             do: a + b
+           ) == [3]
+  end
+
   test "for comprehensions with binary, enum generators and filters" do
     assert for(x <- [1, 2, 3], <<(y <- <<4, 5, 6>>)>>, y / 2 == x, do: x * y) == [8, 18]
   end
