@@ -79,7 +79,8 @@ defmodule Kernel.ExpansionTest do
 
   defp expand_with_version(expr) do
     env = :elixir_env.reset_vars(__ENV__)
-    elem(:elixir_expand.expand(expr, env), 0)
+    {expr, _, _} = :elixir_expand.expand(expr, :elixir_env.env_to_ex(env), env)
+    expr
   end
 
   describe "__block__" do
@@ -154,6 +155,13 @@ defmodule Kernel.ExpansionTest do
       assert_raise CompileError, message, fn ->
         expand(quote(do: alias(:lists, as: :"Elixir.foobar")))
       end
+
+      message =
+        ~r"alias cannot be inferred automatically for module: :lists, please use the :as option"
+
+      assert_raise CompileError, message, fn ->
+        expand(quote(do: alias(:lists)))
+      end
     end
 
     test "invalid expansion" do
@@ -230,7 +238,7 @@ defmodule Kernel.ExpansionTest do
 
     test "raises on conflicting options" do
       message =
-        ~r":only and :except can only be given together to import when :only is either :functions or :macros"
+        ~r":only and :except can only be given together to import when :only is :functions, :macros, or :sigils"
 
       assert_raise CompileError, message, fn ->
         expand(quote(do: import(Kernel, only: [], except: [])))
@@ -265,10 +273,6 @@ defmodule Kernel.ExpansionTest do
   end
 
   describe "=" do
-    test "sets context to match" do
-      assert expand(quote(do: __ENV__.context = :match)) == quote(do: :match = :match)
-    end
-
     test "defines vars" do
       {output, env} = expand_env(quote(do: a = 1), __ENV__)
       assert output == quote(do: a = 1)
@@ -307,6 +311,20 @@ defmodule Kernel.ExpansionTest do
       assert __ENV__.lexical_tracker == nil
       assert __ENV__.tracers == []
     end
+
+    test "on match" do
+      assert_raise CompileError,
+                   ~r"invalid pattern in match, __ENV__ is not allowed in matches",
+                   fn -> expand(quote(do: __ENV__ = :ok)) end
+
+      assert_raise CompileError,
+                   ~r"invalid pattern in match, __CALLER__ is not allowed in matches",
+                   fn -> expand(quote(do: __CALLER__ = :ok)) end
+
+      assert_raise CompileError,
+                   ~r"invalid pattern in match, __STACKTRACE__ is not allowed in matches",
+                   fn -> expand(quote(do: __STACKTRACE__ = :ok)) end
+    end
   end
 
   describe "vars" do
@@ -325,14 +343,13 @@ defmodule Kernel.ExpansionTest do
 
       assert expand(code)
 
-      message = ~r"expected \"a\" to expand to an existing variable or be part of a match"
+      message = ~r"undefined variable \"a\""
 
       assert_raise CompileError, message, fn ->
         expand(quote(do: var!(a)))
       end
 
-      message =
-        ~r"expected \"a\" \(context Unknown\) to expand to an existing variable or be part of a match"
+      message = ~r"undefined variable \"a\" \(context Unknown\)"
 
       assert_raise CompileError, message, fn ->
         expand(quote(do: var!(a, Unknown)))
@@ -401,7 +418,7 @@ defmodule Kernel.ExpansionTest do
     test "in guards" do
       code = quote(do: fn pid when :erlang.==(pid, self) -> pid end)
       expanded_code = quote(do: fn pid when :erlang.==(pid, :erlang.self()) -> pid end)
-      assert clean_meta(expand(code), [:import, :context]) == expanded_code
+      assert clean_meta(expand(code), [:imports, :context]) == expanded_code
 
       message = ~r"cannot find or invoke local foo/1"
 
@@ -458,16 +475,13 @@ defmodule Kernel.ExpansionTest do
 
       assert expand(ast) == ast
 
-      assert_raise CompileError,
-                   ~r"cannot use pin operator \^x inside a data structure as a map key in a pattern",
-                   fn ->
-                     expand(
-                       quote do
-                         x = 1
-                         %{{^x} => 1} = %{}
-                       end
-                     )
-                   end
+      ast =
+        quote do
+          x = 1
+          %{{^x} => 1} = %{{1} => 1}
+        end
+
+      assert expand(ast) == ast
 
       assert_raise CompileError, ~r"cannot use variable x as map key inside a pattern", fn ->
         expand(quote(do: %{x => 1} = %{}))
@@ -625,15 +639,6 @@ defmodule Kernel.ExpansionTest do
       assert expand(before_expansion) == after_expansion
     end
 
-    test "raises when not required" do
-      msg =
-        ~r"you must require Kernel\.ExpansionTarget before invoking the macro Kernel\.ExpansionTarget\.seventeen/0"
-
-      assert_raise CompileError, msg, fn ->
-        expand(quote(do: Kernel.ExpansionTarget.seventeen()))
-      end
-    end
-
     test "in matches" do
       message = ~r"cannot invoke remote function Hello.fun_that_does_not_exist/0 inside a match"
 
@@ -732,6 +737,74 @@ defmodule Kernel.ExpansionTest do
       assert expand(before_expansion) |> clean_meta([:alignment]) == after_expansion
     end
 
+    test "variables inside generator args do not leak" do
+      before_expansion =
+        quote do
+          for(
+            b <-
+              (
+                a = 1
+                [2]
+              ),
+            do: {a, b}
+          )
+
+          a
+        end
+
+      after_expansion =
+        quote do
+          for(
+            b <-
+              (
+                a = 1
+                [2]
+              ),
+            do: {a(), b}
+          )
+
+          a()
+        end
+
+      assert expand(before_expansion) == after_expansion
+
+      before_expansion =
+        quote do
+          for(
+            b <-
+              (
+                a = 1
+                [2]
+              ),
+            d <-
+              (
+                c = 3
+                [4]
+              ),
+            do: {a, b, c, d}
+          )
+        end
+
+      after_expansion =
+        quote do
+          for(
+            b <-
+              (
+                a = 1
+                [2]
+              ),
+            d <-
+              (
+                c = 3
+                [4]
+              ),
+            do: {a(), b, c(), d}
+          )
+        end
+
+      assert expand(before_expansion) == after_expansion
+    end
+
     test "variables inside filters are available in blocks" do
       assert expand(quote(do: for(a <- b, c = a, do: c))) ==
                quote(do: for(a <- b(), c = a, do: c))
@@ -847,6 +920,74 @@ defmodule Kernel.ExpansionTest do
       assert expand(before_expansion) == after_expansion
     end
 
+    test "variables inside args expression do not leak" do
+      before_expansion =
+        quote do
+          with(
+            b <-
+              (
+                a = 1
+                2
+              ),
+            do: {a, b}
+          )
+
+          a
+        end
+
+      after_expansion =
+        quote do
+          with(
+            b <-
+              (
+                a = 1
+                2
+              ),
+            do: {a(), b}
+          )
+
+          a()
+        end
+
+      assert expand(before_expansion) == after_expansion
+
+      before_expansion =
+        quote do
+          with(
+            b <-
+              (
+                a = 1
+                2
+              ),
+            d <-
+              (
+                c = 3
+                4
+              ),
+            do: {a, b, c, d}
+          )
+        end
+
+      after_expansion =
+        quote do
+          with(
+            b <-
+              (
+                a = 1
+                2
+              ),
+            d <-
+              (
+                c = 3
+                4
+              ),
+            do: {a(), b, c(), d}
+          )
+        end
+
+      assert expand(before_expansion) == after_expansion
+    end
+
     test "variables are available in do option" do
       before_expansion =
         quote do
@@ -920,10 +1061,11 @@ defmodule Kernel.ExpansionTest do
 
     test "expands remotes" do
       assert expand(quote(do: &List.flatten/2)) ==
-               quote(do: &:"Elixir.List".flatten/2) |> clean_meta([:import, :context, :no_parens])
+               quote(do: &:"Elixir.List".flatten/2)
+               |> clean_meta([:imports, :context, :no_parens])
 
       assert expand(quote(do: &Kernel.is_atom/1)) ==
-               quote(do: &:erlang.is_atom/1) |> clean_meta([:import, :context, :no_parens])
+               quote(do: &:erlang.is_atom/1) |> clean_meta([:imports, :context, :no_parens])
     end
 
     test "expands macros" do
@@ -939,25 +1081,26 @@ defmodule Kernel.ExpansionTest do
           fn -> 17 end
         end
 
-      assert expand(before_expansion) == after_expansion
+      assert clean_meta(expand(before_expansion), [:import, :context, :no_parens]) ==
+               after_expansion
     end
 
     test "fails on non-continuous" do
-      assert_raise CompileError, ~r"capture &0 is not allowed", fn ->
+      assert_raise CompileError, ~r"capture argument &0 must be numbered between 1 and 255", fn ->
         expand(quote(do: &foo(&0)))
       end
 
-      assert_raise CompileError, ~r"capture &2 cannot be defined without &1", fn ->
+      assert_raise CompileError, ~r"capture argument &2 cannot be defined without &1", fn ->
         expand(quote(do: & &2))
       end
 
-      assert_raise CompileError, ~r"capture &255 cannot be defined without &1", fn ->
+      assert_raise CompileError, ~r"capture argument &255 cannot be defined without &1", fn ->
         expand(quote(do: & &255))
       end
     end
 
     test "fails on block" do
-      message = ~r"invalid args for &, block expressions are not allowed, got: \(\n  1\n  2\n\)"
+      message = ~r"block expressions are not allowed inside the capture operator &, got: 1\n2"
 
       assert_raise CompileError, message, fn ->
         code =
@@ -973,16 +1116,13 @@ defmodule Kernel.ExpansionTest do
     end
 
     test "fails on other types" do
-      message =
-        ~r"invalid args for &, expected an expression in the format of &Mod.fun/arity, &local/arity or a capture containing at least one argument as &1, got: :foo"
-
-      assert_raise CompileError, message, fn ->
+      assert_raise CompileError, ~r"invalid args for &, expected one of:", fn ->
         expand(quote(do: &:foo))
       end
     end
 
     test "fails on invalid arity" do
-      message = ~r"invalid arity for &, expected a number between 0 and 255, got: 256"
+      message = ~r"capture argument &256 must be numbered between 1 and 255"
 
       assert_raise CompileError, message, fn ->
         expand(quote(do: &Mod.fun/256))
@@ -990,24 +1130,21 @@ defmodule Kernel.ExpansionTest do
     end
 
     test "fails when no captures" do
-      message =
-        ~r"invalid args for &, expected an expression in the format of &Mod.fun/arity, &local/arity or a capture containing at least one argument as &1, got: foo()"
-
-      assert_raise CompileError, message, fn ->
+      assert_raise CompileError, ~r"invalid args for &, expected one of:", fn ->
         expand(quote(do: &foo()))
       end
     end
 
     test "fails on nested capture" do
-      assert_raise CompileError, ~r"nested captures via & are not allowed: &\(&1\)", fn ->
+      assert_raise CompileError, ~r"nested captures are not allowed", fn ->
         expand(quote(do: &(& &1)))
       end
     end
 
     test "fails on integers" do
-      assert_raise CompileError, ~r"unhandled &1 outside of a capture", fn ->
-        expand(quote(do: &1))
-      end
+      assert_raise CompileError,
+                   ~r"capture argument &1 must be used within the capture operator &",
+                   fn -> expand(quote(do: &1)) end
     end
   end
 
@@ -2170,7 +2307,7 @@ defmodule Kernel.ExpansionTest do
     end
 
     test "expands size * unit" do
-      import Kernel, except: [-: 2]
+      import Kernel, except: [-: 1, -: 2]
       import Kernel.ExpansionTarget
 
       assert expand(quote(do: <<x::13>>)) |> clean_meta([:alignment]) ==
@@ -2202,7 +2339,7 @@ defmodule Kernel.ExpansionTest do
     end
 
     test "expands binary/bitstring specifiers" do
-      import Kernel, except: [-: 2]
+      import Kernel, except: [-: 1, -: 2]
 
       assert expand(quote(do: <<x::binary>>)) |> clean_meta([:alignment]) ==
                quote(do: <<x()::binary()>>)
@@ -2227,7 +2364,7 @@ defmodule Kernel.ExpansionTest do
     end
 
     test "expands utf* specifiers" do
-      import Kernel, except: [-: 2]
+      import Kernel, except: [-: 1, -: 2]
 
       assert expand(quote(do: <<x::utf8>>)) |> clean_meta([:alignment]) ==
                quote(do: <<x()::utf8()>>)
@@ -2250,7 +2387,7 @@ defmodule Kernel.ExpansionTest do
     end
 
     test "expands numbers specifiers" do
-      import Kernel, except: [-: 2]
+      import Kernel, except: [-: 1, -: 2]
 
       assert expand(quote(do: <<x::integer>>)) |> clean_meta([:alignment]) ==
                quote(do: <<x()::integer()>>)
@@ -2276,7 +2413,7 @@ defmodule Kernel.ExpansionTest do
     end
 
     test "expands macro specifiers" do
-      import Kernel, except: [-: 2]
+      import Kernel, except: [-: 1, -: 2]
       import Kernel.ExpansionTarget
 
       assert expand(quote(do: <<x::seventeen>>)) |> clean_meta([:alignment]) ==
@@ -2288,7 +2425,7 @@ defmodule Kernel.ExpansionTest do
     end
 
     test "expands macro in args" do
-      import Kernel, except: [-: 2]
+      import Kernel, except: [-: 1, -: 2]
 
       before_expansion =
         quote do
@@ -2306,7 +2443,7 @@ defmodule Kernel.ExpansionTest do
     end
 
     test "supports dynamic size" do
-      import Kernel, except: [-: 2]
+      import Kernel, except: [-: 1, -: 2]
 
       before_expansion =
         quote do
@@ -2335,7 +2472,7 @@ defmodule Kernel.ExpansionTest do
     end
 
     test "merges bitstrings" do
-      import Kernel, except: [-: 2]
+      import Kernel, except: [-: 1, -: 2]
 
       assert expand(quote(do: <<x, <<y::signed-native>>, z>>)) |> clean_meta([:alignment]) ==
                quote(do: <<x()::integer(), y()::integer()-native()-signed(), z()::integer()>>)
@@ -2346,7 +2483,7 @@ defmodule Kernel.ExpansionTest do
     end
 
     test "merges binaries" do
-      import Kernel, except: [-: 2]
+      import Kernel, except: [-: 1, -: 2]
 
       assert expand(quote(do: "foo" <> x)) |> clean_meta([:alignment]) ==
                quote(do: <<"foo"::binary(), x()::binary()>>)
@@ -2359,6 +2496,52 @@ defmodule Kernel.ExpansionTest do
                quote(do: <<"foo"::binary(), x()::integer()-size(4), y()::integer()-size(4)>>)
     end
 
+    test "guard expressions on size" do
+      import Kernel, except: [-: 1, -: 2, +: 1, +: 2, length: 1]
+
+      # Arithmetic operations with literals and variables are valid expressions
+      # for bitstring size in OTP 23+
+
+      before_expansion =
+        quote do
+          var = 1
+          <<x::size(var + 3)>>
+        end
+
+      after_expansion =
+        quote do
+          var = 1
+          <<x()::integer()-size(:erlang.+(var, 3))>>
+        end
+
+      assert expand(before_expansion) |> clean_meta([:alignment]) == after_expansion
+
+      # Other valid guard expressions are also legal for bitstring size in OTP 23+
+
+      before_expansion = quote(do: <<x::size(length('test'))>>)
+      after_expansion = quote(do: <<x()::integer()-size(:erlang.length('test'))>>)
+
+      assert expand(before_expansion) |> clean_meta([:alignment]) == after_expansion
+    end
+
+    test "map lookup on size" do
+      import Kernel, except: [-: 1, -: 2]
+
+      before_expansion =
+        quote do
+          var = %{foo: 3}
+          <<x::size(var.foo)>>
+        end
+
+      after_expansion =
+        quote do
+          var = %{foo: 3}
+          <<x()::integer()-size(var.foo)>>
+        end
+
+      assert expand(before_expansion) |> clean_meta([:alignment]) == after_expansion
+    end
+
     test "raises on unaligned binaries in match" do
       message = ~r"its number of bits is not divisible by 8"
 
@@ -2387,10 +2570,10 @@ defmodule Kernel.ExpansionTest do
       end
     end
 
-    # TODO: Simplify when we require OTP 24
+    # TODO: Simplify when we require Erlang/OTP 24
     if System.otp_release() >= "24" do
       test "16-bit floats" do
-        import Kernel, except: [-: 2]
+        import Kernel, except: [-: 1, -: 2]
 
         assert expand(quote(do: <<12.3::float-16>>)) |> clean_meta([:alignment]) ==
                  quote(do: <<12.3::float()-size(16)>>)
@@ -2434,12 +2617,6 @@ defmodule Kernel.ExpansionTest do
     end
 
     test "raises for invalid size" do
-      message = ~r"size in bitstring expects an integer or a variable as argument, got: :oops"
-
-      assert_raise CompileError, message, fn ->
-        expand(quote(do: <<"foo"::size(:oops)>>))
-      end
-
       assert_raise CompileError, ~r/undefined variable "foo"/, fn ->
         code =
           quote do
@@ -2458,7 +2635,7 @@ defmodule Kernel.ExpansionTest do
         expand(code)
       end
 
-      assert_raise CompileError, ~r/undefined variable "foo" in bitstring segment/, fn ->
+      assert_raise CompileError, ~r/undefined variable "foo"/, fn ->
         code =
           quote do
             fn foo, <<_::size(foo)>> -> :ok end
@@ -2467,7 +2644,16 @@ defmodule Kernel.ExpansionTest do
         expand(code)
       end
 
-      message = ~r"size in bitstring expects an integer or a variable as argument, got: foo()"
+      assert_raise CompileError, ~r/undefined variable "foo"/, fn ->
+        code =
+          quote do
+            fn foo, <<_::size(foo + 1)>> -> :ok end
+          end
+
+        expand(code)
+      end
+
+      message = ~r"cannot find or invoke local foo/0 inside bitstring size specifier"
 
       assert_raise CompileError, message, fn ->
         code =
@@ -2477,6 +2663,40 @@ defmodule Kernel.ExpansionTest do
 
         expand(code)
       end
+
+      message = ~r"anonymous call is not allowed in bitstring size specifier"
+
+      assert_raise CompileError, message, fn ->
+        code =
+          quote do
+            fn <<_::size(foo.())>> -> :ok end
+          end
+
+        expand(code)
+      end
+
+      message = ~r"cannot invoke remote function in bitstring size specifier"
+
+      assert_raise CompileError, message, fn ->
+        code =
+          quote do
+            foo = %{bar: true}
+            fn <<_::size(foo.bar())>> -> :ok end
+          end
+
+        expand(code)
+      end
+
+      message = ~r"cannot invoke remote function Foo.bar/0 inside bitstring size specifier"
+
+      assert_raise CompileError, message, fn ->
+        code =
+          quote do
+            fn <<_::size(Foo.bar())>> -> :ok end
+          end
+
+        expand(code)
+      end
     end
 
     test "raises for invalid unit" do
@@ -2538,10 +2758,10 @@ defmodule Kernel.ExpansionTest do
 
   describe "op ambiguity" do
     test "raises when a call is ambiguous" do
+      # We use string_to_quoted! here to avoid the formatter adding parentheses
       message = ~r["a -1" looks like a function call but there is a variable named "a"]
 
       assert_raise CompileError, message, fn ->
-        # We use string_to_quoted! here to avoid the formatter adding parentheses to "a -1".
         code =
           Code.string_to_quoted!("""
           a = 1
@@ -2550,6 +2770,19 @@ defmodule Kernel.ExpansionTest do
 
         expand(code)
       end
+
+      message =
+        ~r["a -1\.\.\(a \+ 1\)" looks like a function call but there is a variable named "a"]
+
+      assert_raise CompileError, message, fn ->
+        code =
+          Code.string_to_quoted!("""
+          a = 1
+          a -1 .. a + 1
+          """)
+
+        expand(code)
+      end
     end
   end
 
@@ -2575,7 +2808,7 @@ defmodule Kernel.ExpansionTest do
       expand(code)
     end
 
-    assert_raise CompileError, ~r"invalid expression in guard", fn ->
+    assert_raise CompileError, ~r"anonymous call is not allowed in guards", fn ->
       code =
         quote do
           x = & &1
@@ -2659,11 +2892,13 @@ defmodule Kernel.ExpansionTest do
 
   defp expand_env(expr, env) do
     ExUnit.CaptureIO.capture_io(:stderr, fn ->
-      send(self(), {:expand_env, :elixir_expand.expand(expr, env)})
+      send(self(), {:expand_env, :elixir_expand.expand(expr, :elixir_env.env_to_ex(env), env)})
     end)
 
     receive do
-      {:expand_env, {expr, env}} -> {clean_meta(expr, [:version, :inferred_bitstring_spec]), env}
+      {:expand_env, {expr, scope, env}} ->
+        env = :elixir_env.to_caller({env.line, scope, env})
+        {clean_meta(expr, [:version, :inferred_bitstring_spec]), env}
     end
   end
 end
