@@ -143,6 +143,43 @@ defmodule Kernel.ImportTest do
     assert flatten([1, [2], 3]) == [1, 2, 3]
   end
 
+  defmodule ModuleWithSigils do
+    def sigil_i(string, []), do: String.to_integer(string)
+
+    defmacro sigil_I(string, []) do
+      quote do
+        String.to_integer(unquote(string))
+      end
+    end
+
+    def sigil_w(_string, []), do: []
+
+    def bnot(x), do: x
+    defmacro bor(x, _), do: x
+  end
+
+  test "import only sigils" do
+    import Kernel, except: [sigil_w: 2]
+    import ModuleWithSigils, only: :sigils
+
+    # Ensure that both function and macro sigils are imported
+    assert ~i'10' == 10
+    assert ~I'10' == 10
+    assert ~w(abc def) == []
+
+    # Ensure that non-sigil functions and macros from ModuleWithSigils were not loaded
+    assert bnot(0) == -1
+    assert bor(0, 1) == 1
+  end
+
+  test "import only sigils with except" do
+    import ModuleWithSigils, only: :sigils, except: [sigil_w: 2]
+
+    assert ~i'10' == 10
+    assert ~I'10' == 10
+    assert ~w(abc def) == ["abc", "def"]
+  end
+
   test "import only removes the non-import part" do
     import List
     import List, only: :macros
@@ -174,6 +211,28 @@ defmodule Kernel.ImportTest do
     end
   end
 
+  test "import lexical on for" do
+    for x <- [1, 2, 3], x > 10 do
+      import List
+      flatten([1, [2], 3])
+      flunk()
+    end
+
+    # Buggy local duplicate is untouched
+    assert duplicate([1], 2) == [1]
+  end
+
+  test "import lexical on with" do
+    with true <- false do
+      import List
+      flatten([1, [2], 3])
+      flunk()
+    end
+
+    # Buggy local duplicate is untouched
+    assert duplicate([1], 2) == [1]
+  end
+
   test "import lexical on try" do
     try do
       import List
