@@ -3,6 +3,18 @@ Code.require_file("../test_helper.exs", __DIR__)
 defmodule Kernel.ParserTest do
   use ExUnit.Case, async: true
 
+  describe "nullary ops" do
+    test "in expressions" do
+      assert parse!("..") == {:.., [line: 1], []}
+    end
+
+    test "raises on ambiguous uses" do
+      assert_raise SyntaxError, ~r/syntax error before: do/, fn ->
+        parse!("if .. do end")
+      end
+    end
+  end
+
   describe "unary ops" do
     test "in keywords" do
       assert parse!("f(!: :ok)") == {:f, [line: 1], [[!: :ok]]}
@@ -55,6 +67,30 @@ defmodule Kernel.ParserTest do
     end
   end
 
+  describe "identifier unicode normalization" do
+    test "stops at ascii codepoints" do
+      assert {:ok, {:ç, _, nil}} = Code.string_to_quoted("ç\n")
+      assert {:ok, {:\\, _, [{:ç, _, nil}, 1]}} = Code.string_to_quoted(~S"ç\\1")
+    end
+
+    test "nfc normalization is performed" do
+      # before elixir 1.14, non-nfc would error
+      #  non-nfc:        "ç" (code points 0x0063 0x0327)
+      #  nfc-normalized: "ç" (code points 0x00E7)
+      assert Code.eval_string("ç = 1; ç") == {1, [ç: 1]}
+    end
+
+    test "elixir's additional normalization is performed" do
+      # Common micro => Greek mu. See code formatter test too.
+      assert Code.eval_string("µs = 1; μs") == {1, [{:μs, 1}]}
+
+      # commented out: math symbols capability in elixir
+      # normalizations, to ensure that we *can* handle codepoints
+      # that are Common-script and non-ASCII
+      # assert Code.eval_string("_ℕ𝕩 = 1") == {1, [{:"_ℕ𝕩", 1}]}
+    end
+  end
+
   describe "strings/sigils" do
     test "delimiter information for sigils is included" do
       string_to_quoted = &Code.string_to_quoted!(&1, token_metadata: false)
@@ -131,7 +167,6 @@ defmodule Kernel.ParserTest do
         assert atom == "there_is_no_such_atom"
         assert meta[:line] == 1
         assert meta[:column] == 1
-        assert meta[:file] == "nofile"
         {:ok, {:my, "atom", ref}}
       end
 
@@ -146,7 +181,6 @@ defmodule Kernel.ParserTest do
         assert atom == "there_is_no_such_var"
         assert meta[:line] == 1
         assert meta[:column] == 1
-        assert meta[:file] == "nofile"
         {:ok, {:my, "atom", ref}}
       end
 
@@ -154,6 +188,22 @@ defmodule Kernel.ParserTest do
                Code.string_to_quoted("there_is_no_such_var", static_atoms_encoder: encoder)
     end
 
+    test "encodes quoted keyword keys" do
+      ref = make_ref()
+
+      encoder = fn atom, meta ->
+        assert atom == "there is no such key"
+        assert meta[:line] == 1
+        assert meta[:column] == 2
+        {:ok, {:my, "atom", ref}}
+      end
+
+      assert {:ok, [{{:my, "atom", ^ref}, true}]} =
+               Code.string_to_quoted(~S(["there is no such key": true]),
+                 static_atoms_encoder: encoder
+               )
+    end
+
     test "addresses ambiguities" do
       encoder = fn string, _meta -> {:ok, {:atom, string}} end
 
@@ -224,6 +274,17 @@ defmodule Kernel.ParserTest do
       foo = {:foo, [line: 1, column: 1], nil}
       bar = {:bar, [line: 1, column: 7], nil}
       assert string_to_quoted.("foo + bar") == {:ok, {:+, [line: 1, column: 5], [foo, bar]}}
+
+      nfc_abba = [225, 98, 98, 224]
+      nfd_abba = [97, 769, 98, 98, 97, 768]
+      context = [line: 1, column: 8]
+      expr = "'ábbà' = 1"
+
+      assert string_to_quoted.(String.normalize(expr, :nfc)) ==
+               {:ok, {:=, context, [nfc_abba, 1]}}
+
+      assert string_to_quoted.(String.normalize(expr, :nfd)) ==
+               {:ok, {:=, context, [nfd_abba, 1]}}
     end
   end
 
@@ -322,10 +383,10 @@ defmodule Kernel.ParserTest do
                 [[{:__block__, [token: "1", line: 1], [1]}]]}
 
       assert string_to_quoted.(~s("""\nhello\n""")) ==
-               {:__block__, [delimiter: ~s["""], line: 1], ["hello\n"]}
+               {:__block__, [delimiter: ~s["""], indentation: 0, line: 1], ["hello\n"]}
 
       assert string_to_quoted.("'''\nhello\n'''") ==
-               {:__block__, [delimiter: ~s['''], line: 1], ['hello\n']}
+               {:__block__, [delimiter: ~s['''], indentation: 0, line: 1], ['hello\n']}
 
       assert string_to_quoted.(~s[fn (1) -> "hello" end]) ==
                {:fn, [closing: [line: 1], line: 1],
@@ -337,64 +398,101 @@ defmodule Kernel.ParserTest do
                    ]}
                 ]}
     end
+
+    test "adds identifier_location for qualified identifiers" do
+      string_to_quoted = &Code.string_to_quoted!(&1, token_metadata: true, columns: true)
+
+      assert string_to_quoted.("foo.\nbar") ==
+               {{:., [line: 1, column: 4],
+                 [
+                   {:foo, [line: 1, column: 1], nil},
+                   :bar
+                 ]}, [no_parens: true, line: 2, column: 1], []}
+
+      assert string_to_quoted.("foo\n.\nbar") ==
+               {{:., [line: 2, column: 1],
+                 [
+                   {:foo, [line: 1, column: 1], nil},
+                   :bar
+                 ]}, [no_parens: true, line: 3, column: 1], []}
+
+      assert string_to_quoted.(~s[Foo.\nbar(1)]) ==
+               {{:., [line: 1, column: 4],
+                 [
+                   {:__aliases__, [last: [line: 1, column: 1], line: 1, column: 1], [:Foo]},
+                   :bar
+                 ]}, [closing: [line: 2, column: 6], line: 2, column: 1], [1]}
+    end
+
+    test "adds metadata for the last alias segment" do
+      string_to_quoted = &Code.string_to_quoted!(&1, token_metadata: true)
+
+      assert string_to_quoted.("Foo") == {:__aliases__, [last: [line: 1], line: 1], [:Foo]}
+
+      assert string_to_quoted.("Foo.\nBar\n.\nBaz") ==
+               {:__aliases__, [last: [line: 4], line: 1], [:Foo, :Bar, :Baz]}
+
+      assert string_to_quoted.("foo.\nBar\n.\nBaz") ==
+               {:__aliases__, [last: [line: 4], line: 1], [{:foo, [line: 1], nil}, :Bar, :Baz]}
+    end
   end
 
   describe "token missing errors" do
     test "missing paren" do
       assert_token_missing(
-        "nofile:1:9: missing terminator: ) (for \"(\" starting at line 1)",
+        ~r/nofile:1:9: missing terminator: \) \(for \"\(\" starting at line 1\)/,
         'case 1 ('
       )
     end
 
     test "dot terminator" do
       assert_token_missing(
-        "nofile:1:9: missing terminator: \" (for function name starting at line 1)",
+        ~r/nofile:1:9: missing terminator: \" \(for function name starting at line 1\)/,
         'foo."bar'
       )
     end
 
     test "sigil terminator" do
       assert_token_missing(
-        "nofile:3:1: missing terminator: \" (for sigil ~r\" starting at line 1)",
+        ~r/nofile:3:1: missing terminator: " \(for sigil ~r" starting at line 1\)/,
         '~r"foo\n\n'
       )
 
       assert_token_missing(
-        "nofile:3:1: missing terminator: } (for sigil ~r{ starting at line 1)",
+        ~r/nofile:3:1: missing terminator: } \(for sigil ~r{ starting at line 1\)/,
         '~r{foo\n\n'
       )
     end
 
     test "string terminator" do
       assert_token_missing(
-        "nofile:1:5: missing terminator: \" (for string starting at line 1)",
+        ~r/nofile:1:5: missing terminator: \" \(for string starting at line 1\)/,
         '"bar'
       )
     end
 
     test "heredoc with incomplete interpolation" do
       assert_token_missing(
-        "nofile:2:1: missing interpolation terminator: \"}\" (for heredoc starting at line 1)",
+        ~r/nofile:2:1: missing interpolation terminator: \"}\" \(for heredoc starting at line 1\)/,
         '"""\n\#{\n'
       )
     end
 
     test "heredoc terminator" do
       assert_token_missing(
-        "nofile:2:4: missing terminator: \"\"\" (for heredoc starting at line 1)",
+        ~r/nofile:2:4: missing terminator: \"\"\" \(for heredoc starting at line 1\)/,
         '"""\nbar'
       )
 
       assert_token_missing(
-        "nofile:2:7: missing terminator: \"\"\" (for heredoc starting at line 1)",
+        ~r/nofile:2:7: missing terminator: \"\"\" \(for heredoc starting at line 1\)/,
         '"""\nbar"""'
       )
     end
 
     test "missing end" do
       assert_token_missing(
-        "nofile:1:9: missing terminator: end (for \"do\" starting at line 1)",
+        ~r/nofile:1:9: missing terminator: end \(for \"do\" starting at line 1\)/,
         'foo do 1'
       )
 
@@ -416,69 +514,101 @@ defmodule Kernel.ParserTest do
   describe "syntax errors" do
     test "invalid heredoc start" do
       assert_syntax_error(
-        "nofile:1:1: heredoc allows only zero or more whitespace characters followed by a new line after \"\"\"",
+        ~r/nofile:1:1: heredoc allows only zero or more whitespace characters followed by a new line after \"\"\"/,
         '"""bar\n"""'
       )
     end
 
     test "invalid fn" do
       assert_syntax_error(
-        "nofile:1: expected anonymous functions to be defined with -> inside: 'fn'",
+        ~r/nofile:1:1: expected anonymous functions to be defined with -> inside: 'fn'/,
         'fn 1 end'
       )
 
       assert_syntax_error(
-        ~r"nofile:2: unexpected operator ->. If you want to define multiple clauses, ",
+        ~r/nofile:2: unexpected operator ->. If you want to define multiple clauses,/,
         'fn 1\n2 -> 3 end'
       )
     end
 
     test "invalid token" do
       assert_syntax_error(
-        "nofile:1:7: unexpected token: \"\u200B\" (column 7, code point U+200B)",
+        ~r/nofile:1:1: unexpected token: "#{"\u3164"}" \(column 1, code point U\+3164\)/,
+        'ㅤ = 1'
+      )
+
+      assert_syntax_error(
+        ~r/nofile:1:7: unexpected token: "#{"\u200B"}" \(column 7, code point U\+200B\)/,
         '[foo: \u200B]\noops'
       )
+
+      assert_syntax_error(
+        ~r/nofile:1:1: unexpected token: carriage return \(column 1, code point U\+000D\)/,
+        '\r'
+      )
+    end
+
+    test "invalid bidi in source" do
+      assert_syntax_error(
+        ~r"nofile:1:1: invalid bidirectional formatting character in comment: \\u202A",
+        '# This is a \u202A'
+      )
+
+      assert_syntax_error(
+        ~r"nofile:1:5: invalid bidirectional formatting character in comment: \\u202A",
+        'foo. # This is a \u202A'
+      )
+
+      assert_syntax_error(
+        ~r"nofile:1:12: invalid bidirectional formatting character in string: \\u202A. If you want to use such character, use it in its escaped \\u202A form instead",
+        '"this is a \u202A"'
+      )
+
+      assert_syntax_error(
+        ~r"nofile:1:13: invalid bidirectional formatting character in string: \\u202A. If you want to use such character, use it in its escaped \\u202A form instead",
+        '"this is a \\\u202A"'
+      )
     end
 
     test "reserved tokens" do
-      assert_syntax_error("nofile:1:1: reserved token: __aliases__", '__aliases__')
-      assert_syntax_error("nofile:1:1: reserved token: __block__", '__block__')
+      assert_syntax_error(~r/nofile:1:1: reserved token: __aliases__/, '__aliases__')
+      assert_syntax_error(~r/nofile:1:1: reserved token: __block__/, '__block__')
     end
 
     test "invalid alias terminator" do
-      assert_syntax_error(~r"nofile:1:5: unexpected \( after alias Foo", 'Foo()')
+      assert_syntax_error(~r/nofile:1:5: unexpected \( after alias Foo/, 'Foo()')
     end
 
     test "invalid quoted token" do
       assert_syntax_error(
-        "nofile:1:9: syntax error before: \"world\"",
+        ~r/nofile:1:9: syntax error before: \"world\"/,
         '"hello" "world"'
       )
 
       assert_syntax_error(
-        "nofile:1:3: syntax error before: 'Foobar'",
+        ~r/nofile:1:3: syntax error before: 'Foobar'/,
         '1 Foobar'
       )
 
       assert_syntax_error(
-        "nofile:1:5: syntax error before: foo",
+        ~r/nofile:1:5: syntax error before: foo/,
         'Foo.:foo'
       )
 
       assert_syntax_error(
-        "nofile:1:5: syntax error before: \"foo\"",
+        ~r/nofile:1:5: syntax error before: \"foo\"/,
         'Foo.:"foo\#{:bar}"'
       )
 
       assert_syntax_error(
-        "nofile:1:5: syntax error before: \"",
+        ~r/nofile:1:5: syntax error before: \"/,
         'Foo.:"\#{:bar}"'
       )
     end
 
     test "invalid identifier" do
       message = fn name ->
-        "nofile:1:1: invalid character \"@\" (code point U+0040) in identifier: #{name}"
+        ~r/nofile:1:1: invalid character "@" \(code point U\+0040\) in identifier: #{name}/
       end
 
       assert_syntax_error(message.("foo@"), 'foo@')
@@ -486,50 +616,116 @@ defmodule Kernel.ParserTest do
       assert_syntax_error(message.("foo@bar"), 'foo@bar')
 
       message = fn name ->
-        "nofile:1:1: invalid character \"@\" (code point U+0040) in alias: #{name}"
+        ~r/nofile:1:1: invalid character "@" \(code point U\+0040\) in alias: #{name}/
       end
 
       assert_syntax_error(message.("Foo@"), 'Foo@')
       assert_syntax_error(message.("Foo@bar"), 'Foo@bar')
 
-      message = "nofile:1:1: invalid character \"!\" (code point U+0021) in alias: Foo!"
+      message =
+        ~r/nofile:1:1: invalid character "\!" \(code point U\+0021\) in alias \(only ASCII characters, without punctuation, are allowed\): Foo\!/
+
       assert_syntax_error(message, 'Foo!')
 
-      message = "nofile:1:1: invalid character \"?\" (code point U+003F) in alias: Foo?"
+      message =
+        ~r/nofile:1:1: invalid character "\?" \(code point U\+003F\) in alias \(only ASCII characters, without punctuation, are allowed\): Foo\?/
+
       assert_syntax_error(message, 'Foo?')
 
       message =
-        "nofile:1:1: invalid character \"ó\" (code point U+00F3) in alias (only ASCII characters are allowed): Foó"
+        ~r/nofile:1:1: invalid character \"ó\" \(code point U\+00F3\) in alias \(only ASCII characters, without punctuation, are allowed\): Foó/
 
       assert_syntax_error(message, 'Foó')
 
-      message = ~r"""
-      Elixir expects unquoted Unicode atoms, variables, and calls to be in NFC form.
+      # token suggestion heuristic:
+      #  "for foO𝚳, NFKC isn't enough because 𝚳 nfkc's to Greek Μ, would be mixed script.
+      #   however the 'confusability skeleton' for that token produces an all-Latin foOM
+      #   and would tokenize -- so suggest that, in case that's what they want"
+      message =
+        String.trim("""
+        Codepoint failed identifier tokenization, but a simpler form was found.
 
-      Got:
+        Got:
 
-          "foó" \(code points 0x0066 0x006F 0x006F 0x0301\)
+            "foO𝚳" (code points 0x00066 0x0006F 0x0004F 0x1D6B3)
 
-      Expected:
+        Hint: You could write the above in a similar way that is accepted by Elixir:
 
-          "foó" \(code points 0x0066 0x006F 0x00F3\)
+            "foOM" (code points 0x00066 0x0006F 0x0004F 0x0004D)
 
-      """
+        See https://hexdocs.pm/elixir/unicode-syntax.html for more information.
+            |
+          1 | foO𝚳
+            |    ^
+        """)
 
-      assert_syntax_error(message, :unicode.characters_to_nfd_list("foó"))
+      assert_syntax_error(~r/#{message}/, 'foO𝚳')
+
+      # token suggestion heuristic:
+      #  "for fooی𝚳, both NKFC and confusability would result in mixed scripts,
+      #   because the Farsi letter is confusable with a different Arabic letter.
+      #   Well, can't fix it all at once -- let's check for a suggestion just on
+      #   the one codepoint that triggered this, the 𝚳 -- that would at least
+      #   nudge them forwards."
+      message =
+        String.trim("""
+        Elixir expects unquoted Unicode atoms, variables, and calls to use allowed codepoints and to be in NFC form.
+
+        Got:
+
+            "𝚳" (code points 0x1D6B3)
+
+        Hint: You could write the above in a compatible format that is accepted by Elixir:
+
+            "Μ" (code points 0x0039C)
+
+        See https://hexdocs.pm/elixir/unicode-syntax.html for more information.
+            |
+          2 | fooی𝚳
+            |     ^
+        """)
+
+      assert_syntax_error(~r/#{message}/, 'fooی𝚳')
     end
 
-    test "kw missing space" do
-      msg = "nofile:1:1: keyword argument must be followed by space after: foo:"
+    test "keyword missing space" do
+      msg = ~r/nofile:1:1: keyword argument must be followed by space after: foo:/
 
       assert_syntax_error(msg, "foo:bar")
       assert_syntax_error(msg, "foo:+")
       assert_syntax_error(msg, "foo:+1")
     end
 
+    test "expression after keyword lists" do
+      assert_syntax_error(
+        ~r"unexpected expression after keyword list",
+        'call foo: 1, :bar'
+      )
+
+      assert_syntax_error(
+        ~r"unexpected expression after keyword list",
+        'call(foo: 1, :bar)'
+      )
+
+      assert_syntax_error(
+        ~r"unexpected expression after keyword list",
+        '[foo: 1, :bar]'
+      )
+
+      assert_syntax_error(
+        ~r"unexpected expression after keyword list",
+        '%{foo: 1, :bar => :bar}'
+      )
+    end
+
+    test "syntax errors include formatted snippet" do
+      message = "nofile:1:5: syntax error before: '*'\n    |\n  1 | 1 + * 3\n    |     ^"
+      assert_syntax_error(message, "1 + * 3")
+    end
+
     test "invalid map start" do
       assert_syntax_error(
-        "nofile:1:7: expected %{ to define a map, got: %[",
+        ~r/nofile:1:7: expected %{ to define a map, got: %\[/,
         "{:ok, %[], %{}}"
       )
     end
@@ -591,12 +787,12 @@ defmodule Kernel.ParserTest do
 
     test "invalid keywords" do
       assert_syntax_error(
-        "nofile:1:2: syntax error before: '.'",
+        ~r/nofile:1:2: syntax error before: '.'/,
         '+.foo'
       )
 
       assert_syntax_error(
-        ~r"nofile:1:1: syntax error before: after. \"after\" is a reserved word",
+        ~r/nofile:1:1: syntax error before: after. \"after\" is a reserved word/,
         'after = 1'
       )
     end
@@ -616,16 +812,16 @@ defmodule Kernel.ParserTest do
         'if true, do\n'
       )
 
-      assert_syntax_error(~r/nofile:1: unexpected keyword: do:./, 'if true do:\n')
+      assert_syntax_error(~r/nofile:1:9: unexpected keyword: do:./, 'if true do:\n')
     end
 
     test "invalid parens call" do
       msg =
-        "nofile:1: unexpected parentheses. If you are making a function call, do not " <>
+        "nofile:1:5: unexpected parentheses. If you are making a function call, do not " <>
           "insert spaces between the function name and the opening parentheses. " <>
-          "Syntax error before: '('"
+          "Syntax error before: '\\('"
 
-      assert_syntax_error(msg, 'foo (hello, world)')
+      assert_syntax_error(~r/#{msg}/, 'foo (hello, world)')
     end
 
     test "invalid nested no parens call" do
@@ -646,27 +842,27 @@ defmodule Kernel.ParserTest do
       assert interpret.("f 1 + g h 2, 3") == "f(1 + g(h(2, 3)))"
 
       assert interpret.("assert [] = TestRepo.all from p in Post, where: p.title in ^[]") ==
-               "assert([] = TestRepo.all(from(p in Post, where: p.title in ^[])))"
+               "assert [] = TestRepo.all(from(p in Post, where: p.title in ^[]))"
     end
 
     test "invalid atom dot alias" do
       msg =
-        "nofile:1: atom cannot be followed by an alias. If the '.' was meant to be " <>
+        "nofile:1:6: atom cannot be followed by an alias. If the '.' was meant to be " <>
           "part of the atom's name, the atom name must be quoted. Syntax error before: '.'"
 
-      assert_syntax_error(msg, ':foo.Bar')
-      assert_syntax_error(msg, ':"+".Bar')
+      assert_syntax_error(~r/#{msg}/, ':foo.Bar')
+      assert_syntax_error(~r/#{msg}/, ':"+".Bar')
     end
 
     test "invalid map/struct" do
-      assert_syntax_error("nofile:1:5: syntax error before: '}'", '%{:a}')
-      assert_syntax_error("nofile:1:11: syntax error before: '}'", '%{{:a, :b}}')
-      assert_syntax_error("nofile:1:8: syntax error before: '{'", '%{a, b}{a: :b}')
+      assert_syntax_error(~r/nofile:1:5: syntax error before: '}'/, '%{:a}')
+      assert_syntax_error(~r/nofile:1:11: syntax error before: '}'/, '%{{:a, :b}}')
+      assert_syntax_error(~r/nofile:1:8: syntax error before: '{'/, '%{a, b}{a: :b}')
     end
 
     test "invalid interpolation" do
       assert_syntax_error(
-        "nofile:1:17: unexpected token: ). The \"do\" at line 1 is missing terminator \"end\"",
+        ~r/nofile:1:17: unexpected token: \). The \"do\" at line 1 is missing terminator \"end\"/,
         '"foo\#{case 1 do )}bar"'
       )
     end
@@ -700,9 +896,9 @@ defmodule Kernel.ParserTest do
       ''')
 
       # All invalid examples
-      assert_syntax_error("nofile:1:3: syntax error before: ';'", '1+;\n2')
+      assert_syntax_error(~r/nofile:1:3: syntax error before: ';'/, '1+;\n2')
 
-      assert_syntax_error("nofile:1:8: syntax error before: ';'", 'max(1, ;2)')
+      assert_syntax_error(~r/nofile:1:8: syntax error before: ';'/, 'max(1, ;2)')
     end
 
     test "invalid new line" do
@@ -726,37 +922,36 @@ defmodule Kernel.ParserTest do
     end
 
     test "numbers are printed correctly in syntax errors" do
-      assert_syntax_error("nofile:1:5: syntax error before: \"12\"", ':ok 12')
-      assert_syntax_error("nofile:1:5: syntax error before: \"0b1\"", ':ok 0b1')
-      assert_syntax_error("nofile:1:5: syntax error before: \"12.3\"", ':ok 12.3')
+      assert_syntax_error(~r/nofile:1:5: syntax error before: \"12\"/, ':ok 12')
+      assert_syntax_error(~r/nofile:1:5: syntax error before: \"0b1\"/, ':ok 0b1')
+      assert_syntax_error(~r/nofile:1:5: syntax error before: \"12.3\"/, ':ok 12.3')
 
       assert_syntax_error(
-        ~r"nofile:1:1: invalid character _ after number 123_456",
+        ~r"nofile:1:1: invalid character \"_\" after number 123_456",
         '123_456_foo'
       )
     end
 
     test "on hex errors" do
       msg =
-        "invalid hex escape character, expected \\xHH where H is a hexadecimal digit. Syntax error after: \\x"
+        "invalid hex escape character, expected \\\\xHH where H is a hexadecimal digit. Syntax error after: \\\\x"
 
-      assert_syntax_error("nofile:1:2: #{msg}", ~S["\x"])
-      assert_syntax_error("nofile:1:1: #{msg}", ~S[:"\x"])
-      assert_syntax_error("nofile:1:2: #{msg}", ~S["\x": 123])
-      assert_syntax_error("nofile:1:1: #{msg}", ~s["""\n\\x\n"""])
+      assert_syntax_error(~r/nofile:1:2: #{msg}/, ~S["\x"])
+      assert_syntax_error(~r/nofile:1:1: #{msg}/, ~S[:"\x"])
+      assert_syntax_error(~r/nofile:1:2: #{msg}/, ~S["\x": 123])
+      assert_syntax_error(~r/nofile:1:1: #{msg}/, ~s["""\n\\x\n"""])
     end
 
     test "on unicode errors" do
-      msg =
-        "invalid Unicode escape character, expected \\uHHHH or \\u{H*} where H is a hexadecimal digit. Syntax error after: \\u"
+      msg = "invalid Unicode escape character"
 
-      assert_syntax_error("nofile:1:2: #{msg}", ~S["\u"])
-      assert_syntax_error("nofile:1:1: #{msg}", ~S[:"\u"])
-      assert_syntax_error("nofile:1:2: #{msg}", ~S["\u": 123])
-      assert_syntax_error("nofile:1:1: #{msg}", ~s["""\n\\u\n"""])
+      assert_syntax_error(~r/nofile:1:2: #{msg}/, ~S["\u"])
+      assert_syntax_error(~r/nofile:1:1: #{msg}/, ~S[:"\u"])
+      assert_syntax_error(~r/nofile:1:2: #{msg}/, ~S["\u": 123])
+      assert_syntax_error(~r/nofile:1:1: #{msg}/, ~s["""\n\\u\n"""])
 
       assert_syntax_error(
-        "nofile:1:2: invalid or reserved Unicode code point \\u{FFFFFF}. Syntax error after: \\u",
+        ~r/nofile:1:2: invalid or reserved Unicode code point \\u\{FFFFFF\}. Syntax error after: \\u/,
         ~S["\u{FFFFFF}"]
       )
     end
@@ -780,32 +975,6 @@ defmodule Kernel.ParserTest do
     end
   end
 
-  describe "down to Erlang" do
-    test "contains of non-literals" do
-      assert to_erl!("{:ok, make_ref()}") ==
-               {:tuple, 1,
-                [
-                  {:atom, 0, :ok},
-                  {:call, 1, {:remote, 1, {:atom, 0, :erlang}, {:atom, 1, :make_ref}}, []}
-                ]}
-
-      assert to_erl!("[:ok, make_ref()]") ==
-               {:cons, 1, {:atom, 0, :ok},
-                {:cons, 1,
-                 {:call, 1, {:remote, 1, {:atom, 0, :erlang}, {:atom, 1, :make_ref}}, []},
-                 {nil, 0}}}
-    end
-
-    defp to_erl!(code) do
-      {expr, _, _} =
-        code
-        |> Code.string_to_quoted!()
-        |> :elixir.quoted_to_erl(:elixir.env_for_eval([]))
-
-      expr
-    end
-  end
-
   defp parse!(string), do: Code.string_to_quoted!(string)
 
   defp assert_token_missing(given_message, string) do
