@@ -26,6 +26,93 @@ defmodule Kernel.WarningTest do
     assert {:error, _} = Code.string_to_quoted(~s[:"foobar" do])
   end
 
+  describe "unicode identifier security" do
+    test "prevents Restricted codepoints in identifiers" do
+      exception = assert_raise SyntaxError, fn -> Code.string_to_quoted!("_shibㅤ = 1") end
+
+      assert Exception.message(exception) =~
+               "unexpected token: \"ㅤ\" (column 6, code point U+3164)"
+    end
+
+    test "warns on confusables" do
+      assert capture_err(fn -> Code.string_to_quoted("а=1; a=1") end) =~
+               "confusable identifier: 'a' looks like 'а' on line 1"
+
+      assert capture_err(fn -> Code.string_to_quoted("[{:а, 1}, {:a, 1}]") end) =~
+               "confusable identifier: 'a' looks like 'а' on line 1"
+
+      assert capture_err(fn -> Code.string_to_quoted("[а: 1, a: 1]") end) =~
+               "confusable identifier: 'a' looks like 'а' on line 1"
+
+      assert capture_err(fn -> Code.string_to_quoted("quote do: [а(1), a(1)]") end) =~
+               "confusable identifier: 'a' looks like 'а' on line 1"
+
+      assert capture_err(fn -> Code.string_to_quoted("力=1; カ=1") end) =~
+               "confusable identifier: 'カ' looks like '力' on line 1"
+
+      # by convention, doesn't warn on ascii-only confusables
+      assert capture_err(fn -> Code.string_to_quoted("x0 = xO = 1") end) == ""
+      assert capture_err(fn -> Code.string_to_quoted("l1 = ll = 1") end) == ""
+
+      # works with a custom atom encoder
+      assert capture_err(fn ->
+               Code.string_to_quoted("[{:а, 1}, {:a, 1}]",
+                 static_atoms_encoder: fn token, _ -> {:ok, {:wrapped, token}} end
+               )
+             end) =~
+               "confusable identifier: 'a' looks like 'а' on line 1"
+    end
+
+    test "prevents unsafe script mixing in identifiers" do
+      exception =
+        assert_raise SyntaxError, fn ->
+          Code.string_to_quoted!("if аdmin_, do: :ok, else: :err")
+        end
+
+      assert Exception.message(exception) =~
+               "nofile:1:9: invalid mixed-script identifier found: аdmin"
+
+      assert Exception.message(exception) =~ """
+               \\u0430 а {Cyrillic}
+               \\u0064 d {Latin}
+               \\u006D m {Latin}
+               \\u0069 i {Latin}
+               \\u006E n {Latin}
+               \\u005F _
+             """
+
+      # includes suggestion about what to change
+      assert Exception.message(exception) =~ """
+             Hint: You could write the above in a similar way that is accepted by Elixir:
+
+                 "admin_" (code points 0x00061 0x00064 0x0006D 0x00069 0x0006E 0x0005F)
+             """
+
+      # a is in cyrillic
+      assert_raise SyntaxError, ~r/mixed/, fn -> Code.string_to_quoted!("[аdmin: 1]") end
+      assert_raise SyntaxError, ~r/mixed/, fn -> Code.string_to_quoted!("[{:аdmin, 1}]") end
+      assert_raise SyntaxError, ~r/mixed/, fn -> Code.string_to_quoted!("quote do: аdmin(1)") end
+      assert_raise SyntaxError, ~r/mixed/, fn -> Code.string_to_quoted!("рос_api = 1") end
+
+      # T is in cyrillic
+      assert_raise SyntaxError, ~r/mixed/, fn -> Code.string_to_quoted!("[Тシャツ: 1]") end
+    end
+
+    test "allows legitimate script mixing" do
+      # writing systems that legitimately mix multiple scripts, and Common chars like _
+      assert capture_err(fn -> Code.eval_string("幻ㄒㄧㄤ = 1") end) == ""
+      assert capture_err(fn -> Code.eval_string("幻ㄒㄧㄤ1 = 1") end) == ""
+      assert capture_err(fn -> Code.eval_string("__सवव_1? = 1") end) == ""
+
+      # uts39 5.2 allowed 'highly restrictive' script mixing, like 't-shirt' in Jpan:
+      assert capture_err(fn -> Code.string_to_quoted!(":Tシャツ") end) == ""
+
+      # elixir's normalizations combine scriptsets of the 'from' and 'to' characters,
+      # ex: {Common} MICRO => {Greek} MU == {Common, Greek}; Common intersects w/all
+      assert capture_err(fn -> Code.string_to_quoted!("μs") end) == ""
+    end
+  end
+
   test "operators formed by many of the same character followed by that character" do
     output =
       capture_err(fn ->
@@ -50,6 +137,7 @@ defmodule Kernel.WarningTest do
       assert capture_err(fn -> Code.eval_string(~s/:"Foo"/) end) == ""
       assert capture_err(fn -> Code.eval_string(~s/:"foo@bar"/) end) == ""
       assert capture_err(fn -> Code.eval_string(~s/:"héllò"/) end) == ""
+      assert capture_err(fn -> Code.eval_string(~s/:"3L1X1R"/) end) == ""
     end
 
     test "warns for unnecessary quotes" do
@@ -64,6 +152,11 @@ defmodule Kernel.WarningTest do
     end
   end
 
+  test "warns on :: as atom" do
+    assert capture_err(fn -> Code.eval_string(~s/:::/) end) =~
+             "atom ::: must be written between quotes, as in :\"::\", to avoid ambiguity"
+  end
+
   test "unused variable" do
     output =
       capture_err(fn ->
@@ -974,9 +1067,9 @@ defmodule Kernel.WarningTest do
     purge(Sample)
   end
 
-  # TODO: Simplify when we require OTP 24
+  # TODO: Simplify when we require Erlang/OTP 24
   if System.otp_release() >= "24" do
-    @argument_error_message "the call to :erlang.atom_to_binary/2"
+    @argument_error_message "the call to Atom.to_string/1"
     @arithmetic_error_message "the call to +/2"
   else
     @argument_error_message "this expression"
@@ -1122,18 +1215,6 @@ defmodule Kernel.WarningTest do
     purge(EmptyBehaviour)
   end
 
-  test "undefined behavior" do
-    assert capture_err(fn ->
-             Code.eval_string("""
-             defmodule Sample do
-               @behavior Hello
-             end
-             """)
-           end) =~ "@behavior attribute is not supported, please use @behaviour instead"
-  after
-    purge(Sample)
-  end
-
   test "undefined function for protocol" do
     assert capture_err(fn ->
              Code.eval_string("""
@@ -1197,10 +1278,24 @@ defmodule Kernel.WarningTest do
     purge(Sample)
   end
 
+  test "warning on unnecessary code point escape" do
+    assert capture_err(fn ->
+             Code.eval_string("?\\n + ?\\\\")
+           end) == ""
+
+    assert capture_err(fn ->
+             Code.eval_string("?\\w")
+           end) =~ "unknown escape sequence ?\\w, use ?w instead"
+  end
+
   test "warning on code point escape" do
     assert capture_err(fn ->
              Code.eval_string("? ")
            end) =~ "found ? followed by code point 0x20 (space), please use ?\\s instead"
+
+    assert capture_err(fn ->
+             Code.eval_string("?\\ ")
+           end) =~ "found ?\\ followed by code point 0x20 (space), please use ?\\s instead"
   end
 
   test "duplicated docs in the same clause" do
@@ -1430,70 +1525,56 @@ defmodule Kernel.WarningTest do
 
       assert capture_err(fn ->
                Code.eval_string("""
-               defmodule Sample do
+               defmodule Sample1 do
                  @type my_type :: ann_type :: nested_ann_type :: atom
                end
                """)
              end) =~ message
 
-      purge(Sample)
+      purge(Sample1)
 
       assert capture_err(fn ->
                Code.eval_string("""
-               defmodule Sample do
+               defmodule Sample2 do
                  @type my_type :: ann_type :: nested_ann_type :: atom | port
                end
                """)
              end) =~ message
 
-      purge(Sample)
+      purge(Sample2)
 
       assert capture_err(fn ->
                Code.eval_string("""
-               defmodule Sample do
+               defmodule Sample3 do
                  @spec foo :: {pid, ann_type :: nested_ann_type :: atom}
                  def foo, do: nil
                end
                """)
              end) =~ message
     after
-      purge(Sample)
+      purge([Sample1, Sample2, Sample3])
     end
 
     test "invalid type annotations" do
-      message =
-        "invalid type annotation. When using the | operator to represent the union of types, " <>
-          "make sure to wrap type annotations in parentheses"
+      assert capture_err(fn ->
+               Code.eval_string("""
+               defmodule Sample1 do
+                 @type my_type :: (pid() :: atom)
+               end
+               """)
+             end) =~ "invalid type annotation. The left side of :: must be a variable, got: pid()"
 
       assert capture_err(fn ->
                Code.eval_string("""
-               defmodule Sample do
+               defmodule Sample2 do
                  @type my_type :: pid | ann_type :: atom
                end
                """)
-             end) =~ message
-
-      purge(Sample)
-
-      assert capture_err(fn ->
-               Code.eval_string("""
-               defmodule Sample do
-                 @type my_type :: pid | ann_type :: atom | port
-               end
-               """)
-             end) =~ message
-
-      purge(Sample)
-
-      assert capture_err(fn ->
-               Code.eval_string("""
-               defmodule Sample do
-                 @type my_type :: {port, pid | ann_type :: atom | port}
-               end
-               """)
-             end) =~ message
+             end) =~
+               "invalid type annotation. The left side of :: must be a variable, got: pid | ann_type. " <>
+                 "Note \"left | right :: ann\" is the same as \"(left | right) :: ann\""
     after
-      purge(Sample)
+      purge([Sample1, Sample2])
     end
   end
 
@@ -1835,6 +1916,11 @@ defmodule Kernel.WarningTest do
     end
   end
 
+  test "warnings on trailing comma on call" do
+    assert capture_err(fn -> Code.eval_string("Keyword.merge([], foo: 1,)") end) =~
+             "trailing commas are not allowed inside function/macro call arguments"
+  end
+
   test "defstruct warns with duplicate keys" do
     assert capture_err(fn ->
              Code.eval_string("""
