@@ -440,6 +440,12 @@ defmodule KernelTest do
     end
   end
 
+  describe ".." do
+    test "returns 0..-1//1" do
+      assert (..) == 0..-1//1
+    end
+  end
+
   describe "in/2" do
     test "too large list in guards" do
       defmodule TooLargeList do
@@ -609,7 +615,7 @@ defmodule KernelTest do
       assert map_dot(%{field: true})
 
       message =
-        "cannot invoke remote function in guard. " <>
+        "cannot invoke remote function in guards. " <>
           "If you want to do a map lookup instead, please remove parens from map.field()"
 
       assert_raise CompileError, Regex.compile!(message), fn ->
@@ -668,7 +674,7 @@ defmodule KernelTest do
     end
 
     test "with a non-literal non-escaped compile-time range in guards" do
-      message = "non-literal range in guard should be escaped with Macro.escape/2"
+      message = ~r"found unescaped value on the right side of in/2: 1..3"
 
       assert_eval_raise(ArgumentError, message, """
       defmodule InErrors do
@@ -723,18 +729,16 @@ defmodule KernelTest do
       result = expand_to_string(quote(do: rand() in 1..2))
       assert result =~ "var = rand()"
 
-      assert result =~
-               ":erlang.andalso(:erlang.is_integer(var), :erlang.andalso(:erlang.>=(var, 1), :erlang.\"=<\"(var, 2)))"
+      assert result =~ """
+             :erlang.andalso(
+               :erlang.is_integer(var),
+               :erlang.andalso(:erlang.>=(var, 1), :erlang.\"=<\"(var, 2))
+             )\
+             """
 
       # Empty list
-      assert expand_to_string(quote(do: :x in [])) =~
-               ~S"""
-                 _ = :x
-                 false
-               """
-
-      assert expand_to_string(quote(do: :x in []), :guard) ==
-               "false"
+      assert expand_to_string(quote(do: :x in [])) =~ "_ = :x\nfalse"
+      assert expand_to_string(quote(do: :x in []), :guard) == "false"
 
       # Lists
       result = expand_to_string(quote(do: rand() in [1, 2]))
@@ -742,30 +746,10 @@ defmodule KernelTest do
       assert result =~ ":erlang.orelse(:erlang.\"=:=\"(var, 1), :erlang.\"=:=\"(var, 2))"
 
       result = expand_to_string(quote(do: rand() in [1 | [2]]))
-      assert result =~ "var = rand()"
-      assert result =~ ":erlang.orelse(:erlang.\"=:=\"(var, 1), :erlang.\"=:=\"(var, 2))"
-
-      result = expand_to_string(quote(do: rand() in [1, 2 | [3]]))
-      assert result =~ "var = rand()"
-
-      assert result =~
-               ":erlang.orelse(:erlang.\"=:=\"(var, 1), :erlang.orelse(:erlang.\"=:=\"(var, 2), :erlang.\"=:=\"(var, 3)))"
-
-      result = expand_to_string(quote(do: rand() in [1 | [2 | [3]]]))
-      assert result =~ "var = rand()"
-
-      assert result =~
-               ":erlang.orelse(:erlang.\"=:=\"(var, 1), :erlang.orelse(:erlang.\"=:=\"(var, 2), :erlang.\"=:=\"(var, 3)))"
+      assert result =~ ":lists.member(rand(), [1 | [2]]"
 
       result = expand_to_string(quote(do: rand() in [1 | some_call()]))
-      assert result =~ "var = rand()"
-      assert result =~ "{arg1} = {some_call()}"
-      assert result =~ ":erlang.orelse(:erlang.\"=:=\"(var, 1), :lists.member(var, arg1))"
-
-      result = expand_to_string(quote(do: rand() in [{1}, {2}, {3} | some_call()]))
-      assert result =~ "var = rand()"
-      assert result =~ "{arg1, arg2, arg3, arg4} = {{1}, {2}, {3}, some_call()}"
-      assert result =~ ":erlang.orelse(:erlang.\"=:=\"(var, arg3), :lists.member(var, arg4)))"
+      assert result =~ ":lists.member(rand(), [1 | some_call()]"
     end
 
     defp quote_case_in(left, right) do
@@ -777,20 +761,29 @@ defmodule KernelTest do
     end
 
     test "is optimized" do
-      assert expand_to_string(quote(do: foo in [])) =~
-               ~S"""
-                 _ = foo
-                 false
-               """
+      assert expand_to_string(quote(do: foo in [])) ==
+               "_ = foo\nfalse"
 
-      assert expand_to_string(quote(do: foo in [foo])) =~
-               ~r/{arg1} = {foo}\n\s+:erlang."=:="\(foo, arg1\)\n/
+      assert expand_to_string(quote(do: foo in [1, 2, 3])) == """
+             :erlang.orelse(
+               :erlang.orelse(:erlang.\"=:=\"(foo, 1), :erlang.\"=:=\"(foo, 2)),
+               :erlang.\"=:=\"(foo, 3)
+             )\
+             """
 
-      assert expand_to_string(quote(do: foo in 0..1)) ==
-               ":erlang.andalso(:erlang.is_integer(foo), :erlang.andalso(:erlang.>=(foo, 0), :erlang.\"=<\"(foo, 1)))"
+      assert expand_to_string(quote(do: foo in 0..1)) == """
+             :erlang.andalso(
+               :erlang.is_integer(foo),
+               :erlang.andalso(:erlang.>=(foo, 0), :erlang.\"=<\"(foo, 1))
+             )\
+             """
 
-      assert expand_to_string(quote(do: foo in -1..0)) ==
-               ":erlang.andalso(:erlang.is_integer(foo), :erlang.andalso(:erlang.>=(foo, -1), :erlang.\"=<\"(foo, 0)))"
+      assert expand_to_string(quote(do: foo in -1..0)) == """
+             :erlang.andalso(
+               :erlang.is_integer(foo),
+               :erlang.andalso(:erlang.>=(foo, -1), :erlang.\"=<\"(foo, 0))
+             )\
+             """
 
       assert expand_to_string(quote(do: foo in 1..1)) ==
                ":erlang.\"=:=\"(foo, 1)"
@@ -818,6 +811,11 @@ defmodule KernelTest do
       refute {:__info__, 1} in Kernel.__info__(:functions)
     end
 
+    test ":struct" do
+      assert Kernel.__info__(:struct) == nil
+      assert hd(URI.__info__(:struct)) == %{field: :scheme, required: false}
+    end
+
     test "others" do
       assert Kernel.__info__(:module) == Kernel
       assert is_list(Kernel.__info__(:compile))
@@ -874,6 +872,18 @@ defmodule KernelTest do
     end
 
     test "raises with non-variable arguments" do
+      assert_raise ArgumentError,
+                   "guards are not allowed in defdelegate/2, got: when is_list(term) or is_binary(term)",
+                   fn ->
+                     string = """
+                     defmodule IntDelegateWithGuards do
+                       defdelegate foo(term) when is_list(term) or is_binary(term), to: List
+                     end
+                     """
+
+                     Code.eval_string(string, [], __ENV__)
+                   end
+
       msg = "defdelegate/2 only accepts function parameters, got: 1"
 
       assert_raise ArgumentError, msg, fn ->
@@ -897,6 +907,16 @@ defmodule KernelTest do
       end
     end
 
+    test "raises when :to targeting the delegating module is given without the :as option" do
+      assert_raise ArgumentError,
+                   ~r/defdelegate function is calling itself, which will lead to an infinite loop. You should either change the value of the :to option or specify the :as option/,
+                   fn ->
+                     defmodule ImplAttributes do
+                       defdelegate foo(), to: __MODULE__
+                     end
+                   end
+    end
+
     defdelegate my_reverse(list \\ []), to: :lists, as: :reverse
     defdelegate my_get(map \\ %{}, key, default \\ ""), to: Map, as: :get
 
@@ -912,6 +932,14 @@ defmodule KernelTest do
   end
 
   describe "defmodule" do
+    test "expects atoms as module names" do
+      msg = ~r"invalid module name: 3"
+
+      assert_raise CompileError, msg, fn ->
+        defmodule 1 + 2, do: :ok
+      end
+    end
+
     test "does not accept special atoms as module names" do
       special_atoms = [nil, true, false]
 
@@ -923,6 +951,16 @@ defmodule KernelTest do
         end
       end)
     end
+
+    test "does not accept slashes in module names" do
+      assert_raise CompileError, ~r(invalid module name: :"foo/bar"), fn ->
+        defmodule :"foo/bar", do: :ok
+      end
+
+      assert_raise CompileError, ~r(invalid module name: :"foo\\\\bar"), fn ->
+        defmodule :"foo\\bar", do: :ok
+      end
+    end
   end
 
   describe "access" do
@@ -939,7 +977,7 @@ defmodule KernelTest do
       map = %{"fruits" => ["banana", "apple", "orange"]}
       assert get_in(map, ["fruits", by_index(0)]) == "banana"
       assert get_in(map, ["fruits", by_index(3)]) == nil
-      assert get_in(map, ["unknown", by_index(3)]) == :oops
+      assert get_in(map, ["unknown", by_index(3)]) == nil
 
       assert_raise FunctionClauseError, fn ->
         get_in(users, [])
@@ -1153,12 +1191,15 @@ defmodule KernelTest do
 
     def by_index(index) do
       fn
-        _, nil, next ->
-          next.(:oops)
+        :get, nil, _next ->
+          raise "won't be invoked"
 
         :get, data, next ->
           next.(Enum.at(data, index))
 
+        :get_and_update, nil, next ->
+          next.(:oops)
+
         :get_and_update, data, next ->
           current = Enum.at(data, index)
 
@@ -1192,6 +1233,10 @@ defmodule KernelTest do
       assert [1] |> (&hd(&1)).() == 1
     end
 
+    test "reverse associativity" do
+      assert [1, [2], 3] |> (List.flatten() |> Enum.map(&(&1 * 2))) == [2, 4, 6]
+    end
+
     defp twice(a), do: a * 2
 
     defp local(list) do
@@ -1383,6 +1428,13 @@ defmodule KernelTest do
     assert match?(x when ceil(x) == 1, 0.2)
   end
 
+  test "binary_slice/2" do
+    assert binary_slice("abc", -1..0) == ""
+    assert binary_slice("abc", -5..-5) == ""
+    assert binary_slice("x", 0..0//2) == "x"
+    assert binary_slice("abcde", 1..3//2) == "bd"
+  end
+
   test "sigil_U/2" do
     assert ~U[2015-01-13 13:00:07.123Z] == %DateTime{
              calendar: Calendar.ISO,
@@ -1415,4 +1467,40 @@ defmodule KernelTest do
       Code.eval_string(~s{~U[2015-01-13 13:00:07+00:30]})
     end
   end
+
+  describe "dbg/2" do
+    import ExUnit.CaptureIO
+
+    test "prints the given expression and returns its value" do
+      output = capture_io(fn -> assert dbg(List.duplicate(:foo, 3)) == [:foo, :foo, :foo] end)
+      assert output =~ "kernel_test.exs"
+      assert output =~ "KernelTest"
+      assert output =~ "List.duplicate(:foo, 3)"
+      assert output =~ ":foo"
+    end
+
+    test "doesn't print any colors if :syntax_colors is []" do
+      output =
+        capture_io(fn ->
+          assert dbg(List.duplicate(:foo, 3), syntax_colors: []) == [:foo, :foo, :foo]
+        end)
+
+      assert output =~ "kernel_test.exs"
+      assert output =~ "KernelTest."
+      assert output =~ "List.duplicate(:foo, 3)"
+      assert output =~ "[:foo, :foo, :foo]"
+      refute output =~ "\\e["
+    end
+
+    test "prints binding() if no arguments are given" do
+      my_var = 1
+      my_other_var = :foo
+
+      output = capture_io(fn -> dbg() end)
+
+      assert output =~ "binding()"
+      assert output =~ "my_var:"
+      assert output =~ "my_other_var:"
+    end
+  end
 end
