@@ -226,6 +226,16 @@ defmodule MacroTest do
       assert Macro.expand_once(expr, __ENV__) == expr
     end
 
+    test "propagates generated" do
+      assert {:||, meta, [1, false]} = Macro.expand_once(quote(do: oror(1, false)), __ENV__)
+      refute meta[:generated]
+
+      assert {:||, meta, [1, false]} =
+               Macro.expand_once(quote(generated: true, do: oror(1, false)), __ENV__)
+
+      assert meta[:generated]
+    end
+
     test "does not expand module attributes" do
       message =
         "could not call Module.get_attribute/2 because the module #{inspect(__MODULE__)} " <>
@@ -259,24 +269,161 @@ defmodule MacroTest do
     assert expand_and_clean(quote(do: oror(1, false)), __ENV__) == quoted
   end
 
+  test "expand_literal" do
+    assert Macro.expand_literal(quote(do: Foo), __ENV__) == Foo
+    assert Macro.expand_literal(quote(do: Foo + Bar), __ENV__) == quote(do: Foo + Bar)
+  end
+
   test "var/2" do
     assert Macro.var(:foo, nil) == {:foo, [], nil}
     assert Macro.var(:foo, Other) == {:foo, [], Other}
   end
 
+  describe "dbg/3" do
+    defmacrop dbg_format(ast, options \\ quote(do: [syntax_colors: []])) do
+      quote do
+        {result, formatted} =
+          ExUnit.CaptureIO.with_io(fn ->
+            unquote(Macro.dbg(ast, options, __CALLER__))
+          end)
+
+        # Make sure there's an empty line after the output.
+        assert String.ends_with?(formatted, "\n\n") or
+                 String.ends_with?(formatted, "\n\n" <> IO.ANSI.reset())
+
+        {result, formatted}
+      end
+    end
+
+    test "with a simple expression" do
+      {result, formatted} = dbg_format(1 + 1)
+      assert result == 2
+      assert formatted =~ "1 + 1 #=> 2"
+    end
+
+    test "with variables" do
+      my_var = 1 + 1
+      {result, formatted} = dbg_format(my_var)
+      assert result == 2
+      assert formatted =~ "my_var #=> 2"
+    end
+
+    test "with a function call" do
+      {result, formatted} = dbg_format(Atom.to_string(:foo))
+
+      assert result == "foo"
+      assert formatted =~ ~s[Atom.to_string(:foo) #=> "foo"]
+    end
+
+    test "with a multiline input" do
+      {result, formatted} =
+        dbg_format(
+          case 1 + 1 do
+            2 -> :two
+            _other -> :math_is_broken
+          end
+        )
+
+      assert result == :two
+
+      assert formatted =~ """
+             case 1 + 1 do
+               2 -> :two
+               _other -> :math_is_broken
+             end #=> :two
+             """
+    end
+
+    test "with a pipeline on a single line" do
+      {result, formatted} = dbg_format([:a, :b, :c] |> tl() |> tl |> Kernel.hd())
+      assert result == :c
+
+      assert formatted =~ "macro_test.exs"
+
+      assert formatted =~ """
+             \n[:a, :b, :c] #=> [:a, :b, :c]
+             |> tl() #=> [:b, :c]
+             |> tl #=> [:c]
+             |> Kernel.hd() #=> :c
+             """
+
+      # Regression for pipes sometimes erroneously ending with three newlines (one
+      # extra than needed).
+      assert formatted =~ ~r/[^\n]\n\n$/
+    end
+
+    test "with a pipeline on multiple lines" do
+      {result, formatted} =
+        dbg_format(
+          [:a, :b, :c]
+          |> tl()
+          |> tl
+          |> Kernel.hd()
+        )
+
+      assert result == :c
+
+      assert formatted =~ "macro_test.exs"
+
+      assert formatted =~ """
+             \n[:a, :b, :c] #=> [:a, :b, :c]
+             |> tl() #=> [:b, :c]
+             |> tl #=> [:c]
+             |> Kernel.hd() #=> :c
+             """
+
+      # Regression for pipes sometimes erroneously ending with three newlines (one
+      # extra than needed).
+      assert formatted =~ ~r/[^\n]\n\n$/
+    end
+
+    test "with \"syntax_colors: []\" it doesn't print any color sequences" do
+      {_result, formatted} = dbg_format("hello")
+      refute formatted =~ "\e["
+    end
+
+    test "with \"syntax_colors: [...]\" it forces color sequences" do
+      {_result, formatted} = dbg_format("hello", syntax_colors: [string: :cyan])
+      assert formatted =~ IO.iodata_to_binary(IO.ANSI.format([:cyan, ~s("hello")]))
+    end
+
+    test "forwards options to the underlying inspect calls" do
+      value = 'hello'
+      assert {^value, formatted} = dbg_format(value, syntax_colors: [], charlists: :as_lists)
+      assert formatted =~ "value #=> [104, 101, 108, 108, 111]\n"
+    end
+
+    test "with the :print_location option set to false, doesn't print any header" do
+      {result, formatted} = dbg_format("hello", print_location: false)
+      assert result == "hello"
+      refute formatted =~ Path.basename(__ENV__.file)
+    end
+  end
+
   describe "to_string/1" do
+    test "converts quoted to string" do
+      assert Macro.to_string(quote do: hello(world)) == "hello(world)"
+    end
+  end
+
+  describe "to_string/2" do
+    defp macro_to_string(var, fun \\ fn _ast, string -> string end) do
+      module = Macro
+      module.to_string(var, fun)
+    end
+
     test "variable" do
-      assert Macro.to_string(quote(do: foo)) == "foo"
+      assert macro_to_string(quote(do: foo)) == "foo"
     end
 
     test "local call" do
-      assert Macro.to_string(quote(do: foo(1, 2, 3))) == "foo(1, 2, 3)"
-      assert Macro.to_string(quote(do: foo([1, 2, 3]))) == "foo([1, 2, 3])"
+      assert macro_to_string(quote(do: foo(1, 2, 3))) == "foo(1, 2, 3)"
+      assert macro_to_string(quote(do: foo([1, 2, 3]))) == "foo([1, 2, 3])"
     end
 
     test "remote call" do
-      assert Macro.to_string(quote(do: foo.bar(1, 2, 3))) == "foo.bar(1, 2, 3)"
-      assert Macro.to_string(quote(do: foo.bar([1, 2, 3]))) == "foo.bar([1, 2, 3])"
+      assert macro_to_string(quote(do: foo.bar(1, 2, 3))) == "foo.bar(1, 2, 3)"
+      assert macro_to_string(quote(do: foo.bar([1, 2, 3]))) == "foo.bar([1, 2, 3])"
 
       quoted =
         quote do
@@ -285,68 +432,70 @@ defmodule MacroTest do
            end).bar([1, 2, 3])
         end
 
-      assert Macro.to_string(quoted) == "(foo do\n  :ok\nend).bar([1, 2, 3])"
+      assert macro_to_string(quoted) == "(foo do\n  :ok\nend).bar([1, 2, 3])"
     end
 
     test "nullary remote call" do
-      assert Macro.to_string(quote do: foo.bar) == "foo.bar"
-      assert Macro.to_string(quote do: foo.bar()) == "foo.bar()"
+      assert macro_to_string(quote do: foo.bar) == "foo.bar"
+      assert macro_to_string(quote do: foo.bar()) == "foo.bar()"
     end
 
     test "atom remote call" do
-      assert Macro.to_string(quote(do: :foo.bar(1, 2, 3))) == ":foo.bar(1, 2, 3)"
+      assert macro_to_string(quote(do: :foo.bar(1, 2, 3))) == ":foo.bar(1, 2, 3)"
     end
 
     test "remote and fun call" do
-      assert Macro.to_string(quote(do: foo.bar().(1, 2, 3))) == "foo.bar().(1, 2, 3)"
-      assert Macro.to_string(quote(do: foo.bar().([1, 2, 3]))) == "foo.bar().([1, 2, 3])"
+      assert macro_to_string(quote(do: foo.bar().(1, 2, 3))) == "foo.bar().(1, 2, 3)"
+      assert macro_to_string(quote(do: foo.bar().([1, 2, 3]))) == "foo.bar().([1, 2, 3])"
     end
 
     test "unusual remote atom fun call" do
-      assert Macro.to_string(quote(do: Foo."42"())) == ~s/Foo."42"()/
-      assert Macro.to_string(quote(do: Foo."Bar"())) == ~s/Foo."Bar"()/
-      assert Macro.to_string(quote(do: Foo."bar baz"().""())) == ~s/Foo."bar baz"().""()/
-      assert Macro.to_string(quote(do: Foo."%{}"())) == ~s/Foo."%{}"()/
-      assert Macro.to_string(quote(do: Foo."..."())) == ~s/Foo."..."()/
+      assert macro_to_string(quote(do: Foo."42"())) == ~s/Foo."42"()/
+      assert macro_to_string(quote(do: Foo."Bar"())) == ~s/Foo."Bar"()/
+      assert macro_to_string(quote(do: Foo."bar baz"().""())) == ~s/Foo."bar baz"().""()/
+      assert macro_to_string(quote(do: Foo."%{}"())) == ~s/Foo."%{}"()/
+      assert macro_to_string(quote(do: Foo."..."())) == ~s/Foo."..."()/
     end
 
     test "atom fun call" do
-      assert Macro.to_string(quote(do: :foo.(1, 2, 3))) == ":foo.(1, 2, 3)"
+      assert macro_to_string(quote(do: :foo.(1, 2, 3))) == ":foo.(1, 2, 3)"
     end
 
     test "aliases call" do
-      assert Macro.to_string(quote(do: Foo.Bar.baz(1, 2, 3))) == "Foo.Bar.baz(1, 2, 3)"
-      assert Macro.to_string(quote(do: Foo.Bar.baz([1, 2, 3]))) == "Foo.Bar.baz([1, 2, 3])"
-      assert Macro.to_string(quote(do: Foo.bar(<<>>, []))) == "Foo.bar(<<>>, [])"
+      assert macro_to_string(quote(do: Elixir)) == "Elixir"
+      assert macro_to_string(quote(do: Foo)) == "Foo"
+      assert macro_to_string(quote(do: Foo.Bar.baz(1, 2, 3))) == "Foo.Bar.baz(1, 2, 3)"
+      assert macro_to_string(quote(do: Foo.Bar.baz([1, 2, 3]))) == "Foo.Bar.baz([1, 2, 3])"
+      assert macro_to_string(quote(do: Foo.bar(<<>>, []))) == "Foo.bar(<<>>, [])"
     end
 
     test "keyword call" do
-      assert Macro.to_string(quote(do: Foo.bar(foo: :bar))) == "Foo.bar(foo: :bar)"
-      assert Macro.to_string(quote(do: Foo.bar("Elixir.Foo": :bar))) == "Foo.bar([{Foo, :bar}])"
+      assert macro_to_string(quote(do: Foo.bar(foo: :bar))) == "Foo.bar(foo: :bar)"
+      assert macro_to_string(quote(do: Foo.bar("Elixir.Foo": :bar))) == "Foo.bar([{Foo, :bar}])"
     end
 
     test "sigil call" do
-      assert Macro.to_string(quote(do: ~r"123")) == ~S/~r"123"/
-      assert Macro.to_string(quote(do: ~r"\n123")) == ~S/~r"\n123"/
-      assert Macro.to_string(quote(do: ~r"12\"3")) == ~S/~r"12\"3"/
-      assert Macro.to_string(quote(do: ~r/12\/3/u)) == ~S"~r/12\/3/u"
-      assert Macro.to_string(quote(do: ~r{\n123})) == ~S/~r{\n123}/
-      assert Macro.to_string(quote(do: ~r((1\)(2\)3))) == ~S/~r((1\)(2\)3)/
-      assert Macro.to_string(quote(do: ~r{\n1{1\}23})) == ~S/~r{\n1{1\}23}/
-      assert Macro.to_string(quote(do: ~r|12\|3|)) == ~S"~r|12\|3|"
+      assert macro_to_string(quote(do: ~r"123")) == ~S/~r"123"/
+      assert macro_to_string(quote(do: ~r"\n123")) == ~S/~r"\n123"/
+      assert macro_to_string(quote(do: ~r"12\"3")) == ~S/~r"12\"3"/
+      assert macro_to_string(quote(do: ~r/12\/3/u)) == ~S"~r/12\/3/u"
+      assert macro_to_string(quote(do: ~r{\n123})) == ~S/~r{\n123}/
+      assert macro_to_string(quote(do: ~r((1\)(2\)3))) == ~S/~r((1\)(2\)3)/
+      assert macro_to_string(quote(do: ~r{\n1{1\}23})) == ~S/~r{\n1{1\}23}/
+      assert macro_to_string(quote(do: ~r|12\|3|)) == ~S"~r|12\|3|"
 
-      assert Macro.to_string(quote(do: ~r[1#{two}3])) == ~S/~r[1#{two}3]/
-      assert Macro.to_string(quote(do: ~r[1[#{two}\]3])) == ~S/~r[1[#{two}\]3]/
-      assert Macro.to_string(quote(do: ~r'1#{two}3'u)) == ~S/~r'1#{two}3'u/
+      assert macro_to_string(quote(do: ~r[1#{two}3])) == ~S/~r[1#{two}3]/
+      assert macro_to_string(quote(do: ~r[1[#{two}\]3])) == ~S/~r[1[#{two}\]3]/
+      assert macro_to_string(quote(do: ~r'1#{two}3'u)) == ~S/~r'1#{two}3'u/
 
-      assert Macro.to_string(quote(do: ~R"123")) == ~S/~R"123"/
-      assert Macro.to_string(quote(do: ~R"123"u)) == ~S/~R"123"u/
-      assert Macro.to_string(quote(do: ~R"\n123")) == ~S/~R"\n123"/
+      assert macro_to_string(quote(do: ~R"123")) == ~S/~R"123"/
+      assert macro_to_string(quote(do: ~R"123"u)) == ~S/~R"123"u/
+      assert macro_to_string(quote(do: ~R"\n123")) == ~S/~R"\n123"/
 
-      assert Macro.to_string(quote(do: ~S["'(123)'"])) == ~S/~S["'(123)'"]/
-      assert Macro.to_string(quote(do: ~s"#{"foo"}")) == ~S/~s"#{"foo"}"/
+      assert macro_to_string(quote(do: ~S["'(123)'"])) == ~S/~S["'(123)'"]/
+      assert macro_to_string(quote(do: ~s"#{"foo"}")) == ~S/~s"#{"foo"}"/
 
-      assert Macro.to_string(
+      assert macro_to_string(
                quote do
                  ~s"""
                  "\""foo"\""
@@ -354,7 +503,7 @@ defmodule MacroTest do
                end
              ) == ~s[~s"""\n"\\""foo"\\""\n"""]
 
-      assert Macro.to_string(
+      assert macro_to_string(
                quote do
                  ~s'''
                  '\''foo'\''
@@ -362,7 +511,7 @@ defmodule MacroTest do
                end
              ) == ~s[~s'''\n'\\''foo'\\''\n''']
 
-      assert Macro.to_string(
+      assert macro_to_string(
                quote do
                  ~s"""
                  "\"foo\""
@@ -370,7 +519,7 @@ defmodule MacroTest do
                end
              ) == ~s[~s"""\n"\\"foo\\""\n"""]
 
-      assert Macro.to_string(
+      assert macro_to_string(
                quote do
                  ~s'''
                  '\"foo\"'
@@ -378,7 +527,7 @@ defmodule MacroTest do
                end
              ) == ~s[~s'''\n'\\"foo\\"'\n''']
 
-      assert Macro.to_string(
+      assert macro_to_string(
                quote do
                  ~S"""
                  "123"
@@ -388,14 +537,14 @@ defmodule MacroTest do
     end
 
     test "tuple call" do
-      assert Macro.to_string(quote(do: alias(Foo.{Bar, Baz, Bong}))) ==
+      assert macro_to_string(quote(do: alias(Foo.{Bar, Baz, Bong}))) ==
                "alias(Foo.{Bar, Baz, Bong})"
 
-      assert Macro.to_string(quote(do: foo(Foo.{}))) == "foo(Foo.{})"
+      assert macro_to_string(quote(do: foo(Foo.{}))) == "foo(Foo.{})"
     end
 
     test "arrow" do
-      assert Macro.to_string(quote(do: foo(1, (2 -> 3)))) == "foo(1, (2 -> 3))"
+      assert macro_to_string(quote(do: foo(1, (2 -> 3)))) == "foo(1, (2 -> 3))"
     end
 
     test "block" do
@@ -424,11 +573,11 @@ defmodule MacroTest do
       )
       """
 
-      assert Macro.to_string(quoted) <> "\n" == expected
+      assert macro_to_string(quoted) <> "\n" == expected
     end
 
     test "not in" do
-      assert Macro.to_string(quote(do: false not in [])) == "false not in []"
+      assert macro_to_string(quote(do: false not in [])) == "false not in []"
     end
 
     test "if else" do
@@ -440,7 +589,7 @@ defmodule MacroTest do
       end
       """
 
-      assert Macro.to_string(quote(do: if(foo, do: bar, else: baz))) <> "\n" == expected
+      assert macro_to_string(quote(do: if(foo, do: bar, else: baz))) <> "\n" == expected
     end
 
     test "case" do
@@ -466,7 +615,7 @@ defmodule MacroTest do
       end
       """
 
-      assert Macro.to_string(quoted) <> "\n" == expected
+      assert macro_to_string(quoted) <> "\n" == expected
     end
 
     test "try" do
@@ -505,12 +654,12 @@ defmodule MacroTest do
       end
       """
 
-      assert Macro.to_string(quoted) <> "\n" == expected
+      assert macro_to_string(quoted) <> "\n" == expected
     end
 
     test "fn" do
-      assert Macro.to_string(quote(do: fn -> 1 + 2 end)) == "fn -> 1 + 2 end"
-      assert Macro.to_string(quote(do: fn x -> x + 1 end)) == "fn x -> x + 1 end"
+      assert macro_to_string(quote(do: fn -> 1 + 2 end)) == "fn -> 1 + 2 end"
+      assert macro_to_string(quote(do: fn x -> x + 1 end)) == "fn x -> x + 1 end"
 
       quoted =
         quote do
@@ -527,7 +676,7 @@ defmodule MacroTest do
       end
       """
 
-      assert Macro.to_string(quoted) <> "\n" == expected
+      assert macro_to_string(quoted) <> "\n" == expected
 
       quoted =
         quote do
@@ -551,9 +700,9 @@ defmodule MacroTest do
       end
       """
 
-      assert Macro.to_string(quoted) <> "\n" == expected
+      assert macro_to_string(quoted) <> "\n" == expected
 
-      assert Macro.to_string(quote(do: (fn x -> x end).(1))) == "(fn x -> x end).(1)"
+      assert macro_to_string(quote(do: (fn x -> x end).(1))) == "(fn x -> x end).(1)"
 
       quoted =
         quote do
@@ -572,23 +721,23 @@ defmodule MacroTest do
       end).(1)
       """
 
-      assert Macro.to_string(quoted) <> "\n" == expected
+      assert macro_to_string(quoted) <> "\n" == expected
     end
 
     test "range" do
-      assert Macro.to_string(quote(do: unquote(-1..+2))) == "-1..2"
-      assert Macro.to_string(quote(do: Foo.integer()..3)) == "Foo.integer()..3"
-      assert Macro.to_string(quote(do: unquote(-1..+2//-3))) == "-1..2//-3"
+      assert macro_to_string(quote(do: unquote(-1..+2))) == "-1..2"
+      assert macro_to_string(quote(do: Foo.integer()..3)) == "Foo.integer()..3"
+      assert macro_to_string(quote(do: unquote(-1..+2//-3))) == "-1..2//-3"
 
-      assert Macro.to_string(quote(do: Foo.integer()..3//Bar.bat())) ==
+      assert macro_to_string(quote(do: Foo.integer()..3//Bar.bat())) ==
                "Foo.integer()..3//Bar.bat()"
     end
 
     test "when" do
-      assert Macro.to_string(quote(do: (() -> x))) == "(() -> x)"
-      assert Macro.to_string(quote(do: (x when y -> z))) == "(x when y -> z)"
-      assert Macro.to_string(quote(do: (x, y when z -> w))) == "((x, y) when z -> w)"
-      assert Macro.to_string(quote(do: (x, y when z -> w))) == "((x, y) when z -> w)"
+      assert macro_to_string(quote(do: (() -> x))) == "(() -> x)"
+      assert macro_to_string(quote(do: (x when y -> z))) == "(x when y -> z)"
+      assert macro_to_string(quote(do: (x, y when z -> w))) == "((x, y) when z -> w)"
+      assert macro_to_string(quote(do: (x, y when z -> w))) == "((x, y) when z -> w)"
     end
 
     test "nested" do
@@ -609,132 +758,132 @@ defmodule MacroTest do
       end
       """
 
-      assert Macro.to_string(quoted) <> "\n" == expected
+      assert macro_to_string(quoted) <> "\n" == expected
     end
 
     test "operator precedence" do
-      assert Macro.to_string(quote(do: (1 + 2) * (3 - 4))) == "(1 + 2) * (3 - 4)"
-      assert Macro.to_string(quote(do: (1 + 2) * 3 - 4)) == "(1 + 2) * 3 - 4"
-      assert Macro.to_string(quote(do: 1 + 2 + 3)) == "1 + 2 + 3"
-      assert Macro.to_string(quote(do: 1 + 2 - 3)) == "1 + 2 - 3"
+      assert macro_to_string(quote(do: (1 + 2) * (3 - 4))) == "(1 + 2) * (3 - 4)"
+      assert macro_to_string(quote(do: (1 + 2) * 3 - 4)) == "(1 + 2) * 3 - 4"
+      assert macro_to_string(quote(do: 1 + 2 + 3)) == "1 + 2 + 3"
+      assert macro_to_string(quote(do: 1 + 2 - 3)) == "1 + 2 - 3"
     end
 
     test "capture operator" do
-      assert Macro.to_string(quote(do: &foo/0)) == "&foo/0"
-      assert Macro.to_string(quote(do: &Foo.foo/0)) == "&Foo.foo/0"
-      assert Macro.to_string(quote(do: &(&1 + &2))) == "&(&1 + &2)"
-      assert Macro.to_string(quote(do: & &1)) == "&(&1)"
-      assert Macro.to_string(quote(do: & &1.(:x))) == "&(&1.(:x))"
-      assert Macro.to_string(quote(do: (& &1).(:x))) == "(&(&1)).(:x)"
+      assert macro_to_string(quote(do: &foo/0)) == "&foo/0"
+      assert macro_to_string(quote(do: &Foo.foo/0)) == "&Foo.foo/0"
+      assert macro_to_string(quote(do: &(&1 + &2))) == "&(&1 + &2)"
+      assert macro_to_string(quote(do: & &1)) == "&(&1)"
+      assert macro_to_string(quote(do: & &1.(:x))) == "&(&1.(:x))"
+      assert macro_to_string(quote(do: (& &1).(:x))) == "(&(&1)).(:x)"
     end
 
     test "containers" do
-      assert Macro.to_string(quote(do: {})) == "{}"
-      assert Macro.to_string(quote(do: [])) == "[]"
-      assert Macro.to_string(quote(do: {1, 2, 3})) == "{1, 2, 3}"
-      assert Macro.to_string(quote(do: [1, 2, 3])) == "[1, 2, 3]"
-      assert Macro.to_string(quote(do: ["Elixir.Foo": :bar])) == "[{Foo, :bar}]"
-      assert Macro.to_string(quote(do: %{})) == "%{}"
-      assert Macro.to_string(quote(do: %{:foo => :bar})) == "%{foo: :bar}"
-      assert Macro.to_string(quote(do: %{:"Elixir.Foo" => :bar})) == "%{Foo => :bar}"
-      assert Macro.to_string(quote(do: %{{1, 2} => [1, 2, 3]})) == "%{{1, 2} => [1, 2, 3]}"
-      assert Macro.to_string(quote(do: %{map | "a" => "b"})) == "%{map | \"a\" => \"b\"}"
-      assert Macro.to_string(quote(do: [1, 2, 3])) == "[1, 2, 3]"
+      assert macro_to_string(quote(do: {})) == "{}"
+      assert macro_to_string(quote(do: [])) == "[]"
+      assert macro_to_string(quote(do: {1, 2, 3})) == "{1, 2, 3}"
+      assert macro_to_string(quote(do: [1, 2, 3])) == "[1, 2, 3]"
+      assert macro_to_string(quote(do: ["Elixir.Foo": :bar])) == "[{Foo, :bar}]"
+      assert macro_to_string(quote(do: %{})) == "%{}"
+      assert macro_to_string(quote(do: %{:foo => :bar})) == "%{foo: :bar}"
+      assert macro_to_string(quote(do: %{:"Elixir.Foo" => :bar})) == "%{Foo => :bar}"
+      assert macro_to_string(quote(do: %{{1, 2} => [1, 2, 3]})) == "%{{1, 2} => [1, 2, 3]}"
+      assert macro_to_string(quote(do: %{map | "a" => "b"})) == "%{map | \"a\" => \"b\"}"
+      assert macro_to_string(quote(do: [1, 2, 3])) == "[1, 2, 3]"
     end
 
     test "struct" do
-      assert Macro.to_string(quote(do: %Test{})) == "%Test{}"
-      assert Macro.to_string(quote(do: %Test{foo: 1, bar: 1})) == "%Test{foo: 1, bar: 1}"
-      assert Macro.to_string(quote(do: %Test{struct | foo: 2})) == "%Test{struct | foo: 2}"
-      assert Macro.to_string(quote(do: %Test{} + 1)) == "%Test{} + 1"
-      assert Macro.to_string(quote(do: %Test{foo(1)} + 2)) == "%Test{foo(1)} + 2"
+      assert macro_to_string(quote(do: %Test{})) == "%Test{}"
+      assert macro_to_string(quote(do: %Test{foo: 1, bar: 1})) == "%Test{foo: 1, bar: 1}"
+      assert macro_to_string(quote(do: %Test{struct | foo: 2})) == "%Test{struct | foo: 2}"
+      assert macro_to_string(quote(do: %Test{} + 1)) == "%Test{} + 1"
+      assert macro_to_string(quote(do: %Test{foo(1)} + 2)) == "%Test{foo(1)} + 2"
     end
 
     test "binary operators" do
-      assert Macro.to_string(quote(do: 1 + 2)) == "1 + 2"
-      assert Macro.to_string(quote(do: [1, 2 | 3])) == "[1, 2 | 3]"
-      assert Macro.to_string(quote(do: [h | t] = [1, 2, 3])) == "[h | t] = [1, 2, 3]"
-      assert Macro.to_string(quote(do: (x ++ y) ++ z)) == "(x ++ y) ++ z"
-      assert Macro.to_string(quote(do: (x +++ y) +++ z)) == "(x +++ y) +++ z"
+      assert macro_to_string(quote(do: 1 + 2)) == "1 + 2"
+      assert macro_to_string(quote(do: [1, 2 | 3])) == "[1, 2 | 3]"
+      assert macro_to_string(quote(do: [h | t] = [1, 2, 3])) == "[h | t] = [1, 2, 3]"
+      assert macro_to_string(quote(do: (x ++ y) ++ z)) == "(x ++ y) ++ z"
+      assert macro_to_string(quote(do: (x +++ y) +++ z)) == "(x +++ y) +++ z"
     end
 
     test "unary operators" do
-      assert Macro.to_string(quote(do: not 1)) == "not(1)"
-      assert Macro.to_string(quote(do: not foo)) == "not(foo)"
-      assert Macro.to_string(quote(do: -1)) == "-1"
-      assert Macro.to_string(quote(do: +(+1))) == "+(+1)"
-      assert Macro.to_string(quote(do: !(foo > bar))) == "!(foo > bar)"
-      assert Macro.to_string(quote(do: @foo(bar))) == "@foo(bar)"
-      assert Macro.to_string(quote(do: identity(&1))) == "identity(&1)"
+      assert macro_to_string(quote(do: not 1)) == "not(1)"
+      assert macro_to_string(quote(do: not foo)) == "not(foo)"
+      assert macro_to_string(quote(do: -1)) == "-1"
+      assert macro_to_string(quote(do: +(+1))) == "+(+1)"
+      assert macro_to_string(quote(do: !(foo > bar))) == "!(foo > bar)"
+      assert macro_to_string(quote(do: @foo(bar))) == "@foo(bar)"
+      assert macro_to_string(quote(do: identity(&1))) == "identity(&1)"
     end
 
     test "access" do
-      assert Macro.to_string(quote(do: a[b])) == "a[b]"
-      assert Macro.to_string(quote(do: a[1 + 2])) == "a[1 + 2]"
-      assert Macro.to_string(quote(do: (a || [a: 1])[:a])) == "(a || [a: 1])[:a]"
-      assert Macro.to_string(quote(do: Map.put(%{}, :a, 1)[:a])) == "Map.put(%{}, :a, 1)[:a]"
+      assert macro_to_string(quote(do: a[b])) == "a[b]"
+      assert macro_to_string(quote(do: a[1 + 2])) == "a[1 + 2]"
+      assert macro_to_string(quote(do: (a || [a: 1])[:a])) == "(a || [a: 1])[:a]"
+      assert macro_to_string(quote(do: Map.put(%{}, :a, 1)[:a])) == "Map.put(%{}, :a, 1)[:a]"
     end
 
     test "keyword list" do
-      assert Macro.to_string(quote(do: [a: a, b: b])) == "[a: a, b: b]"
-      assert Macro.to_string(quote(do: [a: 1, b: 1 + 2])) == "[a: 1, b: 1 + 2]"
-      assert Macro.to_string(quote(do: ["a.b": 1, c: 1 + 2])) == "[\"a.b\": 1, c: 1 + 2]"
+      assert macro_to_string(quote(do: [a: a, b: b])) == "[a: a, b: b]"
+      assert macro_to_string(quote(do: [a: 1, b: 1 + 2])) == "[a: 1, b: 1 + 2]"
+      assert macro_to_string(quote(do: ["a.b": 1, c: 1 + 2])) == "[\"a.b\": 1, c: 1 + 2]"
     end
 
     test "interpolation" do
-      assert Macro.to_string(quote(do: "foo#{bar}baz")) == ~S["foo#{bar}baz"]
+      assert macro_to_string(quote(do: "foo#{bar}baz")) == ~S["foo#{bar}baz"]
     end
 
     test "bit syntax" do
       ast = quote(do: <<1::8*4>>)
-      assert Macro.to_string(ast) == "<<1::8*4>>"
+      assert macro_to_string(ast) == "<<1::8*4>>"
 
       ast = quote(do: @type(foo :: <<_::8, _::_*4>>))
-      assert Macro.to_string(ast) == "@type(foo :: <<_::8, _::_*4>>)"
+      assert macro_to_string(ast) == "@type(foo :: <<_::8, _::_*4>>)"
 
       ast = quote(do: <<69 - 4::bits-size(8 - 4)-unit(1), 65>>)
-      assert Macro.to_string(ast) == "<<69 - 4::bits-size(8 - 4)-unit(1), 65>>"
+      assert macro_to_string(ast) == "<<69 - 4::bits-size(8 - 4)-unit(1), 65>>"
 
       ast = quote(do: <<(<<65>>), 65>>)
-      assert Macro.to_string(ast) == "<<(<<65>>), 65>>"
+      assert macro_to_string(ast) == "<<(<<65>>), 65>>"
 
       ast = quote(do: <<65, (<<65>>)>>)
-      assert Macro.to_string(ast) == "<<65, (<<65>>)>>"
+      assert macro_to_string(ast) == "<<65, (<<65>>)>>"
 
       ast = quote(do: for(<<(a::4 <- <<1, 2>>)>>, do: a))
-      assert Macro.to_string(ast) == "for(<<(a :: 4 <- <<1, 2>>)>>) do\n  a\nend"
+      assert macro_to_string(ast) == "for(<<(a :: 4 <- <<1, 2>>)>>) do\n  a\nend"
     end
 
     test "charlist" do
-      assert Macro.to_string(quote(do: [])) == "[]"
-      assert Macro.to_string(quote(do: 'abc')) == "'abc'"
+      assert macro_to_string(quote(do: [])) == "[]"
+      assert macro_to_string(quote(do: 'abc')) == "'abc'"
     end
 
     test "string" do
-      assert Macro.to_string(quote(do: "")) == ~S/""/
-      assert Macro.to_string(quote(do: "abc")) == ~S/"abc"/
-      assert Macro.to_string(quote(do: "#{"abc"}")) == ~S/"#{"abc"}"/
+      assert macro_to_string(quote(do: "")) == ~S/""/
+      assert macro_to_string(quote(do: "abc")) == ~S/"abc"/
+      assert macro_to_string(quote(do: "#{"abc"}")) == ~S/"#{"abc"}"/
     end
 
     test "last arg keyword list" do
-      assert Macro.to_string(quote(do: foo([]))) == "foo([])"
-      assert Macro.to_string(quote(do: foo(x: y))) == "foo(x: y)"
-      assert Macro.to_string(quote(do: foo(x: 1 + 2))) == "foo(x: 1 + 2)"
-      assert Macro.to_string(quote(do: foo(x: y, p: q))) == "foo(x: y, p: q)"
-      assert Macro.to_string(quote(do: foo(a, x: y, p: q))) == "foo(a, x: y, p: q)"
+      assert macro_to_string(quote(do: foo([]))) == "foo([])"
+      assert macro_to_string(quote(do: foo(x: y))) == "foo(x: y)"
+      assert macro_to_string(quote(do: foo(x: 1 + 2))) == "foo(x: 1 + 2)"
+      assert macro_to_string(quote(do: foo(x: y, p: q))) == "foo(x: y, p: q)"
+      assert macro_to_string(quote(do: foo(a, x: y, p: q))) == "foo(a, x: y, p: q)"
 
-      assert Macro.to_string(quote(do: {[]})) == "{[]}"
-      assert Macro.to_string(quote(do: {[a: b]})) == "{[a: b]}"
-      assert Macro.to_string(quote(do: {x, a: b})) == "{x, [a: b]}"
-      assert Macro.to_string(quote(do: foo(else: a))) == "foo(else: a)"
-      assert Macro.to_string(quote(do: foo(catch: a))) == "foo(catch: a)"
+      assert macro_to_string(quote(do: {[]})) == "{[]}"
+      assert macro_to_string(quote(do: {[a: b]})) == "{[a: b]}"
+      assert macro_to_string(quote(do: {x, a: b})) == "{x, [a: b]}"
+      assert macro_to_string(quote(do: foo(else: a))) == "foo(else: a)"
+      assert macro_to_string(quote(do: foo(catch: a))) == "foo(catch: a)"
     end
 
     test "with fun" do
-      assert Macro.to_string(quote(do: foo(1, 2, 3)), fn _, string -> ":#{string}:" end) ==
+      assert macro_to_string(quote(do: foo(1, 2, 3)), fn _, string -> ":#{string}:" end) ==
                ":foo(:1:, :2:, :3:):"
 
-      assert Macro.to_string(quote(do: Bar.foo(1, 2, 3)), fn _, string -> ":#{string}:" end) ==
+      assert macro_to_string(quote(do: Bar.foo(1, 2, 3)), fn _, string -> ":#{string}:" end) ==
                "::Bar:.foo(:1:, :2:, :3:):"
     end
   end
@@ -778,6 +927,17 @@ defmodule MacroTest do
   end
 
   describe "env" do
+    doctest Macro.Env
+
+    test "inspect" do
+      assert inspect(__ENV__) =~ "#Macro.Env<"
+    end
+
+    test "prune_compile_info" do
+      assert %Macro.Env{lexical_tracker: nil, tracers: []} =
+               Macro.Env.prune_compile_info(%{__ENV__ | lexical_tracker: self(), tracers: [Foo]})
+    end
+
     test "stacktrace" do
       env = %{__ENV__ | file: "foo", line: 12}
 
@@ -797,6 +957,14 @@ defmodule MacroTest do
       defmodule Foo.Bar do
         assert __MODULE__ in __ENV__.context_modules
       end
+
+      assert Foo.Bar in __ENV__.context_modules
+
+      Code.compile_string("""
+      defmodule Foo.Bar.Compiled do
+        true = __MODULE__ in __ENV__.context_modules
+      end
+      """)
     end
 
     test "to_match/1" do
@@ -807,6 +975,11 @@ defmodule MacroTest do
 
       assert Macro.expand_once(quote, Macro.Env.to_match(__ENV__)) == false
     end
+
+    test "prepend_tracer" do
+      assert %Macro.Env{tracers: [MyCustomTracer | _]} =
+               Macro.Env.prepend_tracer(__ENV__, MyCustomTracer)
+    end
   end
 
   ## pipe/unpipe
@@ -821,8 +994,8 @@ defmodule MacroTest do
       Macro.pipe(1, 2, 0)
     end
 
-    assert_raise ArgumentError, ~r"cannot pipe 1 into {:ok}", fn ->
-      Macro.pipe(1, {:ok}, 0)
+    assert_raise ArgumentError, ~r"cannot pipe 1 into \{2, 3\}", fn ->
+      Macro.pipe(1, {2, 3}, 0)
     end
 
     assert_raise ArgumentError, ~r"cannot pipe 1 into 1 \+ 1, the :\+ operator can", fn ->
@@ -987,6 +1160,60 @@ defmodule MacroTest do
     assert Macro.struct!(StructBang, __ENV__) == %{__struct__: StructBang, a: nil, b: nil}
   end
 
+  test "prewalker/1" do
+    ast = quote do: :mod.foo(bar({1, 2}), [3, 4, five])
+    map = Enum.map(Macro.prewalker(ast), & &1)
+
+    assert map == [
+             {{:., [], [:mod, :foo]}, [], [{:bar, [], [{1, 2}]}, [3, 4, {:five, [], MacroTest}]]},
+             {:., [], [:mod, :foo]},
+             :mod,
+             :foo,
+             {:bar, [], [{1, 2}]},
+             {1, 2},
+             1,
+             2,
+             [3, 4, {:five, [], MacroTest}],
+             3,
+             4,
+             {:five, [], MacroTest}
+           ]
+
+    assert map == ast |> Macro.prewalk([], &{&1, [&1 | &2]}) |> elem(1) |> Enum.reverse()
+    assert Enum.zip(Macro.prewalker(ast), []) == Enum.zip(map, [])
+
+    for i <- 0..(length(map) + 1) do
+      assert Enum.take(Macro.prewalker(ast), i) == Enum.take(map, i)
+    end
+  end
+
+  test "postwalker/1" do
+    ast = quote do: :mod.foo(bar({1, 2}), [3, 4, five])
+    map = Enum.map(Macro.postwalker(ast), & &1)
+
+    assert map == [
+             :mod,
+             :foo,
+             {:., [], [:mod, :foo]},
+             1,
+             2,
+             {1, 2},
+             {:bar, [], [{1, 2}]},
+             3,
+             4,
+             {:five, [], MacroTest},
+             [3, 4, {:five, [], MacroTest}],
+             {{:., [], [:mod, :foo]}, [], [{:bar, [], [{1, 2}]}, [3, 4, {:five, [], MacroTest}]]}
+           ]
+
+    assert map == ast |> Macro.postwalk([], &{&1, [&1 | &2]}) |> elem(1) |> Enum.reverse()
+    assert Enum.zip(Macro.postwalker(ast), []) == Enum.zip(map, [])
+
+    for i <- 0..(length(map) + 1) do
+      assert Enum.take(Macro.postwalker(ast), i) == Enum.take(map, i)
+    end
+  end
+
   test "operator?/2" do
     assert Macro.operator?(:+, 2)
     assert Macro.operator?(:+, 1)
