@@ -125,7 +125,7 @@ defmodule Module.Types.IntegrationTest do
       }
 
       warning = """
-      warning: List.old_flatten/1 is undefined or private. Did you mean one of:
+      warning: List.old_flatten/1 is undefined or private. Did you mean:
 
             * flatten/1
             * flatten/2
@@ -176,13 +176,13 @@ defmodule Module.Types.IntegrationTest do
       }
 
       warning = """
-      warning: A.a/1 is undefined or private. Did you mean one of:
+      warning: A.a/1 is undefined or private. Did you mean:
 
             * a/0
 
         a.ex:3: A.b/0
 
-      warning: A.b/1 is undefined or private. Did you mean one of:
+      warning: A.b/1 is undefined or private. Did you mean:
 
             * b/0
 
@@ -201,6 +201,8 @@ defmodule Module.Types.IntegrationTest do
 
           @file "external_source.ex"
           def c, do: E.no_module()
+
+          def i, do: Io.puts "hello"
         end
         """
       }
@@ -212,6 +214,9 @@ defmodule Module.Types.IntegrationTest do
       warning: E.no_module/0 is undefined (module E is not available or is yet to be defined)
         external_source.ex:5: A.c/0
 
+      warning: Io.puts/1 is undefined (module Io is not available or is yet to be defined)
+        a.ex:7: A.i/0
+
       """
 
       assert_warnings(files, warning)
@@ -304,12 +309,12 @@ defmodule Module.Types.IntegrationTest do
 
       warning = """
       warning: A2.no_func/0 is undefined (module A2 is not available or is yet to be defined)
-      Found at 2 locations:
+      Invalid call found at 2 locations:
         a.ex:8: A.d/0
         external_source.ex:5: A.b/0
 
       warning: A.no_func/0 is undefined or private
-      Found at 2 locations:
+      Invalid call found at 2 locations:
         a.ex:2: A.a/0
         a.ex:7: A.c/0
 
@@ -370,7 +375,7 @@ defmodule Module.Types.IntegrationTest do
       }
 
       warning = """
-      warning: A.to_list/1 is undefined or private. Did you mean one of:
+      warning: A.to_list/1 is undefined or private. Did you mean:
 
             * to_charlist/1
 
@@ -451,7 +456,7 @@ defmodule Module.Types.IntegrationTest do
       warning: MissingModule3.func/2 is undefined (module MissingModule3 is not available or is yet to be defined)
         a.ex:8: A.d/0
 
-      warning: B.func/3 is undefined or private. Did you mean one of:
+      warning: B.func/3 is undefined or private. Did you mean:
 
             * func/1
 
@@ -499,7 +504,7 @@ defmodule Module.Types.IntegrationTest do
         warning: MissingModule3.func/2 is undefined (module MissingModule3 is not available or is yet to be defined)
           a.ex:8: A.d/0
 
-        warning: B.func/3 is undefined or private. Did you mean one of:
+        warning: B.func/3 is undefined or private. Did you mean:
 
               * func/1
 
@@ -557,6 +562,29 @@ defmodule Module.Types.IntegrationTest do
     end
   end
 
+  describe "after_verify" do
+    test "reports functions" do
+      files = %{
+        "a.ex" => """
+        defmodule A do
+          @after_verify __MODULE__
+
+          def __after_verify__(__MODULE__) do
+            IO.warn "from after_verify", []
+          end
+        end
+        """
+      }
+
+      warning = """
+      warning: from after_verify
+
+      """
+
+      assert_warnings(files, warning)
+    end
+  end
+
   describe "deprecated" do
     test "reports functions" do
       files = %{
@@ -623,13 +651,13 @@ defmodule Module.Types.IntegrationTest do
       warnings = [
         """
         warning: A.__struct__/0 is deprecated. oops
-        Found at 2 locations:
+        Invalid call found at 2 locations:
           a.ex:4: A.match/1
           a.ex:5: A.build/1
         """,
         """
         warning: A.__struct__/0 is deprecated. oops
-        Found at 2 locations:
+        Invalid call found at 2 locations:
           b.ex:2: B.match/1
           b.ex:3: B.build/1
 
