@@ -17,6 +17,10 @@ defmodule Module.Types.TypesTest do
     end
   end
 
+  defmacro generated(ast) do
+    Macro.prewalk(ast, fn node -> Macro.update_meta(node, &([generated: true] ++ &1)) end)
+  end
+
   def __expr__({patterns, guards, body}) do
     with {:ok, _types, context} <-
            Pattern.of_head(patterns, guards, TypeHelper.new_stack(), TypeHelper.new_context()),
@@ -124,17 +128,17 @@ defmodule Module.Types.TypesTest do
              in expression:
 
                  # types_test.ex:1
-                 <<..., var::binary()>>
+                 <<..., var::binary>>
 
              where "var" was given the type integer() in:
 
                  # types_test.ex:1
-                 <<var::integer(), ...>>
+                 <<var::integer, ...>>
 
              where "var" was given the type binary() in:
 
                  # types_test.ex:1
-                 <<..., var::binary()>>
+                 <<..., var::binary>>
              """
     end
 
@@ -196,23 +200,49 @@ defmodule Module.Types.TypesTest do
                  # types_test.ex:1
                  is_binary(y)
 
-             where "y" was given the same type as "x" in:
+             where "x" was given the same type as "y" in:
 
                  # types_test.ex:1
                  x = y
 
+             where "y" was given the type integer() in:
+
+                 # types_test.ex:1
+                 is_integer(x)
+
              where "y" was given the type binary() in:
 
                  # types_test.ex:1
                  is_binary(y)
-
-             where "x" was given the type integer() in:
-
-                 # types_test.ex:1
-                 is_integer(x)
              """
     end
 
+    test "warns on guards from cases unless generated" do
+      string =
+        warning(
+          [var],
+          [is_integer(var)],
+          case var do
+            _ when is_binary(var) -> :ok
+          end
+        )
+
+      assert is_binary(string)
+
+      string =
+        generated(
+          warning(
+            [var],
+            [is_integer(var)],
+            case var do
+              _ when is_binary(var) -> :ok
+            end
+          )
+        )
+
+      assert string == :none
+    end
+
     test "only show relevant traces in warning" do
       string = warning([x = y, z], [is_integer(x) and is_binary(y) and is_boolean(z)], {x, y, z})
 
@@ -226,20 +256,20 @@ defmodule Module.Types.TypesTest do
                  # types_test.ex:1
                  is_binary(y)
 
-             where "y" was given the same type as "x" in:
+             where "x" was given the same type as "y" in:
 
                  # types_test.ex:1
                  x = y
 
+             where "y" was given the type integer() in:
+
+                 # types_test.ex:1
+                 is_integer(x)
+
              where "y" was given the type binary() in:
 
                  # types_test.ex:1
                  is_binary(y)
-
-             where "x" was given the type integer() in:
-
-                 # types_test.ex:1
-                 is_integer(x)
              """
     end
 
@@ -307,7 +337,7 @@ defmodule Module.Types.TypesTest do
              in expression:
 
                  # types_test.ex:1
-                 <<foo::integer()>>
+                 <<foo::integer>>
 
              where "foo" was given the type binary() in:
 
@@ -317,7 +347,74 @@ defmodule Module.Types.TypesTest do
              where "foo" was given the type integer() in:
 
                  # types_test.ex:1
-                 <<foo::integer()>>
+                 <<foo::integer>>
+             """
+    end
+
+    test "is_tuple warning" do
+      string = warning([foo], [is_tuple(foo)], {_} = foo)
+
+      assert string == """
+             incompatible types:
+
+                 tuple() !~ {dynamic()}
+
+             in expression:
+
+                 # types_test.ex:1
+                 {_} = foo
+
+             where "foo" was given the type tuple() in:
+
+                 # types_test.ex:1
+                 is_tuple(foo)
+
+             where "foo" was given the type {dynamic()} in:
+
+                 # types_test.ex:1
+                 {_} = foo
+
+             HINT: use pattern matching or "is_tuple(foo) and tuple_size(foo) == 1" to guard a sized tuple.
+             """
+    end
+
+    test "function call" do
+      string = warning([foo], [rem(foo, 2.0) == 0], foo)
+
+      assert string == """
+             expected Kernel.rem/2 to have signature:
+
+                 var1, float() -> dynamic()
+
+             but it has signature:
+
+                 integer(), integer() -> integer()
+
+             in expression:
+
+                 # types_test.ex:1
+                 rem(foo, 2.0)
+             """
+    end
+
+    test "operator call" do
+      string = warning([foo], [foo - :bar == 0], foo)
+
+      assert string == """
+             expected Kernel.-/2 to have signature:
+
+                 var1, :bar -> dynamic()
+
+             but it has signature:
+
+                 integer(), integer() -> integer()
+                 float(), integer() | float() -> float()
+                 integer() | float(), float() -> float()
+
+             in expression:
+
+                 # types_test.ex:1
+                 foo - :bar
              """
     end
   end
@@ -346,7 +443,17 @@ defmodule Module.Types.TypesTest do
                  # types_test.ex:5
                  %{"id" => user_id} = user
 
-             where "user" was given the same type as "amount" in:
+             where "amount" was given the type binary() in:
+
+                 # types_test.ex:3
+                 %{"amount" => amount} = event
+
+             where "amount" was given the same type as "user" in:
+
+                 # types_test.ex:4
+                 %{"user" => user} = event
+
+             where "user" was given the type binary() in:
 
                  # types_test.ex:4
                  %{"user" => user} = event
@@ -355,11 +462,6 @@ defmodule Module.Types.TypesTest do
 
                  # types_test.ex:5
                  %{"id" => user_id} = user
-
-             where "amount" was given the type binary() in:
-
-                 # types_test.ex:3
-                 %{"amount" => amount} = event
              """
     end
 
@@ -539,6 +641,41 @@ defmodule Module.Types.TypesTest do
                  %{foo: ^other_key} = event
              """
     end
+
+    test "expands map when maps are nested" do
+      string =
+        warning(
+          [map1, map2],
+          (
+            [_var1, _var2] = [map1, map2]
+            %{} = map1
+            %{} = map2.subkey
+          )
+        )
+
+      assert string == """
+             incompatible types:
+
+                 %{subkey: var1, optional(dynamic()) => dynamic()} !~ %{optional(dynamic()) => dynamic()} | %{optional(dynamic()) => dynamic()}
+
+             in expression:
+
+                 # types_test.ex:5
+                 map2.subkey
+
+             where "map2" was given the type %{optional(dynamic()) => dynamic()} | %{optional(dynamic()) => dynamic()} in:
+
+                 # types_test.ex:3
+                 [_var1, _var2] = [map1, map2]
+
+             where "map2" was given the type %{subkey: var1, optional(dynamic()) => dynamic()} (due to calling var.field) in:
+
+                 # types_test.ex:5
+                 map2.subkey
+
+             HINT: "var.field" (without parentheses) implies "var" is a map() while "var.fun()" (with parentheses) implies "var" is an atom()
+             """
+    end
   end
 
   describe "regressions" do
@@ -555,5 +692,69 @@ defmodule Module.Types.TypesTest do
                end
              ) == :none
     end
+
+    test "no-recursion on guards with map fields" do
+      assert warning(
+               [assigns],
+               (
+                 variable_enum = assigns.variable_enum
+
+                 case true do
+                   _ when variable_enum != nil -> assigns.variable_enum
+                 end
+               )
+             ) == :none
+    end
+
+    test "map patterns with pinned keys and field access" do
+      assert warning(
+               [x, y],
+               (
+                 key_var = y
+                 %{^key_var => _value} = x
+                 key_var2 = y
+                 %{^key_var2 => _value2} = x
+                 y.z
+               )
+             ) == :none
+    end
+
+    test "map patterns with pinned keys" do
+      assert warning(
+               [x, y],
+               (
+                 key_var = y
+                 %{^key_var => _value} = x
+                 key_var2 = y
+                 %{^key_var2 => _value2} = x
+                 key_var3 = y
+                 %{^key_var3 => _value3} = x
+               )
+             ) == :none
+    end
+
+    test "map updates with var key" do
+      assert warning(
+               [state0, key0],
+               (
+                 state1 = %{state0 | key0 => true}
+                 key1 = key0
+                 state2 = %{state1 | key1 => true}
+                 state2
+               )
+             ) == :none
+    end
+
+    test "nested map updates" do
+      assert warning(
+               [state],
+               (
+                 _foo = state.key.user_id
+                 _bar = state.key.user_id
+                 state = %{state | key: %{state.key | other_id: 1}}
+                 _baz = state.key.user_id
+               )
+             ) == :none
+    end
   end
 end
