@@ -16,8 +16,9 @@ defmodule ModuleTest.ToBeUsed do
     quote(do: def(before_compile, do: unquote(Macro.Env.vars(env))))
   end
 
-  defmacro __after_compile__(%Macro.Env{module: ModuleTest.ToUse, vars: []}, bin)
+  defmacro __after_compile__(%Macro.Env{module: ModuleTest.ToUse} = env, bin)
            when is_binary(bin) do
+    [] = Macro.Env.vars(env)
     # IO.puts "HELLO"
   end
 
@@ -32,7 +33,7 @@ end
 
 defmodule ModuleTest.ToUse do
   # Moving the next line around can make tests fail
-  35 = __ENV__.line
+  36 = __ENV__.line
   var = 1
   # Not available in callbacks
   _ = var
@@ -88,25 +89,35 @@ defmodule ModuleTest do
       end
 
     assert_raise ArgumentError,
-                 "could not call Module.__put_attribute__/4 because the module ModuleTest.Raise is in read-only mode (@after_compile)",
+                 "could not call Module.put_attribute/3 because the module ModuleTest.Raise is in read-only mode (@after_compile)",
                  fn ->
                    Module.create(ModuleTest.Raise, contents, __ENV__)
                  end
   end
 
   test "supports read access to module from __after_compile__/2" do
-    contents =
-      quote do
-        @after_compile __MODULE__
+    defmodule ModuleTest.NoRaise do
+      @after_compile __MODULE__
+      @foo 42
 
-        @foo 42
-
-        def __after_compile__(%Macro.Env{module: module}, bin) when is_binary(bin) do
-          Module.get_attribute(module, :foo)
-        end
+      def __after_compile__(%Macro.Env{module: module}, bin) when is_binary(bin) do
+        send(self(), Module.get_attribute(module, :foo))
       end
+    end
 
-    Module.create(ModuleTest.NoRaise, contents, __ENV__)
+    assert_received 42
+  end
+
+  test "supports @after_verify for inlined modules" do
+    defmodule ModuleTest.AfterVerify do
+      @after_verify __MODULE__
+
+      def __after_verify__(ModuleTest.AfterVerify) do
+        send(self(), ModuleTest.AfterVerify)
+      end
+    end
+
+    assert_received ModuleTest.AfterVerify
   end
 
   test "in memory modules are tagged as so" do
@@ -136,7 +147,7 @@ defmodule ModuleTest do
   end
 
   test "retrieves line from use callsite" do
-    assert ModuleTest.ToUse.line() == 40
+    assert ModuleTest.ToUse.line() == 41
   end
 
   ## Callbacks
@@ -151,13 +162,10 @@ defmodule ModuleTest do
   end
 
   def __on_definition__(env, kind, name, args, guards, expr) do
-    Process.put(env.module, :called)
+    Process.put(env.module, {args, guards, expr})
     assert env.module == ModuleTest.OnDefinition
     assert kind == :def
     assert name == :hello
-    assert [{:foo, _, _}, {:bar, _, _}] = args
-    assert [] = guards
-    assert [do: {:+, _, [{:foo, _, nil}, {:bar, _, nil}]}] = expr
     assert Module.defines?(env.module, {:hello, 2})
   end
 
@@ -165,12 +173,17 @@ defmodule ModuleTest do
     defmodule OnDefinition do
       @on_definition ModuleTest
 
+      def hello(foo, bar)
+
+      assert {[{:foo, _, _}, {:bar, _, _}], [], nil} = Process.get(ModuleTest.OnDefinition)
+
       def hello(foo, bar) do
         foo + bar
       end
-    end
 
-    assert Process.get(ModuleTest.OnDefinition) == :called
+      assert {[{:foo, _, _}, {:bar, _, _}], [], [do: {:+, _, [{:foo, _, nil}, {:bar, _, nil}]}]} =
+               Process.get(ModuleTest.OnDefinition)
+    end
   end
 
   defmacro __before_compile__(_) do
@@ -266,7 +279,7 @@ defmodule ModuleTest do
 
   @file "sample.ex"
   test "@file sets __ENV__.file" do
-    assert __ENV__.file == "sample.ex"
+    assert __ENV__.file == Path.absname("sample.ex")
   end
 
   test "@file raises when invalid" do
@@ -342,6 +355,19 @@ defmodule ModuleTest do
     end
   end
 
+  @compile {:no_warn_undefined, ModuleTracersSample}
+
+  test "create with propagated tracers" do
+    contents =
+      quote do
+        def world, do: true
+      end
+
+    env = %{__ENV__ | tracers: [:invalid]}
+    {:module, ModuleTracersSample, _, _} = Module.create(ModuleTracersSample, contents, env)
+    assert ModuleTracersSample.world()
+  end
+
   @compile {:no_warn_undefined, ModuleHygiene}
 
   test "create with aliases/var hygiene" do
@@ -442,17 +468,44 @@ defmodule ModuleTest do
 
   test "definitions in" do
     in_module do
-      def foo(1, 2, 3), do: 4
+      defp bar(), do: :ok
+      def foo(1, 2, 3), do: bar()
+
+      defmacrop macro_bar(), do: 4
+      defmacro macro_foo(1, 2, 3), do: macro_bar()
+
+      assert Module.definitions_in(__MODULE__) |> Enum.sort() ==
+               [{:bar, 0}, {:foo, 3}, {:macro_bar, 0}, {:macro_foo, 3}]
 
-      assert Module.definitions_in(__MODULE__) == [foo: 3]
       assert Module.definitions_in(__MODULE__, :def) == [foo: 3]
-      assert Module.definitions_in(__MODULE__, :defp) == []
+      assert Module.definitions_in(__MODULE__, :defp) == [bar: 0]
+      assert Module.definitions_in(__MODULE__, :defmacro) == [macro_foo: 3]
+      assert Module.definitions_in(__MODULE__, :defmacrop) == [macro_bar: 0]
 
       defoverridable foo: 3
 
-      assert Module.definitions_in(__MODULE__) == []
+      assert Module.definitions_in(__MODULE__) |> Enum.sort() ==
+               [{:bar, 0}, {:macro_bar, 0}, {:macro_foo, 3}]
+
       assert Module.definitions_in(__MODULE__, :def) == []
-      assert Module.definitions_in(__MODULE__, :defp) == []
+    end
+  end
+
+  test "get_definition/2 and delete_definition/2" do
+    in_module do
+      def foo(a, b), do: a + b
+
+      assert {:v1, :def, _,
+              [
+                {_, [{:a, _, nil}, {:b, _, nil}], [],
+                 {{:., _, [:erlang, :+]}, _, [{:a, _, nil}, {:b, _, nil}]}}
+              ]} = Module.get_definition(__MODULE__, {:foo, 2})
+
+      assert {:v1, :def, _, []} = Module.get_definition(__MODULE__, {:foo, 2}, skip_clauses: true)
+
+      assert Module.delete_definition(__MODULE__, {:foo, 2})
+      assert Module.get_definition(__MODULE__, {:foo, 2}) == nil
+      refute Module.delete_definition(__MODULE__, {:foo, 2})
     end
   end
 
@@ -467,10 +520,10 @@ defmodule ModuleTest do
         "tuple, got: {:foo, 256}"
 
     assert_raise ArgumentError, message, fn ->
-      Module.create(Foo, contents, __ENV__)
+      Module.create(MakeOverridable, contents, __ENV__)
     end
   after
-    purge(Foo)
+    purge(MakeOverridable)
   end
 
   test "raise when called with already compiled module" do
@@ -509,7 +562,7 @@ defmodule ModuleTest do
     end
   end
 
-  describe "has_attribute?/2" do
+  describe "has_attribute?/2 and attributes_in/2" do
     test "returns true when attribute has been defined" do
       in_module do
         @foo 1
@@ -521,9 +574,13 @@ defmodule ModuleTest do
         _ = @foo
 
         assert Module.has_attribute?(__MODULE__, :foo)
+        assert :foo in Module.attributes_in(__MODULE__)
         assert Module.has_attribute?(__MODULE__, :bar)
+        assert :bar in Module.attributes_in(__MODULE__)
         assert Module.has_attribute?(__MODULE__, :baz)
+        assert :baz in Module.attributes_in(__MODULE__)
         assert Module.has_attribute?(__MODULE__, :qux)
+        assert :qux in Module.attributes_in(__MODULE__)
       end
     end
 
@@ -542,4 +599,19 @@ defmodule ModuleTest do
       end
     end
   end
+
+  test "@on_load" do
+    Process.register(self(), :on_load_test_process)
+
+    defmodule OnLoadTest do
+      @on_load :on_load
+
+      defp on_load do
+        send(:on_load_test_process, :on_loaded)
+        :ok
+      end
+    end
+
+    assert_received :on_loaded
+  end
 end
