@@ -5,15 +5,6 @@ defmodule StringTest do
 
   doctest String
 
-  test "heredoc with heredoc inside interpolation" do
-    assert """
-           1
-           #{"""
-           2
-           """}
-           """ == "1\n2\n\n"
-  end
-
   test "next_codepoint/1" do
     assert String.next_codepoint("ésoj") == {"é", "soj"}
     assert String.next_codepoint(<<255>>) == {<<255>>, ""}
@@ -82,6 +73,19 @@ defmodule StringTest do
     assert String.split("noël", "") == ["", "n", "o", "ë", "l", ""]
     assert String.split("x-", "-", parts: 2, trim: true) == ["x"]
     assert String.split("x-x-", "-", parts: 3, trim: true) == ["x", "x"]
+
+    assert String.split("hello", []) == ["hello"]
+    assert String.split("hello", [], trim: true) == ["hello"]
+    assert String.split("", []) == [""]
+    assert String.split("", [], trim: true) == []
+
+    assert_raise ArgumentError, fn ->
+      String.split("a,b,c", [""])
+    end
+
+    assert_raise ArgumentError, fn ->
+      String.split("a,b,c", [""])
+    end
   end
 
   test "split/2,3 with regex" do
@@ -113,6 +117,17 @@ defmodule StringTest do
     assert String.splitter(" a b c ", " ", trim: true) |> Enum.to_list() == ["a", "b", "c"]
     assert String.splitter(" a b c ", " ", trim: true) |> Enum.take(1) == ["a"]
     assert String.splitter(" a b c ", " ", trim: true) |> Enum.take(2) == ["a", "b"]
+
+    assert String.splitter("hello", []) |> Enum.to_list() == ["hello"]
+    assert String.splitter("hello", [], trim: true) |> Enum.to_list() == ["hello"]
+    assert String.splitter("", []) |> Enum.to_list() == [""]
+    assert String.splitter("", [], trim: true) |> Enum.to_list() == []
+
+    assert String.splitter("1,2 3,4 5", "") |> Enum.take(4) == ["", "1", ",", "2"]
+
+    assert_raise ArgumentError, fn ->
+      String.splitter("a", [""])
+    end
   end
 
   test "split_at/2" do
@@ -140,6 +155,11 @@ defmodule StringTest do
     end
   end
 
+  test "split_at/2 with invalid guard" do
+    assert String.split_at(<<?a, 195, 10, ?a>>, 2) == {<<?a, 195>>, <<10, ?a>>}
+    assert String.split_at(<<107, 205, 135, 184>>, 1) == {<<107, 205, 135>>, <<184>>}
+  end
+
   test "upcase/1" do
     assert String.upcase("123 abcd 456 efg hij ( %$#) kl mnop @ qrst = -_ uvwxyz") ==
              "123 ABCD 456 EFG HIJ ( %$#) KL MNOP @ QRST = -_ UVWXYZ"
@@ -431,6 +451,14 @@ defmodule StringTest do
       assert String.replace("ELIXIR", "", ".") == ".E.L.I.X.I.R."
       assert String.replace("ELIXIR", "", ".", global: true) == ".E.L.I.X.I.R."
       assert String.replace("ELIXIR", "", ".", global: false) == ".ELIXIR"
+
+      assert_raise ArgumentError, fn ->
+        String.replace("elixir", [""], "")
+      end
+    end
+
+    test "with empty pattern list" do
+      assert String.replace("elixir", [], "anything") == "elixir"
     end
 
     test "with match pattern and string replacement" do
@@ -616,7 +644,7 @@ defmodule StringTest do
     end
   end
 
-  test "slice/2,3" do
+  test "slice/3" do
     assert String.slice("elixir", 1, 3) == "lix"
     assert String.slice("あいうえお", 2, 2) == "うえ"
     assert String.slice("ειξήριολ", 2, 3) == "ξήρ"
@@ -626,9 +654,9 @@ defmodule StringTest do
     assert String.slice("elixir", -3, 2) == "xi"
     assert String.slice("あいうえお", -4, 3) == "いうえ"
     assert String.slice("ειξήριολ", -5, 3) == "ήρι"
-    assert String.slice("elixir", -10, 1) == ""
-    assert String.slice("あいうえお", -10, 2) == ""
-    assert String.slice("ειξήριολ", -10, 3) == ""
+    assert String.slice("elixir", -10, 1) == "e"
+    assert String.slice("あいうえお", -10, 2) == "あい"
+    assert String.slice("ειξήριολ", -10, 3) == "ειξ"
     assert String.slice("elixir", 8, 2) == ""
     assert String.slice("あいうえお", 6, 2) == ""
     assert String.slice("ειξήριολ", 8, 1) == ""
@@ -636,13 +664,17 @@ defmodule StringTest do
     assert String.slice("elixir", 0, 0) == ""
     assert String.slice("elixir", 5, 0) == ""
     assert String.slice("elixir", -5, 0) == ""
+    assert String.slice("elixir", -10, 10) == "elixir"
     assert String.slice("", 0, 1) == ""
     assert String.slice("", 1, 1) == ""
+  end
 
+  test "slice/2" do
     assert String.slice("elixir", 0..-2) == "elixi"
     assert String.slice("elixir", 1..3) == "lix"
     assert String.slice("elixir", -5..-3) == "lix"
     assert String.slice("elixir", -5..3) == "lix"
+    assert String.slice("elixir", -10..10) == "elixir"
     assert String.slice("あいうえお", 2..3) == "うえ"
     assert String.slice("ειξήριολ", 2..4) == "ξήρ"
     assert String.slice("elixir", 3..6) == "xir"
@@ -661,6 +693,34 @@ defmodule StringTest do
     assert String.slice("あいうえお", -10..-15) == ""
     assert String.slice("hello あいうえお Unicode", 8..-1) == "うえお Unicode"
     assert String.slice("abc", -1..14) == "c"
+    assert String.slice("a·̀ͯ‿.⁀:", 0..-2) == "a·̀ͯ‿.⁀"
+
+    assert_raise FunctionClauseError, fn ->
+      String.slice(nil, 0..1)
+    end
+  end
+
+  test "slice/2 with steps" do
+    assert String.slice("elixir", 0..-2//2) == "eii"
+    assert String.slice("elixir", 1..3//2) == "lx"
+    assert String.slice("elixir", -5..-3//2) == "lx"
+    assert String.slice("elixir", -5..3//2) == "lx"
+    assert String.slice("あいうえお", 2..3//2) == "う"
+    assert String.slice("ειξήριολ", 2..4//2) == "ξρ"
+    assert String.slice("elixir", 3..6//2) == "xr"
+    assert String.slice("あいうえお", 3..7//2) == "え"
+    assert String.slice("ειξήριολ", 5..8//2) == "ιλ"
+    assert String.slice("elixir", -3..-2//2) == "x"
+    assert String.slice("あいうえお", -4..-2//2) == "いえ"
+    assert String.slice("ειξήριολ", -5..-3//2) == "ήι"
+    assert String.slice("elixir", 8..9//2) == ""
+    assert String.slice("", 0..0//2) == ""
+    assert String.slice("", 1..1//2) == ""
+    assert String.slice("あいうえお", -2..-4//2) == ""
+    assert String.slice("あいうえお", -10..-15//2) == ""
+    assert String.slice("hello あいうえお Unicode", 8..-1//2) == "うおUioe"
+    assert String.slice("abc", -1..14//2) == "c"
+    assert String.slice("a·̀ͯ‿.⁀:", 0..-2//2) == "a‿⁀"
   end
 
   test "valid?/1" do
@@ -693,6 +753,8 @@ defmodule StringTest do
     assert String.starts_with?("hello", "he")
     assert String.starts_with?("hello", "hello")
     refute String.starts_with?("hello", [])
+    assert String.starts_with?("hello", "")
+    assert String.starts_with?("hello", [""])
     assert String.starts_with?("hello", ["hellö", "hell"])
     assert String.starts_with?("エリクシア", "エリ")
     refute String.starts_with?("hello", "lo")
@@ -718,6 +780,8 @@ defmodule StringTest do
     assert String.contains?("elixir of life", "of")
     assert String.contains?("エリクシア", "シ")
     refute String.contains?("elixir of life", [])
+    assert String.contains?("elixir of life", "")
+    assert String.contains?("elixir of life", [""])
     assert String.contains?("elixir of life", ["mercury", "life"])
     refute String.contains?("elixir of life", "death")
     refute String.contains?("エリクシア", "仙")
@@ -881,4 +945,29 @@ defmodule StringTest do
     # FB13;FB13;FB13;0574 0576;0574 0576;
     assert String.normalize("\u0574\u0576", :nfkc) == "\u0574\u0576"
   end
+
+  # Carriage return can be a grapheme cluster if followed by
+  # newline so we test some corner cases here.
+  test "carriage return" do
+    assert String.at("\r\t\v", 0) == "\r"
+    assert String.at("\r\t\v", 1) == "\t"
+    assert String.at("\r\t\v", 2) == "\v"
+    assert String.at("\xFF\r\t\v", 1) == "\r"
+    assert String.at("\r\xFF\t\v", 2) == "\t"
+    assert String.at("\r\t\xFF\v", 3) == "\v"
+
+    assert String.last("\r\t\v") == "\v"
+    assert String.last("\r\xFF\t\xFF\v") == "\v"
+
+    assert String.next_grapheme("\r\t\v") == {"\r", "\t\v"}
+    assert String.next_grapheme("\t\v") == {"\t", "\v"}
+    assert String.next_grapheme("\v") == {"\v", ""}
+
+    assert String.length("\r\t\v") == 3
+    assert String.length("\r\xFF\t\v") == 4
+    assert String.length("\r\t\xFF\v") == 4
+
+    assert String.bag_distance("\r\t\xFF\v", "\xFF\r\n\xFF") == 0.25
+    assert String.split("\r\t\v", "") == ["", "\r", "\t", "\v", ""]
+  end
 end
