@@ -478,6 +478,21 @@ defmodule TypespecTest do
       assert {:type, _, :map_field_exact, [{:atom, _, :other}, {:type, _, :term, []}]} = arg2
     end
 
+    test "@type with an exception struct" do
+      bytecode =
+        test_module do
+          defexception [:message]
+          @type my_type :: %TypespecSample{}
+        end
+
+      assert [type: {:my_type, type, []}] = types(bytecode)
+      assert {:type, _, :map, [struct, arg1, arg2]} = type
+      assert {:type, _, :map_field_exact, struct_args} = struct
+      assert [{:atom, _, :__struct__}, {:atom, _, TypespecSample}] = struct_args
+      assert {:type, _, :map_field_exact, [{:atom, _, :__exception__}, {:atom, _, true}]} = arg1
+      assert {:type, _, :map_field_exact, [{:atom, _, :message}, {:type, _, :term, []}]} = arg2
+    end
+
     @fields Enum.map(10..42, &{:"f#{&1}", :ok})
 
     test "@type with a large struct" do
@@ -521,12 +536,23 @@ defmodule TypespecTest do
     end
 
     test "@type with a struct with undefined field" do
-      assert_raise CompileError, ~r"undefined field :no_field on struct TypespecSample", fn ->
-        test_module do
-          defstruct [:hello, :eric]
-          @type my_type :: %TypespecSample{no_field: :world}
-        end
-      end
+      assert_raise CompileError,
+                   ~r"undefined field :no_field on struct TypespecTest.TypespecSample",
+                   fn ->
+                     test_module do
+                       defstruct [:hello, :eric]
+                       @type my_type :: %TypespecSample{no_field: :world}
+                     end
+                   end
+
+      assert_raise CompileError,
+                   ~r"undefined field :no_field on struct TypespecTest.TypespecSample",
+                   fn ->
+                     test_module do
+                       defstruct [:hello, :eric]
+                       @type my_type :: %__MODULE__{no_field: :world}
+                     end
+                   end
     end
 
     test "@type when overriding Elixir built-in" do
@@ -1305,11 +1331,11 @@ defmodule TypespecTest do
         end
       end
 
-      msg = ~r/The type two_bad_variables\/2 has an invalid argument\(s\): :ok, Enum.t\(\)/
+      msg = ~r/The type two_bad_variables\/2 has an invalid argument\(s\): :ok, Enumerable.t\(\)/
 
       assert_raise CompileError, msg, fn ->
         test_module do
-          @type two_bad_variables(:ok, Enum.t()) :: {:ok, []}
+          @type two_bad_variables(:ok, Enumerable.t()) :: {:ok, []}
         end
       end
 
@@ -1449,33 +1475,33 @@ defmodule TypespecTest do
         case type do
           # These cases do not translate directly to their own string version.
           {:basic_list_type, _, _} ->
-            assert ast_string == "@type(basic_list_type() :: [integer()])"
+            assert ast_string == "@type basic_list_type() :: [integer()]"
 
           {:basic_nonempty_list_type, _, _} ->
-            assert ast_string == "@type(basic_nonempty_list_type() :: [integer(), ...])"
+            assert ast_string == "@type basic_nonempty_list_type() :: [integer(), ...]"
 
           {:literal_empty_bitstring, _, _} ->
-            assert ast_string == "@type(literal_empty_bitstring() :: <<_::0>>)"
+            assert ast_string == "@type literal_empty_bitstring() :: <<_::0>>"
 
           {:literal_keyword_list_fixed_key, _, _} ->
-            assert ast_string == "@type(literal_keyword_list_fixed_key() :: [{:key, integer()}])"
+            assert ast_string == "@type literal_keyword_list_fixed_key() :: [{:key, integer()}]"
 
           {:literal_keyword_list_fixed_key2, _, _} ->
-            assert ast_string == "@type(literal_keyword_list_fixed_key2() :: [{:key, integer()}])"
+            assert ast_string == "@type literal_keyword_list_fixed_key2() :: [{:key, integer()}]"
 
           {:literal_struct_all_fields_any_type, _, _} ->
             assert ast_string ==
-                     "@type(literal_struct_all_fields_any_type() :: %TypespecTest.SomeStruct{key: term()})"
+                     "@type literal_struct_all_fields_any_type() :: %TypespecTest.SomeStruct{key: term()}"
 
           {:literal_struct_all_fields_key_type, _, _} ->
             assert ast_string ==
-                     "@type(literal_struct_all_fields_key_type() :: %TypespecTest.SomeStruct{key: integer()})"
+                     "@type literal_struct_all_fields_key_type() :: %TypespecTest.SomeStruct{key: integer()}"
 
           {:built_in_fun, _, _} ->
-            assert ast_string == "@type(built_in_fun() :: (... -> any()))"
+            assert ast_string == "@type built_in_fun() :: (... -> any())"
 
           {:built_in_nonempty_list, _, _} ->
-            assert ast_string == "@type(built_in_nonempty_list() :: [...])"
+            assert ast_string == "@type built_in_nonempty_list() :: [...]"
 
           _ ->
             assert ast_string == Macro.to_string(definition)
@@ -1518,4 +1544,63 @@ defmodule TypespecTest do
                ["MACRO-last": 1, bar: 2, first: 1]
     end
   end
+
+  @tag tmp_dir: true
+  test "erlang module", c do
+    erlc(c, :typespec_test_mod, """
+    -module(typespec_test_mod).
+    -export([f/1]).
+    -export_type([t/1]).
+
+    -type t(X) :: list(X).
+
+    -spec f(X) -> X.
+    f(X) -> X.
+    """)
+
+    [type: type] = types(:typespec_test_mod)
+    line = 5
+
+    assert Code.Typespec.type_to_quoted(type) ==
+             {:"::", [], [{:t, [], [{:x, [line: line], nil}]}, [{:x, [line: line], nil}]]}
+
+    [{{:f, 1}, [spec]}] = specs(:typespec_test_mod)
+    line = 7
+
+    assert Code.Typespec.spec_to_quoted(:f, spec) ==
+             {:when, [line: line],
+              [
+                {:"::", [line: line],
+                 [{:f, [line: line], [{:x, [line: line], nil}]}, {:x, [line: line], nil}]},
+                [x: {:var, [line: line], nil}]
+              ]}
+  end
+
+  defp erlc(context, module, code) do
+    dir = context.tmp_dir
+
+    src_path = Path.join([dir, "#{module}.erl"])
+    src_path |> Path.dirname() |> File.mkdir_p!()
+    File.write!(src_path, code)
+
+    ebin_dir = Path.join(dir, "ebin")
+    File.mkdir_p!(ebin_dir)
+
+    {:ok, module} =
+      :compile.file(String.to_charlist(src_path), [
+        :debug_info,
+        outdir: String.to_charlist(ebin_dir)
+      ])
+
+    true = Code.prepend_path(ebin_dir)
+    {:module, ^module} = :code.load_file(module)
+
+    ExUnit.Callbacks.on_exit(fn ->
+      :code.purge(module)
+      :code.delete(module)
+      File.rm_rf!(dir)
+    end)
+
+    :ok
+  end
 end
