@@ -140,19 +140,21 @@ defmodule ExUnit.Case do
   The following tags are set automatically by ExUnit and are
   therefore reserved:
 
-    * `:module`     - the module on which the test was defined
+    * `:module` - the module on which the test was defined
 
-    * `:file`       - the file on which the test was defined
+    * `:file` - the file on which the test was defined
 
-    * `:line`       - the line on which the test was defined
+    * `:line` - the line on which the test was defined
 
-    * `:test`       - the test name
+    * `:test` - the test name
 
-    * `:async`      - if the test case is in async mode
+    * `:async` - if the test case is in async mode
 
     * `:registered` - used for `ExUnit.Case.register_attribute/3` values
 
-    * `:describe`   - the describe block the test belongs to
+    * `:describe` - the describe block the test belongs to
+
+    * `:describe_line` - the line the describe block begins on
 
   The following tags customize how tests behave:
 
@@ -253,11 +255,6 @@ defmodule ExUnit.Case do
 
   @doc false
   defmacro __using__(opts) do
-    unless Process.whereis(ExUnit.Server) do
-      raise "cannot use ExUnit.Case without starting the ExUnit application, " <>
-              "please call ExUnit.start() or explicitly start the :ex_unit app"
-    end
-
     quote do
       unless ExUnit.Case.__register__(__MODULE__, unquote(opts)) do
         use ExUnit.Callbacks
@@ -272,6 +269,12 @@ defmodule ExUnit.Case do
 
   @doc false
   def __register__(module, opts) do
+    unless Keyword.keyword?(opts) do
+      raise ArgumentError,
+            ~s(the argument passed to "use ExUnit.Case" must be a list of options, ) <>
+              ~s(got: #{inspect(opts)})
+    end
+
     registered? = Module.has_attribute?(module, :ex_unit_tests)
 
     unless registered? do
@@ -288,17 +291,18 @@ defmodule ExUnit.Case do
         :moduletag,
         :ex_unit_registered_test_attributes,
         :ex_unit_registered_describe_attributes,
-        :ex_unit_registered_module_attributes,
-        :ex_unit_used_describes
+        :ex_unit_registered_module_attributes
       ]
 
       Enum.each(attributes, &Module.register_attribute(module, &1, accumulate: true))
 
+      persisted_attributes = [:ex_unit_async]
+      Enum.each(persisted_attributes, &Module.register_attribute(module, &1, persist: true))
+
       attributes = [
         before_compile: ExUnit.Case,
         after_compile: ExUnit.Case,
-        ex_unit_async: false,
-        ex_unit_describe: nil
+        ex_unit_async: false
       ]
 
       Enum.each(attributes, fn {k, v} -> Module.put_attribute(module, k, v) end)
@@ -452,48 +456,23 @@ defmodule ExUnit.Case do
   setup steps involved.
   """
   defmacro describe(message, do: block) do
-    quote do
-      ExUnit.Case.__describe__(__MODULE__, __ENV__.line, unquote(message), fn ->
-        unquote(block)
-      end)
-    end
-  end
-
-  @doc false
-  def __describe__(module, line, message, fun) do
-    if Module.get_attribute(module, :ex_unit_describe) do
-      raise "cannot call \"describe\" inside another \"describe\". See the documentation " <>
-              "for ExUnit.Case.describe/2 on named setups and how to handle hierarchies"
-    end
-
-    cond do
-      not is_binary(message) ->
-        raise ArgumentError, "describe name must be a string, got: #{inspect(message)}"
-
-      message in Module.get_attribute(module, :ex_unit_used_describes) ->
-        raise ExUnit.DuplicateDescribeError,
-              "describe #{inspect(message)} is already defined in #{inspect(module)}"
-
-      true ->
-        :ok
-    end
-
-    if Module.get_attribute(module, :describetag) != [] do
-      raise "@describetag must be set inside describe/2 blocks"
-    end
-
-    Module.put_attribute(module, :ex_unit_describe, {line, message})
-    Module.put_attribute(module, :ex_unit_used_describes, message)
-
-    try do
-      fun.()
-    after
-      Module.put_attribute(module, :ex_unit_describe, nil)
-      Module.delete_attribute(module, :describetag)
-
-      for attribute <- Module.get_attribute(module, :ex_unit_registered_describe_attributes) do
-        Module.delete_attribute(module, attribute)
+    definition =
+      quote unquote: false do
+        defp unquote(name)(var!(context)), do: unquote(body)
       end
+
+    quote do
+      ExUnit.Callbacks.__describe__(__MODULE__, __ENV__.line, unquote(message), fn
+        message, describes ->
+          res = unquote(block)
+
+          case ExUnit.Callbacks.__describe__(__MODULE__, message, describes) do
+            {nil, nil} -> :ok
+            {name, body} -> unquote(definition)
+          end
+
+          res
+      end)
     end
   end
 
@@ -508,10 +487,18 @@ defmodule ExUnit.Case do
 
   @doc false
   def __after_compile__(%{module: module}, _) do
-    if Module.get_attribute(module, :ex_unit_async) do
-      ExUnit.Server.add_async_module(module)
-    else
-      ExUnit.Server.add_sync_module(module)
+    cond do
+      Process.whereis(ExUnit.Server) == nil ->
+        unless Code.can_await_module_compilation?() do
+          raise "cannot use ExUnit.Case without starting the ExUnit application, " <>
+                  "please call ExUnit.start() or explicitly start the :ex_unit app"
+        end
+
+      Module.get_attribute(module, :ex_unit_async) ->
+        ExUnit.Server.add_async_module(module)
+
+      true ->
+        ExUnit.Server.add_sync_module(module)
     end
   end
 
@@ -552,12 +539,13 @@ defmodule ExUnit.Case do
 
     {name, describe, describe_line, describetag} =
       case Module.get_attribute(mod, :ex_unit_describe) do
-        {line, describe} ->
-          description = :"#{test_type} #{describe} #{name}"
-          {description, describe, line, Module.get_attribute(mod, :describetag)}
+        {line, describe, _counter} ->
+          test_name = validate_test_name("#{test_type} #{describe} #{name}")
+          {test_name, describe, line, Module.get_attribute(mod, :describetag)}
 
-        _ ->
-          {:"#{test_type} #{name}", nil, nil, []}
+        nil ->
+          test_name = validate_test_name("#{test_type} #{name}")
+          {test_name, nil, nil, []}
       end
 
     if Module.defines?(mod, {name, 1}) do
@@ -566,8 +554,8 @@ defmodule ExUnit.Case do
 
     tags =
       (tags ++ tag ++ describetag ++ moduletag)
-      |> normalize_tags
-      |> validate_tags
+      |> normalize_tags()
+      |> validate_tags()
       |> Map.merge(%{
         line: line,
         file: file,
@@ -756,6 +744,19 @@ defmodule ExUnit.Case do
     tags
   end
 
+  defp validate_test_name(name) do
+    try do
+      String.to_atom(name)
+    rescue
+      SystemLimitError ->
+        raise SystemLimitError, """
+        the computed name of a test (which includes its type, the name of its parent describe \
+        block if present, and the test name itself) must be shorter than 255 characters, \
+        got: #{inspect(name)}
+        """
+    end
+  end
+
   defp normalize_tags(tags) do
     Enum.reduce(Enum.reverse(tags), %{}, fn
       tag, acc when is_atom(tag) -> Map.put(acc, tag, true)
