@@ -5,16 +5,20 @@ defmodule ExUnit.Diff do
   #
   # The Diff struct contains the fields `:equivalent?`, `:left`, `:right`.
   # The `:equivalent?` field represents if the `:left` and `:right` side are
-  # equivalents and contain no diffs. The `:left` and `:right` represent the sides
-  # of the comparison and contain ASTs with some special metas: `:diff` and
-  # `:diff_container`.
+  # equivalents and contain no diffs. The `:left` and `:right` represent the
+  # sides of the comparison as ASTs.
   #
-  # When meta `:diff` is `true`, the AST inside of it has no equivalent on the
-  # other side and should be rendered in a different color. If the AST is a
-  # literal and doesn't contain meta, the `:diff` meta will be placed in a
-  # wrapping block.
+  # The ASTs may be wrapped in blocks with two special metas:
+  #
+  #   * `:diff` - when `true`, the AST inside of it has no equivalent
+  #     on the other side and should be rendered in a different color
+  #
+  #   * `:delimiter` - that particular block should be rendered with
+  #     a delimiter
+  #
+  # Given blocks are do not appear on the left side or right, it is
+  # safe to perform such wrapping.
 
-  alias Code.Identifier
   alias Inspect.Algebra
 
   defstruct equivalent?: true,
@@ -33,10 +37,10 @@ defmodule ExUnit.Diff do
   end
 
   defp context_to_env({:match, pins}),
-    do: %{pins: Map.new(pins), context: :match, current_vars: %{}}
+    do: %{pins: Map.new(pins), context: :match, current_vars: %{}, hints: []}
 
   defp context_to_env(op) when op in [:==, :===],
-    do: %{pins: %{}, context: op, current_vars: %{}}
+    do: %{pins: %{}, context: op, current_vars: %{}, hints: []}
 
   # Main entry point for recursive diff
 
@@ -167,7 +171,7 @@ defmodule ExUnit.Diff do
   end
 
   defp diff_value(left, right, env) when is_binary(left) and is_binary(right) do
-    diff_string(left, right, ?\", env)
+    diff_string(left, right, ?", env)
   end
 
   defp diff_value(left, right, env) do
@@ -193,7 +197,7 @@ defmodule ExUnit.Diff do
     {guard_clause, guard_equivalent?} =
       if diff_expression.equivalent? do
         bindings = Map.merge(post_env.pins, post_env.current_vars)
-        diff_guard_clause(clause, Map.to_list(bindings))
+        diff_guard_clause(clause, bindings)
       else
         {clause, false}
       end
@@ -224,12 +228,24 @@ defmodule ExUnit.Diff do
   defp diff_guard_clause(quoted, bindings) do
     expanded =
       Macro.prewalk(quoted, fn
-        {_, [{:expanded, expanded} | _], _} -> expanded
-        other -> other
+        {_, [{:expanded, expanded} | _], _} ->
+          expanded
+
+        {var, _, context} = expr when is_atom(var) and is_atom(context) ->
+          if Map.has_key?(bindings, var_context(expr)) do
+            expr
+          else
+            throw(:abort)
+          end
+
+        other ->
+          other
       end)
 
-    {equivalent?, _bindings} = Code.eval_quoted(expanded, bindings)
+    {equivalent?, _bindings} = Code.eval_quoted(expanded, Map.to_list(bindings))
     {update_diff_meta(quoted, !equivalent?), equivalent?}
+  catch
+    :abort -> {quoted, false}
   end
 
   # Pins
@@ -245,8 +261,8 @@ defmodule ExUnit.Diff do
 
   # Vars
 
-  defp diff_var({name, meta, context} = left, right, env) do
-    identifier = {name, meta[:counter] || context}
+  defp diff_var(left, right, env) do
+    identifier = var_context(left)
 
     case env.current_vars do
       %{^identifier => ^right} ->
@@ -359,15 +375,17 @@ defmodule ExUnit.Diff do
     left = Enum.reverse(left)
 
     case extract_diff_meta(right) do
-      {[_ | _] = list, _diff?} ->
-        # Inner was escaped, diffs are inside
+      # Outer was escaped. Copy its diff? to its inner element and potentially escape it.
+      {{unescaped}, diff?} ->
+        rebuild_maybe_improper(unescaped, left, &(&1 |> escape() |> update_diff_meta(diff?)))
+
+      # We have a proper list, if there are any diffs, they will be inside, so copy as is.
+      {[_ | _] = list, false} ->
         rebuild_maybe_improper(list, left, & &1)
 
-      {list, diff?} ->
-        # Outer was escaped, move diffs and escape inside
-        list
-        |> unescape()
-        |> rebuild_maybe_improper(left, &(&1 |> escape() |> update_diff_meta(diff?)))
+      # The right itself is improper, so just add it as is.
+      {_, _} ->
+        rebuild_maybe_improper(right, left, & &1)
     end
   end
 
@@ -671,13 +689,11 @@ defmodule ExUnit.Diff do
          {:ok, inspect_left} <- safe_inspect(left),
          {:ok, inspect_right} <- safe_inspect(right) do
       if inspect_left != inspect_right do
-        diff_string(inspect_left, inspect_right, :none, env)
+        diff_string(inspect_left, inspect_right, nil, env)
       else
         # If they are equivalent, still use their inspected form
         case diff_map(kw, right, struct1, struct2, env) do
           {%{equivalent?: true}, ctx} ->
-            left = block_diff_container([inspect_left], :none)
-            right = block_diff_container([inspect_right], :none)
             {%__MODULE__{equivalent?: true, left: left, right: right}, ctx}
 
           diff_ctx ->
@@ -725,20 +741,27 @@ defmodule ExUnit.Diff do
     left = IO.iodata_to_binary(escaped_left)
     right = IO.iodata_to_binary(escaped_right)
 
-    diff =
-      cond do
-        diff_string?(left, right) ->
+    cond do
+      left == right ->
+        {string_script_to_diff([eq: left], delimiter, true, [], []), env}
+
+      diff_string?(left, right) ->
+        diff =
           String.myers_difference(left, right)
           |> string_script_to_diff(delimiter, true, [], [])
 
-        left == right ->
-          string_script_to_diff([eq: left], delimiter, true, [], [])
+        env =
+          if String.equivalent?(left, right) do
+            add_hint(env, :equivalent_but_different_strings)
+          else
+            env
+          end
 
-        true ->
-          string_script_to_diff([del: left, ins: right], delimiter, true, [], [])
-      end
+        {diff, env}
 
-    {diff, env}
+      true ->
+        {string_script_to_diff([del: left, ins: right], delimiter, true, [], []), env}
+    end
   end
 
   # Concat all the literals on `left` and split `right` based on the size of
@@ -751,7 +774,7 @@ defmodule ExUnit.Diff do
   end
 
   defp diff_string_concat(left, nil, indexes, _left_length, right, env) do
-    {parsed_diff, parsed_post_env} = diff_string(left, right, ?\", env)
+    {parsed_diff, parsed_post_env} = diff_string(left, right, ?", env)
     left_diff = rebuild_concat_string(parsed_diff.left, nil, indexes)
 
     diff = %__MODULE__{parsed_diff | left: left_diff}
@@ -760,7 +783,7 @@ defmodule ExUnit.Diff do
 
   defp diff_string_concat(left, quoted, indexes, left_length, right, env) do
     {parsed_right, continue_right} = String.split_at(right, left_length)
-    {parsed_diff, parsed_post_env} = diff_string(left, parsed_right, ?\", env)
+    {parsed_diff, parsed_post_env} = diff_string(left, parsed_right, ?", env)
     {quoted_diff, quoted_post_env} = diff(quoted, continue_right, parsed_post_env)
 
     diff =
@@ -820,7 +843,7 @@ defmodule ExUnit.Diff do
     {:<>, [], [next, rebuilt_right]}
   end
 
-  defp next_concat_result({:__block__, [{:diff_container, _} | _] = meta, list}, index) do
+  defp next_concat_result({:__block__, [{:delimiter, _} | _] = meta, list}, index) do
     {next, continue} = next_concat_result(list, index)
     {{:__block__, meta, next}, {:__block__, meta, continue}}
   end
@@ -846,15 +869,15 @@ defmodule ExUnit.Diff do
     end
   end
 
-  defp block_diff_container(contents, :none),
+  defp block_delimiter(contents, nil),
     do: {:__block__, [], contents}
 
-  defp block_diff_container(contents, container),
-    do: {:__block__, [diff_container: container], contents}
+  defp block_delimiter(contents, container),
+    do: {:__block__, [delimiter: container], contents}
 
   defp string_script_to_diff([], delimiter, equivalent?, left, right) do
-    left = block_diff_container(Enum.reverse(left), delimiter)
-    right = block_diff_container(Enum.reverse(right), delimiter)
+    left = block_delimiter(Enum.reverse(left), delimiter)
+    right = block_delimiter(Enum.reverse(right), delimiter)
     %__MODULE__{equivalent?: equivalent?, left: left, right: right}
   end
 
@@ -873,7 +896,7 @@ defmodule ExUnit.Diff do
   # Numbers
 
   defp diff_number(left, right, env) do
-    diff_string(inspect(left), inspect(right), :none, env)
+    diff_string(inspect(left), inspect(right), nil, env)
   end
 
   # Algebra
@@ -886,9 +909,13 @@ defmodule ExUnit.Diff do
   end
 
   defp safe_to_algebra({:__block__, meta, list}, diff_wrapper) do
-    content_docs = Enum.map(list, &string_to_algebra(&1, diff_wrapper))
+    content_docs =
+      for item <- list do
+        # Each element of the list is either a literal or a wrapped literal in a diff
+        wrap_on_diff(item, fn literal, _ -> literal end, diff_wrapper)
+      end
 
-    if container = meta[:diff_container] do
+    if container = meta[:delimiter] do
       delimiter = to_string([container])
       Algebra.concat([delimiter] ++ content_docs ++ [delimiter])
     else
@@ -941,14 +968,6 @@ defmodule ExUnit.Diff do
     inspect(literal)
   end
 
-  def string_to_algebra(quoted, diff_wrapper) do
-    wrap_on_diff(quoted, &safe_string_to_algebra/2, diff_wrapper)
-  end
-
-  def safe_string_to_algebra(literal, _diff_wrapper) do
-    literal
-  end
-
   defp keyword_to_algebra(quoted, diff_wrapper) do
     wrap_on_diff(quoted, &safe_keyword_to_algebra/2, diff_wrapper)
   end
@@ -968,7 +987,7 @@ defmodule ExUnit.Diff do
   end
 
   defp safe_key_to_algebra(key, _diff_wrapper) do
-    Identifier.inspect_as_key(key)
+    Macro.inspect_atom(:key, key)
   end
 
   defp map_item_to_algebra(quoted, diff_wrapper) do
@@ -1008,8 +1027,12 @@ defmodule ExUnit.Diff do
     wrap_on_diff(quoted, &safe_struct_to_algebra/2, diff_wrapper)
   end
 
+  defp safe_struct_to_algebra({:^, _, _} = name, _diff_wrapper) do
+    Macro.to_string(name)
+  end
+
   defp safe_struct_to_algebra(name, _diff_wrapper) do
-    Code.Identifier.inspect_as_atom(name)
+    Macro.inspect_atom(:literal, name)
   end
 
   defp select_list_item_algebra(list) do
@@ -1031,6 +1054,10 @@ defmodule ExUnit.Diff do
 
   # Diff helpers
 
+  defp add_hint(%{hints: hints} = env, hint) do
+    if hint in hints, do: env, else: %{env | hints: [hint | hints]}
+  end
+
   # The left side is only escaped if it is a value
   defp maybe_escape(other, %{context: :match}), do: other
   defp maybe_escape(other, _env), do: escape(other)
@@ -1041,9 +1068,6 @@ defmodule ExUnit.Diff do
 
   defp escape_pair({key, value}), do: {escape(key), escape(value)}
 
-  defp unescape({other}), do: other
-  defp unescape(other), do: other
-
   defp merge_diff(%__MODULE__{} = result1, %__MODULE__{} = result2, fun) do
     {left, right} = fun.(result1.left, result2.left, result1.right, result2.right)
 
