@@ -131,7 +131,7 @@ defmodule ExUnit.DocTest do
       iex(1)> raise "some error"
       ** (RuntimeError) some error
 
-  Doctest will looking for a line starting with `** (` and it will parse it
+  Doctest will look for a line starting with `** (` and it will parse it
   accordingly to extract the exception name and message. The exception parser
   will consider all following lines part of the exception message until there
   is an empty line or there is a new expression prefixed with `iex>`.
@@ -159,7 +159,7 @@ defmodule ExUnit.DocTest do
       module = Keyword.fetch!(opts, :module)
       message = Keyword.fetch!(opts, :message)
 
-      file = module.__info__(:compile)[:source] |> Path.relative_to_cwd()
+      file = module.module_info(:compile)[:source] |> Path.relative_to_cwd()
       info = Exception.format_file_line(file, opts[:line])
       %__MODULE__{message: info <> " " <> message}
     end
@@ -182,8 +182,7 @@ defmodule ExUnit.DocTest do
     * `:import` - when `true`, one can test a function defined in the module
       without referring to the module name. However, this is not feasible when
       there is a clash with a module like `Kernel`. In these cases, `:import`
-      should be set to `false` and a full `Module.function` construct should be
-      used.
+      should be set to `false` and `Module.function(...)` should be used instead.
 
     * `:tags` - a list of tags to apply to all generated doctests.
 
@@ -223,7 +222,7 @@ defmodule ExUnit.DocTest do
   @doc false
   def __file__(module) do
     source =
-      module.__info__(:compile)[:source] ||
+      module.module_info(:compile)[:source] ||
         raise "#{inspect(module)} does not have compile-time source information"
 
     "(for doctest at) " <> Path.relative_to_cwd(source)
@@ -234,14 +233,14 @@ defmodule ExUnit.DocTest do
     do_import = Keyword.get(opts, :import, false)
 
     extract(module)
-    |> filter_by_opts(opts)
+    |> filter_by_opts(module, opts)
     |> Stream.with_index()
     |> Enum.map(fn {test, acc} ->
       compile_test(test, module, do_import, acc + 1)
     end)
   end
 
-  defp filter_by_opts(tests, opts) do
+  defp filter_by_opts(tests, module, opts) do
     except = Keyword.get(opts, :except, [])
 
     case Keyword.fetch(opts, :only) do
@@ -249,15 +248,43 @@ defmodule ExUnit.DocTest do
         []
 
       {:ok, only} ->
-        tests
-        |> Stream.reject(&(&1.fun_arity in except))
-        |> Stream.filter(&(&1.fun_arity in only))
+        filter_tests(module, tests, except, only)
 
       :error ->
         Stream.reject(tests, &(&1.fun_arity in except))
     end
   end
 
+  defp filter_tests(module, tests, except, only) do
+    {filtered_tests, fun_arities} =
+      for test <- tests,
+          test.fun_arity not in except,
+          test.fun_arity in only,
+          reduce: {[], []} do
+        {tests, fun_arities} -> {[test | tests], [test.fun_arity | fun_arities]}
+      end
+
+    case only -- fun_arities do
+      [] ->
+        filtered_tests
+
+      undefined_fun_arities ->
+        pluralized = pluralize_list_name("function", undefined_fun_arities)
+
+        functions =
+          Enum.map_join(undefined_fun_arities, "\n    ", fn {fun, arity} ->
+            Exception.format_mfa(module, fun, arity)
+          end)
+
+        raise Error,
+          module: module,
+          message: "undefined or private #{pluralized} given to doctest:\n\n    #{functions}\n\n"
+    end
+  end
+
+  defp pluralize_list_name(name, [_]), do: name
+  defp pluralize_list_name(name, _), do: ExUnit.plural_rule(name)
+
   ## Compilation of extracted tests
 
   defp compile_test(test, module, do_import, n) do
@@ -273,7 +300,7 @@ defmodule ExUnit.DocTest do
   end
 
   defp test_content(%{exprs: exprs, line: line}, module, do_import) do
-    file = module.__info__(:compile)[:source] |> Path.relative_to_cwd()
+    file = module.module_info(:compile)[:source] |> Path.relative_to_cwd()
     location = [line: line, file: file]
     stack = Macro.escape([{module, :__MODULE__, 0, location}])
 
@@ -452,7 +479,7 @@ defmodule ExUnit.DocTest do
           if e.__struct__ == TokenMissingError and expr =~ @opaque_type_regex do
             message <>
               """
-              . If you are planning to assert on the result of an iex> expression \
+              \nIf you are planning to assert on the result of an iex> expression \
               which contains a value inspected as #Name<...>, please make sure \
               the inspected value is placed at the beginning of the expression; \
               otherwise Elixir will treat it as a comment due to the leading sign #.\
@@ -524,7 +551,7 @@ defmodule ExUnit.DocTest do
     do: []
 
   defp extract_tests(line_no, doc, module) do
-    all_lines = String.split(doc, "\n", trim: false)
+    all_lines = String.split(doc, ["\r\n", "\n"], trim: false)
     lines = adjust_indent(all_lines, line_no + 1, module)
     extract_tests(lines, "", "", [], true, module, "")
   end
