@@ -197,7 +197,7 @@ defmodule ExUnit.Formatter do
   defp format_exception(test, %FunctionClauseError{} = struct, stack, _width, formatter, _pad) do
     {blamed, stack} = Exception.blame(:error, struct, stack)
     banner = Exception.format_banner(:error, struct)
-    blamed = FunctionClauseError.blame(blamed, &inspect/1, &blame_match(&1, &2, formatter))
+    blamed = FunctionClauseError.blame(blamed, &inspect/1, &blame_match(&1, formatter))
     message = error_info(banner, formatter) <> "\n" <> pad(String.trim_leading(blamed, "\n"))
     {message <> format_code(test, stack, formatter), stack}
   end
@@ -292,15 +292,12 @@ defmodule ExUnit.Formatter do
     nil
   end
 
-  defp blame_match(%{match?: true, node: node}, _, _formatter),
+  defp blame_match(%{match?: true, node: node}, _formatter),
     do: Macro.to_string(node)
 
-  defp blame_match(%{match?: false, node: node}, _, formatter),
+  defp blame_match(%{match?: false, node: node}, formatter),
     do: formatter.(:blame_diff, Macro.to_string(node))
 
-  defp blame_match(_, string, _formatter),
-    do: string
-
   defp format_meta(fields, formatter, padding, padding_size) do
     for {label, value} <- fields, has_value?(value) do
       [padding, format_label(label, formatter, padding_size), value, "\n"]
@@ -391,7 +388,7 @@ defmodule ExUnit.Formatter do
        ) do
     formatted_mailbox =
       for message <- mailbox do
-        {pattern, value} =
+        {pattern, value, _warnings} =
           format_sides(
             left,
             message,
@@ -422,8 +419,8 @@ defmodule ExUnit.Formatter do
          padding_size,
          width
        ) do
-    {left, right} = format_sides(left, right, context, formatter, padding_size, width)
-    [left: left, right: right]
+    {left, right, extras} = format_sides(left, right, context, formatter, padding_size, width)
+    [left: left, right: right] ++ extras
   end
 
   defp format_sides(left, right, context, formatter, padding_size, width) do
@@ -431,7 +428,7 @@ defmodule ExUnit.Formatter do
     content_width = if width == :infinity, do: width, else: width - padding_size
 
     case format_diff(left, right, context, formatter) do
-      {result, _env} ->
+      {result, env} ->
         left =
           result.left
           |> Diff.to_algebra(&colorize_diff_delete(&1, formatter))
@@ -444,13 +441,20 @@ defmodule ExUnit.Formatter do
           |> Algebra.nest(padding_size)
           |> Algebra.format(content_width)
 
-        {left, right}
+        {left, right, Enum.map(env.hints, &{:hint, format_hint(&1)})}
+
+      nil when is_atom(context) ->
+        {if_value(left, inspect), if_value(right, inspect), []}
 
       nil ->
-        {if_value(left, &code_multiline(&1, padding_size)), if_value(right, inspect)}
+        {if_value(left, &code_multiline(&1, padding_size)), if_value(right, inspect), []}
     end
   end
 
+  defp format_hint(:equivalent_but_different_strings) do
+    "you are comparing strings that have the same visual representation but are made of different Unicode codepoints"
+  end
+
   defp format_diff(left, right, context, formatter) do
     if has_value?(left) and has_value?(right) and formatter.(:diff_enabled?, false) do
       find_diff(left, right, context)
