@@ -43,7 +43,7 @@ defmodule ExUnit.Runner do
 
     start_time = System.monotonic_time()
     EM.suite_started(config.manager, opts)
-    async_stop_time = loop(config, :async, %{}, false)
+    async_stop_time = async_loop(config, %{}, false)
     stop_time = System.monotonic_time()
 
     if max_failures_reached?(config) do
@@ -92,52 +92,50 @@ defmodule ExUnit.Runner do
     |> Keyword.put(:include, include)
   end
 
-  defp loop(config, :async, running, async_once?) do
+  defp async_loop(config, running, async_once?) do
     available = config.max_cases - map_size(running)
 
     cond do
       # No modules available, wait for one
       available <= 0 ->
-        wait_until_available(config, :async, running, async_once?)
+        running = wait_until_available(config, running)
+        async_loop(config, running, async_once?)
 
       # Slots are available, start with async modules
       modules = ExUnit.Server.take_async_modules(available) ->
-        spawn_modules(config, modules, :async, running, true)
+        running = spawn_modules(config, modules, running)
+        async_loop(config, running, true)
 
       true ->
-        modules = ExUnit.Server.take_sync_modules()
-        loop(config, modules, running, async_once?)
+        sync_modules = ExUnit.Server.take_sync_modules()
+
+        # Wait for all async modules
+        0 =
+          running
+          |> Enum.reduce(running, fn _, acc -> wait_until_available(config, acc) end)
+          |> map_size()
+
+        async_stop_time = if async_once?, do: System.monotonic_time(), else: nil
+
+        # Run all sync modules directly
+        for module <- sync_modules do
+          running = spawn_modules(config, [module], %{})
+          running != %{} and wait_until_available(config, running)
+        end
+
+        async_stop_time
     end
   end
 
-  defp loop(config, modules, running, async_stop_time) do
-    case modules do
-      _ when running != %{} ->
-        wait_until_available(config, modules, running, async_stop_time)
-
-      # So we can start all sync modules
-      [head | tail] ->
-        spawn_modules(config, [head], tail, running, async_stop_time(async_stop_time))
-
-      # No more modules, we are done!
-      [] ->
-        async_stop_time(async_stop_time)
-    end
-  end
-
-  defp async_stop_time(false = _async_once?), do: nil
-  defp async_stop_time(true = _async_once?), do: System.monotonic_time()
-  defp async_stop_time(async_stop_time), do: async_stop_time
-
-  # Loop expecting down messages from the spawned modules.
+  # Expect down messages from the spawned modules.
   #
   # We first look at the sigquit signal because we don't want
   # to spawn new test cases when we know we will have to handle
   # sigquit next.
   #
   # Otherwise, whenever a module has finished executing, update
-  # the runnig modules and attempt to spawn new ones.
-  defp wait_until_available(config, modules, running, async_timing) do
+  # the running modules and attempt to spawn new ones.
+  defp wait_until_available(config, running) do
     receive do
       {ref, pid, :sigquit} ->
         sigquit(config, ref, pid, running)
@@ -148,25 +146,25 @@ defmodule ExUnit.Runner do
             sigquit(config, ref, pid, running)
 
           {:DOWN, ref, _, _, _} when is_map_key(running, ref) ->
-            loop(config, modules, Map.delete(running, ref), async_timing)
+            Map.delete(running, ref)
         end
     end
   end
 
-  defp spawn_modules(config, [], modules_remaining, running, async_timing) do
-    loop(config, modules_remaining, running, async_timing)
+  defp spawn_modules(_config, [], running) do
+    running
   end
 
-  defp spawn_modules(config, [module | modules], modules_remaining, running, async_timing) do
+  defp spawn_modules(config, [module | modules], running) do
     if max_failures_reached?(config) do
-      loop(config, modules_remaining, running, async_timing)
+      running
     else
       {pid, ref} = spawn_monitor(fn -> run_module(config, module) end)
-      spawn_modules(config, modules, modules_remaining, Map.put(running, ref, pid), async_timing)
+      spawn_modules(config, modules, Map.put(running, ref, pid))
     end
   end
 
-  ## stacktrace
+  ## Stacktrace
 
   # Assertions can pop-up in the middle of the stack
   def prune_stacktrace([{ExUnit.Assertions, _, _, _} | t]), do: prune_stacktrace(t)
@@ -391,10 +389,22 @@ defmodule ExUnit.Runner do
     context
   end
 
+  defp short_hash(module, test_name) do
+    (module <> "/" <> test_name)
+    |> :erlang.md5()
+    |> Base.encode16(case: :lower)
+    |> binary_slice(0..7)
+  end
+
   defp create_tmp_dir!(test, extra_path, context) do
-    module = escape_path(inspect(test.module))
-    name = escape_path(to_string(test.name))
-    path = ["tmp", module, name, extra_path] |> Path.join() |> Path.expand()
+    module_string = inspect(test.module)
+    name_string = to_string(test.name)
+
+    module = escape_path(module_string)
+    name = escape_path(name_string)
+    short_hash = short_hash(module_string, name_string)
+
+    path = ["tmp", module, "#{name}-#{short_hash}", extra_path] |> Path.join() |> Path.expand()
     File.rm_rf!(path)
     File.mkdir_p!(path)
     Map.put(context, :tmp_dir, path)
@@ -444,10 +454,10 @@ defmodule ExUnit.Runner do
     exec_on_exit(test, test_pid, timeout)
   end
 
-  defp spawn_test_monitor(config, test, parent_pid, context) do
+  defp spawn_test_monitor(%{seed: seed}, test, parent_pid, context) do
     spawn_monitor(fn ->
       ExUnit.OnExitHandler.register(self())
-      generate_test_seed(config, test)
+      generate_test_seed(seed, test)
 
       {time, test} =
         :timer.tc(fn ->
@@ -519,7 +529,7 @@ defmodule ExUnit.Runner do
 
   ## Helpers
 
-  defp generate_test_seed(%{seed: seed}, %ExUnit.Test{module: module, name: name}) do
+  defp generate_test_seed(seed, %ExUnit.Test{module: module, name: name}) do
     :rand.seed(@rand_algorithm, {:erlang.phash2(module), :erlang.phash2(name), seed})
   end
 
