@@ -16,6 +16,23 @@ defmodule ExUnit.AssertionsTest.BrokenError do
   end
 end
 
+defmodule ExUnit.AssertionsTest.OverrideOperator do
+  import Kernel, except: [{:=~, 2}]
+
+  defmacro __using__(_) do
+    quote do
+      import Kernel, except: [{:=~, 2}]
+      import ExUnit.AssertionsTest.OverrideOperator
+    end
+  end
+
+  defmacro left =~ right do
+    quote do
+      match?(unquote(left), unquote(right))
+    end
+  end
+end
+
 alias ExUnit.AssertionsTest.{BrokenError, Value}
 
 defmodule ExUnit.AssertionsTest do
@@ -73,7 +90,7 @@ defmodule ExUnit.AssertionsTest do
       "This should never be tested" = assert Value.falsy()
     rescue
       error in [ExUnit.AssertionError] ->
-        "assert(Value.falsy())" = error.expr |> Macro.to_string()
+        "assert Value.falsy()" = error.expr |> Macro.to_string()
         "Expected truthy, got nil" = error.message
     end
   end
@@ -107,7 +124,7 @@ defmodule ExUnit.AssertionsTest do
       error in [ExUnit.AssertionError] ->
         1 = error.right
         2 = error.left
-        "assert(1 + 1 == 1)" = error.expr |> Macro.to_string()
+        "assert 1 + 1 == 1" = error.expr |> Macro.to_string()
     end
   end
 
@@ -118,7 +135,7 @@ defmodule ExUnit.AssertionsTest do
       error in [ExUnit.AssertionError] ->
         1 = error.left
         2 = error.right
-        "assert(1 == 1 + 1)" = error.expr |> Macro.to_string()
+        "assert 1 == 1 + 1" = error.expr |> Macro.to_string()
     end
   end
 
@@ -136,6 +153,15 @@ defmodule ExUnit.AssertionsTest do
     assert argless_macro == 1
   end
 
+  defmodule OperatorOverrideTest do
+    use ExUnit.Case
+    use ExUnit.AssertionsTest.OverrideOperator
+
+    test "assert when operator be overrode" do
+      assert [_a, _b] =~ [1, 2]
+    end
+  end
+
   test "refute when value is falsy" do
     false = refute false
     nil = refute Value.falsy()
@@ -147,7 +173,7 @@ defmodule ExUnit.AssertionsTest do
       raise "refute was supposed to fail"
     rescue
       error in [ExUnit.AssertionError] ->
-        "refute(Value.truthy())" = Macro.to_string(error.expr)
+        "refute Value.truthy()" = Macro.to_string(error.expr)
         "Expected false or nil, got :truthy" = error.message
     end
   end
@@ -203,7 +229,7 @@ defmodule ExUnit.AssertionsTest do
         "match (=) failed\n" <> "The following variables were pinned:\n" <> "  a = 1" =
           error.message
 
-        "assert({^a, 1} = Value.tuple())" = Macro.to_string(error.expr)
+        "assert {^a, 1} = Value.tuple()" = Macro.to_string(error.expr)
     end
   end
 
@@ -216,7 +242,7 @@ defmodule ExUnit.AssertionsTest do
     rescue
       error in [ExUnit.AssertionError] ->
         "match (=) failed" = error.message
-        "assert({^var!(a, Elixir), 1} = Value.tuple())" = Macro.to_string(error.expr)
+        "assert {^var!(a, Elixir), 1} = Value.tuple()" = Macro.to_string(error.expr)
     end
   end
 
@@ -228,7 +254,7 @@ defmodule ExUnit.AssertionsTest do
     rescue
       error in [ExUnit.AssertionError] ->
         "match (match?) failed" = error.message
-        "assert(match?({:ok, _}, error(true)))" = Macro.to_string(error.expr)
+        "assert match?({:ok, _}, error(true))" = Macro.to_string(error.expr)
         "{:error, true}" = Macro.to_string(error.right)
     end
   end
@@ -242,8 +268,7 @@ defmodule ExUnit.AssertionsTest do
       error in [ExUnit.AssertionError] ->
         "match (match?) failed" = error.message
 
-        "assert(match?(tuple when not(is_tuple(tuple)), error(true)))" =
-          Macro.to_string(error.expr)
+        "assert match?(tuple when not is_tuple(tuple), error(true))" = Macro.to_string(error.expr)
 
         "{:error, true}" = Macro.to_string(error.right)
     end
@@ -257,7 +282,7 @@ defmodule ExUnit.AssertionsTest do
     rescue
       error in [ExUnit.AssertionError] ->
         "match (match?) succeeded, but should have failed" = error.message
-        "refute(match?({:error, _}, error(true)))" = Macro.to_string(error.expr)
+        "refute match?({:error, _}, error(true))" = Macro.to_string(error.expr)
         "{:error, true}" = Macro.to_string(error.right)
     end
   end
@@ -271,7 +296,7 @@ defmodule ExUnit.AssertionsTest do
       error in [ExUnit.AssertionError] ->
         "match (match?) failed\nThe following variables were pinned:\n  a = 1" = error.message
 
-        "assert(match?({^a, 1}, Value.tuple()))" = Macro.to_string(error.expr)
+        "assert match?({^a, 1}, Value.tuple())" = Macro.to_string(error.expr)
     end
   end
 
@@ -288,7 +313,7 @@ defmodule ExUnit.AssertionsTest do
           a = 2\
         """ = error.message
 
-        "refute(match?({^a, 1}, Value.tuple()))" = Macro.to_string(error.expr)
+        "refute match?({^a, 1}, Value.tuple())" = Macro.to_string(error.expr)
     end
   end
 
@@ -385,7 +410,7 @@ defmodule ExUnit.AssertionsTest do
         Showing 1 of 1 message in the mailbox\
         """ = error.message
 
-        "assert_received({:status, ^status})" = Macro.to_string(error.expr)
+        "assert_received {:status, ^status}" = Macro.to_string(error.expr)
         "{:status, ^status}" = Macro.to_string(error.left)
     end
   end
@@ -405,7 +430,7 @@ defmodule ExUnit.AssertionsTest do
         Showing 1 of 1 message in the mailbox\
         """ = error.message
 
-        "assert_received({:status, ^status, ^status})" = Macro.to_string(error.expr)
+        "assert_received {:status, ^status, ^status}" = Macro.to_string(error.expr)
         "{:status, ^status, ^status}" = Macro.to_string(error.left)
         "\n\nAssertion failed" <> _ = Exception.message(error)
     end
@@ -428,7 +453,7 @@ defmodule ExUnit.AssertionsTest do
         Showing 1 of 1 message in the mailbox\
         """ = error.message
 
-        "assert_received({:status, ^status, ^other_status})" = Macro.to_string(error.expr)
+        "assert_received {:status, ^status, ^other_status}" = Macro.to_string(error.expr)
         "{:status, ^status, ^other_status}" = Macro.to_string(error.left)
     end
   end
@@ -441,7 +466,7 @@ defmodule ExUnit.AssertionsTest do
         "Assertion failed, no matching message after 0ms\nThe process mailbox is empty." =
           error.message
 
-        "assert_received(:hello)" = Macro.to_string(error.expr)
+        "assert_received :hello" = Macro.to_string(error.expr)
     end
   end
 
@@ -457,7 +482,7 @@ defmodule ExUnit.AssertionsTest do
         Showing 1 of 1 message in the mailbox\
         """ = error.message
 
-        "assert_received(:hello)" = Macro.to_string(error.expr)
+        "assert_received :hello" = Macro.to_string(error.expr)
         ":hello" = Macro.to_string(error.left)
     end
   end
@@ -474,7 +499,7 @@ defmodule ExUnit.AssertionsTest do
         Showing 10 of 11 messages in the mailbox\
         """ = error.message
 
-        "assert_received(x when x == :hello)" = Macro.to_string(error.expr)
+        "assert_received x when x == :hello" = Macro.to_string(error.expr)
         "x when x == :hello" = Macro.to_string(error.left)
     end
   end
@@ -522,7 +547,7 @@ defmodule ExUnit.AssertionsTest do
       error in [ExUnit.AssertionError] ->
         'foo' = error.left
         'bar' = error.right
-        "assert('foo' in 'bar')" = Macro.to_string(error.expr)
+        "assert 'foo' in 'bar'" = Macro.to_string(error.expr)
     end
   end
 
@@ -537,7 +562,7 @@ defmodule ExUnit.AssertionsTest do
       error in [ExUnit.AssertionError] ->
         'foo' = error.left
         ['foo', 'bar'] = error.right
-        "refute('foo' in ['foo', 'bar'])" = Macro.to_string(error.expr)
+        "refute 'foo' in ['foo', 'bar']" = Macro.to_string(error.expr)
     end
   end
 
@@ -556,7 +581,7 @@ defmodule ExUnit.AssertionsTest do
     rescue
       error in [ExUnit.AssertionError] ->
         "match (=) failed" = error.message
-        "assert({:ok, _} = error(true))" = Macro.to_string(error.expr)
+        "assert {:ok, _} = error(true)" = Macro.to_string(error.expr)
         "{:error, true}" = Macro.to_string(error.right)
     end
   end
@@ -567,7 +592,7 @@ defmodule ExUnit.AssertionsTest do
     rescue
       error in [ExUnit.AssertionError] ->
         "match (=) failed" = error.message
-        "assert({:ok, _x} = nil)" = Macro.to_string(error.expr)
+        "assert {:ok, _x} = nil" = Macro.to_string(error.expr)
         "nil" = Macro.to_string(error.right)
     end
   end
@@ -578,7 +603,7 @@ defmodule ExUnit.AssertionsTest do
     rescue
       error in [ExUnit.AssertionError] ->
         "Expected truthy, got nil" = error.message
-        "assert(_x = nil)" = Macro.to_string(error.expr)
+        "assert _x = nil" = Macro.to_string(error.expr)
     end
   end
 
@@ -587,18 +612,18 @@ defmodule ExUnit.AssertionsTest do
       "This should never be tested" = refute _ = ok(true)
     rescue
       error in [ExUnit.AssertionError] ->
-        "refute(_ = ok(true))" = Macro.to_string(error.expr)
+        "refute _ = ok(true)" = Macro.to_string(error.expr)
         "Expected false or nil, got {:ok, true}" = error.message
     end
   end
 
   test "assert regex match" do
-    true = assert "foo" =~ ~r(o)
+    true = assert "foo" =~ ~r/o/
   end
 
   test "assert regex match when no match" do
     try do
-      "This should never be tested" = assert "foo" =~ ~r(a)
+      "This should never be tested" = assert "foo" =~ ~r/a/
     rescue
       error in [ExUnit.AssertionError] ->
         "foo" = error.left
@@ -607,12 +632,12 @@ defmodule ExUnit.AssertionsTest do
   end
 
   test "refute regex match" do
-    false = refute "foo" =~ ~r(a)
+    false = refute "foo" =~ ~r/a/
   end
 
   test "refute regex match when match" do
     try do
-      "This should never be tested" = refute "foo" =~ ~r(o)
+      "This should never be tested" = refute "foo" =~ ~r/o/
     rescue
       error in [ExUnit.AssertionError] ->
         "foo" = error.left
@@ -715,7 +740,7 @@ defmodule ExUnit.AssertionsTest do
     error in [ExUnit.AssertionError] ->
       1 = error.left
       2 = error.right
-      "assert(1 > 2)" = Macro.to_string(error.expr)
+      "assert 1 > 2" = Macro.to_string(error.expr)
   end
 
   test "assert less or equal than operator" do
@@ -726,7 +751,7 @@ defmodule ExUnit.AssertionsTest do
     "This should never be tested" = assert 2 <= 1
   rescue
     error in [ExUnit.AssertionError] ->
-      "assert(2 <= 1)" = Macro.to_string(error.expr)
+      "assert 2 <= 1" = Macro.to_string(error.expr)
       2 = error.left
       1 = error.right
   end
