@@ -32,6 +32,52 @@ defmodule ExUnitTest do
            end) =~ "\n0 failures\n"
   end
 
+  test "supports rerunning given modules" do
+    defmodule SampleAsyncTest do
+      use ExUnit.Case, async: true
+
+      test "true" do
+        assert false
+      end
+    end
+
+    defmodule SampleSyncTest do
+      use ExUnit.Case
+
+      test "true" do
+        assert false
+      end
+    end
+
+    defmodule IgnoreTest do
+      use ExUnit.Case
+
+      test "true" do
+        assert false
+      end
+    end
+
+    configure_and_reload_on_exit([])
+
+    assert capture_io(fn ->
+             assert ExUnit.run() == %{
+                      failures: 3,
+                      skipped: 0,
+                      total: 3,
+                      excluded: 0
+                    }
+           end) =~ "\n3 tests, 3 failures\n"
+
+    assert capture_io(fn ->
+             assert ExUnit.run([SampleSyncTest, SampleAsyncTest]) == %{
+                      failures: 2,
+                      skipped: 0,
+                      total: 2,
+                      excluded: 0
+                    }
+           end) =~ "\n2 tests, 2 failures\n"
+  end
+
   test "prints aborted runs on sigquit", config do
     Process.register(self(), :aborted_on_sigquit)
     line = __ENV__.line + 5
@@ -467,6 +513,25 @@ defmodule ExUnitTest do
     assert output =~ "trying to set reserved field :file"
   end
 
+  test "removes new lines from multiline test name (with --trace option)" do
+    defmodule Multiline do
+      use ExUnit.Case
+
+      test """
+      - line 1
+      - line 2
+      """ do
+        :ok
+      end
+    end
+
+    configure_and_reload_on_exit(trace: true)
+
+    output = capture_io(fn -> ExUnit.run() end)
+
+    assert output =~ "test - line 1 - line 2"
+  end
+
   test "raises on reserved tag :async in setup" do
     defmodule ReservedSetupTagAsync do
       use ExUnit.Case
@@ -792,6 +857,44 @@ defmodule ExUnitTest do
     end
   end
 
+  describe ":exit_status" do
+    test "defaults value to 2" do
+      configure_and_reload_on_exit([])
+      ExUnit.start(autorun: false)
+      config = ExUnit.configuration()
+      assert config[:exit_status] == 2
+    end
+
+    test "sets value of :exit_status" do
+      configure_and_reload_on_exit([])
+      ExUnit.start(exit_status: 5, autorun: false)
+      config = ExUnit.configuration()
+      assert config[:exit_status] == 5
+    end
+  end
+
+  test "prints warning when all tests are excluded" do
+    defmodule OnlyExcludedTests do
+      use ExUnit.Case
+
+      @tag :exclude
+      test "excluded test", do: assert(false)
+
+      @tag :exclude
+      test "one more excluded test", do: assert(false)
+    end
+
+    configure_and_reload_on_exit([])
+
+    output =
+      capture_io(fn ->
+        assert ExUnit.run() == %{total: 2, failures: 0, excluded: 2, skipped: 0}
+      end)
+
+    assert output =~ "All tests have been excluded.\n"
+    assert output =~ "2 tests, 0 failures, 2 excluded\n"
+  end
+
   ##  Helpers
 
   defp run_with_filter(filters, cases) do
