@@ -382,6 +382,10 @@ defmodule IEx.HelpersTest do
       assert capture_io(fn -> h(def) end) =~ def_h
     end
 
+    test "prints sigil documentation" do
+      assert capture_io(fn -> h(~w//) end) =~ "Handles the sigil `~w` for list of words"
+    end
+
     test "prints __info__ documentation" do
       h_output_module = capture_io(fn -> h(Module.__info__()) end)
       assert capture_io(fn -> h(Module.UnlikelyTo.Exist.__info__()) end) == h_output_module
@@ -497,6 +501,7 @@ defmodule IEx.HelpersTest do
       impl = """
       defmodule Impl do
         @behaviour MyBehaviour
+        @doc delegate_to: {Foo, :bar, 3}
         def first(0), do: 0
         @doc "Docs for Impl.second/1"
         def second(0), do: 0
@@ -511,6 +516,14 @@ defmodule IEx.HelpersTest do
         assert c(files, ".") |> Enum.sort() == [Impl, MyBehaviour]
 
         assert capture_io(fn -> h(Impl.first() / 1) end) == """
+
+                                                def first(int)
+
+               delegate_to: Foo.bar/3
+
+               Impl.first/1 has no docs but is a callback for behaviour MyBehaviour. Showing
+               callback docs instead.
+
                @callback first(integer()) :: integer()
 
                Docs for MyBehaviour.first
@@ -533,12 +546,7 @@ defmodule IEx.HelpersTest do
 
                """
 
-        assert capture_io(fn -> h(Impl.first()) end) == """
-               @callback first(integer()) :: integer()
-
-               Docs for MyBehaviour.first
-
-               """
+        assert capture_io(fn -> h(Impl.first()) end) == capture_io(fn -> h(Impl.first() / 1) end)
 
         assert capture_io(fn -> h(Impl.second()) end) == """
 
@@ -696,9 +704,9 @@ defmodule IEx.HelpersTest do
     test "lists all callbacks for an Erlang module" do
       output = capture_io(fn -> b(:gen_server) end)
 
-      assert output =~ "@callback handle_cast(request :: term(), state :: term()) ::"
-      assert output =~ "@callback handle_info(info :: :timeout | term(), state :: term()) ::"
       assert output =~ "@callback init(args :: term()) ::"
+      assert output =~ "@callback handle_cast("
+      assert output =~ "@callback handle_info("
     end
 
     test "lists all macrocallbacks for a module" do
@@ -780,7 +788,7 @@ defmodule IEx.HelpersTest do
                "@callback message(t()) :: String.t()\n\n"
 
       assert capture_io(fn -> b(:gen_server.handle_cast() / 2) end) =~
-               "@callback handle_cast(request :: term(), state :: term()) ::"
+               "@callback handle_cast("
     end
 
     test "prints callback documentation metadata" do
@@ -831,7 +839,7 @@ defmodule IEx.HelpersTest do
 
                @macrocallback optional_macrocallback(:bar) :: atom()
 
-               @optional_callbacks [optional_callback: 1, optional_macrocallback: 1]
+               @optional_callbacks optional_callback: 1, optional_macrocallback: 1
 
                """
       end)
@@ -898,7 +906,7 @@ defmodule IEx.HelpersTest do
     end
 
     test "prints private types" do
-      assert capture_io(fn -> t(Date.Range) end) =~ "@typep iso_days"
+      assert capture_io(fn -> t(Date.Range) end) =~ "@typep days"
     end
 
     test "prints type information" do
@@ -1035,6 +1043,13 @@ defmodule IEx.HelpersTest do
       exports = capture_io(fn -> exports(IEx.Autocomplete) end)
       assert exports == "expand/1      expand/2      exports/1     remsh/1       \n"
     end
+
+    test "handles long function names" do
+      exports = capture_io(fn -> exports(Calendar.UTCOnlyTimeZoneDatabase) end)
+
+      assert exports ==
+               "time_zone_period_from_utc_iso_days/2 time_zone_periods_from_wall_datetime/2 \n"
+    end
   end
 
   describe "import_file" do
@@ -1106,14 +1121,12 @@ defmodule IEx.HelpersTest do
   describe "use_if_available" do
     test "uses a module only if available" do
       assert "nil" == capture_iex("use_if_available NoSuchModule")
-      assert "1" == capture_iex("use_if_available Bitwise; 1 &&& 1")
-      assert "1" == capture_iex("use_if_available Bitwise, only_operators: true; 1 &&& 1")
     end
   end
 
   describe "iex> |> (and other binary operators)" do
     test "passes previous result to the pipe" do
-      Enum.each([:~>>, :<<~, :~>, :<~, :<~>, :<|>], fn op ->
+      Enum.each([:~>>, :<<~, :~>, :<~, :<~>], fn op ->
         assert capture_iex("42\n  #{op} IO.puts()") =~ "undefined function #{op}/2"
       end)
 
@@ -1123,6 +1136,17 @@ defmodule IEx.HelpersTest do
       assert capture_iex("[42]\n++ [24]\n|> IO.inspect(label: \"foo\")") =~ "foo: [42, 24]"
       assert capture_iex("|> IO.puts()") =~ "(RuntimeError) v(-1) is out of bounds"
     end
+
+    test "raises if previous expression was a match" do
+      assert capture_iex("x = 42\n|> IO.puts()") =~
+               "surround the whole pipeline with parentheses '|>'"
+
+      assert capture_iex("%{x: x} = %{x: 42}\n|> IO.puts()") =~
+               "surround the whole pipeline with parentheses '|>'"
+
+      assert capture_iex("%{x: x} = map = %{x: 42}\n|> IO.puts()") =~
+               "surround the whole pipeline with parentheses '|>'"
+    end
   end
 
   describe "c" do
@@ -1293,7 +1317,8 @@ defmodule IEx.HelpersTest do
                  assert Sample.run() == :run
 
                  File.write!(filename, "defmodule Sample do end")
-                 assert {:reloaded, Sample, [Sample]} = r(Sample)
+                 assert {:reloaded, [Sample]} = r(Sample)
+                 assert {:reloaded, [Sample]} = r([Sample])
 
                  message = "function Sample.run/0 is undefined or private"
 
@@ -1320,7 +1345,8 @@ defmodule IEx.HelpersTest do
         assert :sample.hello() == :world
 
         File.write!(filename, other_erlang_module_code())
-        assert {:reloaded, :sample, [:sample]} = r(:sample)
+        assert {:reloaded, [:sample]} = r(:sample)
+        assert {:reloaded, [:sample]} = r([:sample])
         assert :sample.hello() == :bye
       end)
     after
@@ -1338,6 +1364,14 @@ defmodule IEx.HelpersTest do
       end
     end
 
+    test "builds a PID from atom" do
+      assert inspect(pid(:init)) == "#PID<0.0.0>"
+
+      assert_raise ArgumentError, fn ->
+        pid(:random_atom_ZM6pX6VwQx)
+      end
+    end
+
     test "builds a PID from integers" do
       assert inspect(pid(0, 32767, 3276)) == "#PID<0.32767.3276>"
       assert inspect(pid(0, 5, 6)) == "#PID<0.5.6>"
