@@ -19,6 +19,9 @@ defmodule Logger do
     * Integrates with Erlang's [`:logger`](`:logger`)
       to convert terms to Elixir syntax.
 
+    * Allows overriding the logging level for a specific module,
+      application or process.
+
   Logging is useful for tracking when an event of interest happens in your
   system. For example, it may be helpful to log whenever a user is deleted.
 
@@ -61,11 +64,37 @@ defmodule Logger do
   be passed to backends. If your log level is set to `:alert`, only
   `:alert` and `:emergency` will be printed.
 
+  ## Message
+
+  Logger can be used for logging both unstructured and structured data.
+
+  Unstructured data is a string or a list of strings:
+
+      Logger.info("hello world!")
+      Logger.info(["hello ", "world!"])
+
+  Structured data, also known as reports, are keyword lists and maps:
+
+      Logger.info([new_user: user.id, account_type: :admin])
+      Logger.info(%{new_user: user.id, account_type: :admin})
+
+  Log functions also accept a zero-arity anonymous function as a message:
+
+      Logger.info(fn -> "hello world!" end)
+
+  The anonymous function can return a message or a tuple containing
+  the message and additional metadata (to be described in the next
+  section).
+
+  In all cases, the arguments given to the `Logger` macros are only
+  evaluated if required by the current log level. The exception is
+  the `bare_log/3` function, which is the raw mechanism for logging.
+
   ## Metadata
 
   Whenever a message is logged, additional information can be given
   via metadata. Each log operation, such as `Logger.info/2`, allows
-  metadata to be given as argument.
+  metadata to be given as an argument.
 
   Furthermore, metadata can be set per process with `Logger.metadata/1`.
 
@@ -194,7 +223,7 @@ defmodule Logger do
       with severity less than the configured level will simply
       cause the message to be ignored. Keep in mind that each backend
       may have its specific level, too. In addition to levels mentioned
-      above it also support 2 "meta-levels":
+      above it also supports 2 "meta-levels":
 
         - `:all` - all messages will be logged, conceptually identical to
           `:debug`
@@ -282,15 +311,18 @@ defmodule Logger do
   The available backends by default are:
 
     * `:console` - logs messages to the console (enabled by default).
-      `:console` is simply a shortcut for `Logger.Backends.Console`.
+      `:console` is simply a shortcut for `Logger.Backends.Console`
+      (see its documentation for more information)
 
   Developers may also implement their own backends, an option that
   is explored in more detail below.
 
   The initial backends are loaded via the `:backends` configuration,
   which must be set before the `:logger` application is started.
-  However, it preferred to add and remove backends via `add_backend/2`
-  and `remove_backend/2` functions. This is often done in your
+  However, by the time the Logger application starts, the code for your
+  own and third-party backends may not yet be available. For this reason,
+  it is preferred to add and remove backends via `add_backend/2` and
+  `remove_backend/2` functions. This is often done in your
   `c:Application.start/2` callback:
 
       @impl true
@@ -437,11 +469,11 @@ defmodule Logger do
   @type metadata :: keyword()
   @levels [:emergency, :alert, :critical, :error, :warning, :notice, :info, :debug]
 
-  @metadata :logger_enabled
+  @metadata :logger_level
   @compile {:inline, enabled?: 1}
 
   @doc """
-  Alters the current process metadata according the given keyword list.
+  Alters the current process metadata according to the given keyword list.
 
   This function will merge the given keyword list into the existing metadata,
   with the exception of setting a key to `nil`, which will remove that key
@@ -491,33 +523,44 @@ defmodule Logger do
   Enables logging for the current process.
 
   Currently the only accepted PID is `self()`.
+
+  Equivalent of:
+
+      delete_process_level(pid)
   """
+  # TODO: Deprecate me on v1.18
+  @doc deprecated: "Use Logger.delete_process_level(pid) instead"
   @spec enable(pid) :: :ok
   def enable(pid) when pid == self() do
-    Process.delete(@metadata)
-    :ok
+    delete_process_level(pid)
   end
 
   @doc """
   Disables logging for the current process.
 
   Currently the only accepted PID is `self()`.
+
+  Equivalent of:
+
+      put_process_level(pid, :none)
   """
+  # TODO: Deprecate me on v1.18
+  @doc deprecated: "Use Logger.put_process_level(pid, :none) instead"
   @spec disable(pid) :: :ok
   def disable(pid) when pid == self() do
-    Process.put(@metadata, false)
-    :ok
+    put_process_level(pid, :none)
   end
 
   @doc """
-  Returns whether the logging is enabled for given process.
+  Returns whether the logging is enabled for a given process.
 
   Currently the only accepted PID is `self()`.
   """
-  @doc since: "1.10.0"
+  # TODO: Deprecate me on v1.18
+  @doc deprecated: "Use Logger.get_process_level(pid) instead"
   @spec enabled?(pid) :: boolean
   def enabled?(pid) when pid == self() do
-    Process.get(@metadata, true)
+    get_process_level(pid) != :none
   end
 
   @doc """
@@ -622,14 +665,14 @@ defmodule Logger do
 
   """
   @doc since: "1.11.0"
-  @spec put_module_level(module(), level() | :all | :none) :: :ok | {:error, term()}
+  @spec put_module_level(module() | [module()], level() | :all | :none) :: :ok | {:error, term()}
   defdelegate put_module_level(mod, level), to: :logger, as: :set_module_level
 
   @doc """
   Gets logging level for given module.
 
-  Returned value will be the effective value used. If no value
-  was set for given module, then it will not be present in
+  The returned value will be the effective value used. If no value
+  was set for a given module, then it will not be present in
   the returned list.
   """
   @doc since: "1.11.0"
@@ -637,19 +680,83 @@ defmodule Logger do
   defdelegate get_module_level(mod), to: :logger
 
   @doc """
-  Deletes logging level for given module to primary level.
+  Resets the logging level for a given module to the primary level.
   """
   @doc since: "1.11.0"
   @spec delete_module_level(module() | [module()]) :: :ok
   defdelegate delete_module_level(module), to: :logger, as: :unset_module_level
 
   @doc """
-  Deletes logging level for all modules to primary level
+  Resets the logging level for all modules to the primary level.
   """
   @doc since: "1.11.0"
   @spec delete_all_module_levels() :: :ok
   defdelegate delete_all_module_levels(), to: :logger, as: :unset_module_level
 
+  @doc """
+  Puts logging level for modules in a given application.
+
+  This will take priority over the primary level set, so it can be
+  used to increase or decrease verbosity of some parts of the project.
+
+  Equivalent of:
+
+      appname |> Application.spec(:modules) |> Logger.put_module_level(level)
+  """
+  @doc since: "1.13.0"
+  @spec put_application_level(atom(), level() | :all | :none) :: :ok | {:error, :not_loaded}
+  defdelegate put_application_level(appname, level), to: :logger, as: :set_application_level
+
+  @doc """
+  Resets logging level for all modules in the given application to the primary level.
+
+  Equivalent of:
+
+      appname |> Application.spec(:modules) |> Logger.delete_module_level()
+  """
+  @doc since: "1.13.0"
+  @spec delete_application_level(application) :: :ok | {:error, {:not_loaded, application}}
+        when application: atom()
+  defdelegate delete_application_level(appname), to: :logger, as: :unset_application_level
+
+  @doc """
+  Puts logging level for the current process.
+
+  Currently the only accepted PID is `self()`.
+
+  This will take priority over the primary level set, so it can be
+  used to increase or decrease verbosity of some parts of the running system.
+  """
+  @spec put_process_level(pid(), level() | :all | :none) :: :ok
+  def put_process_level(pid, level) when pid == self() do
+    Process.put(@metadata, Logger.Handler.elixir_level_to_erlang_level(level))
+    :ok
+  end
+
+  @doc """
+  Gets logging level for the current process.
+
+  Currently the only accepted PID is `self()`.
+
+  The returned value will be the effective value used. If no value
+  was set for a given process, then `nil` is returned.
+  """
+  @spec get_process_level(pid) :: level() | :all | :none | nil
+  def get_process_level(pid) when pid == self() do
+    Process.get(@metadata, nil)
+  end
+
+  @doc """
+  Resets logging level for the current process to the primary level.
+
+  Currently the only accepted PID is `self()`.
+  """
+  @spec delete_process_level(pid()) :: :ok
+  def delete_process_level(pid) when pid == self() do
+    Process.delete(@metadata)
+    :ok
+  end
+
   @doc """
   Adds a new backend.
 
@@ -764,7 +871,7 @@ defmodule Logger do
   def __should_log__(level, module) do
     level = Logger.Handler.elixir_level_to_erlang_level(level)
 
-    if enabled?(self()) and :logger.allow(level, module) do
+    if :logger.allow(level, module) do
       level
     end
   end
@@ -788,9 +895,9 @@ defmodule Logger do
     if is_msg(msg) do
       :logger.macro_log(location, level, msg, add_elixir_domain(metadata))
     else
-      # TODO: Remove this branch in Elixir v2.0
       IO.warn(
-        "passing #{inspect(msg)} to Logger is deprecated, expected a map, a keyword list, a binary, or an iolist"
+        "passing #{inspect(msg)} to Logger is deprecated, expected a map, a keyword list, " <>
+          "a string, a list of strings, or a zero-arity anonymous function"
       )
 
       :logger.macro_log(location, level, to_string(msg), add_elixir_domain(metadata))
@@ -949,7 +1056,15 @@ defmodule Logger do
   defp compile_time_purge_matching?(level, compile_metadata) do
     matching = Application.get_env(:logger, :compile_time_purge_matching, [])
 
+    if not is_list(matching) do
+      bad_compile_time_purge_matching!(matching)
+    end
+
     Enum.any?(matching, fn filter ->
+      if not is_list(filter) do
+        bad_compile_time_purge_matching!(matching)
+      end
+
       Enum.all?(filter, fn
         {:level_lower_than, min_level} ->
           compare_levels(level, min_level) == :lt
@@ -967,12 +1082,16 @@ defmodule Logger do
           Map.fetch(compile_metadata, k) == {:ok, v}
 
         _ ->
-          raise "expected :compile_time_purge_matching to be a list of keyword lists, " <>
-                  "got: #{inspect(matching)}"
+          bad_compile_time_purge_matching!(matching)
       end)
     end)
   end
 
+  defp bad_compile_time_purge_matching!(matching) do
+    raise "expected :compile_time_purge_matching to be a list of keyword lists, " <>
+            "got: #{inspect(matching)}"
+  end
+
   defp maybe_log(level, data, metadata, caller) do
     min_level =
       if env_level = Application.get_env(:logger, :compile_time_purge_level) do
