@@ -107,9 +107,8 @@ defmodule Logger.Translator do
     case message do
       # This is no longer emitted by Erlang/OTP but it may be
       # manually emitted by libraries like connection.
-      # TODO: Remove this translation on Elixir v1.14
       {'** Generic server ' ++ _, [name, last, state, reason | client]} ->
-        opts = Application.get_env(:logger, :translator_inspect_opts)
+        opts = Application.fetch_env!(:logger, :translator_inspect_opts)
         {formatted, reason} = format_reason(reason)
         metadata = [crash_reason: reason] ++ registered_name(name)
 
@@ -181,7 +180,7 @@ defmodule Logger.Translator do
   end
 
   defp report_gen_server_terminate(min_level, report) do
-    inspect_opts = Application.get_env(:logger, :translator_inspect_opts)
+    inspect_opts = Application.fetch_env!(:logger, :translator_inspect_opts)
 
     %{
       client_info: client,
@@ -207,7 +206,7 @@ defmodule Logger.Translator do
   end
 
   defp report_gen_event_terminate(min_level, report) do
-    inspect_opts = Application.get_env(:logger, :translator_inspect_opts)
+    inspect_opts = Application.fetch_env!(:logger, :translator_inspect_opts)
 
     %{
       handler: handler,
@@ -582,7 +581,7 @@ defmodule Logger.Translator do
   defp registered_name(_name), do: []
 
   defp format_mfa(mod, fun, :undefined),
-    do: [inspect(mod), ?., Code.Identifier.inspect_as_function(fun) | "/?"]
+    do: [inspect(mod), ?., Macro.inspect_atom(:remote_call, fun) | "/?"]
 
   defp format_mfa(mod, fun, args),
     do: Exception.format_mfa(mod, fun, args)
