@@ -14,7 +14,6 @@ defmodule Logger.MixProject do
       registered: [Logger, Logger.BackendSupervisor, Logger.Supervisor, Logger.Watcher],
       mod: {Logger.App, []},
       env: [
-        level: :debug,
         utc_log: false,
         truncate: 8096,
         backends: [:console],
@@ -25,11 +24,11 @@ defmodule Logger.MixProject do
         handle_sasl_reports: false,
         discard_threshold_periodic_check: 30_000,
         discard_threshold_for_error_logger: 500,
-        compile_time_purge_level: :debug,
         compile_time_purge_matching: [],
         compile_time_application: nil,
         translator_inspect_opts: [],
-        console: []
+        console: [],
+        start_options: []
       ]
     ]
   end
