@@ -37,7 +37,7 @@ defmodule Logger.HandlerTest do
 
     assert capture_log(fn ->
              :error_logger.info_msg('world: ~p', [:ok])
-           end) =~ "[info]  rewritten"
+           end) =~ "[notice] rewritten"
   after
     assert Logger.remove_translator({CustomTranslator, :t})
   end
@@ -51,7 +51,7 @@ defmodule Logger.HandlerTest do
 
     assert capture_log(fn ->
              :logger.info('world: ~p', [:ok])
-           end) =~ "[info]  rewritten"
+           end) =~ "[info] rewritten"
 
     assert capture_log(fn ->
              :logger.info(%{hello: :ok})
@@ -59,7 +59,7 @@ defmodule Logger.HandlerTest do
 
     assert capture_log(fn ->
              :logger.info(%{world: :ok})
-           end) =~ "[info]  rewritten"
+           end) =~ "[info] rewritten"
   after
     assert Logger.remove_translator({CustomTranslator, :t})
   end
@@ -68,7 +68,7 @@ defmodule Logger.HandlerTest do
     assert Logger.add_translator({CustomTranslator, :t})
 
     message = capture_log(fn -> :logger.info(%{error: "oops"}) end)
-    assert message =~ "[info]  Failure while translating Erlang's logger event\n"
+    assert message =~ "[info] Failure while translating Erlang's logger event\n"
     assert message =~ "** (RuntimeError) oops\n"
   after
     assert Logger.remove_translator({CustomTranslator, :t})
@@ -94,16 +94,17 @@ defmodule Logger.HandlerTest do
     assert capture_log(fn ->
              callback = fn %{hello: :world} -> {"~p~n", [:formatted]} end
              :logger.info(%{hello: :world}, %{report_cb: callback})
-           end) =~ "[info]  :formatted"
+           end) =~ "[info] :formatted"
   end
 
-  test "converts log levels" do
-    assert capture_log(fn -> :logger.emergency('ok') end) =~ "[error] ok"
-    assert capture_log(fn -> :logger.alert('ok') end) =~ "[error] ok"
-    assert capture_log(fn -> :logger.critical('ok') end) =~ "[error] ok"
+  test "uses Erlang log levels" do
+    assert capture_log(fn -> :logger.emergency('ok') end) =~ "[emergency] ok"
+    assert capture_log(fn -> :logger.alert('ok') end) =~ "[alert] ok"
+    assert capture_log(fn -> :logger.critical('ok') end) =~ "[critical] ok"
     assert capture_log(fn -> :logger.error('ok') end) =~ "[error] ok"
-    assert capture_log(fn -> :logger.warning('ok') end) =~ "[warn]  ok"
-    assert capture_log(fn -> :logger.info('ok') end) =~ "[info]  ok"
+    assert capture_log(fn -> :logger.warning('ok') end) =~ "[warning] ok"
+    assert capture_log(fn -> :logger.notice('ok') end) =~ "[notice] ok"
+    assert capture_log(fn -> :logger.info('ok') end) =~ "[info] ok"
     assert capture_log(fn -> :logger.debug('ok') end) =~ "[debug] ok"
   end
 
@@ -135,6 +136,15 @@ defmodule Logger.HandlerTest do
     assert_received {:format, ^report}
   end
 
+  test "respects translator_inspect_opts for reports" do
+    Application.put_env(:logger, :translator_inspect_opts, printable_limit: 1)
+
+    assert capture_log(fn -> :logger.error(%{foo: "bar"}) end) =~
+             ~S([error] [foo: "b" <> ...])
+  after
+    Application.put_env(:logger, :translator_inspect_opts, [])
+  end
+
   test "calls report_cb/2 when supplied" do
     report = %{foo: "bar"}
 
@@ -155,9 +165,11 @@ defmodule Logger.HandlerTest do
     assert %{chars_limit: 1000} = opts
     assert %{depth: 10} = opts
     assert %{single_line: false} = opts
+  after
+    Application.put_env(:logger, :translator_inspect_opts, [])
   end
 
-  test "calls report_cb when passed %{label: term(), report: term()}" do
+  test "calls report_cb/2 when passed %{label: term(), report: term()}" do
     report = %{label: :foo, report: %{bar: 1, baz: 2}}
 
     assert capture_log(fn -> :logger.error(report, %{report_cb: &format_report/1}) end)
