@@ -1,6 +1,13 @@
 Logger.configure_backend(:console, colors: [enabled: false])
 
-ExUnit.start()
+{line_exclude, line_include} =
+  if line = System.get_env("LINE"), do: {[:test], [line: line]}, else: {[], []}
+
+ExUnit.start(
+  trace: !!System.get_env("TRACE"),
+  include: line_include,
+  exclude: line_exclude
+)
 
 defmodule Logger.Case do
   use ExUnit.CaseTemplate
