@@ -1,7 +1,8 @@
 defmodule Mix.Compilers.Elixir do
   @moduledoc false
 
-  @manifest_vsn 8
+  @manifest_vsn 14
+  @checkpoint_vsn 2
 
   import Record
 
@@ -10,6 +11,7 @@ defmodule Mix.Compilers.Elixir do
   defrecord :source,
     source: nil,
     size: 0,
+    digest: nil,
     compile_references: [],
     export_references: [],
     runtime_references: [],
@@ -22,43 +24,115 @@ defmodule Mix.Compilers.Elixir do
   Compiles stale Elixir files.
 
   It expects a `manifest` file, the source directories, the destination
-  directory, an option to know if compilation is being forced or not, and a
-  list of any additional compiler options.
+  directory, the cache key based on compiler configuration, external
+  manifests, and external modules, followed by opts.
 
   The `manifest` is written down with information including dependencies
   between modules, which helps it recompile only the modules that
   have changed at runtime.
   """
-  def compile(manifest, srcs, dest, exts, force, opts) do
+  def compile(manifest, srcs, dest, new_cache_key, new_parent_manifests, new_parents, opts) do
+    modified = Mix.Utils.last_modified(manifest)
+    new_parents = :ordsets.from_list(new_parents)
+
     # We fetch the time from before we read files so any future
     # change to files are still picked up by the compiler. This
     # timestamp is used when writing BEAM files and the manifest.
     timestamp = System.os_time(:second)
-    all_paths = MapSet.new(Mix.Utils.extract_files(srcs, exts))
+    all_paths = Mix.Utils.extract_files(srcs, [:ex])
 
-    {all_modules, all_sources, all_local_exports} = parse_manifest(manifest, dest)
-    modified = Mix.Utils.last_modified(manifest)
+    {all_modules, all_sources, all_local_exports, old_parents, old_cache_key, old_deps_config} =
+      parse_manifest(manifest, dest)
 
-    {stale_local_deps, stale_local_mods, stale_local_exports, all_local_exports} =
-      stale_local_deps(manifest, modified, all_local_exports)
+    # If modules have been added or removed from the Erlang compiler,
+    # we need to recompile all references to old and new modules.
+    stale =
+      if old_parents != new_parents or
+           Mix.Utils.stale?(new_parent_manifests, [modified]) do
+        :ordsets.union(old_parents, new_parents)
+      else
+        []
+      end
 
-    prev_paths = for source(source: source) <- all_sources, into: MapSet.new(), do: source
-    removed = prev_paths |> MapSet.difference(all_paths) |> MapSet.to_list()
+    # If mix.exs has changed, recompile anything that calls Mix.Project.
+    stale =
+      if Mix.Utils.stale?([Mix.Project.project_file()], [modified]),
+        do: [Mix.Project | stale],
+        else: stale
+
+    # If the dependencies have changed, we need to traverse lock/config files.
+    deps_changed? = Mix.Utils.stale?([Mix.Project.config_mtime()], [modified])
+
+    # If a configuration is only accessed at compile-time, we don't need to
+    # track modules, only the compile env. So far this is only true for Elixir's
+    # dbg callback.
+    compile_env_apps = deps_config_compile_env_apps(old_deps_config)
+
+    # The app tracer will return information about apps before this compilation.
+    app_tracer = Mix.Compilers.ApplicationTracer.init()
+
+    {force?, stale, new_deps_config} =
+      cond do
+        !!opts[:force] or is_nil(old_deps_config) or old_cache_key != new_cache_key ->
+          {true, stale, deps_config()}
+
+        deps_changed? or compile_env_apps != [] ->
+          new_deps_config = deps_config()
+          config_apps = merge_appset(old_deps_config.config, new_deps_config.config, [])
+          apps = merge_appset(old_deps_config.lock, new_deps_config.lock, config_apps)
+
+          if Mix.Project.config()[:app] in apps do
+            {true, stale, new_deps_config}
+          else
+            apps_stale =
+              apps
+              |> deps_on()
+              |> Enum.flat_map(fn {app, _} ->
+                new_modules = Application.spec(app, :modules) || []
+
+                if old_modules = Mix.Compilers.ApplicationTracer.app_modules(app_tracer, app) do
+                  :ordsets.union(old_modules, :ordsets.from_list(new_modules))
+                else
+                  new_modules
+                end
+              end)
+
+            compile_env_apps = compile_env_apps ++ config_apps
+
+            compile_env_stale =
+              for source(compile_env: compile_env, modules: modules) <- all_sources,
+                  Enum.any?(compile_env_apps, &List.keymember?(compile_env, &1, 0)),
+                  module <- modules,
+                  do: module
+
+            stale = (stale ++ compile_env_stale) ++ apps_stale
+            {false, stale, new_deps_config}
+          end
+
+        true ->
+          {false, stale, old_deps_config}
+      end
+
+    {stale_modules, stale_exports, all_local_exports} =
+      stale_local_deps(manifest, stale, modified, all_local_exports)
+
+    prev_paths = for source(source: source) <- all_sources, do: source
+    removed = prev_paths -- all_paths
     {sources, removed_modules} = remove_removed_sources(all_sources, removed)
 
     {modules, exports, changed, sources_stats} =
-      if force do
+      if force? do
         compiler_info_from_force(manifest, all_paths, all_modules, dest)
       else
         compiler_info_from_updated(
+          manifest,
           modified,
-          all_paths,
+          all_paths -- prev_paths,
           all_modules,
           all_sources,
-          prev_paths,
           removed,
-          stale_local_mods,
-          Map.merge(stale_local_exports, removed_modules),
+          Map.merge(stale_modules, removed_modules),
+          Map.merge(stale_exports, removed_modules),
           dest
         )
       end
@@ -68,54 +142,110 @@ defmodule Mix.Compilers.Elixir do
     {sources, removed_modules} =
       update_stale_sources(sources, stale, removed_modules, sources_stats)
 
-    if opts[:all_warnings], do: show_warnings(sources)
+    if stale != [] do
+      Mix.Utils.compiling_n(length(stale), :ex)
+      Mix.Project.ensure_structure()
+      true = Code.prepend_path(dest)
 
-    cond do
-      stale != [] ->
-        Mix.Utils.compiling_n(length(stale), hd(exts))
-        Mix.Project.ensure_structure()
-        true = Code.prepend_path(dest)
+      {pending_tracer, tracer_opts} = Mix.Compilers.ApplicationTracer.prepare(app_tracer, opts)
+      previous_opts = set_compiler_opts(tracer_opts)
 
-        previous_opts =
-          {stale_local_deps, opts}
-          |> Mix.Compilers.ApplicationTracer.init()
-          |> set_compiler_opts()
+      # Stores state for keeping track which files were compiled
+      # and the dependencies between them.
+      put_compiler_info({[], exports, sources, modules, removed_modules})
 
-        # Stores state for keeping track which files were compiled
-        # and the dependencies between them.
-        put_compiler_info({modules, exports, sources, modules, removed_modules})
+      try do
+        compile_path(stale, dest, timestamp, opts)
+      else
+        {:ok, _, warnings} ->
+          {modules, _exports, sources, pending_modules, _pending_exports} = get_compiler_info()
 
-        try do
-          compile_path(stale, dest, timestamp, opts)
-        else
-          {:ok, _, warnings} ->
-            {modules, _exports, sources, _pending_modules, _pending_exports} = get_compiler_info()
-            sources = apply_warnings(sources, warnings)
-            write_manifest(manifest, modules, sources, all_local_exports, timestamp)
-            put_compile_env(sources)
-            {:ok, Enum.map(warnings, &diagnostic(&1, :warning))}
+          # We only collect the warnings if --all-warnings is given.
+          # In this case, we print them too. Then we apply the new warnings.
+          previous_warnings =
+            if opts[:all_warnings], do: previous_warnings(sources, true), else: []
 
-          {:error, errors, warnings} ->
-            # In case of errors, we show all previous warnings and all new ones
-            {_, _, sources, _, _} = get_compiler_info()
-            errors = Enum.map(errors, &diagnostic(&1, :error))
-            warnings = Enum.map(warnings, &diagnostic(&1, :warning))
-            {:error, warning_diagnostics(sources) ++ warnings ++ errors}
-        after
-          Code.compiler_options(previous_opts)
-          Mix.Compilers.ApplicationTracer.stop()
-          Code.purge_compiler_modules()
-          delete_compiler_info()
-        end
+          sources = apply_warnings(sources, warnings)
 
-      # We need to return ok if stale_local_mods changed
-      # because we want that to propagate to compile.protocols
-      removed != [] or stale_local_mods != %{} ->
-        write_manifest(manifest, modules, sources, all_local_exports, timestamp)
-        {:ok, warning_diagnostics(sources)}
+          write_manifest(
+            manifest,
+            modules ++ pending_modules,
+            sources,
+            all_local_exports,
+            new_parents,
+            new_cache_key,
+            new_deps_config,
+            timestamp
+          )
 
-      true ->
-        {:noop, warning_diagnostics(sources)}
+          put_compile_env(sources)
+          all_warnings = previous_warnings ++ Enum.map(warnings, &diagnostic(&1, :warning))
+          unless_previous_warnings_as_errors(previous_warnings, opts, {:ok, all_warnings})
+
+        {:error, errors, warnings} ->
+          # In case of errors, we show all previous warnings and all new ones.
+          # Print the new ones if --all-warnings was given.
+          {_, _, sources, _, _} = get_compiler_info()
+          errors = Enum.map(errors, &diagnostic(&1, :error))
+          warnings = Enum.map(warnings, &diagnostic(&1, :warning))
+          {:error, previous_warnings(sources, opts[:all_warnings]) ++ warnings ++ errors}
+      after
+        Code.compiler_options(previous_opts)
+        Mix.Compilers.ApplicationTracer.stop(pending_tracer)
+        Code.purge_compiler_modules()
+        delete_compiler_info()
+      end
+    else
+      # We need to return ok if deps_changed? or stale_modules changed,
+      # even if no code was compiled, because we need to propagate the changed
+      # status to compile.protocols. This will be the case whenever:
+      #
+      #   * the lock file or a config changes
+      #   * any module in a path dependency changes
+      #   * the mix.exs changes
+      #   * the Erlang manifest updates (Erlang files are compiled)
+      #
+      # In the first case, we will consolidate from scratch. In the remaining, we
+      # will only compute the diff with current protocols. In fact, there is no
+      # need to reconsolidate if an Erlang file changes and it doesn't trigger
+      # any other change, but the diff check should be reasonably fast anyway.
+      status = if removed != [] or deps_changed? or stale_modules != %{}, do: :ok, else: :noop
+
+      # If nothing changed but there is one more recent mtime, bump the manifest
+      if status != :noop or Enum.any?(Map.values(sources_stats), &(elem(&1, 0) > modified)) do
+        write_manifest(
+          manifest,
+          modules,
+          sources,
+          all_local_exports,
+          new_parents,
+          new_cache_key,
+          new_deps_config,
+          timestamp
+        )
+      end
+
+      previous_warnings = previous_warnings(sources, opts[:all_warnings])
+      unless_previous_warnings_as_errors(previous_warnings, opts, {status, previous_warnings})
+    end
+  after
+    Mix.Compilers.ApplicationTracer.stop()
+  end
+
+  defp deps_config do
+    # If you change this config, you need to bump @manifest_vsn
+    %{
+      lock: Enum.sort(Mix.Dep.Lock.read()),
+      config: Enum.sort(Mix.Tasks.Loadconfig.read_compile()),
+      dbg: Application.fetch_env!(:elixir, :dbg_callback)
+    }
+  end
+
+  defp deps_config_compile_env_apps(deps_config) do
+    if deps_config[:dbg] != Application.fetch_env!(:elixir, :dbg_callback) do
+      [:elixir]
+    else
+      []
     end
   end
 
@@ -150,15 +280,13 @@ defmodule Mix.Compilers.Elixir do
     rescue
       _ -> {[], []}
     else
-      {@manifest_vsn, modules, sources, _local_exports} -> {modules, sources}
+      {@manifest_vsn, modules, sources, _, _, _, _} -> {modules, sources}
       _ -> {[], []}
     end
   end
 
   defp compiler_info_from_force(manifest, all_paths, all_modules, dest) do
     # A config, path dependency or manifest has changed, let's just compile everything
-    all_paths = MapSet.to_list(all_paths)
-
     for module(module: module) <- all_modules,
         do: remove_and_purge(beam_path(dest, module), module)
 
@@ -175,43 +303,63 @@ defmodule Mix.Compilers.Elixir do
     {[], %{}, all_paths, sources_stats}
   end
 
+  # If any .beam file is missing, the first one will the first to miss,
+  # so we always check that. If there are no modules, then we can rely
+  # purely on digests.
+  defp missing_beam_file?(dest, [mod | _]), do: not File.exists?(beam_path(dest, mod))
+  defp missing_beam_file?(_dest, []), do: false
+
   defp compiler_info_from_updated(
+         manifest,
          modified,
-         all_paths,
+         new_paths,
          all_modules,
          all_sources,
-         prev_paths,
          removed,
-         stale_local_mods,
-         stale_local_exports,
+         stale_modules,
+         stale_exports,
          dest
        ) do
-    # Otherwise let's start with the new sources
-    new_paths =
-      all_paths
-      |> MapSet.difference(prev_paths)
-      |> MapSet.to_list()
+    modules_to_recompile =
+      for module(module: module, recompile?: recompile?) <- all_modules,
+          recompile_module?(module, recompile?) or Map.has_key?(stale_modules, module),
+          do: module
+
+    {checkpoint_stale_modules, checkpoint_stale_exports, checkpoint_modules} =
+      parse_checkpoint(manifest)
+
+    modules_to_recompile =
+      Map.merge(checkpoint_modules, Map.from_keys(modules_to_recompile, true))
+
+    stale_modules = Map.merge(checkpoint_stale_modules, stale_modules)
+    stale_exports = Map.merge(checkpoint_stale_exports, stale_exports)
+
+    if map_size(stale_modules) != map_size(checkpoint_stale_modules) or
+         map_size(stale_exports) != map_size(checkpoint_stale_exports) or
+         map_size(modules_to_recompile) != map_size(checkpoint_modules) do
+      write_checkpoint(manifest, stale_modules, stale_exports, modules_to_recompile)
+    end
 
     sources_stats =
       for path <- new_paths,
           into: mtimes_and_sizes(all_sources),
           do: {path, Mix.Utils.last_modified_and_size(path)}
 
-    modules_to_recompile =
-      for module(module: module, recompile?: true) <- all_modules,
-          recompile_module?(module),
-          into: %{},
-          do: {module, true}
-
     # Sources that have changed on disk or
     # any modules associated with them need to be recompiled
     changed =
-      for source(source: source, external: external, size: size, modules: modules) <-
+      for source(source: source, external: external, size: size, digest: digest, modules: modules) <-
             all_sources,
           {last_mtime, last_size} = Map.fetch!(sources_stats, source),
-          times = Enum.map(external, &(sources_stats |> Map.fetch!(&1) |> elem(0))),
-          size != last_size or Mix.Utils.stale?([last_mtime | times], [modified]) or
-            Enum.any?(modules, &Map.has_key?(modules_to_recompile, &1)),
+          # If the user does a change, compilation fails, and then they revert
+          # the change, the mtime will have changed but the .beam files will
+          # be missing and the digest is the same, so we need to check if .beam
+          # files are available.
+          size != last_size or
+            Enum.any?(modules, &Map.has_key?(modules_to_recompile, &1)) or
+            Enum.any?(external, &stale_external?(&1, modified, sources_stats)) or
+            (last_mtime > modified and
+               (missing_beam_file?(dest, modules) or digest != digest(source))),
           do: source
 
     changed = new_paths ++ changed
@@ -221,8 +369,8 @@ defmodule Mix.Compilers.Elixir do
         all_modules,
         all_sources,
         removed ++ changed,
-        stale_local_mods,
-        stale_local_exports,
+        stale_modules,
+        stale_exports,
         dest
       )
 
@@ -238,14 +386,29 @@ defmodule Mix.Compilers.Elixir do
     {modules, exports, changed, sources_stats}
   end
 
+  defp stale_external?({external, existed?}, modified, sources_stats) do
+    case sources_stats do
+      %{^external => {0, 0}} -> existed?
+      %{^external => {mtime, _}} -> mtime > modified
+    end
+  end
+
   defp mtimes_and_sizes(sources) do
     Enum.reduce(sources, %{}, fn source(source: source, external: external), map ->
-      Enum.reduce([source | external], map, fn file, map ->
+      map = Map.put_new_lazy(map, source, fn -> Mix.Utils.last_modified_and_size(source) end)
+
+      Enum.reduce(external, map, fn {file, _}, map ->
         Map.put_new_lazy(map, file, fn -> Mix.Utils.last_modified_and_size(file) end)
       end)
     end)
   end
 
+  defp digest(file) do
+    file
+    |> File.read!()
+    |> :erlang.md5()
+  end
+
   defp compile_path(stale, dest, timestamp, opts) do
     cwd = File.cwd!()
     long_compilation_threshold = opts[:long_compilation_threshold] || 10
@@ -297,14 +460,14 @@ defmodule Mix.Compilers.Elixir do
     end
 
     if changed == [] do
-      runtime_modules = dependent_runtime_modules(sources, modules, pending_modules)
+      modules = Enum.map(modules, &module(&1, :module))
       warnings = Mix.Compilers.ApplicationTracer.warnings(modules)
+
+      modules_set = Map.from_keys(modules, true)
+      {_, runtime_modules} = fixpoint_runtime_modules(sources, modules_set)
       {:runtime, runtime_modules, warnings}
     else
-      modules =
-        for module(sources: source_files) = module <- modules do
-          module(module, sources: source_files -- changed)
-        end
+      Mix.Utils.compiling_n(length(changed), :ex)
 
       # If we have a compile time dependency to a module, as soon as its file
       # change, we will detect the compile time dependency and recompile. However,
@@ -317,41 +480,6 @@ defmodule Mix.Compilers.Elixir do
     end
   end
 
-  defp dependent_runtime_modules(sources, all_modules, pending_modules) do
-    changed_modules =
-      for module(module: module) = entry <- all_modules,
-          entry not in pending_modules,
-          into: %{},
-          do: {module, true}
-
-    fixpoint_runtime_modules(sources, changed_modules, %{}, pending_modules)
-  end
-
-  defp fixpoint_runtime_modules(sources, changed, dependent, not_dependent) do
-    {new_dependent, not_dependent} =
-      Enum.reduce(not_dependent, {dependent, []}, fn module, {new_dependent, not_dependent} ->
-        depending? =
-          Enum.any?(module(module, :sources), fn file ->
-            source(runtime_references: runtime_refs) =
-              List.keyfind(sources, file, source(:source))
-
-            has_any_key?(changed, runtime_refs)
-          end)
-
-        if depending? do
-          {Map.put(new_dependent, module(module, :module), true), not_dependent}
-        else
-          {new_dependent, [module | not_dependent]}
-        end
-      end)
-
-    if map_size(dependent) != map_size(new_dependent) do
-      fixpoint_runtime_modules(sources, new_dependent, new_dependent, not_dependent)
-    else
-      Map.keys(new_dependent)
-    end
-  end
-
   defp each_module(file, module, _binary, cwd) do
     {modules, exports, sources, pending_modules, pending_exports} = get_compiler_info()
 
@@ -403,8 +531,8 @@ defmodule Mix.Compilers.Elixir do
     :ok
   end
 
-  defp recompile_module?(module) do
-    Code.ensure_loaded?(module) and
+  defp recompile_module?(module, recompile?) do
+    recompile? and Code.ensure_loaded?(module) and
       function_exported?(module, :__mix_recompile__?, 0) and
       module.__mix_recompile__?()
   end
@@ -433,7 +561,9 @@ defmodule Mix.Compilers.Elixir do
   end
 
   defp get_external_resources(module, cwd) do
-    for file <- Module.get_attribute(module, :external_resource), do: Path.relative_to(file, cwd)
+    for file <- Module.get_attribute(module, :external_resource) do
+      {Path.relative_to(file, cwd), File.exists?(file)}
+    end
   end
 
   defp each_file(file, lexical, cwd, verbose) do
@@ -460,6 +590,8 @@ defmodule Mix.Compilers.Elixir do
     source =
       source(
         source,
+        # We preserve the digest if the file is recompiled but not changed
+        digest: source(source, :digest) || digest(file),
         compile_references: compile_references,
         export_references: export_references,
         runtime_references: runtime_references,
@@ -509,11 +641,11 @@ defmodule Mix.Compilers.Elixir do
   # to be recompiled (but were not changed on disk)
   defp update_stale_sources(sources, changed) do
     Enum.reduce(changed, {sources, %{}}, fn file, {acc_sources, acc_modules} ->
-      {source(size: size, modules: modules), acc_sources} =
+      {source(size: size, digest: digest, modules: modules), acc_sources} =
         List.keytake(acc_sources, file, source(:source))
 
       acc_modules = Enum.reduce(modules, acc_modules, &Map.put(&2, &1, true))
-      {[source(source: file, size: size) | acc_sources], acc_modules}
+      {[source(source: file, size: size, digest: digest) | acc_sources], acc_modules}
     end)
   end
 
@@ -521,15 +653,15 @@ defmodule Mix.Compilers.Elixir do
   # files that have changed. Then it recursively figures out
   # all the files that changed (via the module dependencies) and
   # return the non-changed entries and the removed sources.
-  defp update_stale_entries(modules, _sources, [], stale_mods, stale_exports, _compile_path)
-       when stale_mods == %{} and stale_exports == %{} do
+  defp update_stale_entries(modules, _sources, [], stale_modules, stale_exports, _compile_path)
+       when stale_modules == %{} and stale_exports == %{} do
     {modules, %{}, []}
   end
 
-  defp update_stale_entries(modules, sources, changed, stale_mods, stale_exports, compile_path) do
-    changed = Enum.into(changed, %{}, &{&1, true})
+  defp update_stale_entries(modules, sources, changed, stale_modules, stale_exports, compile_path) do
+    changed = Map.from_keys(changed, true)
     reducer = &remove_stale_entry(&1, &2, sources, stale_exports, compile_path)
-    remove_stale_entries(modules, %{}, changed, stale_mods, reducer)
+    remove_stale_entries(modules, %{}, changed, stale_modules, reducer)
   end
 
   defp remove_stale_entries(modules, exports, old_changed, old_stale, reducer) do
@@ -583,23 +715,38 @@ defmodule Mix.Compilers.Elixir do
     Enum.any?(enumerable, &Map.has_key?(map, &1))
   end
 
-  defp stale_local_deps(manifest, modified, old_exports) do
+  defp stale_local_deps(manifest, stale_modules, modified, old_exports) do
     base = Path.basename(manifest)
 
+    # The stale modules so far will become both stale_modules and stale_exports,
+    # as any export from a dependency needs to be recompiled.
+    stale_modules = Map.from_keys(stale_modules, true)
+
     for %{scm: scm, opts: opts} = dep <- Mix.Dep.cached(),
         not scm.fetchable?,
-        Mix.Utils.last_modified(Path.join([opts[:build], ".mix", base])) > modified,
-        reduce: {%{}, %{}, %{}, old_exports} do
-      {deps, modules, exports, new_exports} ->
-        {modules, exports, new_exports} =
+        manifest = Path.join([opts[:build], ".mix", base]),
+        Mix.Utils.last_modified(manifest) > modified,
+        reduce: {stale_modules, stale_modules, old_exports} do
+      {modules, exports, new_exports} ->
+        {_manifest_modules, dep_sources} = read_manifest(manifest)
+
+        dep_modules =
           for path <- Mix.Dep.load_paths(dep),
               beam <- Path.wildcard(Path.join(path, "*.beam")),
               Mix.Utils.last_modified(beam) > modified,
-              reduce: {modules, exports, new_exports} do
-            {modules, exports, new_exports} ->
-              module = beam |> Path.basename() |> Path.rootname() |> String.to_atom()
+              do: beam |> Path.basename() |> Path.rootname() |> String.to_atom()
+
+        # If any module has a compile time dependency on a changed module
+        # within the dependnecy, they will be recompiled. However, export
+        # and runtime dependencies won't have recompiled so we need to
+        # propagate them to the parent app.
+        {dep_modules, _} = fixpoint_runtime_modules(dep_sources, Map.from_keys(dep_modules, true))
+
+        # Update exports
+        {exports, new_exports} =
+          for {module, _} <- dep_modules, reduce: {exports, new_exports} do
+            {exports, new_exports} ->
               export = exports_md5(module, false)
-              modules = Map.put(modules, module, true)
 
               # If the exports are the same, then the API did not change,
               # so we do not mark the export as stale. Note this has to
@@ -618,13 +765,42 @@ defmodule Mix.Compilers.Elixir do
                   do: Map.put(new_exports, module, export),
                   else: Map.delete(new_exports, module)
 
-              {modules, exports, new_exports}
+              {exports, new_exports}
           end
 
-        {Map.put(deps, dep.app, true), modules, exports, new_exports}
+        {Map.merge(modules, dep_modules), exports, new_exports}
     end
   end
 
+  defp fixpoint_runtime_modules(sources, modules) when modules != %{} do
+    fixpoint_runtime_modules(sources, modules, false, [], [])
+  end
+
+  defp fixpoint_runtime_modules(_sources, modules) do
+    {modules, []}
+  end
+
+  defp fixpoint_runtime_modules([source | sources], modules, new?, acc_modules, acc_sources) do
+    source(export_references: export_refs, runtime_references: runtime_refs) = source
+
+    if has_any_key?(modules, export_refs) or has_any_key?(modules, runtime_refs) do
+      new_modules = Enum.reject(source(source, :modules), &Map.has_key?(modules, &1))
+      modules = Enum.reduce(new_modules, modules, &Map.put(&2, &1, true))
+      new? = new? or new_modules != []
+      acc_modules = new_modules ++ acc_modules
+      fixpoint_runtime_modules(sources, modules, new?, acc_modules, acc_sources)
+    else
+      fixpoint_runtime_modules(sources, modules, new?, acc_modules, [source | acc_sources])
+    end
+  end
+
+  defp fixpoint_runtime_modules([], modules, new?, acc_modules, acc_sources)
+       when new? == false or acc_sources == [],
+       do: {modules, acc_modules}
+
+  defp fixpoint_runtime_modules([], modules, true, acc_modules, acc_sources),
+    do: fixpoint_runtime_modules(acc_sources, modules, false, acc_modules, [])
+
   defp exports_md5(module, use_attributes?) do
     cond do
       function_exported?(module, :__info__, 1) ->
@@ -653,13 +829,13 @@ defmodule Mix.Compilers.Elixir do
     _ = :code.delete(module)
   end
 
-  defp show_warnings(sources) do
-    for source(source: source, warnings: warnings) <- sources do
-      file = Path.absname(source)
-
-      for {line, message} <- warnings do
-        :elixir_errors.erl_warn(line, file, message)
-      end
+  defp previous_warnings(sources, print?) do
+    for source(source: source, warnings: warnings) <- sources,
+        file = Path.absname(source),
+        {location, message} <- warnings do
+      warning = {file, location, message}
+      print? && Kernel.ParallelCompiler.print_warning(warning)
+      diagnostic(warning, :warning)
     end
   end
 
@@ -671,45 +847,92 @@ defmodule Mix.Compilers.Elixir do
     end
   end
 
-  defp warning_diagnostics(sources) do
-    for source(source: source, warnings: warnings) <- sources,
-        {line, message} <- warnings,
-        do: diagnostic({Path.absname(source), line, message}, :warning)
-  end
-
-  defp diagnostic({file, line, message}, severity) do
+  defp diagnostic({file, location, message}, severity) do
     %Mix.Task.Compiler.Diagnostic{
       file: file,
-      position: line,
+      position: location,
       message: message,
       severity: severity,
       compiler_name: "Elixir"
     }
   end
 
+  ## Merging of lock and config files
+
+  # Value for app didn't change
+  defp merge_appset([{app, value} | old_set], [{app, value} | new_set], apps),
+    do: merge_appset(old_set, new_set, apps)
+
+  # Value for app changed
+  defp merge_appset([{app, _} | old_set], [{app, _} | new_set], apps),
+    do: merge_appset(old_set, new_set, [app | apps])
+
+  # Added value for app
+  defp merge_appset([{app1, _} | _] = old_set, [{app2, _} | new_set], apps)
+       when app1 > app2,
+       do: merge_appset(old_set, new_set, [app2 | apps])
+
+  # Removed value for app
+  defp merge_appset([{app1, _} | old_set], [{app2, _} | _] = new_set, apps)
+       when app1 < app2,
+       do: merge_appset(old_set, new_set, [app1 | apps])
+
+  # One of them is done, add the others
+  defp merge_appset(old_set, new_set, apps) do
+    apps = Enum.reduce(old_set, apps, fn {app, _}, apps -> [app | apps] end)
+    Enum.reduce(new_set, apps, fn {app, _}, apps -> [app | apps] end)
+  end
+
+  defp deps_on(apps) do
+    apps = Map.from_keys(apps, true)
+    deps_on(Mix.Dep.cached(), apps, [], false)
+  end
+
+  defp deps_on([%{app: app, deps: deps} = dep | cached_deps], apps, acc, stored?) do
+    cond do
+      # We have already seen this dep
+      Map.has_key?(apps, app) ->
+        deps_on(cached_deps, apps, acc, stored?)
+
+      # It depends on one of the apps, store it
+      Enum.any?(deps, &Map.has_key?(apps, &1.app)) ->
+        deps_on(cached_deps, Map.put(apps, app, true), acc, true)
+
+      # Otherwise we will check it later
+      true ->
+        deps_on(cached_deps, apps, [dep | acc], stored?)
+    end
+  end
+
+  defp deps_on([], apps, cached_deps, true), do: deps_on(cached_deps, apps, [], false)
+  defp deps_on([], apps, _cached_deps, false), do: apps
+
   ## Manifest handling
 
+  @default_manifest {[], [], %{}, [], nil, nil}
+
   # Similar to read_manifest, but for internal consumption and with data migration support.
   defp parse_manifest(manifest, compile_path) do
     try do
       manifest |> File.read!() |> :erlang.binary_to_term()
     rescue
       _ ->
-        {[], [], %{}}
+        @default_manifest
     else
-      {@manifest_vsn, modules, sources, local_exports} ->
-        {modules, sources, local_exports}
+      {@manifest_vsn, modules, sources, local_exports, parent, cache_key, deps_config} ->
+        {modules, sources, local_exports, parent, cache_key, deps_config}
 
-      # From v5 and later
-      {vsn, modules, _sources} when is_integer(vsn) ->
-        purge_old_manifest(compile_path, modules)
+      # {vsn, modules, sources} v5-v7 (v1.10)
+      # {vsn, modules, sources, local_exports} v8-v10 (v1.11)
+      manifest when is_tuple(manifest) and is_integer(elem(manifest, 0)) ->
+        purge_old_manifest(compile_path, elem(manifest, 1))
 
-      # From v4 and before
+      # v1-v4
       [vsn | data] when is_integer(vsn) ->
         purge_old_manifest(compile_path, data)
 
       _ ->
-        {[], [], %{}}
+        @default_manifest
     end
   end
 
@@ -728,21 +951,31 @@ defmodule Mix.Compilers.Elixir do
         )
     end
 
-    {[], [], %{}}
+    @default_manifest
   end
 
-  defp write_manifest(manifest, [], [], _exports, _timestamp) do
+  defp write_manifest(manifest, [], [], _exports, _parents, _cache_key, _deps_config, _timestamp) do
     File.rm(manifest)
     :ok
   end
 
-  defp write_manifest(manifest, modules, sources, exports, timestamp) do
+  defp write_manifest(
+         manifest,
+         modules,
+         sources,
+         exports,
+         parents,
+         cache_key,
+         deps_config,
+         timestamp
+       ) do
     File.mkdir_p!(Path.dirname(manifest))
 
-    term = {@manifest_vsn, modules, sources, exports}
+    term = {@manifest_vsn, modules, sources, exports, parents, cache_key, deps_config}
     manifest_data = :erlang.term_to_binary(term, [:compressed])
     File.write!(manifest, manifest_data)
     File.touch!(manifest, timestamp)
+    delete_checkpoint(manifest)
 
     # Since Elixir is a dependency itself, we need to touch the lock
     # so the current Elixir version, used to compile the files above,
@@ -753,4 +986,51 @@ defmodule Mix.Compilers.Elixir do
   defp beam_path(compile_path, module) do
     Path.join(compile_path, Atom.to_string(module) <> ".beam")
   end
+
+  # Once we added semantic recompilation, the following can happen:
+  #
+  # 1. The user changes config/mix.exs/__mix_recompile__?
+  # 2. We detect the change, remove .beam files and start recompilation
+  # 3. Recompilation fails
+  # 4. The user reverts the change
+  # 5. The compiler no longer recompiles and the .beam files are missing
+  #
+  # Therefore, it is important for us to checkpoint any state that may
+  # have lead to a compilation and which can now be reverted.
+
+  defp parse_checkpoint(manifest) do
+    try do
+      (manifest <> ".checkpoint") |> File.read!() |> :erlang.binary_to_term()
+    rescue
+      _ ->
+        {%{}, %{}, %{}}
+    else
+      {@checkpoint_vsn, stale_modules, stale_exports, recompile_modules} ->
+        {stale_modules, stale_exports, recompile_modules}
+
+      _ ->
+        {%{}, %{}, %{}}
+    end
+  end
+
+  defp write_checkpoint(manifest, stale_modules, stale_exports, recompile_modules) do
+    File.mkdir_p!(Path.dirname(manifest))
+    term = {@checkpoint_vsn, stale_modules, stale_exports, recompile_modules}
+    checkpoint_data = :erlang.term_to_binary(term, [:compressed])
+    File.write!(manifest <> ".checkpoint", checkpoint_data)
+  end
+
+  defp delete_checkpoint(manifest) do
+    File.rm(manifest <> ".checkpoint")
+  end
+
+  defp unless_previous_warnings_as_errors(previous_warnings, opts, {status, all_warnings}) do
+    if previous_warnings != [] and opts[:warnings_as_errors] do
+      message = "Compilation failed due to warnings while using the --warnings-as-errors option"
+      IO.puts(:stderr, message)
+      {:error, all_warnings}
+    else
+      {status, all_warnings}
+    end
+  end
 end
