@@ -11,7 +11,8 @@ defmodule Mix.Compilers.Test do
     runtime_references: [],
     external: []
 
-  @compile {:no_warn_undefined, ExUnit.Server}
+  # Necessary to avoid warnings during bootstrap
+  @compile {:no_warn_undefined, ExUnit}
   @stale_manifest "compile.test_stale"
   @manifest_vsn 1
 
@@ -21,7 +22,7 @@ defmodule Mix.Compilers.Test do
   It expects all of the test patterns, the test files that were matched for the
   test patterns, the test paths, and the opts from the test task.
   """
-  def require_and_run(matched_test_files, test_paths, opts) do
+  def require_and_run(matched_test_files, test_paths, elixirc_opts, opts) do
     stale = opts[:stale]
 
     {test_files, stale_manifest_pid, parallel_require_callbacks} =
@@ -31,34 +32,70 @@ defmodule Mix.Compilers.Test do
         {matched_test_files, nil, []}
       end
 
-    if test_files == [] do
-      :noop
-    else
-      task = Task.async(ExUnit, :run, [])
+    cond do
+      test_files == [] ->
+        # Make sure we run the after_suite callbacks but with no feedback
+        formatters = Application.fetch_env!(:ex_unit, :formatters)
 
-      try do
-        case Kernel.ParallelCompiler.require(test_files, parallel_require_callbacks) do
-          {:ok, _, _} -> :ok
-          {:error, _, _} -> exit({:shutdown, 1})
+        try do
+          Application.put_env(:ex_unit, :formatters, [])
+          _ = ExUnit.run()
+          :noop
+        after
+          Application.put_env(:ex_unit, :formatters, formatters)
         end
 
-        ExUnit.Server.modules_loaded()
-        %{failures: failures} = results = Task.await(task, :infinity)
+      Keyword.get(opts, :profile_require) == "time" ->
+        require(test_files, [profile: :time], elixirc_opts)
+        :noop
 
-        if failures == 0 do
-          agent_write_manifest(stale_manifest_pid)
+      true ->
+        task = ExUnit.async_run()
+        warnings_as_errors? = Keyword.get(opts, :warnings_as_errors, false)
+
+        try do
+          failed? =
+            case require(test_files, parallel_require_callbacks, elixirc_opts) do
+              {:ok, _, [_ | _]} when warnings_as_errors? -> true
+              {:ok, _, _} -> false
+              {:error, _, _} -> exit({:shutdown, 1})
+            end
+
+          %{failures: failures} = results = ExUnit.await_run(task)
+
+          if failures == 0 do
+            if failed? do
+              message =
+                "\nERROR! Test suite aborted after successful execution due to warnings while using the --warnings-as-errors option"
+
+              IO.puts(:stderr, IO.ANSI.format([:red, message]))
+              exit({:shutdown, 1})
+            end
+
+            agent_write_manifest(stale_manifest_pid)
+          end
+
+          {:ok, results}
+        catch
+          kind, reason ->
+            # In case there is an error, shut down the runner task
+            # before the error propagates up and trigger links.
+            Task.shutdown(task)
+            :erlang.raise(kind, reason, __STACKTRACE__)
+        after
+          agent_stop(stale_manifest_pid)
         end
+    end
+  end
 
-        {:ok, results}
-      catch
-        kind, reason ->
-          # In case there is an error, shut down the runner task
-          # before the error propagates up and trigger links.
-          Task.shutdown(task)
-          :erlang.raise(kind, reason, __STACKTRACE__)
-      after
-        agent_stop(stale_manifest_pid)
-      end
+  defp require(files, require_opts, elixirc_opts) do
+    elixirc_opts = Keyword.merge([docs: false, debug_info: false], elixirc_opts)
+    previous_opts = Code.compiler_options(elixirc_opts)
+
+    try do
+      Kernel.ParallelCompiler.require(files, require_opts)
+    after
+      Code.compiler_options(previous_opts)
     end
   end
 
@@ -70,9 +107,9 @@ defmodule Mix.Compilers.Test do
     removed =
       for source(source: source) <- all_sources, source not in matched_test_files, do: source
 
-    config_mtime = Mix.Project.config_mtime()
     test_helpers = Enum.map(test_paths, &Path.join(&1, "test_helper.exs"))
-    force = opts[:force] || Mix.Utils.stale?([config_mtime | test_helpers], [modified])
+    sources = [Mix.Project.config_mtime(), Mix.Project.project_file() | test_helpers]
+    force = opts[:force] || Mix.Utils.stale?(sources, [modified])
 
     changed =
       if force do
@@ -230,9 +267,9 @@ defmodule Mix.Compilers.Test do
           source: source,
           runtime_references: r,
           compile_references: c,
-          struct_references: s
+          export_references: e
         ) <- sources,
-        module in r or module in c or module in s,
+        module in r or module in c or module in e,
         CE.module(sources: sources, module: dependent_module) <- modules,
         source in sources,
         do: dependent_module
@@ -259,13 +296,13 @@ defmodule Mix.Compilers.Test do
       file = Path.relative_to(file, cwd)
       {source, sources} = List.keytake(sources, file, source(:source))
 
-      {compile_references, struct_references, runtime_references, _compile_env} =
+      {compile_references, export_references, runtime_references, _compile_env} =
         Kernel.LexicalTracker.references(lexical)
 
       source =
         source(
           source,
-          compile_references: compile_references ++ struct_references,
+          compile_references: compile_references ++ export_references,
           runtime_references: runtime_references
         )
 
