@@ -23,7 +23,7 @@ defmodule Mix.Dep do
     * `top_level` - true if dependency was defined in the top-level project
 
     * `manager` - the project management, possible values:
-      `:rebar` | `:rebar3` | `:mix` | `:make` | `nil`
+      `:rebar3` | `:mix` | `:make` | `nil`
 
     * `from` - path to the file where the dependency was defined
 
@@ -69,7 +69,7 @@ defmodule Mix.Dep do
           status: {:ok, String.t() | nil} | atom | tuple,
           opts: keyword,
           top_level: boolean,
-          manager: :rebar | :rebar3 | :mix | :make | nil,
+          manager: :rebar3 | :mix | :make | nil,
           from: String.t(),
           extra: term,
           system_env: keyword
@@ -259,8 +259,10 @@ defmodule Mix.Dep do
       |> Keyword.put(:app_path, opts[:build])
       |> Keyword.put(:build_scm, scm)
 
-    config = Keyword.take(opts, [:inherit_parent_config_files]) ++ config
-
+    # If the dependency is not fetchable, then it is never compiled
+    # from scratch and therefore it needs the parent configuration
+    # files to know when to recompile.
+    config = [inherit_parent_config_files: not scm.fetchable?] ++ config
     env = opts[:env] || :prod
     old_env = Mix.env()
 
@@ -508,36 +510,9 @@ defmodule Mix.Dep do
 
   Automatically derived from source paths.
   """
-  def load_paths(%Mix.Dep{opts: opts} = dep) do
+  def load_paths(%Mix.Dep{app: app, opts: opts}) do
     build_path = Path.dirname(opts[:build])
-
-    Enum.map(source_paths(dep), fn {_, base} ->
-      Path.join([build_path, base, "ebin"])
-    end)
-  end
-
-  @doc """
-  Returns all source paths.
-
-  Source paths are the directories that contain ebin files for a given
-  dependency. All managers, except `:rebar`, have only one source path.
-  """
-  def source_paths(%Mix.Dep{manager: :rebar, app: app, opts: opts, extra: extra}) do
-    sub_dirs = extra[:sub_dirs] || []
-    dest = opts[:dest]
-
-    # Add root dir and all sub dirs with ebin/ directory
-    in_sub_dirs =
-      for sub_dir <- sub_dirs,
-          path <- Path.wildcard(Path.join(dest, sub_dir)),
-          File.dir?(Path.join(path, "ebin")),
-          do: {path, Path.basename(path)}
-
-    [{opts[:dest], Atom.to_string(app)}] ++ in_sub_dirs
-  end
-
-  def source_paths(%Mix.Dep{app: app, opts: opts}) do
-    [{opts[:dest], Atom.to_string(app)}]
+    [Path.join([build_path, Atom.to_string(app), "ebin"])]
   end
 
   @doc """
@@ -551,7 +526,7 @@ defmodule Mix.Dep do
   Returns `true` if dependency is a Rebar project.
   """
   def rebar?(%Mix.Dep{manager: manager}) do
-    manager in [:rebar, :rebar3]
+    manager == :rebar3
   end
 
   @doc """
