@@ -91,6 +91,33 @@ defmodule Mix.Project do
   to be part of the configuration returned by `project/0`. This setting also
   makes sure Elixir is not added as a dependency to the generated `.app` file or
   to the escript generated with `mix escript.build`, and so on.
+
+  ## Invoking this module
+
+  This module contains many functions that return project information and
+  metadata. However, since Mix is not included nor configured during releases,
+  we recommend using the functions in this module only inside Mix tasks.
+  If you need to configure your own app, consider using the application
+  environment instead. For example, don't do this:
+
+      def some_config do
+        Mix.Project.config()[:some_config]
+      end
+
+  Nor this:
+
+      @some_config Mix.Project.config()[:some_config]
+
+  Instead, do this:
+
+      def some_config do
+        Application.get_env(:my_app, :some_config)
+      end
+
+  Or this:
+
+      @some_config Application.compile_env(:my_app, :some_config)
+
   """
 
   @doc false
@@ -109,17 +136,17 @@ defmodule Mix.Project do
   # Push a project onto the project stack.
   # Only the top of the stack can be accessed.
   @doc false
-  def push(atom, file \\ nil, app \\ nil) when is_atom(atom) do
-    file = file || (atom && List.to_string(atom.__info__(:compile)[:source]))
-    config = Keyword.merge([app: app] ++ default_config(), get_project_config(atom))
+  def push(module, file \\ nil, app \\ nil) when is_atom(module) do
+    file = file || (module && List.to_string(module.__info__(:compile)[:source]))
+    config = Keyword.merge([app: app] ++ default_config(), get_project_config(module))
 
-    case Mix.ProjectStack.push(atom, config, file) do
+    case Mix.ProjectStack.push(module, config, file) do
       :ok ->
         :ok
 
       {:error, other} when is_binary(other) ->
         Mix.raise(
-          "Trying to load #{inspect(atom)} from #{inspect(file)}" <>
+          "Trying to load #{inspect(module)} from #{inspect(file)}" <>
             " but another project with the same name was already defined at #{inspect(other)}"
         )
     end
@@ -140,7 +167,8 @@ defmodule Mix.Project do
       consolidate_protocols: false,
       consolidation_path: consolidation_path(config),
       deps_path: deps_path(config),
-      env_path: build_path(config)
+      env_path: build_path(config),
+      lockfile: Path.expand(config[:lockfile])
     ]
   end
 
@@ -177,6 +205,16 @@ defmodule Mix.Project do
     get() || raise Mix.NoProjectError, []
   end
 
+  @doc """
+  Returns the path to the file that defines the current project.
+
+  The majority of the time, it will point to a `mix.exs` file.
+  Returns `nil` if not inside a project.
+  """
+  @doc since: "1.13.0"
+  @spec project_file() :: binary | nil
+  defdelegate project_file(), to: Mix.ProjectStack
+
   @doc """
   Returns the project configuration.
 
@@ -206,13 +244,17 @@ defmodule Mix.Project do
   This function is usually used in compilation tasks to trigger
   a full recompilation whenever such configuration files change.
 
-  It returns the `mix.exs` file, the lock manifest, and all config
-  files in the `config` directory that do not start with a leading
-  period (for example, `.my_config.exs`).
+  It returns the lock manifest, and all config files in the `config`
+  directory that do not start with a leading period (for example,
+  `.my_config.exs`).
+
+  Note: before Elixir v1.13.0, the `mix.exs` file was also included
+  as a config file, but since then it has been moved to its own
+  function called `project_file/0`.
   """
   @spec config_files() :: [Path.t()]
   def config_files do
-    [Mix.Tasks.WillRecompile.manifest() | Mix.ProjectStack.config_files()]
+    Mix.ProjectStack.config_files()
   end
 
   @doc """
@@ -221,13 +263,15 @@ defmodule Mix.Project do
   This function is usually used in compilation tasks to trigger
   a full recompilation whenever such configuration files change.
   For this reason, the mtime is cached to avoid file system lookups.
+
+  Note: before Elixir v1.13.0, the `mix.exs` file was also included
+  in the mtimes, but not anymore. You can compute its modification
+  date by calling `project_file/0`.
   """
   @doc since: "1.7.0"
   @spec config_mtime() :: posix_mtime when posix_mtime: integer()
   def config_mtime do
-    Mix.Tasks.WillRecompile.manifest()
-    |> Mix.Utils.last_modified()
-    |> max(Mix.ProjectStack.config_mtime())
+    Mix.ProjectStack.config_mtime()
   end
 
   @doc """
@@ -402,7 +446,7 @@ defmodule Mix.Project do
   ## Options
 
     * `:depth` - only returns dependencies to the depth level,
-      a depth of 1 will only return top-level dependencies
+      a depth of `1` will only return top-level dependencies
     * `:parents` - starts the dependency traversal from the
       given parents instead of the application root
 
@@ -414,7 +458,7 @@ defmodule Mix.Project do
   """
   @doc since: "1.10.0"
   @spec deps_scms(keyword) :: %{optional(atom) => Mix.SCM.t()}
-  def deps_scms(opts \\ []) do
+  def deps_scms(opts \\ []) when is_list(opts) do
     traverse_deps(opts, fn %{scm: scm} -> scm end)
   end
 
@@ -424,7 +468,7 @@ defmodule Mix.Project do
   ## Options
 
     * `:depth` - only returns dependencies to the depth level,
-      a depth of 1 will only return top-level dependencies
+      a depth of `1` will only return top-level dependencies
     * `:parents` - starts the dependency traversal from the
       given parents instead of the application root
 
@@ -435,7 +479,7 @@ defmodule Mix.Project do
 
   """
   @spec deps_paths(keyword) :: %{optional(atom) => Path.t()}
-  def deps_paths(opts \\ []) do
+  def deps_paths(opts \\ []) when is_list(opts) do
     traverse_deps(opts, fn %{opts: opts} -> opts[:dest] end)
   end
 
@@ -564,8 +608,10 @@ defmodule Mix.Project do
 
   ## Examples
 
+  If your project defines the app `my_app`:
+
       Mix.Project.manifest_path()
-      #=> "/path/to/project/_build/shared/lib/app/.mix"
+      #=> "/path/to/project/_build/shared/lib/my_app/.mix"
 
   """
   @spec manifest_path(keyword) :: Path.t()
@@ -588,8 +634,10 @@ defmodule Mix.Project do
 
   ## Examples
 
+  If your project defines the app `my_app`:
+
       Mix.Project.app_path()
-      #=> "/path/to/project/_build/shared/lib/app"
+      #=> "/path/to/project/_build/shared/lib/my_app"
 
   """
   @spec app_path(keyword) :: Path.t()
@@ -600,7 +648,7 @@ defmodule Mix.Project do
           Path.join([build_path(config), "lib", Atom.to_string(app)])
 
         config[:apps_path] ->
-          raise "trying to access Mix.Project.app_path for an umbrella project but umbrellas have no app"
+          raise "trying to access Mix.Project.app_path/1 for an umbrella project but umbrellas have no app"
 
         true ->
           Mix.raise(
@@ -620,8 +668,10 @@ defmodule Mix.Project do
 
   ## Examples
 
+  If your project defines the app `my_app`:
+
       Mix.Project.compile_path()
-      #=> "/path/to/project/_build/dev/lib/app/ebin"
+      #=> "/path/to/project/_build/dev/lib/my_app/ebin"
 
   """
   @spec compile_path(keyword) :: Path.t()
@@ -636,6 +686,8 @@ defmodule Mix.Project do
 
   ## Examples
 
+  If your project defines the app `my_app`:
+
       Mix.Project.consolidation_path()
       #=> "/path/to/project/_build/dev/lib/my_app/consolidated"
 
@@ -671,45 +723,39 @@ defmodule Mix.Project do
   """
   @spec build_structure(keyword, keyword) :: :ok
   def build_structure(config \\ config(), opts \\ []) do
-    app = app_path(config)
-    File.mkdir_p!(app)
+    source = opts[:source] || File.cwd!()
+    target = app_path(config)
+    File.mkdir_p!(target)
 
-    source = Path.expand("ebin")
-    target = Path.join(app, "ebin")
+    target_ebin = Path.join(target, "ebin")
+    hard_copy? = config[:build_embedded]
 
     _ =
       cond do
         opts[:symlink_ebin] ->
-          _ = symlink_or_copy(config, source, target)
+          _ = Mix.Utils.symlink_or_copy(hard_copy?, Path.join(source, "ebin"), target_ebin)
 
-        match?({:ok, _}, :file.read_link(target)) ->
-          _ = File.rm_rf!(target)
-          File.mkdir_p!(target)
+        match?({:ok, _}, :file.read_link(target_ebin)) ->
+          _ = File.rm_rf!(target_ebin)
+          File.mkdir_p!(target_ebin)
 
         true ->
-          File.mkdir_p!(target)
+          File.mkdir_p!(target_ebin)
       end
 
-    _ = symlink_or_copy(config, Path.expand("include"), Path.join(app, "include"))
-    _ = symlink_or_copy(config, Path.expand("priv"), Path.join(app, "priv"))
-    :ok
-  end
-
-  defp symlink_or_copy(config, source, target) do
-    if config[:build_embedded] do
-      if File.exists?(source) do
-        File.rm_rf!(target)
-        File.cp_r!(source, target)
-      end
-    else
-      Mix.Utils.symlink_or_copy(source, target)
+    for dir <- ~w(include priv) do
+      Mix.Utils.symlink_or_copy(hard_copy?, Path.join(source, dir), Path.join(target, dir))
     end
+
+    :ok
   end
 
   @doc """
   Ensures the project structure for the given project exists.
 
   In case it does exist, it is a no-op. Otherwise, it is built.
+
+  `opts` are the same options that can be passed to `build_structure/2`.
   """
   @spec ensure_structure(keyword, keyword) :: :ok
   def ensure_structure(config \\ config(), opts \\ []) do
