@@ -7,13 +7,12 @@ defmodule Mix.Release do
   The Mix.Release struct has the following read-only fields:
 
     * `:name` - the name of the release as an atom
-    * `:version` - the version of the release as a string or
-       `{:from_app, app_name}`
+    * `:version` - the version of the release as a string
     * `:path` - the path to the release root
     * `:version_path` - the path to the release version inside the release
     * `:applications` - a map of application with their definitions
-    * `:erts_source` - the erts source as a charlist (or nil)
-    * `:erts_version` - the erts version as a charlist
+    * `:erts_source` - the ERTS source as a charlist (or nil)
+    * `:erts_version` - the ERTS version as a charlist
 
   The following fields may be modified as long as they keep their defined types:
 
@@ -54,7 +53,7 @@ defmodule Mix.Release do
           name: atom(),
           version: String.t(),
           path: String.t(),
-          version_path: String.t() | {:from_app, application()},
+          version_path: String.t(),
           applications: %{application() => keyword()},
           boot_scripts: %{atom() => [{application(), mode()}]},
           erts_version: charlist(),
@@ -92,14 +91,15 @@ defmodule Mix.Release do
     {erts_source, erts_lib_dir, erts_version} = erts_data(include_erts)
 
     deps_apps = Mix.Project.deps_apps()
-    loaded_apps = apps |> Keyword.keys() |> load_apps(deps_apps, %{}, erts_lib_dir, :maybe)
+    loaded_apps = load_apps(apps, deps_apps, %{}, erts_lib_dir, [], apps)
 
     # Make sure IEx is either an active part of the release or add it as none.
     {loaded_apps, apps} =
       if Map.has_key?(loaded_apps, :iex) do
         {loaded_apps, apps}
       else
-        {load_apps([:iex], deps_apps, loaded_apps, erts_lib_dir, :maybe), apps ++ [iex: :none]}
+        {load_apps([iex: :none], deps_apps, loaded_apps, erts_lib_dir, [], apps),
+         apps ++ [iex: :none]}
       end
 
     start_boot = build_start_boot(loaded_apps, apps)
@@ -266,69 +266,110 @@ defmodule Mix.Release do
     end
   end
 
-  defp load_apps(apps, deps_apps, seen, otp_root, included) do
-    for app <- apps, reduce: seen do
+  defp load_apps(apps, deps_apps, seen, otp_root, optional, overrides) do
+    for {app, mode} <- apps, reduce: seen do
       seen ->
-        if reentrant_seen = reentrant(seen, app, included) do
-          reentrant_seen
-        else
-          load_app(app, deps_apps, seen, otp_root, included)
+        properties = seen[app]
+
+        cond do
+          is_nil(properties) ->
+            load_app(app, overrides[app] || mode, deps_apps, seen, otp_root, optional, overrides)
+
+          Keyword.has_key?(overrides, app) ->
+            seen
+
+          new_mode = merge_mode!(app, mode, properties[:mode]) ->
+            apps =
+              properties
+              |> Keyword.get(:applications, [])
+              |> Enum.map(&{&1, new_mode})
+
+            seen = put_in(seen[app][:mode], new_mode)
+            load_apps(apps, deps_apps, seen, otp_root, [], overrides)
+
+          true ->
+            seen
         end
     end
   end
 
-  defp reentrant(seen, app, included) do
-    properties = seen[app]
+  defp merge_mode!(_app, mode, mode), do: nil
 
+  defp merge_mode!(app, left, right) do
+    if left == :included or right == :included do
+      Mix.raise(
+        "#{inspect(app)} is listed both as a regular application and as an included application"
+      )
+    else
+      merge_mode(left, right)
+    end
+  end
+
+  defp merge_mode(:none, other), do: other
+  defp merge_mode(other, :none), do: other
+  defp merge_mode(:load, other), do: other
+  defp merge_mode(other, :load), do: other
+  defp merge_mode(:temporary, other), do: other
+  defp merge_mode(other, :temporary), do: other
+  defp merge_mode(:transient, other), do: other
+  defp merge_mode(other, :transient), do: other
+  defp merge_mode(:permanent, other), do: other
+  defp merge_mode(other, :permanent), do: other
+
+  defp load_app(app, mode, deps_apps, seen, otp_root, optional, overrides) do
     cond do
-      is_nil(properties) ->
-        nil
+      path = app not in deps_apps && otp_path(otp_root, app) ->
+        do_load_app(app, mode, path, deps_apps, seen, otp_root, true, overrides)
 
-      included != :maybe and properties[:included] != included ->
-        if properties[:included] == :maybe do
-          put_in(seen[app][:included], included)
-        else
-          Mix.raise(
-            "#{inspect(app)} is listed both as a regular application and as an included application"
-          )
-        end
+      path = code_path(app) ->
+        do_load_app(app, mode, path, deps_apps, seen, otp_root, false, overrides)
+
+      app in optional ->
+        seen
 
       true ->
-        seen
+        Mix.raise("Could not find application #{inspect(app)}")
     end
   end
 
-  defp load_app(app, deps_apps, seen, otp_root, included) do
-    {path, otp_app?} = if app in deps_apps, do: code_path(app), else: otp_path(otp_root, app)
-    do_load_app(app, path, deps_apps, seen, otp_root, otp_app?, included)
-  end
-
   defp otp_path(otp_root, app) do
     path = Path.join(otp_root, "#{app}-*")
 
     case Path.wildcard(path) do
-      [] -> code_path(app)
-      paths -> {paths |> Enum.sort() |> List.last() |> to_charlist(), true}
+      [] -> nil
+      paths -> paths |> Enum.sort() |> List.last() |> to_charlist()
     end
   end
 
   defp code_path(app) do
     case :code.lib_dir(app) do
-      {:error, :bad_name} -> Mix.raise("Could not find application #{inspect(app)}")
-      path -> {path, false}
+      {:error, :bad_name} -> nil
+      path -> path
     end
   end
 
-  defp do_load_app(app, path, deps_apps, seen, otp_root, otp_app?, included) do
+  defp do_load_app(app, mode, path, deps_apps, seen, otp_root, otp_app?, overrides) do
     case :file.consult(Path.join(path, "ebin/#{app}.app")) do
       {:ok, terms} ->
         [{:application, ^app, properties}] = terms
-        value = [path: path, otp_app?: otp_app?, included: included] ++ properties
+        value = [path: path, otp_app?: otp_app?, mode: mode] ++ properties
         seen = Map.put(seen, app, value)
-        applications = Keyword.get(properties, :applications, [])
-        seen = load_apps(applications, deps_apps, seen, otp_root, false)
-        included_applications = Keyword.get(properties, :included_applications, [])
-        load_apps(included_applications, deps_apps, seen, otp_root, true)
+        child_mode = if mode == :included, do: :load, else: mode
+
+        applications =
+          properties
+          |> Keyword.get(:applications, [])
+          |> Enum.map(&{&1, child_mode})
+
+        optional = Keyword.get(properties, :optional_applications, [])
+        seen = load_apps(applications, deps_apps, seen, otp_root, optional, overrides)
+
+        included_applications =
+          properties
+          |> Keyword.get(:included_applications, [])
+          |> Enum.map(&{&1, :included})
+
+        load_apps(included_applications, deps_apps, seen, otp_root, [], overrides)
 
       {:error, reason} ->
         Mix.raise("Could not load #{app}.app. Reason: #{inspect(reason)}")
@@ -341,14 +382,13 @@ defmodule Mix.Release do
         for(
           {app, props} <- all_apps,
           not List.keymember?(specified_apps, app, 0),
-          do: {app, default_mode(props)}
+          do: {app, boot_mode(props[:mode])}
         )
       )
   end
 
-  defp default_mode(props) do
-    if props[:included] == true, do: :load, else: :permanent
-  end
+  defp boot_mode(:included), do: :load
+  defp boot_mode(mode), do: mode
 
   defp build_start_clean_boot(boot) do
     for({app, _mode} <- boot, do: {app, :none})
@@ -445,13 +485,12 @@ defmodule Mix.Release do
     end
   end
 
-  defp valid_config?(n) when is_number(n), do: true
-  defp valid_config?(a) when is_atom(a), do: true
-  defp valid_config?(b) when is_binary(b), do: true
+  defp valid_config?(m) when is_map(m),
+    do: Enum.all?(Map.delete(m, :__struct__), &valid_config?/1)
+
   defp valid_config?(l) when is_list(l), do: Enum.all?(l, &valid_config?/1)
-  defp valid_config?(m) when is_map(m), do: Enum.all?(m, &valid_config?/1)
   defp valid_config?(t) when is_tuple(t), do: Enum.all?(Tuple.to_list(t), &valid_config?/1)
-  defp valid_config?(_), do: false
+  defp valid_config?(o), do: is_number(o) or is_atom(o) or is_binary(o)
 
   defp merge_provider_config(%{config_providers: []}, sys_config, _), do: {sys_config, false}
 
@@ -512,7 +551,8 @@ defmodule Mix.Release do
   def make_cookie(release, path) do
     cond do
       cookie = release.options[:cookie] ->
-        Mix.Generator.create_file(path, cookie, quiet: true)
+        force? = Keyword.get(release.options, :overwrite, false)
+        Mix.Generator.create_file(path, cookie, quiet: true, force: force?)
         :ok
 
       File.exists?(path) ->
@@ -598,13 +638,24 @@ defmodule Mix.Release do
   end
 
   defp build_release_spec(release, modes) do
-    %{name: name, version: version, erts_version: erts_version, applications: apps} = release
+    %{
+      name: name,
+      version: version,
+      erts_version: erts_version,
+      applications: apps,
+      options: options
+    } = release
+
+    skip_mode_validation_for =
+      options
+      |> Keyword.get(:skip_mode_validation_for, [])
+      |> MapSet.new()
 
     rel_apps =
       for {app, mode} <- modes do
         properties = Map.get(apps, app) || throw({:error, "Unknown application #{inspect(app)}"})
         children = Keyword.get(properties, :applications, [])
-        validate_mode!(app, mode, modes, children)
+        app in skip_mode_validation_for || validate_mode!(app, mode, modes, children)
         build_app_for_release(app, mode, properties)
       end
 
@@ -722,7 +773,7 @@ defmodule Mix.Release do
 
     release.erts_source
     |> Path.join("bin")
-    |> File.cp_r!(destination, fn _, _ -> false end)
+    |> File.cp_r!(destination, on_conflict: fn _, _ -> false end)
 
     _ = File.rm(Path.join(destination, "erl"))
     _ = File.rm(Path.join(destination, "erl.ini"))
@@ -772,14 +823,7 @@ defmodule Mix.Release do
       for dir <- @copy_app_dirs do
         source_dir = Path.join(source_app, dir)
         target_dir = Path.join(target_app, dir)
-
-        source_dir =
-          case File.read_link(source_dir) do
-            {:ok, link_target} -> Path.expand(link_target, source_app)
-            _ -> source_dir
-          end
-
-        File.exists?(source_dir) && File.cp_r!(source_dir, target_dir)
+        File.exists?(source_dir) && File.cp_r!(source_dir, target_dir, dereference_symlinks: true)
       end
 
       true
@@ -834,12 +878,18 @@ defmodule Mix.Release do
   def strip_beam(binary, options \\ []) when is_list(options) do
     chunks_to_keep = options[:keep] |> List.wrap() |> Enum.map(&String.to_charlist/1)
     all_chunks = Enum.uniq(@significant_chunks ++ chunks_to_keep)
+    compress? = Keyword.get(options, :compress, false)
 
     case :beam_lib.chunks(binary, all_chunks, [:allow_missing_chunks]) do
       {:ok, {_, chunks}} ->
         chunks = for {name, chunk} <- chunks, is_binary(chunk), do: {name, chunk}
         {:ok, binary} = :beam_lib.build_module(chunks)
-        {:ok, :zlib.gzip(binary)}
+
+        if compress? do
+          {:ok, :zlib.gzip(binary)}
+        else
+          {:ok, binary}
+        end
 
       {:error, _, _} = error ->
         error
