@@ -4,7 +4,7 @@ defmodule Mix.Tasks.Deps.Loadpaths do
   import Mix.Dep, only: [format_dep: 1, format_status: 1, check_lock: 1]
 
   @moduledoc """
-  Checks and loads all dependencies along the way.
+  Checks, compiles, and loads all dependencies along the way.
 
   If there is an invalid dependency, its status is printed
   before aborting.
@@ -14,9 +14,11 @@ defmodule Mix.Tasks.Deps.Loadpaths do
 
   ## Command line options
 
-    * `--no-deps-check` - does not check or compile deps, only load available ones
     * `--no-compile` - does not compile dependencies
-    * `--no-load-deps` - does not load deps from the code path
+    * `--no-deps-check` - does not check or compile deps, only load available ones
+    * `--no-elixir-version-check` - does not check Elixir version
+    * `--no-load-deps` - does not add deps loadpaths to the code path
+    * `--no-optional-deps` - does not compile or load optional deps
 
   """
 
@@ -24,58 +26,51 @@ defmodule Mix.Tasks.Deps.Loadpaths do
   def run(args) do
     all = Mix.Dep.load_and_cache()
 
+    all =
+      if "--no-optional-deps" in args do
+        for dep <- all, dep.opts[:optional] != true, do: dep
+      else
+        all
+      end
+
+    config = Mix.Project.config()
+
+    unless "--no-elixir-version-check" in args do
+      check_elixir_version(config)
+    end
+
     unless "--no-deps-check" in args do
       deps_check(all, "--no-compile" in args)
     end
 
     unless "--no-load-deps" in args do
-      load_paths =
-        for dep <- all,
-            path <- Mix.Dep.load_paths(dep) do
-          _ = Code.prepend_path(path)
-          path
-        end
-
-      prune_deps(load_paths, "--no-deps-check" in args)
+      for dep <- all,
+          path <- Mix.Dep.load_paths(dep) do
+        _ = Code.prepend_path(path)
+        path
+      end
     end
   end
 
-  # If the build is per environment, we should be able to look
-  # at all dependencies and remove the builds that no longer
-  # have a dependency defined for them.
-  #
-  # Note that we require the build_path to be nil. If it is not nil,
-  # it means the build_path is shared so we don't delete entries.
-  #
-  # We also expect env_path to be nil. If it is not nil, it means
-  # it was set by a parent application and the parent application
-  # should be the one doing the pruning.
-  defp prune_deps(load_paths, no_check?) do
-    config = Mix.Project.config()
+  defp check_elixir_version(config) do
+    if req = config[:elixir] do
+      case Version.parse_requirement(req) do
+        {:ok, req} ->
+          unless Version.match?(System.version(), req) do
+            raise Mix.ElixirVersionError,
+              target: config[:app] || Mix.Project.get(),
+              expected: req,
+              actual: System.version()
+          end
 
-    shared_build? =
-      no_check? or config[:build_path] != nil or config[:build_per_environment] == false
-
-    config
-    |> Mix.Project.build_path()
-    |> Path.join("lib/*/ebin")
-    |> Path.wildcard()
-    |> List.delete(config[:app] && Mix.Project.compile_path(config))
-    |> Kernel.--(load_paths)
-    |> Enum.each(&prune_path(&1, shared_build?))
-  end
-
-  defp prune_path(path, shared_build?) do
-    _ = Code.delete_path(path)
-
-    unless shared_build? do
-      path |> Path.dirname() |> File.rm_rf!()
+        :error ->
+          Mix.raise("Invalid Elixir version requirement #{req} in mix.exs file")
+      end
     end
   end
 
   defp deps_check(all, no_compile?) do
     all = Enum.map(all, &check_lock/1)
-
     {not_ok, compile} = partition(all, [], [])
 
     cond do
@@ -119,7 +114,7 @@ defmodule Mix.Tasks.Deps.Loadpaths do
 
   # Those are compiled by umbrella.
   defp from_umbrella?(dep) do
-    dep.opts()[:from_umbrella]
+    dep.opts[:from_umbrella]
   end
 
   # Every local dependency (i.e. that are not fetchable)
