@@ -8,8 +8,8 @@ defmodule Mix.Tasks.Format do
 
       mix format mix.exs "lib/**/*.{ex,exs}" "test/**/*.{ex,exs}"
 
-  If any of the files is `-`, then the output is read from stdin
-  and written to stdout.
+  If any of the files is `-`, then the input is read from stdin and the output
+  is written to stdout.
 
   ## Formatting options
 
@@ -29,6 +29,10 @@ defmodule Mix.Tasks.Format do
       to be used by this task. For example, `["mix.exs", "{config,lib,test}/**/*.{ex,exs}"]`.
       Patterns are expanded with `Path.wildcard/2`.
 
+    * `:plugins` (a list of modules) (since v1.13.0) - specifies a list of
+      modules to customize how the formatter works. See the "Plugins" section
+      below for more information.
+
     * `:subdirectories` (a list of paths and patterns) - specifies subdirectories
       that have their own formatting rules. Each subdirectory should have a
       `.formatter.exs` that configures how entries in that subdirectory should be
@@ -41,9 +45,8 @@ defmodule Mix.Tasks.Format do
 
     * `:import_deps` (a list of dependencies as atoms) - specifies a list
        of dependencies whose formatter configuration will be imported.
-       When specified, the formatter should run in the same directory as
-       the `mix.exs` file that defines those dependencies. See the "Importing
-       dependencies configuration" section below for more information.
+       See the "Importing dependencies configuration" section below for more
+       information.
 
     * `:export` (a keyword list) - specifies formatter configuration to be exported.
       See the "Importing dependencies configuration" section below.
@@ -52,29 +55,27 @@ defmodule Mix.Tasks.Format do
 
     * `--check-formatted` - checks that the file is already formatted.
       This is useful in pre-commit hooks and CI scripts if you want to
-      reject contributions with unformatted code. However keep in mind
+      reject contributions with unformatted code. If the check fails,
+      the formatted contents are not written to disk. Keep in mind
       that the formatted output may differ between Elixir versions as
       improvements and fixes are applied to the formatter.
 
-    * `--check-equivalent` - checks if the files after formatting have the
-      same AST as before formatting. If the ASTs are not equivalent, it is
-      a bug in the code formatter. This option is useful if you suspect you
-      have ran into a formatter bug and you would like confirmation.
-
     * `--dry-run` - does not save files after formatting.
 
     * `--dot-formatter` - path to the file with formatter configuration.
-      Defaults to `.formatter.exs` if one is available. See the "Formatting options"
-      section above for more information.
+      Defaults to `.formatter.exs` if one is available. See the
+      "Formatting options" section above for more information.
 
-  If any of the `--check-*` options are given and a check fails, the formatted
-  contents won't be written to disk nor printed to standard output.
+    * `--stdin-filename` - path to the file being formatted on stdin.
+      This is useful if you are using plugins to support custom filetypes such
+      as `.heex`. Without passing this flag, it is assumed that the code being
+      passed via stdin is valid Elixir code. Defaults to "stdin.exs".
 
   ## When to format code
 
   We recommend developers to format code directly in their editors, either
   automatically when saving a file or via an explicit command or key binding. If
-  such option is not yet available in your editor of choice, adding the required
+  such option is not available in your editor of choice, adding the required
   integration is usually a matter of invoking:
 
       cd $project && mix format $file
@@ -87,13 +88,54 @@ defmodule Mix.Tasks.Format do
   documentation. This list can also be set in the `.formatter.exs` file under the
   `:inputs` key.
 
+  ## Plugins
+
+  It is possible to customize how the formatter behaves. Plugins must implement
+  the `Mix.Tasks.Format` behaviour. For example, imagine that your project uses
+  Markdown in two distinct ways: via a custom `~M` sigil and via files with the
+  `.md` and `.markdown` extensions. A custom plugin would look like this:
+
+      defmodule MixMarkdownFormatter do
+        @behaviour Mix.Tasks.Format
+
+        def features(_opts) do
+          [sigils: [:M], extensions: [".md", ".markdown"]]
+        end
+
+        def format(contents, opts) do
+          # logic that formats markdown
+        end
+      end
+
+  The `opts` passed to `format/2` contains all the formatting options and either:
+
+    * `:sigil` (atom) - the sigil being formatted, e.g. `:M`.
+
+    * `:modifiers` (charlist) - list of sigil modifiers.
+
+    * `:extension` (string) - the extension of the file being formatted, e.g. `".md"`.
+
+  Now any application can use your formatter as follows:
+
+      # .formatters.exs
+      [
+        # Define the desired plugins
+        plugins: [MixMarkdownFormatter],
+        # Remember to update the inputs list to include the new extensions
+        inputs: ["{mix,.formatter}.exs", "{config,lib,test}/**/*.{ex,exs}", "posts/*.{md,markdown}"]
+      ]
+
+  Remember that, when running the formatter with plugins, you must make
+  sure that your dependencies and your application have been compiled,
+  so the relevant plugin code can be loaded. Otherwise a warning is logged.
+
   ## Importing dependencies configuration
 
   This task supports importing formatter configuration from dependencies.
 
   A dependency that wants to export formatter configuration needs to have a
   `.formatter.exs` file at the root of the project. In this file, the dependency
-  can export a `:export` option with configuration to export. For now, only one
+  can list an `:export` option with configuration to export. For now, only one
   option is supported under `:export`: `:locals_without_parens` (whose value has
   the same shape as the value of the `:locals_without_parens` in `Code.format_string!/2`).
 
@@ -129,38 +171,68 @@ defmodule Mix.Tasks.Format do
     check_equivalent: :boolean,
     check_formatted: :boolean,
     dot_formatter: :string,
-    dry_run: :boolean
+    dry_run: :boolean,
+    stdin_filename: :string
   ]
 
   @manifest "cached_dot_formatter"
   @manifest_vsn 1
 
+  @doc """
+  Returns which features this plugin should plug into.
+  """
+  @callback features(Keyword.t()) :: [sigils: [atom()], extensions: [binary()]]
+
+  @doc """
+  Receives a string to be formatted with options and returns said string.
+  """
+  @callback format(String.t(), Keyword.t()) :: String.t()
+
   @impl true
   def run(args) do
     {opts, args} = OptionParser.parse!(args, strict: @switches)
     {dot_formatter, formatter_opts} = eval_dot_formatter(opts)
 
+    if opts[:check_equivalent] do
+      IO.warn("--check-equivalent has been deprecated and has no effect")
+    end
+
     {formatter_opts_and_subs, _sources} =
       eval_deps_and_subdirectories(dot_formatter, [], formatter_opts, [dot_formatter])
 
     args
-    |> expand_args(dot_formatter, formatter_opts_and_subs)
+    |> expand_args(dot_formatter, formatter_opts_and_subs, opts)
     |> Task.async_stream(&format_file(&1, opts), ordered: false, timeout: 30000)
-    |> Enum.reduce({[], [], []}, &collect_status/2)
+    |> Enum.reduce({[], []}, &collect_status/2)
     |> check!()
   end
 
   @doc """
-  Returns formatter options to be used for the given file.
+  Returns a formatter function and the formatter options to
+  be used for the given file.
+
+  The function must be called with the contents of the file
+  to be formatted. The options are returned for reflection
+  purposes.
   """
-  def formatter_opts_for_file(file, opts \\ []) do
+  @doc since: "1.13.0"
+  def formatter_for_file(file, opts \\ []) do
     {dot_formatter, formatter_opts} = eval_dot_formatter(opts)
 
     {formatter_opts_and_subs, _sources} =
       eval_deps_and_subdirectories(dot_formatter, [], formatter_opts, [dot_formatter])
 
-    split = file |> Path.relative_to_cwd() |> Path.split()
-    find_formatter_opts_for_file(split, formatter_opts_and_subs)
+    find_formatter_and_opts_for_file(file, formatter_opts_and_subs)
+  end
+
+  @doc """
+  Returns formatter options to be used for the given file.
+  """
+  # TODO: Deprecate on Elixir v1.17
+  @doc deprecated: "Use formatter_for_file/2 instead"
+  def formatter_opts_for_file(file, opts \\ []) do
+    {_, formatter_opts} = formatter_for_file(file, opts)
+    formatter_opts
   end
 
   defp eval_dot_formatter(opts) do
@@ -182,6 +254,7 @@ defmodule Mix.Tasks.Format do
   defp eval_deps_and_subdirectories(dot_formatter, prefix, formatter_opts, sources) do
     deps = Keyword.get(formatter_opts, :import_deps, [])
     subs = Keyword.get(formatter_opts, :subdirectories, [])
+    plugins = Keyword.get(formatter_opts, :plugins, [])
 
     if not is_list(deps) do
       Mix.raise("Expected :import_deps to return a list of dependencies, got: #{inspect(deps)}")
@@ -191,6 +264,41 @@ defmodule Mix.Tasks.Format do
       Mix.raise("Expected :subdirectories to return a list of directories, got: #{inspect(subs)}")
     end
 
+    if not is_list(plugins) do
+      Mix.raise("Expected :plugins to return a list of modules, got: #{inspect(plugins)}")
+    end
+
+    if plugins != [] do
+      Mix.Task.run("loadpaths", [])
+    end
+
+    if not Enum.all?(plugins, &Code.ensure_loaded?/1) do
+      Mix.Task.run("compile", [])
+    end
+
+    for plugin <- plugins do
+      cond do
+        not Code.ensure_loaded?(plugin) ->
+          Mix.raise("Formatter plugin #{inspect(plugin)} cannot be found")
+
+        not function_exported?(plugin, :features, 1) ->
+          Mix.raise("Formatter plugin #{inspect(plugin)} does not define features/1")
+
+        true ->
+          :ok
+      end
+    end
+
+    sigils =
+      for plugin <- plugins,
+          sigil <- find_sigils_from_plugins(plugin, formatter_opts),
+          do: {sigil, &plugin.format(&1, &2 ++ formatter_opts)}
+
+    formatter_opts =
+      formatter_opts
+      |> Keyword.put(:plugins, plugins)
+      |> Keyword.put(:sigils, sigils)
+
     if deps == [] and subs == [] do
       {{formatter_opts, []}, sources}
     else
@@ -316,7 +424,7 @@ defmodule Mix.Tasks.Format do
     opts
   end
 
-  defp expand_args([], dot_formatter, formatter_opts_and_subs) do
+  defp expand_args([], dot_formatter, formatter_opts_and_subs, _opts) do
     if no_entries_in_formatter_opts?(formatter_opts_and_subs) do
       Mix.raise(
         "Expected one or more files/patterns to be given to mix format " <>
@@ -326,10 +434,12 @@ defmodule Mix.Tasks.Format do
 
     dot_formatter
     |> expand_dot_inputs([], formatter_opts_and_subs, %{})
-    |> Enum.map(fn {file, {_dot_formatter, formatter_opts}} -> {file, formatter_opts} end)
+    |> Enum.map(fn {file, {_dot_formatter, formatter_opts}} ->
+      {file, find_formatter_for_file(file, formatter_opts)}
+    end)
   end
 
-  defp expand_args(files_and_patterns, _dot_formatter, {formatter_opts, subs}) do
+  defp expand_args(files_and_patterns, _dot_formatter, {formatter_opts, subs}, opts) do
     files =
       for file_or_pattern <- files_and_patterns,
           file <- stdin_or_wildcard(file_or_pattern),
@@ -345,10 +455,15 @@ defmodule Mix.Tasks.Format do
 
     for file <- files do
       if file == :stdin do
-        {file, formatter_opts}
+        stdin_filename = Keyword.get(opts, :stdin_filename, "stdin.exs")
+
+        {formatter, _opts} =
+          find_formatter_and_opts_for_file(stdin_filename, {formatter_opts, subs})
+
+        {file, formatter}
       else
-        split = file |> Path.relative_to_cwd() |> Path.split()
-        {file, find_formatter_opts_for_file(split, {formatter_opts, subs})}
+        {formatter, _opts} = find_formatter_and_opts_for_file(file, {formatter_opts, subs})
+        {file, formatter}
       end
     end
   end
@@ -390,10 +505,40 @@ defmodule Mix.Tasks.Format do
     end
   end
 
-  defp find_formatter_opts_for_file(split, {formatter_opts, subs}) do
+  defp find_formatter_for_file(file, formatter_opts) do
+    ext = Path.extname(file)
+
+    cond do
+      plugin = find_plugin_for_extension(formatter_opts, ext) ->
+        &plugin.format(&1, [extension: ext, file: file] ++ formatter_opts)
+
+      ext in ~w(.ex .exs) ->
+        &elixir_format(&1, [file: file] ++ formatter_opts)
+
+      true ->
+        & &1
+    end
+  end
+
+  defp find_plugin_for_extension(formatter_opts, ext) do
+    plugins = Keyword.get(formatter_opts, :plugins, [])
+
+    Enum.find(plugins, fn plugin ->
+      Code.ensure_loaded?(plugin) and function_exported?(plugin, :features, 1) and
+        ext in List.wrap(plugin.features(formatter_opts)[:extensions])
+    end)
+  end
+
+  defp find_formatter_and_opts_for_file(file, formatter_opts_and_subs) do
+    split = file |> Path.relative_to_cwd() |> Path.split()
+    formatter_opts = recur_formatter_opts_for_file(split, formatter_opts_and_subs)
+    {find_formatter_for_file(file, formatter_opts), formatter_opts}
+  end
+
+  defp recur_formatter_opts_for_file(split, {formatter_opts, subs}) do
     Enum.find_value(subs, formatter_opts, fn {sub, formatter_opts_and_subs} ->
       if List.starts_with?(split, Path.split(sub)) do
-        find_formatter_opts_for_file(split, formatter_opts_and_subs)
+        recur_formatter_opts_for_file(split, formatter_opts_and_subs)
       end
     end)
   end
@@ -405,26 +550,31 @@ defmodule Mix.Tasks.Format do
   defp stdin_or_wildcard("-"), do: [:stdin]
   defp stdin_or_wildcard(path), do: path |> Path.expand() |> Path.wildcard(match_dot: true)
 
-  defp read_file(:stdin) do
-    {IO.stream() |> Enum.to_list() |> IO.iodata_to_binary(), file: "stdin"}
+  defp elixir_format(content, formatter_opts) do
+    case Code.format_string!(content, formatter_opts) do
+      [] -> ""
+      formatted_content -> IO.iodata_to_binary([formatted_content, ?\n])
+    end
   end
 
-  defp read_file(file) do
-    {File.read!(file), file: file}
+  defp find_sigils_from_plugins(plugin, formatter_opts) do
+    if Code.ensure_loaded?(plugin) and function_exported?(plugin, :features, 1) do
+      List.wrap(plugin.features(formatter_opts)[:sigils])
+    else
+      []
+    end
   end
 
-  defp format_file({file, formatter_opts}, task_opts) do
-    {input, extra_opts} = read_file(file)
-    output = IO.iodata_to_binary([Code.format_string!(input, extra_opts ++ formatter_opts), ?\n])
+  defp read_file(:stdin), do: IO.stream() |> Enum.to_list() |> IO.iodata_to_binary()
+  defp read_file(file), do: File.read!(file)
 
-    check_equivalent? = Keyword.get(task_opts, :check_equivalent, false)
+  defp format_file({file, formatter}, task_opts) do
+    input = read_file(file)
+    output = formatter.(input)
     check_formatted? = Keyword.get(task_opts, :check_formatted, false)
     dry_run? = Keyword.get(task_opts, :dry_run, false)
 
     cond do
-      check_equivalent? and not equivalent?(input, output) ->
-        {:not_equivalent, file}
-
       check_formatted? ->
         if input == output, do: :ok, else: {:not_formatted, file}
 
@@ -451,44 +601,29 @@ defmodule Mix.Tasks.Format do
 
   defp collect_status({:ok, :ok}, acc), do: acc
 
-  defp collect_status({:ok, {:exit, _, _, _} = exit}, {exits, not_equivalent, not_formatted}) do
-    {[exit | exits], not_equivalent, not_formatted}
+  defp collect_status({:ok, {:exit, _, _, _} = exit}, {exits, not_formatted}) do
+    {[exit | exits], not_formatted}
   end
 
-  defp collect_status({:ok, {:not_equivalent, file}}, {exits, not_equivalent, not_formatted}) do
-    {exits, [file | not_equivalent], not_formatted}
+  defp collect_status({:ok, {:not_formatted, file}}, {exits, not_formatted}) do
+    {exits, [file | not_formatted]}
   end
 
-  defp collect_status({:ok, {:not_formatted, file}}, {exits, not_equivalent, not_formatted}) do
-    {exits, not_equivalent, [file | not_formatted]}
-  end
-
-  defp check!({[], [], []}) do
+  defp check!({[], []}) do
     :ok
   end
 
-  defp check!({[{:exit, :stdin, exception, stacktrace} | _], _not_equivalent, _not_formatted}) do
+  defp check!({[{:exit, :stdin, exception, stacktrace} | _], _not_formatted}) do
     Mix.shell().error("mix format failed for stdin")
     reraise exception, stacktrace
   end
 
-  defp check!({[{:exit, file, exception, stacktrace} | _], _not_equivalent, _not_formatted}) do
+  defp check!({[{:exit, file, exception, stacktrace} | _], _not_formatted}) do
     Mix.shell().error("mix format failed for file: #{Path.relative_to_cwd(file)}")
     reraise exception, stacktrace
   end
 
-  defp check!({_exits, [_ | _] = not_equivalent, _not_formatted}) do
-    Mix.raise("""
-    mix format failed due to --check-equivalent.
-    The following files are not equivalent:
-
-    #{to_bullet_list(not_equivalent)}
-
-    Please report this bug with the input files at https://github.com/elixir-lang/elixir/issues
-    """)
-  end
-
-  defp check!({_exits, _not_equivalent, [_ | _] = not_formatted}) do
+  defp check!({_exits, [_ | _] = not_formatted}) do
     Mix.raise("""
     mix format failed due to --check-formatted.
     The following files are not formatted:
@@ -500,8 +635,4 @@ defmodule Mix.Tasks.Format do
   defp to_bullet_list(files) do
     Enum.map_join(files, "\n", &"  * #{&1 |> to_string() |> Path.relative_to_cwd()}")
   end
-
-  defp equivalent?(input, output) do
-    Code.Formatter.equivalent(input, output) == :ok
-  end
 end
