@@ -7,6 +7,9 @@ defmodule Mix.Tasks.Loadpaths do
   This task is never directly invoked from the command line,
   but it is rather used as building block by other tasks.
 
+  Dependencies are checked, compiled, and loaded. Each step
+  can be explicitly disabled with flags.
+
   ## Configuration
 
     * `:elixir` - matches the current Elixir version against the
@@ -14,26 +17,27 @@ defmodule Mix.Tasks.Loadpaths do
 
   ## Command line options
 
-    * `--no-archives-check` - does not check archive
-    * `--no-deps-check` - does not check dependencies
+    * `--no-archives-check` - does not check archives
+    * `--no-compile` - does not compile dependencies, only check and load them
+    * `--no-deps-check` - does not check dependencies, only load available ones
     * `--no-elixir-version-check` - does not check Elixir version
+    * `--no-load-deps` - does not add deps loadpaths to the code path
+    * `--no-optional-deps` - does not compile or load optional deps
 
   """
   @impl true
   def run(args) do
     config = Mix.Project.config()
 
-    unless "--no-elixir-version-check" in args do
-      check_elixir_version(config, args)
-    end
-
     unless "--no-archives-check" in args do
       Mix.Task.run("archive.check", args)
     end
 
-    # --no-deps-loading is used only internally. It has no
-    # purpose from Mix.CLI because running a task may load deps.
-    unless "--no-deps-loading" in args do
+    # --from-mix-deps-compile is used only internally to avoid
+    # recursively checking and loading dependencies that have
+    # already been loaded. It has no purpose from Mix.CLI
+    # because running a task may load deps.
+    unless "--from-mix-deps-compile" in args do
       Mix.Task.run("deps.loadpaths", args)
     end
 
@@ -44,23 +48,6 @@ defmodule Mix.Tasks.Loadpaths do
     :ok
   end
 
-  defp check_elixir_version(config, _) do
-    if req = config[:elixir] do
-      case Version.parse_requirement(req) do
-        {:ok, req} ->
-          unless Version.match?(System.version(), req) do
-            raise Mix.ElixirVersionError,
-              target: config[:app] || Mix.Project.get(),
-              expected: req,
-              actual: System.version()
-          end
-
-        :error ->
-          Mix.raise("Invalid Elixir version requirement #{req} in mix.exs file")
-      end
-    end
-  end
-
   defp load_project(config, _args) do
     vsn = {System.version(), :erlang.system_info(:otp_release)}
     scm = config[:build_scm]
