@@ -6,8 +6,8 @@ defmodule Mix.Tasks.Release do
   @moduledoc """
   Assembles a self-contained release for the current project:
 
-      MIX_ENV=prod mix release
-      MIX_ENV=prod mix release NAME
+      $ MIX_ENV=prod mix release
+      $ MIX_ENV=prod mix release NAME
 
   Once a release is assembled, it can be packaged and deployed to a
   target, as long as the target runs on the same operating system (OS)
@@ -34,7 +34,7 @@ defmodule Mix.Tasks.Release do
   and the value is a keyword list with the release configuration.
   Releasing a certain name is done with:
 
-      MIX_ENV=prod mix release demo
+      $ MIX_ENV=prod mix release demo
 
   If the given name does not exist, an error is raised.
 
@@ -54,14 +54,13 @@ defmodule Mix.Tasks.Release do
       interactive and embedded. By default, it runs in the interactive
       mode which dynamically loads modules when they are used for the
       first time. The first time your application calls `Enum.map/2`,
-      the VM will find the `Enum` module and load it. There’s a downside.
-      When you start a new server in production, it may need to load
+      the VM will find the `Enum` module and load it. There's a downside:
+      when you start a new server in production, it may need to load
       many other modules, causing the first requests to have an unusual
-      spike in response time. When running in Erlang/OTP earlier than 23,
-      the system always runs in embedded mode. When using Erlang/OTP 23+,
-      they run in interactive mode while being configured and then it
-      swaps to embedded mode, guaranteeing your system is ready to handle
-      requests after booting.
+      spike in response time. With releases, the system is configured in
+      interactive mode and then it swaps to embedded mode, which preloads
+      all modules and guarantees your system is ready to handle requests
+      after booting.
 
     * Configuration and customization. Releases give developers fine
       grained control over system configuration and the VM flags used
@@ -87,8 +86,8 @@ defmodule Mix.Tasks.Release do
   Once a release is assembled, you can start it by calling
   `bin/RELEASE_NAME start` inside the release. In production, you would do:
 
-      MIX_ENV=prod mix release
-      _build/prod/rel/my_app/bin/my_app start
+      $ MIX_ENV=prod mix release
+      $ _build/prod/rel/my_app/bin/my_app start
 
   `bin/my_app start` will start the system connected to the current standard
   input/output, where logs are also written to by default. This is the
@@ -108,8 +107,8 @@ defmodule Mix.Tasks.Release do
   If you want to invoke specific modules and functions in your release,
   you can do so in two ways: using `eval` or `rpc`.
 
-      bin/RELEASE_NAME eval "IO.puts(:hello)"
-      bin/RELEASE_NAME rpc "IO.puts(:hello)"
+      $ bin/RELEASE_NAME eval "IO.puts(:hello)"
+      $ bin/RELEASE_NAME rpc "IO.puts(:hello)"
 
   The `eval` command starts its own instance of the VM but without
   starting any of the applications in the release and without starting
@@ -161,18 +160,18 @@ defmodule Mix.Tasks.Release do
 
   And to run them:
 
-      bin/RELEASE_NAME eval "MyApp.ReleaseTasks.eval_purge_stale_data()"
-      bin/RELEASE_NAME rpc "MyApp.ReleaseTasks.rpc_print_connected_users()"
+      $ bin/RELEASE_NAME eval "MyApp.ReleaseTasks.eval_purge_stale_data()"
+      $ bin/RELEASE_NAME rpc "MyApp.ReleaseTasks.rpc_print_connected_users()"
 
   ### Daemon mode (Unix-like)
 
   You can run the release in daemon mode with the command:
 
-      bin/RELEASE_NAME daemon
+      $ bin/RELEASE_NAME daemon
 
   In daemon mode, the system is started on the background via
-  [`run_erl`](https://erlang.org/doc/man/run_erl.html). You may also
-  want to enable [`heart`](https://erlang.org/doc/man/heart.html)
+  [`run_erl`](https://www.erlang.org/doc/man/run_erl.html). You may also
+  want to enable [`heart`](https://www.erlang.org/doc/man/heart.html)
   in daemon mode so it automatically restarts the system in case
   of crashes. See the generated `releases/RELEASE_VSN/env.sh` file.
 
@@ -195,10 +194,10 @@ defmodule Mix.Tasks.Release do
 
   While daemons are not available on Windows, it is possible to install a
   released system as a service on Windows with the help of
-  [`erlsrv`](https://erlang.org/doc/man/erlsrv.html). This can be done by
+  [`erlsrv`](https://www.erlang.org/doc/man/erlsrv.html). This can be done by
   running:
 
-      bin/RELEASE_NAME install
+      $ bin/RELEASE_NAME install
 
   Once installed, the service must be explicitly managed via the `erlsrv`
   executable, which is included in the `erts-VSN/bin` directory.
@@ -207,8 +206,8 @@ defmodule Mix.Tasks.Release do
   For example, if you have a release named `demo`, you can install
   the service and then start it from the release root as follows:
 
-      bin/demo install
-      erts-VSN/bin/erlsrv.exe start demo_demo
+      $ bin/demo install
+      $ erts-VSN/bin/erlsrv.exe start demo_demo
 
   The name of the service is `demo_demo` because the name is built
   by concatenating the node name with the release name. Since Elixir
@@ -221,18 +220,20 @@ defmodule Mix.Tasks.Release do
 
   The following commands are supported by `bin/RELEASE_NAME`:
 
-      start        Starts the system
-      start_iex    Starts the system with IEx attached
-      daemon       Starts the system as a daemon (Unix-like only)
-      daemon_iex   Starts the system as a daemon with IEx attached (Unix-like only)
-      install      Installs this system as a Windows service (Windows only)
-      eval "EXPR"  Executes the given expression on a new, non-booted system
-      rpc "EXPR"   Executes the given expression remotely on the running system
-      remote       Connects to the running system via a remote shell
-      restart      Restarts the running system via a remote command
-      stop         Stops the running system via a remote command
-      pid          Prints the operating system PID of the running system via a remote command
-      version      Prints the release name and version to be booted
+  ```text
+  start        Starts the system
+  start_iex    Starts the system with IEx attached
+  daemon       Starts the system as a daemon (Unix-like only)
+  daemon_iex   Starts the system as a daemon with IEx attached (Unix-like only)
+  install      Installs this system as a Windows service (Windows only)
+  eval "EXPR"  Executes the given expression on a new, non-booted system
+  rpc "EXPR"   Executes the given expression remotely on the running system
+  remote       Connects to the running system via a remote shell
+  restart      Restarts the running system via a remote command
+  stop         Stops the running system via a remote command
+  pid          Prints the operating system PID of the running system via a remote command
+  version      Prints the release name and version to be booted
+  ```
 
   ## Deployments
 
@@ -304,22 +305,22 @@ defmodule Mix.Tasks.Release do
   compile the code and assemble the release on the target itself. It would
   be something like this:
 
-      git clone remote://path/to/my_app.git my_app_source
-      cd my_app_source
-      mix deps.get --only prod
-      MIX_ENV=prod mix release
-      _build/prod/rel/my_app/bin/my_app start
+      $ git clone remote://path/to/my_app.git my_app_source
+      $ cd my_app_source
+      $ mix deps.get --only prod
+      $ MIX_ENV=prod mix release
+      $ _build/prod/rel/my_app/bin/my_app start
 
   If you prefer, you can also compile the release to a separate directory,
   so you can erase all source after the release is assembled:
 
-      git clone remote://path/to/my_app.git my_app_source
-      cd my_app_source
-      mix deps.get --only prod
-      MIX_ENV=prod mix release --path ../my_app_release
-      cd ../my_app_release
-      rm -rf ../my_app_source
-      bin/my_app start
+      $ git clone remote://path/to/my_app.git my_app_source
+      $ cd my_app_source
+      $ mix deps.get --only prod
+      $ MIX_ENV=prod mix release --path ../my_app_release
+      $ cd ../my_app_release
+      $ rm -rf ../my_app_source
+      $ bin/my_app start
 
   However, this option can be expensive if you have multiple production
   nodes or if the release assembling process is a long one, as each node
@@ -364,10 +365,16 @@ defmodule Mix.Tasks.Release do
 
   The following options can be set inside your `mix.exs` on each release definition:
 
-    * `:applications` - a keyword list that configures and adds new applications
-      to the release. The key is the application name and the value is one of:
+    * `:applications` - a keyword list with application names as keys and their
+      mode as value. By default `:applications` includes the current application and
+      all applications the current application depends on, recursively. You can include
+      new applications or change the mode of existing ones by listing them here.
 
-        * `:permanent` - the application is started and the node shuts down
+      The order of the applications given will be preserved as much as possible, with
+      only `:kernel`, `:stdlib`, `:sasl`, and `:elixir` listed before the given application
+      list. The supported values are:
+
+        * `:permanent` (default) - the application is started and the node shuts down
           if the application terminates, regardless of reason
         * `:transient` - the application is started and the node shuts down
           if the application terminates abnormally
@@ -377,23 +384,18 @@ defmodule Mix.Tasks.Release do
         * `:none` - the application is part of the release but it is neither
           loaded nor started
 
-      All applications default to `:permanent`.
-
-      By default `:applications` includes the current application and all
-      applications the current application depends on, recursively.
-      You can include new applications or change the mode of
-      existing ones by listing them here. The order of the applications given
-      in `:applications` will be preserved as much as possible, with only
-      `:kernel`, `:stdlib`, `:sasl`, and `:elixir` listed before the given
-      application list.
-
-      Releases assembled from an umbrella project require this configuration
-      to be explicitly given.
+      If you change the mode of an application, the mode will apply to all its child
+      applications. However, if an application has two parents, the mode of the parent
+      with highest priority wins (where `:permanent` has the highest priority, according
+      to the list above).
 
     * `:strip_beams` - controls if BEAM files should have their debug information,
       documentation chunks, and other non-essential metadata removed. Defaults to
       `true`. May be set to `false` to disable stripping. Also accepts
-      `[keep: ["Docs", "Dbgi"]]` to keep certain chunks that are usually stripped.
+     `[keep: ["Docs", "Dbgi"]]` to keep certain chunks that are usually stripped.
+     You can also set the `:compress` option to true to enable individual
+     compression of BEAM files, although it is typically preferred to compress
+     the whole release instead.
 
     * `:cookie` - a string representing the Erlang Distribution cookie. If this
       option is not set, a random cookie is written to the `releases/COOKIE` file
@@ -460,6 +462,14 @@ defmodule Mix.Tasks.Release do
     * `:steps` - a list of steps to execute when assembling the release. See
       the "Steps" section for more information.
 
+    * `:skip_mode_validation_for` - a list of application names
+      (atoms) specifying applications to skip strict validation of
+      "unsafe" modes. An "unsafe" case is when a parent application
+      mode is `:permanent` but one of the applications it depends on
+      is set to `:load`. Use this with care, as a release with
+      invalid modes may no longer boot without additional tweaks.
+      Defaults to `[]`.
+
   Note each release definition can be given as an anonymous function. This
   is useful if some release attributes are expensive to compute:
 
@@ -529,7 +539,7 @@ defmodule Mix.Tasks.Release do
   `@release`, with the `Mix.Release` struct.
 
   The `vm.args` and `remote.vm.args` files may contain any of the VM flags
-  accepted by the [`erl` command](https://erlang.org/doc/man/erl.html).
+  accepted by the [`erl` command](https://www.erlang.org/doc/man/erl.html).
 
   The `env.sh` and `env.bat` is used to set environment variables.
   In there, you can set vars such as `RELEASE_NODE`, `RELEASE_COOKIE`,
@@ -544,14 +554,16 @@ defmodule Mix.Tasks.Release do
   if you want to make sure the Erlang Distribution listens only on
   a given port known at runtime, you can set the following:
 
-      case $RELEASE_COMMAND in
-        start*|daemon*)
-          ELIXIR_ERL_OPTIONS="-kernel inet_dist_listen_min $BEAM_PORT inet_dist_listen_max $BEAM_PORT"
-          export ELIXIR_ERL_OPTIONS
-          ;;
-        *)
-          ;;
-      esac
+  ```bash
+  case $RELEASE_COMMAND in
+    start*|daemon*)
+      ELIXIR_ERL_OPTIONS="-kernel inet_dist_listen_min $BEAM_PORT inet_dist_listen_max $BEAM_PORT"
+      export ELIXIR_ERL_OPTIONS
+      ;;
+    *)
+      ;;
+  esac
+  ```
 
   Note we only set the port on start/daemon commands. If you also limit
   the port on other commands, such as `rpc`, then you will be unable
@@ -560,36 +572,37 @@ defmodule Mix.Tasks.Release do
 
   On Windows, your `env.bat` would look like this:
 
-      IF NOT %RELEASE_COMMAND:start=%==%RELEASE_COMMAND% (
-        set ELIXIR_ERL_OPTIONS="-kernel inet_dist_listen_min %BEAM_PORT% inet_dist_listen_max %BEAM_PORT%"
-      )
+  ```bash
+  IF NOT %RELEASE_COMMAND:start=%==%RELEASE_COMMAND% (
+    set ELIXIR_ERL_OPTIONS="-kernel inet_dist_listen_min %BEAM_PORT% inet_dist_listen_max %BEAM_PORT%"
+  )
+  ```
 
   ## Application configuration
 
-  Releases provides two mechanisms for configuring OTP applications:
-  build-time and runtime.
+  Mix provides two mechanisms for configuring the application environment
+  of your application and your dependencies: build-time and runtime. On this
+  section, we will learn how those mechanisms apply to releases. An introduction
+  to this topic can be found in the "Configuration" section of the `Mix` module.
 
   ### Build-time configuration
 
-  Whenever you invoke a `mix` command, Mix loads the configuration
-  in `config/config.exs`, if said file exists. It is common for the
-  `config/config.exs` file itself to import other configuration based
-  on the current `MIX_ENV`, such as `config/dev.exs`, `config/test.exs`,
-  and `config/prod.exs`. We say that this configuration is a build-time
-  configuration as it is evaluated whenever you compile your code or
-  whenever you assemble the release.
+  Whenever you invoke a `mix` command, Mix loads the configuration in
+  `config/config.exs`, if said file exists. We say that this configuration
+  is a build-time configuration as it is evaluated whenever you compile your
+  code or whenever you assemble the release.
 
   In other words, if your configuration does something like:
 
+      import Config
       config :my_app, :secret_key, System.fetch_env!("MY_APP_SECRET_KEY")
 
   The `:secret_key` key under `:my_app` will be computed on the
-  host machine, whenever the release is built. Setting the
-  `MY_APP_SECRET_KEY` right before starting your release will have
-  no effect.
-
-  Luckily, releases also provide runtime configuration, which we will
-  see next.
+  host machine, whenever the release is built. Therefore if the machine
+  assembling the release not have access to all environment variables used
+  to run your code, loading the configuration will fail as the environment
+  variable is missing. Luckily, Mix also provides runtime configuration,
+  which should be preferred and we will see next.
 
   ### Runtime configuration
 
@@ -626,6 +639,9 @@ defmodule Mix.Tasks.Release do
         ]
       ]
 
+  By setting `:runtime_config_path` to `false` it can be used to prevent
+  a runtime configuration file to be included in the release.
+
   Finally, in order for runtime configuration to work properly (as well
   as any other "Config provider" as defined next), it needs to be able
   to persist the newly computed configuration to disk. The computed config
@@ -638,10 +654,10 @@ defmodule Mix.Tasks.Release do
 
   Releases also supports custom mechanisms, called config providers, to load
   any sort of runtime configuration to the system while it boots. For instance,
-  if you need to access a vault or load configuration from a JSON file, it
-  can be achieved with config providers. The runtime configuration outlined
-  in the previous section, which is handled by the `Config.Reader` provider.
-  See the `Config.Provider` module for more information and more examples.
+  if you need to access a vault or load configuration from a JSON file, it can
+  be achieved with config providers. The runtime configuration outlined in the
+  previous section is handled by the `Config.Reader` provider. See the
+  `Config.Provider` module for more information and more examples.
 
   The following options can be set inside your releases key in your `mix.exs`
   to control how config providers work:
@@ -667,7 +683,7 @@ defmodule Mix.Tasks.Release do
     * `:start_distribution_during_config` - if `:reboot_system_after_config` is
       set, releases only start the Erlang VM distribution features after the config
       files are evaluated. You can set it to `true` if you need distribution during
-      configuration. Defaults to `false` from Erlang/OTP 22+.
+      configuration. Defaults to `false`.
 
     * `:config_providers` - a list of tuples with custom config providers.
       See `Config.Provider` for more information. Defaults to `[]`.
@@ -701,34 +717,36 @@ defmodule Mix.Tasks.Release do
 
   A release is organized as follows:
 
-      bin/
-        RELEASE_NAME
-      erts-ERTS_VSN/
-      lib/
-        APP_NAME-APP_VSN/
-          ebin/
-          include/
-          priv/
-      releases/
-        RELEASE_VSN/
-          consolidated/
-          elixir
-          elixir.bat
-          env.bat
-          env.sh
-          iex
-          iex.bat
-          remote.vm.args
-          runtime.exs
-          start.boot
-          start.script
-          start_clean.boot
-          start_clean.script
-          sys.config
-          vm.args
-        COOKIE
-        start_erl.data
-      tmp/
+  ```text
+  bin/
+    RELEASE_NAME
+  erts-ERTS_VSN/
+  lib/
+    APP_NAME-APP_VSN/
+      ebin/
+      include/
+      priv/
+  releases/
+    RELEASE_VSN/
+      consolidated/
+      elixir
+      elixir.bat
+      env.bat
+      env.sh
+      iex
+      iex.bat
+      remote.vm.args
+      runtime.exs
+      start.boot
+      start.script
+      start_clean.boot
+      start_clean.script
+      sys.config
+      vm.args
+    COOKIE
+    start_erl.data
+  tmp/
+  ```
 
   We document this structure for completeness. In practice, developers
   should not modify any of those files after the release is assembled.
@@ -769,9 +787,10 @@ defmodule Mix.Tasks.Release do
     * `RELEASE_COOKIE` - the release cookie. By default uses the value
       in `releases/COOKIE`. It can be set to a custom value
 
-    * `RELEASE_NODE` - the release node name, in the format `name@host`.
-      It can be set to a custom value. The name part must be made only
-      of letters, digits, underscores, and hyphens
+    * `RELEASE_NODE` - the release node name, in the format `name` or
+      optionally `name@host` if running in distributed mode. It can be
+      set to a custom value. The name part must be made only of letters,
+      digits, underscores, and hyphens
 
     * `RELEASE_SYS_CONFIG` - the location of the sys.config file. It can
       be set to a custom path and it must not include the `.config` extension
@@ -786,17 +805,17 @@ defmodule Mix.Tasks.Release do
       files to. It can be set to a custom directory. It defaults to
       `$RELEASE_ROOT/tmp`
 
-    * `RELEASE_MODE` - if the release should start in embedded or
-      interactive mode. Defaults to "embedded". It applies only to
-      start/daemon/install commands
+    * `RELEASE_MODE` - if the release should load code on demand (interactive)
+      or preload it (embedded). Defaults to "embedded", which increases boot
+      time but it means the runtime will respond faster as it doesn't have to
+      load code. Choose interactive if you need to decrease boot time and reduce
+      memory usage on boot. It applies only to start/daemon/install commands
 
     * `RELEASE_DISTRIBUTION` - how do we want to run the distribution.
       May be `name` (long names), `sname` (short names) or `none`
       (distribution is not started automatically). Defaults to
       `sname` which allows access only within the current system.
-      `name` allows external connections. If `name` is used and you are
-      not running on Erlang/OTP 22 or later, you must set `RELEASE_NODE`
-      to `RELEASE_NAME@127.0.0.1` with an IP or a known host
+      `name` allows external connections
 
     * `RELEASE_BOOT_SCRIPT` - the name of the boot script to use when starting
       the release. This script is used when running commands such as `start` and
@@ -819,11 +838,13 @@ defmodule Mix.Tasks.Release do
   your release and the starting point for each release. For example,
   imagine this umbrella applications:
 
-      my_app_umbrella/
-        apps/
-          my_app_core/
-          my_app_event_processing/
-          my_app_web/
+  ```text
+  my_app_umbrella/
+    apps/
+      my_app_core/
+      my_app_event_processing/
+      my_app_web/
+  ```
 
   where both `my_app_event_processing` and `my_app_web` depend on
   `my_app_core` but they do not depend on each other.
@@ -972,13 +993,13 @@ defmodule Mix.Tasks.Release do
   hot code upgrade it. This is one of the many steps necessary
   to perform hot code upgrades and it must be taken into account by
   every process and application being upgraded in the system.
-  The [`.appup` cookbook](https://erlang.org/doc/design_principles/appup_cookbook.html)
+  The [`.appup` cookbook](https://www.erlang.org/doc/design_principles/appup_cookbook.html)
   provides a good reference and more examples.
 
   Once `.appup`s are created, the next step is to create a `.relup`
   file with all instructions necessary to update the release itself.
   Erlang documentation does provide a chapter on
-  [Creating and upgrading a target system](https://erlang.org/doc/system_principles/create_target.html).
+  [Creating and upgrading a target system](https://www.erlang.org/doc/system_principles/create_target.html).
   [Learn You Some Erlang has a chapter on hot code upgrades](https://learnyousomeerlang.com/relups).
 
   Overall, there are many steps, complexities and assumptions made
@@ -1203,6 +1224,9 @@ defmodule Mix.Tasks.Release do
 
     {path, reboot?} =
       cond do
+        opts[:runtime_config_path] == false ->
+          {false, false}
+
         path = opts[:runtime_config_path] ->
           {path, false}
 
@@ -1217,7 +1241,10 @@ defmodule Mix.Tasks.Release do
           {default_path, false}
 
         File.exists?(deprecated_path) ->
-          # TODO: Warn from Elixir v1.13 onwards
+          IO.warn(
+            "config/releases.exs is deprecated, use config/runtime.exs or set :runtime_config_path in your release configuration instead"
+          )
+
           {deprecated_path, true}
 
         true ->
@@ -1234,7 +1261,7 @@ defmodule Mix.Tasks.Release do
         release = update_in(release.config_providers, &[{Config.Reader, opts} | &1])
         update_in(release.options, &Keyword.put_new(&1, :reboot_system_after_config, reboot?))
 
-      release.config_providers == [] ->
+      release.config_providers == [] and path != false ->
         skipping("runtime configuration (#{default_path} not found)")
         release
 
@@ -1288,7 +1315,7 @@ defmodule Mix.Tasks.Release do
 
     info(release, """
 
-    Release created at #{path}!
+    Release created at #{path}
 
         # To start your system
         #{cmd} start
@@ -1427,9 +1454,11 @@ defmodule Mix.Tasks.Release do
     ]
   end
 
+  @erts_bin [~s[ERTS_BIN="$ERTS_BIN"], ~s[ERTS_BIN=!ERTS_BIN!]]
+
   defp replace_erts_bin(contents, release, new_path) do
     if release.erts_source do
-      String.replace(contents, ~s[ERTS_BIN=], ~s[ERTS_BIN=#{new_path}])
+      String.replace(contents, @erts_bin, ~s[ERTS_BIN=#{new_path}])
     else
       contents
     end
@@ -1440,11 +1469,9 @@ defmodule Mix.Tasks.Release do
   # Helper functions
 
   defp release_mode(release, env_var) do
-    # TODO: Remove otp_release check once we require Erlang/OTP 23+
-    otp_gte_23? = :erlang.system_info(:otp_release) >= '23'
     reboot? = Keyword.get(release.options, :reboot_system_after_config, false)
 
-    if otp_gte_23? and reboot? and release.config_providers != [] do
+    if reboot? and release.config_providers != [] do
       "-elixir -config_provider_reboot_mode #{env_var}"
     else
       "-mode #{env_var}"
