@@ -71,7 +71,7 @@ defmodule Mix.Tasks.Test do
   tests run concurrently.
 
   Finally, how many tests we have run, how many of them
-  failed, how many were invalid, etc.
+  failed, how many were invalid, and so on.
 
   ### Understanding test failures
 
@@ -113,6 +113,9 @@ defmodule Mix.Tasks.Test do
 
     * `--exclude` - excludes tests that match the filter
 
+    * `--exit-status` - use an alternate exit status to use when the test suite
+      fails (default is 2).
+
     * `--export-coverage` - the name of the file to export coverage results to.
       Only has an effect when used with `--cover`
 
@@ -149,12 +152,17 @@ defmodule Mix.Tasks.Test do
 
     * `--only` - runs only tests that match the filter
 
-    * `--partitions` - sets the amount of partitions to split tests in. This option
-      requires the `MIX_TEST_PARTITION` environment variable to be set. See the
-      "Operating system process partitioning" section for more information
+    * `--partitions` - sets the amount of partitions to split tests in. It must be
+      a number greater than zero. If set to one, it acts a no-op. If more than one,
+      then you must also set the `MIX_TEST_PARTITION` environment variable with the
+      partition to use in the current test run. See the "Operating system process
+      partitioning" section for more information
 
     * `--preload-modules` - preloads all modules defined in applications
 
+    * `--profile-require` - profiles the time spent to require test files.
+      Used only for debugging. The test suite does not run.
+
     * `--raise` - raises if the test suite failed
 
     * `--seed` - seeds the random number generator used to randomize the order of tests;
@@ -174,14 +182,21 @@ defmodule Mix.Tasks.Test do
       Note that in trace mode test timeouts will be ignored as timeout is set to `:infinity`
 
     * `--warnings-as-errors` - (since v1.12.0) treats warnings as errors and returns a non-zero
-      exit code. This option only applies to test files. To treat warnings as errors during
+      exit status. This option only applies to test files. To treat warnings as errors during
       compilation and during tests, run:
-          MIX_ENV=test mix do compile --warnings-as-errors, test --warnings-as-errors
+          MIX_ENV=test mix do compile --warnings-as-errors + test --warnings-as-errors
 
   ## Configuration
 
   These configurations can be set in the `def project` section of your `mix.exs`:
 
+    * `:test_coverage` - a set of options to be passed down to the coverage
+      mechanism. See the "Coverage" section for more information
+
+    * `:test_elixirc_options` - the compiler options to used when
+      loading/compiling test files. By default it disables the debug chunk
+      and docs chunk
+
     * `:test_paths` - list of paths containing test files. Defaults to
       `["test"]` if the `test` directory exists; otherwise, it defaults to `[]`.
       It is expected that all test paths contain a `test_helper.exs` file
@@ -191,9 +206,6 @@ defmodule Mix.Tasks.Test do
     * `:warn_test_pattern` - a pattern to match potentially misnamed test files
       and display a warning. Defaults to `*_test.ex`
 
-    * `:test_coverage` - a set of options to be passed down to the coverage
-      mechanism
-
   ## Coloring
 
   Coloring is enabled by default on most Unix terminals. They are also
@@ -262,26 +274,37 @@ defmodule Mix.Tasks.Test do
 
   ## Coverage
 
-  The `:test_coverage` configuration accepts the following options:
+  The `:test_coverage` configures the coverage tool and reports generated
+  by the `--cover` flag. It accepts the following options:
 
-    * `:output` - the output directory for cover results. Defaults to `"cover"`
-    * `:tool` - the coverage tool
-    * `:summary` - summary output configuration; can be either a boolean
-      or a keyword list. When a keyword list is passed, it can specify a
-      `:threshold`, which is a boolean or numeric value that enables coloring
-      of code coverage results in red or green depending on whether the
-      percentage is below or above the specified threshold, respectively.
-      Defaults to `[threshold: 90]`
-    * `:export` - a file name to export results to instead of generating
-      the result on the fly. The `.coverdata` extension is automatically
-      added to the given file. This option is automatically set via the
-      `--export-coverage` option or when using process partitioning.
+    * `:output` - the output directory for cover results. Defaults to `"cover"`.
+
+    * `:tool` - a module specifying the coverage tool to use.
+
+    * `:summary` - at the end of each coverage run, a summary of each
+      module is printed, with results in red or green depending on whether
+      the percentage is below or above a given threshold. The task will
+      exit with status of 1 if the total coverage is below the threshold.
+      The `:summary` option allows you to customize the summary generation
+      and defaults to `[threshold: 90]`, but it may be set to `false` to
+      disable such reports.
+
+    * `:export` - a filename to export results to instead of generating
+      the coverage result on the fly. The `.coverdata` extension is
+      automatically added to the given file. This option is automatically
+      set via the `--export-coverage` option or when using process partitioning.
       See `mix test.coverage` to compile a report from multiple exports.
-    * `:ignore_modules` - modules to ignore from generating reports and
-      in summaries
 
-  By default, a very simple wrapper around OTP's `cover` is used as a tool,
-  but it can be overridden as follows:
+    * `:ignore_modules` - modules to ignore from generating reports and
+      in summaries. It is a list of module names as atoms and regular
+      expressions that are matched against the module names.
+
+    * `:local_only` - by default coverage only tracks local calls,
+      set this option to false if you plan to run coverage across nodes.
+
+  By default, a wrapper around OTP's `cover` is used as the default coverage
+  tool. You can learn more about how it works in the docs for
+  `mix test.coverage`. Your tool of choice can be given as follows:
 
       def project() do
         [
@@ -293,8 +316,8 @@ defmodule Mix.Tasks.Test do
 
   `CoverModule` can be any module that exports `start/2`, receiving the
   compilation path and the `test_coverage` options as arguments.
-  It must return either `nil` or an anonymous function of zero arity that will
-  be run after the test suite is done.
+  It must return either `nil` or an anonymous function of zero arity that
+  will run after the test suite is done.
 
   ## Operating system process partitioning
 
@@ -391,7 +414,9 @@ defmodule Mix.Tasks.Test do
     slowest: :integer,
     partitions: :integer,
     preload_modules: :boolean,
-    warnings_as_errors: :boolean
+    warnings_as_errors: :boolean,
+    profile_require: :string,
+    exit_status: :integer
   ]
 
   @cover [output: "cover", tool: Mix.Tasks.Test.Coverage]
@@ -404,7 +429,10 @@ defmodule Mix.Tasks.Test do
       do_run(opts, args, files)
     else
       {files_in_apps_path, files_not_in_apps_path} =
-        Enum.split_with(files, &String.starts_with?(&1, "apps/"))
+        files
+        |> Enum.map(&Path.expand/1)
+        |> Enum.map(&Path.relative_to_cwd/1)
+        |> Enum.split_with(&String.starts_with?(&1, "apps/"))
 
       app = Mix.Project.config()[:app]
       current_app_path = "apps/#{app}/"
@@ -454,25 +482,23 @@ defmodule Mix.Tasks.Test do
       """)
     end
 
-    # Load ExUnit before we compile anything
+    # Load ExUnit before we compile anything in case we are compiling
+    # helper modules that depend on ExUnit.
     Application.ensure_loaded(:ex_unit)
 
-    old_warnings_as_errors = Code.get_compiler_option(:warnings_as_errors)
-    args = args -- ["--warnings-as-errors"]
-
-    Mix.Task.run("compile", args)
-
-    if opts[:warnings_as_errors] do
-      Code.put_compiler_option(:warnings_as_errors, true)
-    end
+    # --warnings-as-errors in test does not pass down to compile,
+    # if you need this, call compile explicitly before.
+    Mix.Task.run("compile", args -- ["--warnings-as-errors"])
 
     project = Mix.Project.config()
+    {partitions, opts} = Keyword.pop(opts, :partitions)
+    partitioned? = is_integer(partitions) and partitions > 1
 
     # Start cover after we load deps but before we start the app.
     cover =
       if opts[:cover] do
         compile_path = Mix.Project.compile_path(project)
-        partition = opts[:partitions] && System.get_env("MIX_TEST_PARTITION")
+        partition = partitioned? && System.get_env("MIX_TEST_PARTITION")
 
         cover =
           @cover
@@ -501,6 +527,7 @@ defmodule Mix.Tasks.Test do
     ExUnit.configure(merge_helper_opts(ex_unit_opts))
 
     # Finally parse, require and load the files
+    test_elixirc_options = project[:test_elixirc_options] || []
     test_files = parse_files(files, shell, test_paths)
     test_pattern = project[:test_pattern] || "*_test.exs"
     warn_test_pattern = project[:warn_test_pattern] || "*_test.ex"
@@ -509,49 +536,47 @@ defmodule Mix.Tasks.Test do
       test_files
       |> Mix.Utils.extract_files(test_pattern)
       |> filter_to_allowed_files(allowed_files)
-      |> filter_by_partition(shell, opts)
+      |> filter_by_partition(shell, partitions)
 
     display_warn_test_pattern(test_files, test_pattern, matched_test_files, warn_test_pattern)
 
-    result =
-      case CT.require_and_run(matched_test_files, test_paths, opts) do
-        {:ok, %{excluded: excluded, failures: failures, total: total}} ->
-          Mix.shell(shell)
-          cover && cover.()
+    case CT.require_and_run(matched_test_files, test_paths, test_elixirc_options, opts) do
+      {:ok, %{excluded: excluded, failures: failures, total: total}} ->
+        Mix.shell(shell)
+        cover && cover.()
 
-          cond do
-            failures > 0 and opts[:raise] ->
-              raise_with_shell(shell, "\"mix test\" failed")
+        cond do
+          failures > 0 and opts[:raise] ->
+            raise_with_shell(shell, "\"mix test\" failed")
 
-            failures > 0 ->
-              System.at_exit(fn _ -> exit({:shutdown, 1}) end)
+          failures > 0 ->
+            System.at_exit(fn _ ->
+              exit({:shutdown, Keyword.fetch!(ex_unit_opts, :exit_status)})
+            end)
 
-            excluded == total and Keyword.has_key?(opts, :only) ->
-              message = "The --only option was given to \"mix test\" but no test was executed"
-              raise_or_error_at_exit(shell, message, opts)
+          excluded == total and Keyword.has_key?(opts, :only) ->
+            message = "The --only option was given to \"mix test\" but no test was executed"
+            raise_or_error_at_exit(shell, message, opts)
 
-            true ->
-              :ok
-          end
+          true ->
+            :ok
+        end
 
-        :noop ->
-          cond do
-            opts[:stale] ->
-              Mix.shell().info("No stale tests")
+      :noop ->
+        cond do
+          opts[:stale] ->
+            Mix.shell().info("No stale tests")
 
-            files == [] ->
-              Mix.shell().info("There are no tests to run")
+          files == [] ->
+            Mix.shell().info("There are no tests to run")
 
-            true ->
-              message = "Paths given to \"mix test\" did not match any directory/file: "
-              raise_or_error_at_exit(shell, message <> Enum.join(files, ", "), opts)
-          end
+          true ->
+            message = "Paths given to \"mix test\" did not match any directory/file: "
+            raise_or_error_at_exit(shell, message <> Enum.join(files, ", "), opts)
+        end
 
-          :ok
-      end
-
-    Code.put_compiler_option(:warnings_as_errors, old_warnings_as_errors)
-    result
+        :ok
+    end
   end
 
   defp raise_with_shell(shell, message) do
@@ -596,7 +621,8 @@ defmodule Mix.Tasks.Test do
     :slowest,
     :failures_manifest_file,
     :only_test_ids,
-    :test_location_relative_path
+    :test_location_relative_path,
+    :exit_status
   ]
 
   @doc false
@@ -610,6 +636,7 @@ defmodule Mix.Tasks.Test do
       |> filter_opts(:only)
       |> formatter_opts()
       |> color_opts()
+      |> exit_status_opts()
       |> Keyword.take(@option_keys)
       |> default_opts()
 
@@ -713,34 +740,40 @@ defmodule Mix.Tasks.Test do
     Enum.filter(matched_test_files, &MapSet.member?(allowed_files, Path.expand(&1)))
   end
 
-  defp filter_by_partition(files, shell, opts) do
-    if total = opts[:partitions] do
-      partition = System.get_env("MIX_TEST_PARTITION")
+  defp filter_by_partition(files, _shell, total) when total in [nil, 1],
+    do: files
 
-      case partition && Integer.parse(partition) do
-        {partition, ""} when partition in 1..total ->
-          partition = partition - 1
+  defp filter_by_partition(files, shell, total) when total > 1 do
+    partition = System.get_env("MIX_TEST_PARTITION")
 
-          # We sort the files because Path.wildcard does not guarantee
-          # ordering, so different OSes could return a different order,
-          # meaning run across OSes on different partitions could run
-          # duplicate files.
-          for {file, index} <- Enum.with_index(Enum.sort(files)),
-              rem(index, total) == partition,
-              do: file
+    case partition && Integer.parse(partition) do
+      {partition, ""} when partition in 1..total ->
+        partition = partition - 1
 
-        _ ->
-          raise_with_shell(
-            shell,
-            "The MIX_TEST_PARTITION environment variable must be set to an integer between " <>
-              "1..#{total} when the --partitions option is set, got: #{inspect(partition)}"
-          )
-      end
-    else
-      files
+        # We sort the files because Path.wildcard does not guarantee
+        # ordering, so different OSes could return a different order,
+        # meaning run across OSes on different partitions could run
+        # duplicate files.
+        for {file, index} <- Enum.with_index(Enum.sort(files)),
+            rem(index, total) == partition,
+            do: file
+
+      _ ->
+        raise_with_shell(
+          shell,
+          "The MIX_TEST_PARTITION environment variable must be set to an integer between " <>
+            "1..#{total} when the --partitions option is set, got: #{inspect(partition)}"
+        )
     end
   end
 
+  defp filter_by_partition(_files, shell, total) do
+    raise_with_shell(
+      shell,
+      "--partitions : expected to be positive integer, got #{total}"
+    )
+  end
+
   defp color_opts(opts) do
     case Keyword.fetch(opts, :color) do
       {:ok, enabled?} ->
@@ -751,6 +784,10 @@ defmodule Mix.Tasks.Test do
     end
   end
 
+  defp exit_status_opts(opts) do
+    Keyword.put_new(opts, :exit_status, 2)
+  end
+
   defp require_test_helper(shell, dir) do
     file = Path.join(dir, "test_helper.exs")
 
