@@ -1,15 +1,6 @@
 defmodule Mix.Tasks.WillRecompile do
   use Mix.Task
-
   @moduledoc false
-  @manifest "compile.lock"
-
-  @doc """
-  Returns the will_recompile manifest for the project.
-  """
-  def manifest(path \\ Mix.Project.manifest_path()) do
-    Path.join(path, @manifest)
-  end
 
   @doc """
   Annotates the current project will recompile.
@@ -33,10 +24,12 @@ defmodule Mix.Tasks.WillRecompile do
         [Mix.Project.app_path(config)]
       end
 
+    filename = Mix.ProjectStack.reset_config_mtime()
+
     for path <- paths do
       path = Path.join(path, ".mix")
       File.mkdir_p!(path)
-      File.touch!(manifest(path))
+      File.touch!(Path.join(path, filename))
     end
 
     :ok
