@@ -11,50 +11,104 @@ defmodule Mix.Tasks.Xref do
   @moduledoc """
   Prints cross reference information between modules.
 
-  This task is automatically reenabled, so you can print information
-  multiple times in the same Mix invocation.
-
-  ## Xref modes
-
   The `xref` task expects a mode as first argument:
 
-      mix xref MODE
+      $ mix xref MODE
 
   All available modes are discussed below.
 
-  ### callers CALLEE
+  This task is automatically re-enabled, so you can print
+  information multiple times in the same Mix invocation.
+
+  ## mix xref callers MODULE
 
   Prints all callers of the given `MODULE`. Example:
 
-      mix xref callers MyMod
+      $ mix xref callers MyMod
 
-  ### graph
+  ## mix xref trace FILE
+
+  Compiles the given file listing all dependencies within the same app.
+  It includes the type and line for each one. Example:
+
+      $ mix xref trace lib/my_app/router.ex
+
+  The `--label` option may be given to keep only certain traces
+  (compile, runtime or export):
+
+      $ mix xref trace lib/my_app/router.ex --label compile
+
+  If you have an umbrella application, we also recommend using the
+  `--include-siblings` flag to see the dependencies on other
+  umbrella applications.
+
+  ### Example
+
+  Imagine the given file lib/b.ex:
+
+      defmodule B do
+        import A
+        A.macro()
+        macro()
+        A.fun()
+        fun()
+        def calls_macro, do: A.macro()
+        def calls_fun, do: A.fun()
+        def calls_struct, do: %A{}
+      end
+
+  `mix xref trace` will print:
+
+      lib/b.ex:2: require A (export)
+      lib/b.ex:3: call A.macro/0 (compile)
+      lib/b.ex:4: import A.macro/0 (compile)
+      lib/b.ex:5: call A.fun/0 (compile)
+      lib/b.ex:6: call A.fun/0 (compile)
+      lib/b.ex:6: import A.fun/0 (compile)
+      lib/b.ex:7: call A.macro/0 (compile)
+      lib/b.ex:8: call A.fun/0 (runtime)
+      lib/b.ex:9: struct A (export)
+
+  ## mix xref graph
 
   Prints a file dependency graph where an edge from `A` to `B` indicates
   that `A` (source) depends on `B` (sink).
 
-      mix xref graph --format stats
+      $ mix xref graph --format stats
 
   The following options are accepted:
 
-    * `--exclude` - paths to exclude
+    * `--exclude` - path to exclude. Can be repeated to exclude multiple paths.
 
     * `--label` - only shows relationships with the given label.
-      By default, it keeps all labels that are transitive.
-      The labels are "compile", "export" and "runtime". See
-      "Dependencies types" section below
+      The labels are "compile", "export" and "runtime". By default,
+      the `--label` option simply filters the printed graph to show
+      only relationships with the given label. You can pass `--only-direct`
+      to trim the graph to only the nodes that have the direct
+      relationship given by label. There is also a special label
+      called "compile-connected" that keeps only compile-time files
+      with at least one transitive dependency. See "Dependencies types"
+      section below.
+
+    * `--group` - provide comma-separated paths to consider as a group. Dependencies
+      from and into multiple files of the group are considered a single dependency.
+      Dependencies between the group elements are ignored. This is useful when you
+      are computing compile and compile-connected dependencies and you want a
+      series of files to be treated as one. The group is printed using the first path,
+      with a `+` suffix. Can be repeated to create multiple groups.
+
+    * `--only-direct` - keeps only files with the direct relationship
+      given by `--label`
 
     * `--only-nodes` - only shows the node names (no edges).
       Generally useful with the `--sink` flag
 
-    * `--only-direct` - the `--label` option will restrict itself
-      to only direct dependencies instead of transitive ones
-
     * `--source` - displays all files that the given source file
-      references (directly or indirectly)
+      references (directly or indirectly). Can be repeated to display
+      references from multiple sources.
 
     * `--sink` - displays all files that reference the given file
-      (directly or indirectly)
+      (directly or indirectly). Can be repeated.
 
     * `--min-cycle-size` - controls the minimum cycle size on formats
       like `stats` and `cycles`
@@ -80,8 +134,8 @@ defmodule Mix.Tasks.Xref do
   those options with `--label` and `--only-nodes` to get all files that exhibit a certain
   property, for example:
 
-      # To get all files and their direct compile time dependencies
-      mix xref graph --label compile --only-direct
+      # To show all compile-time relationships
+      mix xref graph --label compile
 
       # To get the tree that depend on lib/foo.ex at compile time
       mix xref graph --label compile --sink lib/foo.ex
@@ -95,51 +149,81 @@ defmodule Mix.Tasks.Xref do
       # To show general statistics about the graph
       mix xref graph --format stats
 
-      # To limit statistics only to certain labels
-      mix xref graph --format stats --label compile
-
-  #### Understanding the printed graph
+  ### Understanding the printed graph
 
   When `mix xref graph` runs, it will print a tree of the following
-  format:
+  format. Imagine the following code:
 
+      # lib/a.ex
+      defmodule A do
+        IO.puts B.hello()
+      end
+
+      # lib/b.ex
+      defmodule B do
+        def hello, do: C.world()
+      end
+
+      # lib/c.ex
+      defmodule C do
+        def world, do: "hello world"
+      end
+
+  It will print:
+
+      $ mix xref graph
       lib/a.ex
-      `-- lib/b.ex (compile)
-          `-- lib/c.ex
+      └── lib/b.ex (compile)
+      lib/b.ex
+      └── lib/c.ex
+      lib/c.ex
 
   This tree means that `lib/a.ex` depends on `lib/b.ex` at compile
-  time which then depends on `lib/c.ex` at runtime. This is often
+  time. And `lib/b.ex` depends on `lib/c.ex` at runtime. This is often
   problematic because if `lib/c.ex` changes, `lib/a.ex` also has to
-  recompile due to this indirect compile time dependency.
-
-  The flags `--source` or `--sink` does not change how you read the
-  graph. For example, if we use the `--sink lib/c.ex` flag, we would
-  see the same tree:
+  recompile due to this indirect compile time dependency. When you pass
+  `--label compile`, the graph shows only the compile-time dependencies:
 
+      $ mix xref graph --label compile
       lib/a.ex
-      `-- lib/b.ex (compile)
-          `-- lib/c.ex
+      └── lib/b.ex (compile)
 
-  If the `--label compile` flag is given with `--sink`, then `lib/c.ex`
-  won't be shown, because no module has a compile time dependency on
-  `lib/c.ex` but `lib/a.ex` still has an indirect compile time dependency
-  on `lib/c.ex` via `lib/b.ex`:
+  The `--label compile` flag removes all non-compile dependencies. However,
+  this can be misleading because having direct compile time dependencies is
+  not necessarily an issue. The biggest concern, as mentioned above, are the
+  transitive compile time dependencies. You can get all compile time
+  dependencies that cause transitive compile time dependencies by using
+  `--label compile-connected`:
 
+      $ mix xref graph --label compile-connected
       lib/a.ex
-      `-- lib/b.ex (compile)
+      └── lib/b.ex (compile)
 
-  Therefore, using a combination of `--sink` with `--label` is useful to
-  find all files that will change once the sink changes, alongside the
-  transitive dependencies that will cause said recompilations.
+  The above says `lib/a.ex` depends on `lib/b.ex` and that causes transitive
+  compile time dependencies - as we know, `lib/a.ex` also depends on `lib/c.ex`.
+  We can retrieve those transitive dependencies by passing `lib/b.ex` as
+  `--source` to `mix xref graph`:
 
-  #### Dependencies types
+      $ mix xref graph --source lib/b.ex
+      lib/b.ex
+      └── lib/c.ex
+
+  Similarly, you can use the `--label compile` and the `--sink` flag to find
+  all compile time dependencies that will recompile once the sink changes:
+
+      $ mix xref graph --label compile --sink lib/c.ex
+      lib/a.ex
+      └── lib/b.ex (compile)
+
+  ### Dependencies types
 
   Elixir tracks three types of dependencies between modules: compile,
   exports, and runtime. If a module has a compile time dependency on
   another module, the caller module has to be recompiled whenever the
   callee changes. Compile-time dependencies are typically added when
   using macros or when invoking functions in the module body (outside
-  of functions).
+  of functions). You can list all dependencies in a file by running
+  `mix xref trace path/to/file.ex`.
 
   Exports dependencies are compile time dependencies on the module API,
   namely structs and its public definitions. For example, if you import
@@ -152,12 +236,16 @@ defmodule Mix.Tasks.Xref do
   Runtime dependencies are added whenever you invoke another module
   inside a function. Modules with runtime dependencies do not have
   to be compiled when the callee changes, unless there is a transitive
-  compile or export time dependency between them.
+  compile or an outdated export time dependency between them. The option
+  `--label compile-connected` can be used to find the first case.
 
   ## Shared options
 
   Those options are shared across all modes:
 
+    * `--fail-above` - generates a failure if the relevant metric is above the
+      given threshold. Applies to all modes except `mix xref graph --format stats`.
+
     * `--include-siblings` - includes dependencies that have `:in_umbrella` set
       to true in the current project in the reports. This can be used to find
       callers or to analyze graphs between projects
@@ -173,19 +261,20 @@ defmodule Mix.Tasks.Xref do
   """
 
   @switches [
-    abort_if_any: :boolean,
     archives_check: :boolean,
     compile: :boolean,
     deps_check: :boolean,
     elixir_version_check: :boolean,
     exclude: :keep,
+    fail_above: :integer,
     format: :string,
+    group: :keep,
     include_siblings: :boolean,
     label: :string,
     only_nodes: :boolean,
     only_direct: :boolean,
-    sink: :string,
-    source: :string,
+    sink: :keep,
+    source: :keep,
     min_cycle_size: :integer
   ]
 
@@ -197,11 +286,14 @@ defmodule Mix.Tasks.Xref do
     {opts, args} = OptionParser.parse!(args, strict: @switches)
 
     case args do
-      ["callers", callee] ->
-        callers(callee, opts)
+      ["callers", module] ->
+        handle_callers(module, opts)
+
+      ["trace", file] ->
+        handle_trace(file, opts)
 
       ["graph"] ->
-        graph(opts)
+        handle_graph(opts)
 
       # TODO: Remove on v2.0
       ["deprecated"] ->
@@ -235,8 +327,7 @@ defmodule Mix.Tasks.Xref do
   information from. To get the function calls of each child in an umbrella,
   execute the function at the root of each individual application.
   """
-  # TODO: Remove on v2.0
-  @doc deprecated: "Use compilation tracers described in the Code module"
+  @deprecated "Use compilation tracers described in the Code module"
   @spec calls(keyword()) :: [
           %{
             callee: {module(), atom(), arity()},
@@ -348,8 +439,8 @@ defmodule Mix.Tasks.Xref do
 
   ## Modes
 
-  defp callers(callee, opts) do
-    module = parse_callee(callee)
+  defp handle_callers(module, opts) do
+    module = parse_module(module)
 
     file_callers =
       for source <- sources(opts),
@@ -360,25 +451,57 @@ defmodule Mix.Tasks.Xref do
       Mix.shell().info([file, " (", type, ")"])
     end
 
-    :ok
+    check_failure(:references, length(file_callers), opts[:fail_above])
   end
 
-  defp graph(opts) do
-    filter = label_filter(opts[:label])
+  defp handle_trace(file, opts) do
+    set =
+      for app <- apps(opts),
+          modules = Application.spec(app, :modules),
+          module <- modules,
+          into: MapSet.new(),
+          do: module
+
+    new = [ignore_already_consolidated: true, ignore_module_conflict: true, tracers: [__MODULE__]]
+    old = Code.compiler_options(new)
+    ets = :ets.new(__MODULE__, [:named_table, :duplicate_bag, :public])
+    :ets.insert(ets, [{:config, set, trace_label(opts[:label])}])
+
+    try do
+      Code.compile_file(file)
+    else
+      _ ->
+        :ets.delete(ets, :modules)
+
+        traces =
+          try do
+            print_traces(Enum.sort(:ets.lookup_element(__MODULE__, :entry, 2)))
+          rescue
+            _ -> []
+          end
+
+        check_failure(:traces, length(traces), opts[:fail_above])
+    after
+      :ets.delete(ets)
+      Code.compiler_options(old)
+    end
+  end
+
+  defp handle_graph(opts) do
+    label = label_filter(opts[:label])
 
     {direct_filter, transitive_filter} =
-      if opts[:only_direct], do: {filter, :all}, else: {:all, filter}
+      if opts[:only_direct], do: {label, :all}, else: {:all, label}
 
     write_graph(file_references(direct_filter, opts), transitive_filter, opts)
-    :ok
   end
 
   ## Callers
 
-  defp parse_callee(callee) do
-    case Mix.Utils.parse_mfa(callee) do
+  defp parse_module(module) do
+    case Mix.Utils.parse_mfa(module) do
       {:ok, [module]} -> module
-      _ -> Mix.raise("xref callers MODULE expects a MODULE, got: " <> callee)
+      _ -> Mix.raise("xref callers MODULE expects a MODULE, got: " <> module)
     end
   end
 
@@ -391,19 +514,178 @@ defmodule Mix.Tasks.Xref do
     end
   end
 
+  ## Trace
+
+  @doc false
+  def trace({:require, meta, module, _opts}, env),
+    do: add_trace(require_mode(meta), :require, module, module, meta, env)
+
+  def trace({:struct_expansion, meta, module, _keys}, env),
+    do: add_trace(:export, :struct, module, module, meta, env)
+
+  def trace({:alias_reference, meta, module}, env) when env.module != module,
+    do: add_trace(mode(env), :alias, module, module, meta, env)
+
+  def trace({:remote_function, meta, module, function, arity}, env),
+    do: add_trace(mode(env), :call, module, {module, function, arity}, meta, env)
+
+  def trace({:remote_macro, meta, module, function, arity}, env),
+    do: add_trace(:compile, :call, module, {module, function, arity}, meta, env)
+
+  def trace({:imported_function, meta, module, function, arity}, env),
+    do: add_trace(mode(env), :import, module, {module, function, arity}, meta, env)
+
+  def trace({:imported_macro, meta, module, function, arity}, env),
+    do: add_trace(:compile, :import, module, {module, function, arity}, meta, env)
+
+  def trace(_event, _env),
+    do: :ok
+
+  defp require_mode(meta), do: if(meta[:from_macro], do: :compile, else: :export)
+
+  defp mode(%{function: nil}), do: :compile
+  defp mode(_), do: :runtime
+
+  defp add_trace(mode, type, module, module_or_mfa, meta, env) do
+    [{:config, modules, label}] = :ets.lookup(__MODULE__, :config)
+
+    if module in modules and (label == nil or mode == label) do
+      line = meta[:line] || env.line
+      :ets.insert(__MODULE__, {:entry, {env.file, line, module_or_mfa, mode, type}})
+    end
+
+    :ok
+  end
+
+  defp print_traces(entries) do
+    # We don't want to show aliases if there is an entry of the same type
+    non_aliases =
+      for {_file, _line, module_or_mfa, mode, type} <- entries,
+          type != :alias,
+          into: %{},
+          do: {{trace_module(module_or_mfa), mode}, []}
+
+    shell = Mix.shell()
+
+    for {file, line, module_or_mfa, mode, type} <- entries,
+        type != :alias or not Map.has_key?(non_aliases, {module_or_mfa, mode}) do
+      shell.info([
+        Exception.format_file_line(Path.relative_to_cwd(file), line),
+        ?\s,
+        Atom.to_string(type),
+        ?\s,
+        format_module_or_mfa(module_or_mfa),
+        " (#{mode})"
+      ])
+
+      :ok
+    end
+  end
+
+  defp trace_label(nil), do: nil
+  defp trace_label("compile"), do: :compile
+  defp trace_label("export"), do: :export
+  defp trace_label("runtime"), do: :runtime
+  defp trace_label(other), do: Mix.raise("Unknown --label #{other} in mix xref trace")
+
+  defp trace_module({m, _, _}), do: m
+  defp trace_module(m), do: m
+
+  defp format_module_or_mfa({m, f, a}), do: Exception.format_mfa(m, f, a)
+  defp format_module_or_mfa(m), do: inspect(m)
+
   ## Graph
 
-  defp excluded(opts) do
-    opts
-    |> Keyword.get_values(:exclude)
-    |> Enum.flat_map(&[{&1, nil}, {&1, :compile}, {&1, :export}])
+  defp merge_groups(file_references, comma_separated_groups) do
+    for group_paths <- comma_separated_groups,
+        reduce: {file_references, %{}} do
+      {file_references, aliases} ->
+        group_paths
+        |> String.split(",")
+        |> check_files(file_references, :group)
+        |> group(file_references, aliases)
+    end
+  end
+
+  @type_order %{
+    compile: 0,
+    export: 1,
+    nil: 2
+  }
+
+  # Group the given paths.
+  # In graph theory vocabulary, this is done by vertex identification
+  # and removal of edges between contracting vertices.
+  defp group(paths, file_references, aliases) do
+    group_name = hd(paths) <> "+"
+    aliases = paths |> Map.new(&{&1, group_name}) |> Map.merge(aliases)
+
+    # Merge the references *from* the paths to group
+    {from_group, file_references} = Map.split(file_references, paths)
+
+    file_references =
+      Map.put(file_references, group_name, merge_references_from_group(from_group))
+
+    # Remap the references *to* the merged group
+    file_references =
+      Map.new(file_references, fn {file, references} ->
+        {file, remap_references_to_group(references, aliases, group_name)}
+      end)
+
+    # Remove the resulting reference from the merged group to itself, if there is one
+    file_references = Map.update!(file_references, group_name, &List.keydelete(&1, group_name, 0))
+
+    {file_references, aliases}
+  end
+
+  # Calculate the references from the merged group by concatenating all the references
+  # from its components; in case of duplicates keep the one with the most important type.
+  defp merge_references_from_group(file_references_to_merge) do
+    file_references_to_merge
+    |> Map.values()
+    |> Enum.concat()
+    |> Enum.sort_by(fn {_ref, type} -> @type_order[type] end)
+    |> Enum.uniq_by(fn {ref, _type} -> ref end)
+    |> Enum.sort()
+  end
+
+  defp remap_references_to_group(references, aliases, group_name) do
+    case Enum.split_with(references, fn {ref, _type} -> Map.has_key?(aliases, ref) end) do
+      {[], _all_references} ->
+        references
+
+      {refs_to_merge, other_refs} ->
+        type =
+          refs_to_merge
+          |> Enum.map(fn {_ref, type} -> type end)
+          |> Enum.min_by(&@type_order[&1])
+
+        Enum.sort([{group_name, type} | other_refs])
+    end
+  end
+
+  defp exclude(file_references, nil), do: file_references
+
+  defp exclude(file_references, excluded) do
+    excluded_set = MapSet.new(excluded)
+
+    file_references
+    |> Map.drop(excluded)
+    |> Map.new(fn {key, list} ->
+      {key, Enum.reject(list, fn {ref, _kind} -> MapSet.member?(excluded_set, ref) end)}
+    end)
   end
 
   defp label_filter(nil), do: :all
   defp label_filter("compile"), do: :compile
   defp label_filter("export"), do: :export
   defp label_filter("runtime"), do: nil
-  defp label_filter(other), do: Mix.raise("unknown --label #{other}")
+  defp label_filter("compile-connected"), do: :compile_connected
+  defp label_filter(other), do: Mix.raise("Unknown --label #{other} in mix xref graph")
+
+  defp file_references(:compile_connected, _opts) do
+    Mix.raise("Cannot use --only-direct with --label=compile-connected")
+  end
 
   defp file_references(filter, opts) do
     module_sources =
@@ -456,130 +738,174 @@ defmodule Mix.Tasks.Xref do
         into: %{}
   end
 
+  @humanize_option %{
+    group: "Group files",
+    source: "Sources",
+    sink: "Sinks",
+    exclude: "Excluded files"
+  }
+
+  defp get_files(what, opts, file_references, aliases) do
+    files =
+      for file <- Keyword.get_values(opts, what) do
+        Map.get(aliases, file, file)
+      end
+
+    check_files(files, file_references, what)
+  end
+
+  defp check_files(files, file_references, what) do
+    case files -- Map.keys(file_references) do
+      [_ | _] = missing ->
+        Mix.raise("#{@humanize_option[what]} could not be found: #{Enum.join(missing, ", ")}")
+
+      _ ->
+        :ok
+    end
+
+    if files == [], do: nil, else: files
+  end
+
   defp write_graph(file_references, filter, opts) do
-    excluded = excluded(opts)
-    source = opts[:source]
-    sink = opts[:sink]
-
-    if source && is_nil(file_references[source]) do
-      Mix.raise("Source could not be found: #{source}")
-    end
-
-    if sink && is_nil(file_references[sink]) do
-      Mix.raise("Sink could not be found: #{sink}")
-    end
+    {file_references, aliases} = merge_groups(file_references, Keyword.get_values(opts, :group))
 
     file_references =
-      if sink = opts[:sink] do
-        filter_for_sink(file_references, sink, filter)
+      exclude(file_references, get_files(:exclude, opts, file_references, aliases))
+
+    sources = get_files(:source, opts, file_references, aliases)
+    sinks = get_files(:sink, opts, file_references, aliases)
+
+    file_references =
+      cond do
+        sinks -> sink_tree(file_references, sinks)
+        sources -> source_tree(file_references, sources)
+        true -> file_references
+      end
+
+    # Filter according to non direct label
+    file_references = filter(file_references, filter)
+
+    # If a label is given, remove empty root nodes
+    file_references =
+      if opts[:label] do
+        for {_, [_ | _]} = pair <- file_references, into: %{}, do: pair
       else
-        filter_for_source(file_references, filter)
+        file_references
       end
 
     roots =
-      if source = opts[:source] do
-        %{source => nil}
+      if sources do
+        Enum.map(sources, &{&1, nil})
       else
         file_references
-        |> Map.delete(opts[:sink])
+        |> Map.drop(sinks || [])
         |> Enum.map(&{elem(&1, 0), nil})
-        |> Kernel.--(excluded)
-        |> Map.new()
       end
 
     callback = fn {file, type} ->
       children = if opts[:only_nodes], do: [], else: Map.get(file_references, file, [])
       type = type && "(#{type})"
-      {{file, type}, Enum.sort(children -- excluded)}
+      {{file, type}, Enum.sort(children)}
     end
 
-    case opts[:format] do
-      "dot" ->
-        Mix.Utils.write_dot_graph!(
-          "xref_graph.dot",
-          "xref graph",
-          Enum.sort(roots),
-          callback,
-          opts
-        )
+    {found, count} =
+      case opts[:format] do
+        "dot" ->
+          Mix.Utils.write_dot_graph!(
+            "xref_graph.dot",
+            "xref graph",
+            Enum.sort(roots),
+            callback,
+            opts
+          )
 
-        """
-        Generated "xref_graph.dot" in the current directory. To generate a PNG:
+          """
+          Generated "xref_graph.dot" in the current directory. To generate a PNG:
 
-           dot -Tpng xref_graph.dot -o xref_graph.png
+             dot -Tpng xref_graph.dot -o xref_graph.png
 
-        For more options see http://www.graphviz.org/.
-        """
-        |> String.trim_trailing()
-        |> Mix.shell().info()
+          For more options see http://www.graphviz.org/.
+          """
+          |> String.trim_trailing()
+          |> Mix.shell().info()
 
-      "stats" ->
-        print_stats(file_references, opts)
+          {:references, count_references(file_references)}
 
-      "cycles" ->
-        print_cycles(file_references, opts)
+        "stats" ->
+          print_stats(file_references, opts)
+          {:stats, 0}
 
-      _ ->
-        Mix.Utils.print_tree(Enum.sort(roots), callback, opts)
-    end
+        "cycles" ->
+          {:cycles, print_cycles(file_references, opts)}
+
+        other when other in [nil, "plain", "pretty"] ->
+          Mix.Utils.print_tree(Enum.sort(roots), callback, opts)
+
+          {:references, count_references(file_references)}
+
+        other ->
+          Mix.raise("Unknown --format #{other} in mix xref graph")
+      end
+
+    check_failure(found, count, opts[:fail_above])
   end
 
-  defp filter_for_source(file_references, :all), do: file_references
+  defp count_references(file_references) do
+    Enum.reduce(file_references, 0, fn {_, refs}, total -> total + length(refs) end)
+  end
 
-  defp filter_for_source(file_references, filter) do
-    Enum.reduce(file_references, %{}, fn {key, _}, acc ->
-      {children, _} = filter_for_source(file_references, key, %{}, %{}, filter)
-      Map.put(acc, key, children |> Map.delete(key) |> Map.to_list())
+  defp filter_fn(file_references, :compile_connected),
+    do: fn {key, type} ->
+      type == :compile and match?([_ | _], file_references[key] || [])
+    end
+
+  defp filter_fn(_file_references, filter),
+    do: fn {_key, type} -> type == filter end
+
+  defp filter(file_references, :all), do: file_references
+
+  defp filter(file_references, filter) do
+    filter_fn = filter_fn(file_references, filter)
+
+    for {key, children} <- file_references,
+        into: %{},
+        do: {key, Enum.filter(children, filter_fn)}
+  end
+
+  defp source_tree(file_references, keys) do
+    keys
+    |> Enum.reduce({%{}, %{}}, fn key, {acc, seen} ->
+      source_tree(file_references, key, acc, seen)
     end)
+    |> elem(0)
   end
 
-  defp filter_for_source(references, key, acc, seen, filter) do
-    nodes = references[key]
+  defp source_tree(file_references, key, acc, seen) do
+    nodes = file_references[key]
 
-    if is_nil(nodes) || seen[key] do
+    if is_nil(nodes) or Map.has_key?(seen, key) do
       {acc, seen}
     else
+      acc = Map.put(acc, key, nodes)
       seen = Map.put(seen, key, true)
 
-      Enum.reduce(nodes, {acc, seen}, fn {child_key, type}, {acc, seen} ->
-        if type == filter do
-          {Map.put(acc, child_key, type), Map.put(seen, child_key, true)}
-        else
-          filter_for_source(references, child_key, acc, seen, filter)
-        end
+      Enum.reduce(nodes, {acc, seen}, fn {key, _type}, {acc, seen} ->
+        source_tree(file_references, key, acc, seen)
       end)
     end
   end
 
-  defp filter_for_sink(file_references, sink, filter) do
-    fun = if filter == :all, do: fn _ -> true end, else: fn type -> type == filter end
-
+  defp sink_tree(file_references, keys) do
     file_references
-    |> invert_references(fn _ -> true end)
-    |> depends_on_sink([{sink, nil}], %{})
-    |> invert_references(fun)
+    |> invert_references()
+    |> source_tree(keys)
+    |> invert_references()
   end
 
-  defp depends_on_sink(file_references, new_nodes, acc) do
-    Enum.reduce(new_nodes, acc, fn {new_node_name, _type}, acc ->
-      new_nodes = file_references[new_node_name]
-
-      if acc[new_node_name] || !new_nodes do
-        acc
-      else
-        depends_on_sink(file_references, new_nodes, Map.put(acc, new_node_name, new_nodes))
-      end
-    end)
-  end
-
-  defp invert_references(file_references, fun) do
+  defp invert_references(file_references) do
     Enum.reduce(file_references, %{}, fn {file, references}, acc ->
-      Enum.reduce(references, acc, fn {reference, type}, acc ->
-        if fun.(type) do
-          Map.update(acc, reference, [{file, type}], &[{file, type} | &1])
-        else
-          acc
-        end
+      Enum.reduce(references, acc, fn {file_reference, type}, acc ->
+        Map.update(acc, file_reference, [{file, type}], &[{file, type} | &1])
       end)
     end)
   end
@@ -663,6 +989,7 @@ defmodule Mix.Tasks.Xref do
       case graph |> cycles(opts) |> Enum.sort(:desc) do
         [] ->
           shell.info("No cycles found")
+          0
 
         cycles ->
           shell.info("#{length(cycles)} cycles found. Showing them in decreasing size:\n")
@@ -676,6 +1003,8 @@ defmodule Mix.Tasks.Xref do
 
             shell.info("")
           end
+
+          length(cycles)
       end
     end)
   end
@@ -688,6 +1017,19 @@ defmodule Mix.Tasks.Xref do
         do: source
   end
 
+  defp apps(opts) do
+    siblings =
+      if opts[:include_siblings] do
+        for %{scm: Mix.SCM.Path, opts: opts, app: app} <- Mix.Dep.cached(),
+            opts[:in_umbrella],
+            do: app
+      else
+        []
+      end
+
+    [Mix.Project.config()[:app] | siblings]
+  end
+
   defp manifests(opts) do
     siblings =
       if opts[:include_siblings] do
@@ -700,4 +1042,13 @@ defmodule Mix.Tasks.Xref do
 
     [Path.join(Mix.Project.manifest_path(), @manifest) | siblings]
   end
+
+  defp check_failure(found, count, max_count)
+       when not is_nil(max_count) and count > max_count do
+    Mix.raise("Too many #{found} (found: #{count}, permitted: #{max_count})")
+  end
+
+  defp check_failure(_, _, _) do
+    :ok
+  end
 end
