@@ -435,10 +435,28 @@ defmodule Mix.Utils do
     |> Enum.map_join(".", &Macro.camelize/1)
   end
 
+  @doc """
+  Symlinks or copy with the option to force a hard copy.
+  See `symlink_or_copy/2`.
+  """
+  def symlink_or_copy(hard_copy?, source, target) do
+    if hard_copy? do
+      if File.exists?(source) do
+        File.rm_rf!(target)
+        File.cp_r!(source, target)
+      end
+    else
+      symlink_or_copy(source, target)
+    end
+  end
+
   @doc """
   Symlinks directory `source` to `target` or copies it recursively
   in case symlink fails.
 
+  In case of conflicts, it copies files only if they have been
+  recently touched.
+
   Expects source and target to be absolute paths as it generates
   a relative symlink.
   """
@@ -450,9 +468,8 @@ defmodule Mix.Utils do
           {:win32, _} -> source
           _ -> make_relative_path(source, target)
         end
-        |> String.to_charlist()
 
-      case :file.read_link(target) do
+      case File.read_link(target) do
         {:ok, ^link} ->
           :ok
 
@@ -495,17 +512,21 @@ defmodule Mix.Utils do
   end
 
   defp do_symlink_or_copy(source, target, link) do
-    case :file.make_symlink(link, target) do
+    case File.ln_s(link, target) do
       :ok ->
         :ok
 
       {:error, _} ->
-        file =
-          File.cp_r!(source, target, fn orig, dest ->
-            File.stat!(orig).mtime > File.stat!(dest).mtime
-          end)
+        files =
+          File.cp_r!(source, target,
+            on_conflict: fn orig, dest ->
+              {orig_mtime, orig_size} = last_modified_and_size(orig)
+              {dest_mtime, dest_size} = last_modified_and_size(dest)
+              orig_mtime > dest_mtime or orig_size != dest_size
+            end
+          )
 
-        {:ok, file}
+        {:ok, files}
     end
   end
 
@@ -648,7 +669,8 @@ defmodule Mix.Utils do
     try do
       case httpc_request(request, http_options) do
         {:error, {:failed_connect, [{:to_address, _}, {inet, _, reason}]}}
-        when inet in [:inet, :inet6] and reason in [:ehostunreach, :enetunreach] ->
+        when inet in [:inet, :inet6] and
+               reason in [:ehostunreach, :enetunreach, :eprotonosupport, :nxdomain] ->
           :httpc.set_options([ipfamily: fallback(inet)], :mix)
           request |> httpc_request(http_options) |> httpc_response()
 
