@@ -8,7 +8,7 @@ defmodule Bar.MixProject do
       # Choose something besides *_test.exs so that these test files don't
       # get accidentally swept up into the actual Mix test suite.
       test_pattern: "*_tests.exs",
-      test_coverage: [ignore_modules: [Bar]]
+      test_coverage: [ignore_modules: [Bar, ~r/Ignore/]]
     ]
   end
 end
