@@ -56,9 +56,9 @@ defmodule Mix.DepTest do
   end
 
   test "extracts all dependencies from the given project" do
-    Mix.Project.push(DepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(DepsApp)
+
       deps = Mix.Dep.load_on_environment([])
       assert length(deps) == 6
       assert Enum.find(deps, &match?(%Mix.Dep{app: :ok, status: {:ok, _}}, &1))
@@ -71,9 +71,9 @@ defmodule Mix.DepTest do
   end
 
   test "extracts all dependencies paths/scms from the given project" do
-    Mix.Project.push(DepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(DepsApp)
+
       apps = Mix.Project.deps_apps()
       assert length(apps) == 6
       assert :ok in apps
@@ -194,7 +194,7 @@ defmodule Mix.DepTest do
     Process.put(:custom_deps_git_repo_opts, manager: :make)
 
     deps = [
-      {:deps_repo, "0.1.0", path: "custom/deps_repo", manager: :rebar},
+      {:deps_repo, "0.1.0", path: "custom/deps_repo", manager: :rebar3},
       {:git_repo, "0.2.0", git: MixTest.Case.fixture_path("git_repo")}
     ]
 
@@ -202,7 +202,7 @@ defmodule Mix.DepTest do
       in_fixture("deps_status", fn ->
         [dep1, dep2] = Mix.Dep.load_on_environment([])
         assert dep1.manager == nil
-        assert dep2.manager == :rebar
+        assert dep2.manager == :rebar3
       end)
     end)
   end
@@ -310,7 +310,7 @@ defmodule Mix.DepTest do
     dep_path = tmp_path("rebar_dep")
 
     system_env = [{"FILE_FROM_ENV", file_path}, {"CONTENTS_FROM_ENV", "contents dep test"}]
-    deps = [{:rebar_dep, path: dep_path, app: false, manager: :rebar, system_env: system_env}]
+    deps = [{:rebar_dep, path: dep_path, app: false, manager: :rebar3, system_env: system_env}]
 
     with_deps(deps, fn ->
       in_tmp("load dependency with env vars", fn ->
@@ -477,7 +477,7 @@ defmodule Mix.DepTest do
       Mix.RemoteConverger.register(RaiseRemoteConverger)
 
       in_fixture("deps_cycle", fn ->
-        assert_raise Mix.Error, ~r/cycles in the dependency graph/, fn ->
+        assert_raise Mix.Error, ~r/Could not sort dependencies/, fn ->
           Mix.Tasks.Deps.Get.run([])
         end
 
@@ -496,8 +496,11 @@ defmodule Mix.DepTest do
 
     with_deps(deps, fn ->
       in_fixture("deps_status", fn ->
-        assert Enum.map(Mix.Dep.load_on_environment([]), & &1.app) ==
-                 [:git_repo, :abc_repo, :deps_repo]
+        # Both orders below are valid after topological sort
+        assert Enum.map(Mix.Dep.load_on_environment([]), & &1.app) in [
+                 [:git_repo, :abc_repo, :deps_repo],
+                 [:abc_repo, :git_repo, :deps_repo]
+               ]
 
         assert Map.keys(Mix.Project.deps_paths()) == [:abc_repo, :deps_repo, :git_repo]
 
