@@ -84,7 +84,25 @@ defmodule Mix.LocalTest do
       File.write!("csv.signed", @csv_signed)
 
       assert {"1.0.0", "1.2.4", "GHI"} =
-               Mix.Local.find_matching_versions_from_signed_csv!("name", "csv")
+               Mix.Local.find_matching_versions_from_signed_csv!("name", nil, "csv")
+    end)
+  end
+
+  @tag :tmp_dir
+  test "select specific version from csv", %{tmp_dir: tmp_dir} do
+    File.cd!(tmp_dir, fn ->
+      File.write!("csv", @csv)
+      File.write!("csv.signed", @csv_signed)
+
+      assert {"0.9.0", "1.2.5", "ABC"} =
+               Mix.Local.find_matching_versions_from_signed_csv!("name", "1.2.5", "csv")
+
+      assert {"1.0.0", "1.2.3", "DEF"} =
+               Mix.Local.find_matching_versions_from_signed_csv!("name", "1.2.3", "csv")
+
+      assert_raise Mix.Error, "Could not find a matching version of name", fn ->
+        Mix.Local.find_matching_versions_from_signed_csv!("name", "1.3.0", "csv")
+      end
     end)
   end
 
@@ -96,7 +114,7 @@ defmodule Mix.LocalTest do
       File.write!("csv.signed", csv_signed)
 
       assert_raise Mix.Error, fn ->
-        Mix.Local.find_matching_versions_from_signed_csv!("name", "csv")
+        Mix.Local.find_matching_versions_from_signed_csv!("name", nil, "csv")
       end
     end)
   end
