@@ -133,7 +133,7 @@ defmodule Mix.ReleaseTest do
 
     test "uses the locked version of an app", context do
       in_tmp(context.test, fn ->
-        # install newer version of the app in the custom erts
+        # install newer version of the app in the custom ERTS
         custom_erts_path = Path.join([File.cwd!(), "erts-#{@erts_version}"])
         File.cp_r!(@erts_source, custom_erts_path)
 
@@ -329,6 +329,46 @@ defmodule Mix.ReleaseTest do
       assert release.boot_scripts.start[:iex] == :temporary
     end
 
+    test "configures other applications in cascade", context do
+      in_tmp(context.test, fn ->
+        app =
+          {:application, :my_sample_mode,
+           applications: [:kernel, :stdlib, :elixir, :runtime_tools, :compiler],
+           description: 'my_sample_mode',
+           modules: [],
+           vsn: '1.0.0'}
+
+        File.mkdir_p!("my_sample_mode/ebin")
+        Code.prepend_path("my_sample_mode/ebin")
+        format = :io_lib.format("%% coding: utf-8~n~p.~n", [app])
+        File.write!("my_sample_mode/ebin/my_sample_mode.app", format)
+
+        apps = [my_sample_mode: :temporary]
+        release = release(applications: apps)
+        assert release.boot_scripts.start[:my_sample_mode] == :temporary
+        assert release.boot_scripts.start[:runtime_tools] == :temporary
+        assert release.boot_scripts.start[:compiler] == :permanent
+
+        apps = [my_sample_mode: :temporary, elixir: :permanent]
+        release = release(applications: apps)
+        assert release.boot_scripts.start[:my_sample_mode] == :temporary
+        assert release.boot_scripts.start[:runtime_tools] == :temporary
+        assert release.boot_scripts.start[:compiler] == :permanent
+
+        apps = [my_sample_mode: :temporary, runtime_tools: :none, compiler: :none]
+        release = release(applications: apps)
+        assert release.boot_scripts.start[:my_sample_mode] == :temporary
+        assert release.boot_scripts.start[:runtime_tools] == :none
+        assert release.boot_scripts.start[:compiler] == :none
+
+        apps = [my_sample_mode: :temporary, elixir: :permanent, compiler: :none]
+        release = release(applications: apps)
+        assert release.boot_scripts.start[:my_sample_mode] == :temporary
+        assert release.boot_scripts.start[:runtime_tools] == :temporary
+        assert release.boot_scripts.start[:compiler] == :none
+      end)
+    end
+
     test "generates a start_clean script with only kernel and stdlib starting up" do
       release = release([])
 
@@ -479,6 +519,13 @@ defmodule Mix.ReleaseTest do
       assert message =~
                "Application :mix has mode :permanent but it depends on :elixir which is set to :none"
     end
+
+    test "does not raise on child unsafe mode if parent is in `skip_mode_validation_for`" do
+      # `:mix` is a parent that depends on `:elixir`
+      release = release(applications: [elixir: :load], skip_mode_validation_for: [:mix])
+      assert make_boot_script(release, @boot_script_path, release.boot_scripts.start) == :ok
+      assert make_boot_script(release, @boot_script_path, release.boot_scripts.start_clean) == :ok
+    end
   end
 
   describe "make_cookie/1" do
@@ -507,12 +554,19 @@ defmodule Mix.ReleaseTest do
       assert make_cookie(release(cookie: "lmnopqrstuv"), @cookie_path) == :ok
       assert File.read!(@cookie_path) == "lmnopqrstuv"
     end
+
+    test "does not ask to change if set to overwrite" do
+      assert make_cookie(release([]), @cookie_path) == :ok
+      send(self(), {:mix_shell_input, :yes?, false})
+      assert make_cookie(release([cookie: "lmnopqrstuv"], overwrite: true), @cookie_path) == :ok
+      assert File.read!(@cookie_path) == "lmnopqrstuv"
+    end
   end
 
   describe "make_start_erl/1" do
     @start_erl_path tmp_path("mix_release/start_erl.data")
 
-    test "writes erts and release versions" do
+    test "writes ERTS and release versions" do
       assert make_start_erl(release([]), @start_erl_path) == :ok
       assert File.read!(@start_erl_path) == "#{@erts_version} 0.1.0"
     end
@@ -532,13 +586,13 @@ defmodule Mix.ReleaseTest do
     test "writes sys_config with encoding" do
       assert make_sys_config(
                release([]),
-               [encoding: {:time_μs, :"£", "£", '£'}],
+               [encoding: {:_μ, :"£", "£", '£'}],
                "unused/runtime/path"
              ) ==
                :ok
 
       {:ok, contents} = :file.consult(@sys_config)
-      assert contents == [[encoding: {:time_μs, :"£", "£", '£'}]]
+      assert contents == [[encoding: {:_μ, :"£", "£", '£'}]]
     end
 
     test "writes the given sys_config with config providers" do
@@ -585,6 +639,11 @@ defmodule Mix.ReleaseTest do
     test "errors on bad config" do
       assert {:error, "Could not read configuration file." <> _} =
                make_sys_config(release([]), [foo: [bar: self()]], "unused/runtime/path")
+
+      env = %{__ENV__ | lexical_tracker: self()}
+
+      assert {:error, "Could not read configuration file." <> _} =
+               make_sys_config(release([]), [foo: [bar: env]], "unused/runtime/path")
     end
   end
 
@@ -720,6 +779,15 @@ defmodule Mix.ReleaseTest do
       assert {:ok, {EEx, [{'Dbgi', _}]}} = :beam_lib.chunks(beam, ['Dbgi'])
       assert {:ok, {EEx, [{'Docs', _}]}} = :beam_lib.chunks(beam, ['Docs'])
     end
+
+    test "strip beams without compression" do
+      {:ok, beam} =
+        Path.join(@eex_ebin, "Elixir.EEx.beam")
+        |> File.read!()
+        |> strip_beam(compress: false)
+
+      assert match?(<<"FOR1", _::binary>>, beam)
+    end
   end
 
   describe "included applications" do
@@ -739,11 +807,9 @@ defmodule Mix.ReleaseTest do
         File.write!("my_sample1/ebin/my_sample1.app", format)
 
         release = release(applications: [my_sample1: :permanent])
-        assert release.applications.runtime_tools[:included]
         assert release.boot_scripts.start[:runtime_tools] == :load
 
         release = release(applications: [my_sample1: :permanent, runtime_tools: :none])
-        assert release.applications.runtime_tools[:included]
         assert release.boot_scripts.start[:runtime_tools] == :none
       end)
     end
@@ -770,6 +836,28 @@ defmodule Mix.ReleaseTest do
     end
   end
 
+  describe "optional applications" do
+    test "are ignored if not available", context do
+      in_tmp(context.test, fn ->
+        app =
+          {:application, :my_sample1,
+           applications: [:kernel, :stdlib, :elixir, :unknown],
+           optional_applications: [:unknown],
+           description: 'my_sample1',
+           modules: [],
+           vsn: '1.0.0'}
+
+        File.mkdir_p!("my_sample1/ebin")
+        Code.prepend_path("my_sample1/ebin")
+        format = :io_lib.format("%% coding: utf-8~n~p.~n", [app])
+        File.write!("my_sample1/ebin/my_sample1.app", format)
+
+        release = release(applications: [my_sample1: :permanent])
+        assert release.boot_scripts.start[:unknown] == nil
+      end)
+    end
+  end
+
   defp size!(path) do
     File.stat!(path).size
   end
@@ -778,8 +866,8 @@ defmodule Mix.ReleaseTest do
     File.stat!(path).mode
   end
 
-  defp release(config) do
-    from_config!(nil, config(releases: [demo: config]), [])
+  defp release(config, overrides \\ []) do
+    from_config!(nil, config(releases: [demo: config]), overrides)
   end
 
   defp config(extra \\ []) do
