@@ -3,6 +3,12 @@ Code.require_file("../../test_helper.exs", __DIR__)
 defmodule Mix.Tasks.CompileTest do
   use MixTest.Case
 
+  if System.otp_release() >= "24" do
+    defmacro position(line, column), do: {line, column}
+  else
+    defmacro position(line, _column), do: line
+  end
+
   defmodule CustomCompilers do
     def project do
       [compilers: [:elixir, :app, :custom]]
@@ -50,9 +56,10 @@ defmodule Mix.Tasks.CompileTest do
 
   test "compiles a project with cached deps information" do
     Mix.Project.pop()
-    Mix.Project.push(DepsApp)
 
     in_fixture("deps_status", fn ->
+      Mix.Project.push(DepsApp)
+
       File.mkdir_p!("lib")
 
       File.write!("lib/a.ex", """
@@ -66,7 +73,7 @@ defmodule Mix.Tasks.CompileTest do
       end)
       """)
 
-      assert Mix.Task.run("compile", ["--force", "--no-deps-loading"]) == {:ok, []}
+      assert Mix.Task.run("compile", ["--force", "--from-mix-deps-compile"]) == {:ok, []}
     end)
   end
 
@@ -140,7 +147,7 @@ defmodule Mix.Tasks.CompileTest do
         assert %Mix.Task.Compiler.Diagnostic{
                  file: ^file,
                  severity: :error,
-                 position: 2,
+                 position: {2, 20},
                  message: "** (SyntaxError) lib/a.ex:2:" <> _,
                  compiler_name: "Elixir"
                } = diagnostic
@@ -202,7 +209,7 @@ defmodule Mix.Tasks.CompileTest do
                  compiler_name: "erl_parse",
                  file: ^file,
                  message: "syntax error before: b",
-                 position: 2,
+                 position: position(2, 5),
                  severity: :error
                } = diagnostic
       end)
@@ -221,6 +228,21 @@ defmodule Mix.Tasks.CompileTest do
     end)
   end
 
+  test "loads consolidated protocols even on --no-compile" do
+    in_fixture("no_mixfile", fn ->
+      File.rm("_build/dev/lib/sample/.mix/compile.protocols")
+      consolidated = "_build/dev/lib/sample/consolidated" |> Path.expand() |> to_charlist()
+
+      assert Mix.Task.run("compile") == {:ok, []}
+      assert File.regular?("_build/dev/lib/sample/consolidated/Elixir.Enumerable.beam")
+      assert consolidated in :code.get_path()
+
+      Code.delete_path("_build/dev/lib/sample/consolidated")
+      assert Mix.Task.rerun("compile", ["--no-compile"]) == {:noop, []}
+      assert consolidated in :code.get_path()
+    end)
+  end
+
   test "loads Mix config with --erl-config" do
     in_fixture("no_mixfile", fn ->
       File.write!("mix.config", "{erl_config_app, [{value, true}]}.")
@@ -271,6 +293,7 @@ defmodule Mix.Tasks.CompileTest do
 
       File.write!("lib/a.ex", """
       defmodule A do
+        require Application
         _ = Application.compile_env(:sample, :hello)
       end
       """)
@@ -283,7 +306,7 @@ defmodule Mix.Tasks.CompileTest do
       Mix.Tasks.Loadconfig.run([])
       Application.unload(:sample)
 
-      assert ExUnit.CaptureIO.capture_io(:stderr, fn ->
+      assert ExUnit.CaptureIO.capture_io(fn ->
                assert_raise ErlangError, fn -> Mix.Tasks.Compile.All.run([]) end
              end) =~
                " the application :sample has a different value set for key :hello during runtime compared to compile time"
