@@ -45,12 +45,24 @@ defmodule Mix.Tasks.DepsTest do
     end
   end
 
+  defmodule MissingLocalDepsApp do
+    def project do
+      [
+        app: :missing_local_deps,
+        version: "0.1.0",
+        deps: [
+          {:ok, path: "missing/dep"}
+        ]
+      ]
+    end
+  end
+
   ## deps
 
   test "prints list of dependencies and their status" do
-    Mix.Project.push(DepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(DepsApp)
+
       Mix.Tasks.Deps.run([])
 
       assert_received {:mix_shell, :info, ["* ok (https://github.com/elixir-lang/ok.git) (mix)"]}
@@ -73,9 +85,9 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "prints list of dependencies and their status, including req mismatches and custom apps" do
-    Mix.Project.push(ReqDepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(ReqDepsApp)
+
       Mix.Tasks.Deps.run([])
 
       assert_received {:mix_shell, :info, ["* ok (deps/ok) (mix)"]}
@@ -91,9 +103,9 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "prints misspelled dependency name hint" do
-    Mix.Project.push(DepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(DepsApp)
+
       other_app_path = Path.join(Mix.Project.build_path(), "lib/noappfile/ebin/other_app.app")
       File.mkdir_p!(Path.dirname(other_app_path))
       File.write!(other_app_path, "")
@@ -112,9 +124,9 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "prints Elixir req mismatches" do
-    Mix.Project.push(ReqDepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(ReqDepsApp)
+
       File.write!("deps/ok/mix.exs", """
       defmodule Deps.OkApp do
         use Mix.Project
@@ -138,9 +150,9 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "prints list of dependencies and their lock status" do
-    Mix.Project.push(DepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(DepsApp)
+
       File.cd!("deps/ok", fn ->
         System.cmd("git", ~w[-c core.hooksPath='' init])
       end)
@@ -180,9 +192,9 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "cleans and recompiles artifacts if --force is given" do
-    Mix.Project.push(SuccessfulDepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(SuccessfulDepsApp)
+
       Mix.Tasks.Deps.Compile.run([])
       File.touch!("_build/dev/lib/ok/clean-me")
 
@@ -202,29 +214,41 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "doesn't compile any path deps if --skip-local-deps is given" do
-    Mix.Project.push(SuccessfulDepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(SuccessfulDepsApp)
+
       File.rm_rf!("_build/dev/lib/ok/ebin")
       Mix.Tasks.Deps.Compile.run(["--skip-local-deps"])
       refute File.exists?("_build/dev/lib/ok/ebin")
     end)
   end
 
+  test "checks if local dependencies are available before compiling" do
+    Mix.Project.push(MissingLocalDepsApp)
+
+    error_message =
+      "Cannot compile dependency :ok because it isn't available, please ensure the dependency " <>
+        "is at \"missing/dep\""
+
+    assert_raise Mix.Error, error_message, fn ->
+      Mix.Tasks.Deps.Compile.run([])
+    end
+  end
+
   ## deps.loadpaths
 
   test "checks list of dependencies and their status with success" do
-    Mix.Project.push(SuccessfulDepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(SuccessfulDepsApp)
+
       Mix.Tasks.Deps.Loadpaths.run([])
     end)
   end
 
   test "checks list of dependencies and their status on failure" do
-    Mix.Project.push(DepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(DepsApp)
+
       assert_raise Mix.Error, fn ->
         Mix.Tasks.Deps.Loadpaths.run([])
       end
@@ -246,49 +270,10 @@ defmodule Mix.Tasks.DepsTest do
     end)
   end
 
-  test "compiles and prunes builds per environment" do
-    Mix.Project.push(SuccessfulDepsApp)
-
+  test "does not load deps with --no-load-deps" do
     in_fixture("deps_status", fn ->
-      # Start from scratch!
-      File.rm_rf("_build")
-
-      Mix.Tasks.Deps.Compile.run([])
-      Mix.Tasks.Deps.Loadpaths.run([])
-      assert File.exists?("_build/dev/lib/ok/ebin/ok.app")
-      assert File.exists?("_build/dev/lib/ok/priv/sample")
-
-      Mix.Tasks.Compile.run([])
-      assert to_charlist(Path.expand("_build/dev/lib/ok/ebin/")) in :code.get_path()
-      assert File.exists?("_build/dev/lib/sample/ebin/sample.app")
-
-      # Remove the deps but set build_path, deps won't be pruned, but load paths are
-      Mix.ProjectStack.post_config(deps: [], build_path: "_build")
-      Mix.State.clear_cache()
-      Mix.Project.pop()
       Mix.Project.push(SuccessfulDepsApp)
 
-      Mix.Tasks.Deps.Loadpaths.run([])
-      refute to_charlist(Path.expand("_build/dev/lib/ok/ebin/")) in :code.get_path()
-      assert File.exists?("_build/dev/lib/ok/ebin/ok.app")
-      assert File.exists?("_build/dev/lib/sample/ebin/sample.app")
-
-      # Remove the deps without build_path, deps will be pruned
-      Mix.ProjectStack.post_config(deps: [])
-      Mix.State.clear_cache()
-      Mix.Project.pop()
-      Mix.Project.push(SuccessfulDepsApp)
-
-      Mix.Tasks.Deps.Loadpaths.run([])
-      refute File.exists?("_build/dev/lib/ok/ebin/ok.app")
-      assert File.exists?("_build/dev/lib/sample/ebin/sample.app")
-    end)
-  end
-
-  test "does not load or prune builds with --no-load-deps" do
-    Mix.Project.push(SuccessfulDepsApp)
-
-    in_fixture("deps_status", fn ->
       # Start from scratch!
       File.rm_rf("_build")
 
@@ -318,9 +303,9 @@ defmodule Mix.Tasks.DepsTest do
   ## deps.unlock
 
   test "unlocks all deps", context do
-    Mix.Project.push(DepsApp)
-
     in_tmp(context.test, fn ->
+      Mix.Project.push(DepsApp)
+
       Mix.Dep.Lock.write(%{git_repo: "abcdef"})
       assert Mix.Dep.Lock.read() == %{git_repo: "abcdef"}
       Mix.Tasks.Deps.Unlock.run(["--all"])
@@ -329,9 +314,9 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "checks lock file has unused deps with --check-unused", context do
-    Mix.Project.push(DepsApp)
-
     in_tmp(context.test, fn ->
+      Mix.Project.push(DepsApp)
+
       Mix.Dep.Lock.write(%{whatever: "0.2.0", something_else: "1.2.3", ok: "0.1.0"})
       assert Mix.Dep.Lock.read() == %{whatever: "0.2.0", something_else: "1.2.3", ok: "0.1.0"}
 
@@ -355,9 +340,9 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "unlocks unused deps", context do
-    Mix.Project.push(DepsApp)
-
     in_tmp(context.test, fn ->
+      Mix.Project.push(DepsApp)
+
       Mix.Dep.Lock.write(%{whatever: "abcdef", ok: "abcdef"})
       assert Mix.Dep.Lock.read() == %{whatever: "abcdef", ok: "abcdef"}
       Mix.Tasks.Deps.Unlock.run(["--unused"])
@@ -372,10 +357,20 @@ defmodule Mix.Tasks.DepsTest do
     end)
   end
 
-  test "unlocks specific deps", context do
-    Mix.Project.push(DepsApp)
-
+  test "unlocking a dep that is not locked is a no-op", context do
     in_tmp(context.test, fn ->
+      Mix.Project.push(DepsApp)
+      Mix.Tasks.Deps.Unlock.run(["unlocked_dep"])
+
+      assert_received {:mix_shell, :error, ["warning: unlocked_dep dependency is not locked"]}
+      refute_received {:mix_shell, :info, [_]}
+    end)
+  end
+
+  test "unlocks specific deps", context do
+    in_tmp(context.test, fn ->
+      Mix.Project.push(DepsApp)
+
       Mix.Dep.Lock.write(%{git_repo: "abcdef", another: "hash"})
       Mix.Tasks.Deps.Unlock.run(["git_repo", "unknown"])
       assert Mix.Dep.Lock.read() == %{another: "hash"}
@@ -392,9 +387,9 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "unlocks filtered deps", context do
-    Mix.Project.push(DepsApp)
-
     in_tmp(context.test, fn ->
+      Mix.Project.push(DepsApp)
+
       Mix.Dep.Lock.write(%{git_repo: "abcdef", another: "hash", another_one: "hash"})
       Mix.Tasks.Deps.Unlock.run(["--filter", "another"])
       assert Mix.Dep.Lock.read() == %{git_repo: "abcdef"}
@@ -444,18 +439,18 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "sets deps env to prod by default" do
-    Mix.Project.push(DepsEnvApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(DepsEnvApp)
+
       Mix.Tasks.Deps.Update.run(["--all"])
       assert_received {:mix_shell, :info, [":raw_repo env is prod"]}
     end)
   end
 
   test "can customize environment" do
-    Mix.Project.push(CustomDepsEnvApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(CustomDepsEnvApp)
+
       Mix.Tasks.Deps.Update.run(["--all"])
       assert_received {:mix_shell, :info, [":raw_repo env is dev"]}
     end)
@@ -529,9 +524,9 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "fails on missing dependencies" do
-    Mix.Project.push(SuccessfulDepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(SuccessfulDepsApp)
+
       assert_raise Mix.Error, ~r/Unknown dependency invalid for environment dev/, fn ->
         Mix.Tasks.Deps.Update.run(["invalid"])
       end
@@ -541,9 +536,9 @@ defmodule Mix.Tasks.DepsTest do
   @overriding_msg "  the dependency git_repo in mix.exs is overriding a child dependency"
 
   test "fails on diverged dependencies on get/update" do
-    Mix.Project.push(ConflictDepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(ConflictDepsApp)
+
       assert_raise Mix.Error, fn ->
         Mix.Tasks.Deps.Loadpaths.run([])
       end
@@ -565,9 +560,9 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "fails on diverged dependencies on check" do
-    Mix.Project.push(DivergedDepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(DivergedDepsApp)
+
       assert_raise Mix.Error, fn ->
         Mix.Tasks.Deps.Loadpaths.run([])
       end
@@ -581,9 +576,9 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "fails on diverged dependencies by requirement" do
-    Mix.Project.push(ConvergedDepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(ConvergedDepsApp)
+
       File.write!("custom/deps_repo/mix.exs", """
       defmodule DepsRepo do
         use Mix.Project
@@ -616,9 +611,9 @@ defmodule Mix.Tasks.DepsTest do
   @overriding_msg "  the dependency git_repo in mix.exs is overriding"
 
   test "fails on diverged dependencies even when optional" do
-    Mix.Project.push(ConvergedDepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(ConvergedDepsApp)
+
       File.write!("custom/deps_repo/mix.exs", """
       defmodule DepsRepo do
         use Mix.Project
@@ -645,9 +640,9 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "works with converged dependencies" do
-    Mix.Project.push(ConvergedDepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(ConvergedDepsApp)
+
       Mix.Tasks.Deps.Get.run([])
       message = "* Getting git_repo (#{fixture_path("git_repo")})"
       assert_received {:mix_shell, :info, [^message]}
@@ -667,9 +662,9 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "does not check dependencies if --no-deps-check is provided" do
-    Mix.Project.push(SuccessfulDepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(SuccessfulDepsApp)
+
       Mix.Tasks.Deps.Get.run([])
       File.rm_rf!("deps/ok")
 
@@ -682,9 +677,9 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "works with overridden dependencies" do
-    Mix.Project.push(OverriddenDepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(OverriddenDepsApp)
+
       Mix.Tasks.Deps.Get.run([])
       message = "* Getting git_repo (#{fixture_path("git_repo")})"
       assert_received {:mix_shell, :info, [^message]}
@@ -704,9 +699,9 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "converged dependencies errors if not overriding" do
-    Mix.Project.push(NonOverriddenDepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(NonOverriddenDepsApp)
+
       assert_raise Mix.Error, fn ->
         Mix.Tasks.Deps.Loadpaths.run([])
       end
@@ -726,9 +721,9 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "checks if dependencies are using old Elixir version" do
-    Mix.Project.push(SuccessfulDepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(SuccessfulDepsApp)
+
       Mix.Tasks.Deps.Compile.run([])
       Mix.Tasks.Deps.Loadpaths.run([])
 
@@ -753,9 +748,9 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "checks if dependencies are using old scm version" do
-    Mix.Project.push(SuccessfulDepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(SuccessfulDepsApp)
+
       Mix.Tasks.Deps.Compile.run([])
       Mix.Tasks.Deps.Loadpaths.run([])
 
@@ -793,9 +788,9 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "does not compile deps that have explicit option" do
-    Mix.Project.push(NonCompilingDeps)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(NonCompilingDeps)
+
       Mix.Tasks.Deps.Compile.run([])
       refute_received {:mix_shell, :info, ["==> git_repo"]}
     end)
@@ -816,9 +811,9 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "warns and converges duplicated deps at the same level" do
-    Mix.Project.push(DupDeps)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(DupDeps)
+
       Mix.Tasks.Deps.run([])
 
       msg =
@@ -848,9 +843,9 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "cleans dependencies" do
-    Mix.Project.push(CleanDepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(CleanDepsApp)
+
       File.mkdir_p!("_build/dev/lib/raw_sample")
       File.mkdir_p!("_build/dev/lib/git_repo")
       File.mkdir_p!("_build/test/lib/git_repo")
@@ -881,9 +876,9 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "cleans unused dependencies" do
-    Mix.Project.push(CleanDepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(CleanDepsApp)
+
       File.mkdir_p!("_build/dev/lib/raw_sample")
       File.mkdir_p!("deps/git_repo")
       File.mkdir_p!("_build/dev/lib/git_repo")
@@ -900,9 +895,9 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "cleans dependencies build" do
-    Mix.Project.push(CleanDepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(CleanDepsApp)
+
       File.mkdir_p!("deps/raw_sample")
       File.mkdir_p!("_build/dev/lib/raw_sample")
 
@@ -913,9 +908,9 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "warns on invalid path on clean dependencies" do
-    Mix.Project.push(CleanDepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(CleanDepsApp)
+
       File.mkdir_p!("deps/raw_sample")
       File.mkdir_p!("_build/dev/lib/raw_sample")
 
@@ -928,9 +923,9 @@ defmodule Mix.Tasks.DepsTest do
   end
 
   test "does not remove dependency source when using :path" do
-    Mix.Project.push(CleanDepsApp)
-
     in_fixture("deps_status", fn ->
+      Mix.Project.push(CleanDepsApp)
+
       assert File.exists?("deps/ok")
 
       Mix.Tasks.Deps.Clean.run(["raw_sample", "--all"])
