@@ -296,7 +296,7 @@ defmodule Mix.Tasks.ReleaseTest do
         assert_received {:mix_shell, :info, ["* assembling release_test-0.1.0 on MIX_ENV=dev"]}
 
         assert_received {:mix_shell, :info,
-                         ["\nRelease created at _build/dev/rel/release_test!" <> _]}
+                         ["\nRelease created at _build/dev/rel/release_test\n" <> _]}
 
         assert_received {:mix_shell, :info, ["* skipping runtime configuration" <> _]}
 
@@ -330,7 +330,7 @@ defmodule Mix.Tasks.ReleaseTest do
         assert %{
                  app_dir: app_dir,
                  cookie_env: ^cookie,
-                 encoding: {:time_μs, :"£", "£", '£'},
+                 encoding: {:_μ, :"£", "£", '£'},
                  mode: :embedded,
                  node: release_node("release_test"),
                  protocols_consolidated?: true,
@@ -388,7 +388,7 @@ defmodule Mix.Tasks.ReleaseTest do
           config_env: config_env(),
           config_target: config_target()
 
-        config :release_test, :encoding, {:runtime, :time_μs, :"£", "£", '£'}
+        config :release_test, :encoding, {:runtime, :_μ, :"£", "£", '£'}
         """)
 
         root = Path.absname("_build/dev/rel/runtime_config")
@@ -414,7 +414,7 @@ defmodule Mix.Tasks.ReleaseTest do
         open_port(Path.join(root, "bin/runtime_config"), ['start'])
 
         assert %{
-                 encoding: {:runtime, :time_μs, :"£", "£", '£'},
+                 encoding: {:runtime, :_μ, :"£", "£", '£'},
                  mode: :embedded,
                  node: release_node("runtime_config"),
                  protocols_consolidated?: true,
@@ -528,6 +528,7 @@ defmodule Mix.Tasks.ReleaseTest do
       config = [releases: [compile_env_config: []], config_path: "different_config/config.exs"]
 
       File.write!("lib/compile_env.ex", """
+      require Application
       _ = Application.compile_env(:release_test, :static)
       """)
 
@@ -567,6 +568,7 @@ defmodule Mix.Tasks.ReleaseTest do
       config = [releases: [no_compile_env_config: [validate_compile_env: false]]]
 
       File.write!("lib/compile_env.ex", """
+      require Application
       _ = Application.compile_env(:release_test, :static)
       """)
 
@@ -596,6 +598,7 @@ defmodule Mix.Tasks.ReleaseTest do
 
       # We write the compile env to guarantee rpc still works
       File.write!("lib/compile_env.ex", """
+      require Application
       _ = Application.compile_env(:release_test, :static)
       """)
 
@@ -611,14 +614,17 @@ defmodule Mix.Tasks.ReleaseTest do
         Mix.Task.run("release")
         script = Path.join(root, "bin/permanent1")
 
-        open_port(script, ['start'])
+        env = [{"RELEASE_DISTRIBUTION", "name"}, {"RELEASE_NODE", "permanent1@127.0.0.1"}]
+        open_port(script, ['start'], env)
         wait_until_decoded(Path.join(root, "RELEASE_BOOTED"))
-        assert System.cmd(script, ["rpc", "ReleaseTest.hello_world()"]) == {"hello world\n", 0}
 
-        assert {pid, 0} = System.cmd(script, ["pid"])
+        assert System.cmd(script, ["rpc", "ReleaseTest.hello_world()"], env: env) ==
+                 {"hello world\n", 0}
+
+        assert {pid, 0} = System.cmd(script, ["pid"], env: env)
         assert pid != "\n"
 
-        assert System.cmd(script, ["stop"]) == {"", 0}
+        assert System.cmd(script, ["stop"], env: env) == {"", 0}
       end)
     end)
   end
@@ -686,16 +692,17 @@ defmodule Mix.Tasks.ReleaseTest do
         Mix.Task.run("release")
 
         script = Path.join(root, "bin/permanent2")
-        open_port(script, ['daemon_iex'])
+        env = [{"RELEASE_DISTRIBUTION", "name"}, {"RELEASE_NODE", "permanent2@127.0.0.1"}]
+        open_port(script, ['daemon_iex'], env)
 
         assert %{
                  app_dir: app_dir,
                  cookie_env: "abcdefghij",
                  mode: :embedded,
-                 node: release_node("permanent2"),
+                 node: :"permanent2@127.0.0.1",
                  protocols_consolidated?: true,
                  release_name: "permanent2",
-                 release_node: "permanent2",
+                 release_node: "permanent2@127.0.0.1",
                  release_root: ^root,
                  release_vsn: "0.1.0",
                  root_dir: root_dir,
@@ -712,11 +719,20 @@ defmodule Mix.Tasks.ReleaseTest do
 
         assert wait_until(fn ->
                  File.read!(Path.join(root, "tmp/log/erlang.log.1")) =~
-                   "iex(permanent2@#{@hostname})1> "
+                   "iex(permanent2@127.0.0.1)1> "
                end)
 
-        assert System.cmd(script, ["rpc", "ReleaseTest.hello_world()"]) == {"hello world\n", 0}
-        assert System.cmd(script, ["stop"]) == {"", 0}
+        assert System.cmd(script, ["rpc", "ReleaseTest.hello_world()"], env: env) ==
+                 {"hello world\n", 0}
+
+        assert System.cmd(script, ["restart"], env: env) == {"", 0}
+
+        assert wait_until(fn ->
+                 File.read!(Path.join(root, "tmp/log/erlang.log.1")) =~
+                   "RESTARTED!!!"
+               end)
+
+        assert System.cmd(script, ["stop"], env: env) == {"", 0}
       end)
     end)
   end
@@ -741,6 +757,7 @@ defmodule Mix.Tasks.ReleaseTest do
   end
 
   defp open_port(command, args, env \\ []) do
+    env = for {k, v} <- env, do: {to_charlist(k), to_charlist(v)}
     Port.open({:spawn_executable, to_charlist(command)}, [:hide, args: args, env: env])
   end
 
