@@ -3,29 +3,10 @@ Code.require_file("../../test_helper.exs", __DIR__)
 defmodule Mix.Tasks.RunTest do
   use MixTest.Case
 
-  import ExUnit.CaptureIO
-
   setup do
     Mix.Project.push(MixTest.Case.Sample)
   end
 
-  test "loads configuration", context do
-    in_tmp(context.test, fn ->
-      config = fixture_path("configs/good_config.exs")
-      expr = "IO.puts(Application.get_env(:my_app, :key))"
-
-      output =
-        capture_io(:stderr, fn ->
-          assert capture_io(fn -> Mix.Task.run("run", ["--config", config, "--eval", expr]) end) ==
-                   "value\n"
-        end)
-
-      assert output =~ "deprecated"
-    end)
-  after
-    Application.delete_env(:my_app, :key)
-  end
-
   test "run requires files before evaling commands", context do
     git_repo = fixture_path("git_repo/lib/git_repo.ex")
 
