@@ -92,33 +92,42 @@ defmodule Mix.Tasks.XrefTest do
 
         Mix.Task.run("compile")
         after_compile.()
-        assert Enum.sort(Mix.Tasks.Xref.calls()) == Enum.sort(expected)
+        xref = Mix.Tasks.Xref
+        assert Enum.sort(xref.calls()) == Enum.sort(expected)
       end)
     end
   end
 
   describe "mix xref callers MODULE" do
-    test "prints callers of specified Module" do
-      files = %{
-        "lib/a.ex" => """
-        defmodule A do
-          def a, do: :ok
-        end
-        """,
-        "lib/b.ex" => """
-        defmodule B do
-          def b, do: A.a()
-        end
-        """
-      }
-
-      output = """
-      Compiling 2 files (.ex)
-      Generated sample app
-      lib/b.ex (runtime)
+    @callers_files %{
+      "lib/a.ex" => """
+      defmodule A do
+        def a, do: :ok
+      end
+      """,
+      "lib/b.ex" => """
+      defmodule B do
+        def b, do: A.a()
+      end
       """
+    }
 
-      assert_callers("A", files, output)
+    @callers_output """
+    Compiling 2 files (.ex)
+    Generated sample app
+    lib/b.ex (runtime)
+    """
+
+    test "prints callers of specified Module" do
+      assert_callers("A", @callers_files, @callers_output)
+    end
+
+    test "filter by compile-connected label with fail-above" do
+      message = "Too many references (found: 1, permitted: 0)"
+
+      assert_raise Mix.Error, message, fn ->
+        assert_callers(~w[--fail-above 0], "A", @callers_files, @callers_output)
+      end
     end
 
     test "handles aliases" do
@@ -203,14 +212,155 @@ defmodule Mix.Tasks.XrefTest do
       end)
     end
 
-    defp assert_callers(module, files, expected) do
+    defp assert_callers(opts \\ [], module, files, expected) do
       in_fixture("no_mixfile", fn ->
         for {file, contents} <- files do
           File.write!(file, contents)
         end
 
         capture_io(:stderr, fn ->
-          assert Mix.Task.run("xref", ["callers", module]) == :ok
+          assert Mix.Task.run("xref", opts ++ ["callers", module]) == :ok
+        end)
+
+        assert ^expected = receive_until_no_messages([])
+      end)
+    end
+  end
+
+  describe "mix xref trace FILE" do
+    test "shows labelled traces" do
+      files = %{
+        "lib/a.ex" => ~S"""
+        defmodule A do
+          defstruct [:foo, :bar]
+          defmacro macro, do: :ok
+          def fun, do: :ok
+        end
+        """,
+        "lib/b.ex" => ~S"""
+        defmodule B do
+          import A
+          A.macro()
+          macro()
+          A.fun()
+          fun()
+          def calls_macro, do: A.macro()
+          def calls_fun, do: A.fun()
+          def calls_struct, do: %A{}
+        end
+        """
+      }
+
+      output = """
+      Compiling 2 files (.ex)
+      Generated sample app
+      lib/b.ex:2: require A (export)
+      lib/b.ex:3: call A.macro/0 (compile)
+      lib/b.ex:4: import A.macro/0 (compile)
+      lib/b.ex:5: call A.fun/0 (compile)
+      lib/b.ex:6: call A.fun/0 (compile)
+      lib/b.ex:6: import A.fun/0 (compile)
+      lib/b.ex:7: call A.macro/0 (compile)
+      lib/b.ex:8: call A.fun/0 (runtime)
+      lib/b.ex:9: struct A (export)
+      """
+
+      assert_trace("lib/b.ex", files, output)
+    end
+
+    test "shows traces for module callbacks" do
+      files = %{
+        "lib/a.ex" => ~S"""
+        defmodule A do
+          @before_compile :"Elixir.B"
+          @after_compile :"Elixir.B"
+          @after_verify :"Elixir.B"
+        end
+        """,
+        "lib/b.ex" => ~S"""
+        defmodule B do
+          defmacro __before_compile__(_env), do: :ok
+          defmacro __after_compile__(_env, _binary), do: :ok
+          def __after_verify__(_module), do: :ok
+        end
+        """
+      }
+
+      output = """
+      Compiling 2 files (.ex)
+      Generated sample app
+      lib/a.ex:1: call B.__after_compile__/2 (compile)
+      lib/a.ex:1: call B.__after_verify__/1 (compile)
+      lib/a.ex:1: call B.__before_compile__/1 (compile)
+      """
+
+      assert_trace("lib/a.ex", files, output)
+    end
+
+    test "filters per label" do
+      files = %{
+        "lib/a.ex" => ~S"""
+        defmodule A do
+          defmacro macro, do: :ok
+          def fun, do: :ok
+        end
+        """,
+        "lib/b.ex" => ~S"""
+        defmodule B do
+          require A
+          def calls_macro, do: A.macro()
+          def calls_fun, do: A.fun()
+        end
+        """
+      }
+
+      output = """
+      Compiling 2 files (.ex)
+      Generated sample app
+      lib/b.ex:3: call A.macro/0 (compile)
+      """
+
+      assert_trace(~w[--label compile], "lib/b.ex", files, output)
+    end
+
+    test "fails if above limit per label" do
+      files = %{
+        "lib/a.ex" => ~S"""
+        defmodule A do
+          defmacro macro, do: :ok
+          def fun, do: :ok
+        end
+        """,
+        "lib/b.ex" => ~S"""
+        defmodule B do
+          require A
+          def calls_macro, do: A.macro()
+          def calls_fun, do: A.fun()
+        end
+        """
+      }
+
+      output = """
+      Compiling 2 files (.ex)
+      Generated sample app
+      lib/b.ex:3: call A.macro/0 (compile)
+      """
+
+      message = "Too many traces (found: 1, permitted: 0)"
+
+      assert_raise Mix.Error, message, fn ->
+        assert_trace(~w[--label compile --fail-above 0], "lib/b.ex", files, output)
+      end
+    end
+
+    defp assert_trace(opts \\ [], file, files, expected) do
+      in_fixture("no_mixfile", fn ->
+        for {file, contents} <- files do
+          File.write!(file, contents)
+        end
+
+        capture_io(:stderr, fn ->
+          assert Mix.Task.run("xref", opts ++ ["trace", file]) == :ok
         end)
 
         assert ^expected = receive_until_no_messages([])
@@ -272,12 +422,41 @@ defmodule Mix.Tasks.XrefTest do
       """)
     end
 
+    test "cycles with `--fail-above`" do
+      message = "Too many cycles (found: 1, permitted: 0)"
+
+      assert_raise Mix.Error, message, fn ->
+        assert_graph(["--format", "cycles", "--fail-above", "0"], """
+        1 cycles found. Showing them in decreasing size:
+
+        Cycle of length 3:
+
+            lib/b.ex
+            lib/a.ex
+            lib/b.ex
+
+        """)
+      end
+    end
+
     test "cycles with min cycle size" do
       assert_graph(["--format", "cycles", "--min-cycle-size", "3"], """
       No cycles found
       """)
     end
 
+    test "unknown label" do
+      assert_raise Mix.Error, "Unknown --label bad in mix xref graph", fn ->
+        assert_graph(["--label", "bad"], "")
+      end
+    end
+
+    test "unknown format" do
+      assert_raise Mix.Error, "Unknown --format bad in mix xref graph", fn ->
+        assert_graph(["--format", "bad"], "")
+      end
+    end
+
     test "exclude many" do
       assert_graph(~w[--exclude lib/c.ex --exclude lib/b.ex], """
       lib/a.ex
@@ -300,6 +479,39 @@ defmodule Mix.Tasks.XrefTest do
       """)
     end
 
+    @abc_linear_files %{
+      "lib/a.ex" => "defmodule A, do: def a, do: B.b()",
+      "lib/b.ex" => "defmodule B, do: def b, do: C.c()",
+      "lib/c.ex" => "defmodule C, do: def c, do: true"
+    }
+
+    test "exclude one from linear case" do
+      assert_graph(
+        ~w[--exclude lib/b.ex],
+        """
+        lib/a.ex
+        lib/c.ex
+        """,
+        files: @abc_linear_files
+      )
+    end
+
+    test "exclude one with source from linear case" do
+      assert_graph(
+        ~w[--exclude lib/b.ex --source lib/a.ex],
+        """
+        lib/a.ex
+        """,
+        files: @abc_linear_files
+      )
+    end
+
+    test "invalid exclude" do
+      assert_raise Mix.Error, "Excluded files could not be found: lib/a2.ex, lib/a3.ex", fn ->
+        assert_graph(~w[--exclude lib/a2.ex --exclude lib/a.ex --exclude lib/a3.ex], "")
+      end
+    end
+
     test "only nodes" do
       assert_graph(~w[--only-nodes], """
       lib/a.ex
@@ -310,21 +522,68 @@ defmodule Mix.Tasks.XrefTest do
       """)
     end
 
+    test "only nodes with compile direct label" do
+      assert_graph(~w[--label compile --only-direct --only-nodes], """
+      lib/a.ex
+      lib/b.ex
+      lib/c.ex
+      """)
+    end
+
     test "filter by compile label" do
       assert_graph(~w[--label compile], """
       lib/a.ex
       `-- lib/b.ex (compile)
       lib/b.ex
-      |-- lib/d.ex (compile)
       `-- lib/e.ex (compile)
       lib/c.ex
       `-- lib/d.ex (compile)
-      lib/d.ex
-      lib/e.ex
       """)
     end
 
-    test "filter by compile label with only direct" do
+    test "filter by compile-connected label" do
+      assert_graph(~w[--label compile-connected], """
+      lib/a.ex
+      `-- lib/b.ex (compile)
+      lib/c.ex
+      `-- lib/d.ex (compile)
+      """)
+    end
+
+    test "filter by compile-connected label with exclusions" do
+      assert_graph(~w[--label compile-connected --exclude lib/e.ex], """
+      lib/a.ex
+      `-- lib/b.ex (compile)
+      """)
+    end
+
+    test "filter by compile-connected label with fail-above" do
+      message = "Too many references (found: 2, permitted: 1)"
+
+      assert_raise Mix.Error, message, fn ->
+        assert_graph(~w[--label compile-connected --fail-above 1], """
+        lib/a.ex
+        `-- lib/b.ex (compile)
+        lib/c.ex
+        `-- lib/d.ex (compile)
+        """)
+      end
+    end
+
+    test "exclude many with fail-above" do
+      message = "Too many references (found: 1, permitted: 0)"
+
+      assert_raise Mix.Error, message, fn ->
+        assert_graph(~w[--exclude lib/c.ex --exclude lib/b.ex --fail-above 0], """
+        lib/a.ex
+        lib/d.ex
+        `-- lib/e.ex
+        lib/e.ex
+        """)
+      end
+    end
+
+    test "filter by compile direct label" do
       assert_graph(~w[--label compile --only-direct], """
       lib/a.ex
       `-- lib/b.ex (compile)
@@ -332,48 +591,29 @@ defmodule Mix.Tasks.XrefTest do
       `-- lib/e.ex (compile)
       lib/c.ex
       `-- lib/d.ex (compile)
-      lib/d.ex
-      lib/e.ex
       """)
     end
 
     test "filter by runtime label" do
       assert_graph(~w[--label runtime], """
-      lib/a.ex
-      `-- lib/c.ex
       lib/b.ex
       |-- lib/a.ex
       `-- lib/c.ex
-      lib/c.ex
-      `-- lib/e.ex
       lib/d.ex
       `-- lib/e.ex
-      lib/e.ex
       """)
     end
 
-    test "filter by runtime label with only direct" do
-      assert_graph(~w[--label runtime --only-direct], """
-      lib/a.ex
-      lib/b.ex
-      |-- lib/a.ex
-      `-- lib/c.ex
-      lib/c.ex
-      lib/d.ex
-      `-- lib/e.ex
-      lib/e.ex
-      """)
-    end
-
-    test "source" do
-      assert_graph(~w[--source lib/a.ex], """
+    test "sources" do
+      assert_graph(~w[--source lib/a.ex --source lib/c.ex], """
       lib/a.ex
       `-- lib/b.ex (compile)
           |-- lib/a.ex
           |-- lib/c.ex
-          |   `-- lib/d.ex (compile)
-          |       `-- lib/e.ex
           `-- lib/e.ex (compile)
+      lib/c.ex
+      `-- lib/d.ex (compile)
+          `-- lib/e.ex
       """)
     end
 
@@ -381,12 +621,18 @@ defmodule Mix.Tasks.XrefTest do
       assert_graph(~w[--source lib/a.ex --label compile], """
       lib/a.ex
       `-- lib/b.ex (compile)
-          |-- lib/d.ex (compile)
           `-- lib/e.ex (compile)
       """)
     end
 
-    test "source with compile label and only direct" do
+    test "source with compile-connected label" do
+      assert_graph(~w[--source lib/a.ex --label compile-connected], """
+      lib/a.ex
+      `-- lib/b.ex (compile)
+      """)
+    end
+
+    test "source with compile direct label" do
       assert_graph(~w[--source lib/a.ex --label compile --only-direct], """
       lib/a.ex
       `-- lib/b.ex (compile)
@@ -394,9 +640,9 @@ defmodule Mix.Tasks.XrefTest do
       """)
     end
 
-    test "invalid source" do
-      assert_raise Mix.Error, "Source could not be found: lib/a2.ex", fn ->
-        assert_graph(~w[--source lib/a2.ex], "")
+    test "invalid sources" do
+      assert_raise Mix.Error, "Sources could not be found: lib/a2.ex, lib/a3.ex", fn ->
+        assert_graph(~w[--source lib/a2.ex --source lib/a.ex --source lib/a3.ex], "")
       end
     end
 
@@ -426,7 +672,16 @@ defmodule Mix.Tasks.XrefTest do
       """)
     end
 
-    test "sink with compile label and only direct" do
+    test "sink with compile-connected label" do
+      assert_graph(~w[--sink lib/e.ex --label compile-connected], """
+      lib/a.ex
+      `-- lib/b.ex (compile)
+      lib/c.ex
+      `-- lib/d.ex (compile)
+      """)
+    end
+
+    test "sink with compile direct label" do
       assert_graph(~w[--sink lib/e.ex --label compile --only-direct], """
       lib/a.ex
       `-- lib/b.ex (compile)
@@ -435,9 +690,25 @@ defmodule Mix.Tasks.XrefTest do
       """)
     end
 
+    test "multiple sinks" do
+      assert_graph(~w[--sink lib/a.ex --sink lib/c.ex], """
+      lib/b.ex
+      |-- lib/a.ex
+      |   `-- lib/b.ex (compile)
+      `-- lib/c.ex
+      """)
+    end
+
+    test "multiple sinks with only nodes" do
+      assert_graph(~w[--sink lib/a.ex --sink lib/c.ex --sink lib/e.ex --only-nodes], """
+      lib/b.ex
+      lib/d.ex
+      """)
+    end
+
     test "invalid sink" do
-      assert_raise Mix.Error, "Sink could not be found: lib/b2.ex", fn ->
-        assert_graph(~w[--sink lib/b2.ex], "")
+      assert_raise Mix.Error, "Sinks could not be found: lib/b2.ex, lib/b3.ex", fn ->
+        assert_graph(~w[--sink lib/b2.ex --sink lib/b.ex --sink lib/b3.ex], "")
       end
     end
 
@@ -537,7 +808,7 @@ defmodule Mix.Tasks.XrefTest do
                digraph "xref graph" {
                  "lib/a.ex"
                  "lib/a.ex" -> "lib/b.ex" [label="(compile)"]
-                 "lib/b.ex" -> "lib/a.ex" [label="(export)"]
+                 "lib/b.ex" -> "lib/a.ex" [label="(compile)"]
                  "lib/b.ex"
                }
                """
@@ -587,7 +858,7 @@ defmodule Mix.Tasks.XrefTest do
       end)
     end
 
-    test "compiles project first by default" do
+    test "skip project compilation with --no-compile" do
       in_fixture("no_mixfile", fn ->
         File.write!("lib/a.ex", """
         defmodule A do
@@ -595,73 +866,106 @@ defmodule Mix.Tasks.XrefTest do
         end
         """)
 
-        Mix.Tasks.Xref.run(["graph"])
-
-        assert "Compiling" <> _ = receive_until_no_messages([])
-      end)
-    end
-
-    test "passes args over to compile task" do
-      in_fixture("no_mixfile", fn ->
-        File.write!("lib/a.ex", """
-        defmodule A do
-          def a, do: :ok
-        end
-        """)
-
-        Mix.Task.run("compile")
-        Mix.Task.reenable("compile")
-        Mix.shell().flush()
-
         Mix.Tasks.Xref.run(["graph", "--no-compile"])
-
-        refute String.starts_with?(receive_until_no_messages([]), "Compiling")
+        refute receive_until_no_messages([]) =~ "lib/a.ex"
       end)
     end
 
-    defp assert_graph(opts \\ [], expected) do
+    test "group with multiple unconnected files" do
+      assert_graph(~w[--group lib/a.ex,lib/c.ex,lib/e.ex], """
+      lib/a.ex+
+      |-- lib/b.ex (compile)
+      `-- lib/d.ex (compile)
+      lib/b.ex
+      `-- lib/a.ex+ (compile)
+      lib/d.ex
+      `-- lib/a.ex+
+      """)
+    end
+
+    test "group with directly dependent files and cycle" do
+      assert_graph(~w[--group lib/a.ex,lib/b.ex], """
+      lib/a.ex+
+      |-- lib/c.ex
+      `-- lib/e.ex (compile)
+      lib/c.ex
+      `-- lib/d.ex (compile)
+      lib/d.ex
+      `-- lib/e.ex
+      lib/e.ex
+      """)
+    end
+
+    test "multiple groups" do
+      assert_graph(~w[--group lib/a.ex,lib/b.ex --group lib/c.ex,lib/e.ex], """
+      lib/a.ex+
+      `-- lib/c.ex+ (compile)
+      lib/c.ex+
+      `-- lib/d.ex (compile)
+      lib/d.ex
+      `-- lib/c.ex+
+      """)
+    end
+
+    test "group with sink" do
+      assert_graph(~w[--group lib/a.ex,lib/c.ex,lib/e.ex --sink lib/e.ex], """
+      lib/b.ex
+      `-- lib/a.ex+ (compile)
+          |-- lib/b.ex (compile)
+          `-- lib/d.ex (compile)
+      lib/d.ex
+      `-- lib/a.ex+
+      """)
+    end
+
+    @default_files %{
+      "lib/a.ex" => """
+      defmodule A do
+        def a, do: :ok
+        B.b2()
+      end
+      """,
+      "lib/b.ex" => """
+      defmodule B do
+        def b1, do: A.a() == C.c()
+        def b2, do: :ok
+        :e.e()
+      end
+      """,
+      "lib/c.ex" => """
+      defmodule C do
+        def c, do: :ok
+        :d.d()
+      end
+      """,
+      "lib/d.ex" => """
+      defmodule :d do
+        def d, do: :ok
+        def e, do: :e.e()
+      end
+      """,
+      "lib/e.ex" => """
+      defmodule :e do
+        def e, do: :ok
+      end
+      """
+    }
+
+    defp assert_graph(opts \\ [], expected, params \\ []) do
       in_fixture("no_mixfile", fn ->
-        File.write!("lib/a.ex", """
-        defmodule A do
-          def a, do: :ok
-          B.b2()
-        end
-        """)
-
-        File.write!("lib/b.ex", """
-        defmodule B do
-          def b1, do: A.a() == C.c()
-          def b2, do: :ok
-          :e.e()
-        end
-        """)
-
-        File.write!("lib/c.ex", """
-        defmodule C do
-          def c, do: :ok
-          :d.d()
-        end
-        """)
-
-        File.write!("lib/d.ex", """
-        defmodule :d do
-          def d, do: :ok
-          def e, do: :e.e()
-        end
-        """)
-
-        File.write!("lib/e.ex", """
-        defmodule :e do
-          def e, do: :ok
-        end
-        """)
+        nb_files =
+          Enum.count(params[:files] || @default_files, fn {path, content} ->
+            File.write!(path, content)
+          end)
 
         assert Mix.Task.run("xref", opts ++ ["graph"]) == :ok
+        first_line = "Compiling #{nb_files} files (.ex)"
 
-        assert "Compiling 5 files (.ex)\nGenerated sample app\n" <> result =
-                 receive_until_no_messages([])
+        assert [
+                 ^first_line | ["Generated sample app" | result]
+               ] = receive_until_no_messages([]) |> String.split("\n")
 
-        assert normalize_graph_output(result) == expected
+        assert normalize_graph_output(result |> Enum.join("\n")) == expected
       end)
     end
 
