@@ -94,23 +94,6 @@ defmodule Mix.UmbrellaTest do
     end)
   end
 
-  test "recompiles umbrella on config change" do
-    in_fixture("umbrella_dep/deps/umbrella", fn ->
-      Mix.Project.in_project(:umbrella, ".", fn _ ->
-        Mix.Task.run("compile", [])
-        bar = File.stat!("_build/dev/lib/bar/.mix/compile.elixir").mtime
-        foo = File.stat!("_build/dev/lib/foo/.mix/compile.elixir").mtime
-
-        ensure_touched("mix.exs", max(foo, bar))
-
-        Mix.Task.clear()
-        Mix.Task.run("compile", [])
-        assert File.stat!("_build/dev/lib/bar/.mix/compile.elixir").mtime > bar
-        assert File.stat!("_build/dev/lib/foo/.mix/compile.elixir").mtime > foo
-      end)
-    end)
-  end
-
   test "recursively compiles umbrella with protocol consolidation" do
     in_fixture("umbrella_dep/deps/umbrella", fn ->
       Mix.Project.in_project(:umbrella, ".", fn _ ->
@@ -143,9 +126,9 @@ defmodule Mix.UmbrellaTest do
   end
 
   test "loads umbrella dependencies" do
-    Mix.Project.push(UmbrellaDeps)
-
     in_fixture("umbrella_dep/deps/umbrella", fn ->
+      Mix.Project.push(UmbrellaDeps)
+
       File.mkdir_p!("deps/some_dep/ebin")
       File.mkdir_p!("_build/dev/lib/some_dep/ebin")
       File.mkdir_p!("_build/dev/lib/foo/ebin")
@@ -304,9 +287,9 @@ defmodule Mix.UmbrellaTest do
   end
 
   test "handles dependencies with cycles" do
-    Mix.Project.push(CycleDeps)
-
     in_fixture("umbrella_dep", fn ->
+      Mix.Project.push(CycleDeps)
+
       assert Enum.map(Mix.Dep.load_on_environment([]), & &1.app) == [:foo, :bar, :umbrella]
     end)
   end
@@ -442,6 +425,7 @@ defmodule Mix.UmbrellaTest do
 
         mtime = File.stat!("_build/dev/lib/bar/.mix/compile.elixir").mtime
         ensure_touched("_build/dev/lib/foo/.mix/compile.elixir", mtime)
+        ensure_touched("_build/dev/lib/foo/ebin/foo.app", mtime)
         ensure_touched("lib/bar.ex", mtime)
 
         assert Mix.Tasks.Compile.Elixir.run(["--verbose"]) == {:ok, []}
@@ -455,12 +439,9 @@ defmodule Mix.UmbrellaTest do
       Mix.Project.in_project(:bar, "bar", fn _ ->
         File.write!("../foo/lib/foo.ex", "defmodule Foo, do: defstruct [:bar]")
 
-        Mix.Task.run("compile", ["--verbose"])
-
         # Add struct dependency
         File.write!("lib/bar.ex", "defmodule Bar, do: %Foo{bar: true}")
-
-        assert Mix.Tasks.Compile.Elixir.run(["--verbose"]) == {:ok, []}
+        Mix.Task.run("compile", ["--verbose"])
         assert_receive {:mix_shell, :info, ["Compiled lib/bar.ex"]}
 
         # Recompiles for struct dependencies
@@ -474,6 +455,37 @@ defmodule Mix.UmbrellaTest do
     end)
   end
 
+  test "recompiles after compile through runtime path dependency changes" do
+    in_fixture("umbrella_dep/deps/umbrella/apps", fn ->
+      Mix.Project.in_project(:bar, "bar", fn _ ->
+        File.write!("../foo/lib/foo.bar.ex", """
+        defmodule Foo.Bar do
+          def hello, do: Foo.Baz.hello()
+        end
+        """)
+
+        File.write!("../foo/lib/foo.baz.ex", """
+        defmodule Foo.Baz do
+          def hello, do: "from bar"
+        end
+        """)
+
+        # Add compile time to Foo.Bar
+        File.write!("lib/bar.ex", "defmodule Bar, do: Foo.Bar.hello()")
+        Mix.Task.run("compile", ["--verbose"])
+        assert_receive {:mix_shell, :info, ["Compiled lib/bar.ex"]}
+
+        # Recompiles for due to compile dependency via runtime dependencies
+        mtime = File.stat!("_build/dev/lib/bar/.mix/compile.elixir").mtime
+        ensure_touched("_build/dev/lib/foo/ebin/Elixir.Foo.Baz.beam", mtime)
+        ensure_touched("_build/dev/lib/foo/.mix/compile.elixir", mtime)
+
+        assert Mix.Tasks.Compile.Elixir.run(["--verbose"]) == {:ok, []}
+        assert_receive {:mix_shell, :info, ["Compiled lib/bar.ex"]}
+      end)
+    end)
+  end
+
   test "reloads app in app tracer if .app changes" do
     in_fixture("umbrella_dep/deps/umbrella/apps", fn ->
       deps = [{:foo, in_umbrella: true}]
@@ -495,9 +507,7 @@ defmodule Mix.UmbrellaTest do
 
         Mix.Task.clear()
         Application.unload(:foo)
-
-        mtime = File.stat!("_build/dev/lib/bar/.mix/compile.elixir").mtime
-        ensure_touched("../foo/lib/foo.ex", mtime)
+        ensure_touched("../foo/lib/foo.ex", "_build/dev/lib/bar/.mix/compile.app_tracer")
 
         assert Mix.Task.run("compile", ["--verbose"]) == {:ok, []}
         assert_receive {:mix_shell, :info, ["Compiled lib/bar.ex"]}
@@ -525,8 +535,10 @@ defmodule Mix.UmbrellaTest do
         # Mark protocol as outdated
         File.touch!("_build/dev/lib/bar/consolidated/Elixir.Foo.beam", {{2010, 1, 1}, {0, 0, 0}})
 
-        mtime = File.stat!("_build/dev/lib/bar/.mix/compile.protocols").mtime
-        ensure_touched("_build/dev/lib/foo/ebin/Elixir.Foo.beam", mtime)
+        ensure_touched(
+          "_build/dev/lib/foo/ebin/Elixir.Foo.beam",
+          "_build/dev/lib/bar/.mix/compile.protocols"
+        )
 
         assert Mix.Tasks.Compile.Protocols.run([]) == :ok
 
